Require Import Coq.omega.Omega.
Require Export DeclarationEvaluation.
Require Export DeclarationTyping.
Require Export InfrastructureOperations.
Require Export InfrastructureLemmasTerms.
Require Export InfrastructureLemmasContext.
Set Implicit Arguments.

(******************************************************************************)
(* Record field lookup                                                        *)
(******************************************************************************)

Lemma option_map_id {A} (ma: option A) :
  option_map (λ x, x) ma = ma.
Proof. destruct ma; auto. Qed.

Lemma option_map_com {A B C} (f: A → B) (g : B → C) (ma: option A) :
  option_map g (option_map f ma) = option_map (λ x, g (f x)) ma.
Proof.
  destruct ma; auto.
Qed.

Ltac get_comm :=
  let t:=fresh in
    intro t; induction t; simpl;
    [ auto | intros; case (eq_lab_dec _ _); auto ].

Lemma get_tshifttrec :
  ∀ (Tr : TRec) l c, tshiftTRec c Tr l = option_map (tshiftTy c) (Tr l).
Proof. get_comm. Qed.

Lemma get_tsubsttrec :
  ∀ (Tr : TRec) l X S, tsubstTRec X S Tr l = option_map (tsubstTy X S) (Tr l).
Proof. get_comm. Qed.

Lemma get_shiftrec :
  ∀ (tr : Rec) l c, shiftRec c tr l = option_map (shiftTm c) (tr l).
Proof. get_comm. Qed.

Lemma get_tshiftrec :
  ∀ (tr : Rec) l c, tshiftRec c tr l = option_map (tshiftTm c) (tr l).
Proof. get_comm. Qed.

Lemma get_weaken_rec h :
  ∀ (tr : Rec) l,
    weakenRec tr h l = option_map (fun t => weakenTm t h) (tr l).
Proof.
  induction h; simpl; intros; try rewrite option_map_id; auto.
  - rewrite <- option_map_com, get_shiftrec; f_equal; auto.
Qed.

Lemma incl_map (Sr Tr: TRec) (f : Ty → Ty) l
   (hyp: ∀ T, Tr l = Some T → ∃ S : Ty, Sr l = Some S) :
   ∀ T, option_map f (Tr l) = Some T → ∃ S : Ty, option_map f (Sr l) = Some S.
Proof.
  induction (Tr l) as [T'|]; simpl; try discriminate; intros.
  destruct (hyp T' eq_refl) as [S' Srl]; rewrite Srl; simpl; eauto.
Qed.

(******************************************************************************)
(* Well-formedness                                                            *)
(******************************************************************************)

Lemma lookup_tref_wf {h Tr} (wTr: wfTRec h Tr) :
  ∀ l T, Tr l = Some T → wfTy h T.
Proof.
  induction wTr; simpl; intros k Tk; try discriminate.
  - destruct (eq_lab_dec k _); inversion 1; subst; eauto.
Qed.

Ltac specIH :=
  match goal with
    | H: wf_env ?G, IH : wf_env ?G → _ |- _ =>
      specialize (IH H); destruct_conjs
  end.

Lemma sub_wf {Γ U V} (sub: Sub Γ U V) (wΓ: wf_env Γ) : wfTy Γ U ∧ wfTy Γ V.
Proof.
  induction sub; repeat specIH; eauto.
  - assert (wf_env (etvar Γ T1)); eauto; specIH.
    eauto using replace_etvar_wfty.
Qed.

Hint Extern 2 (wfTy _ _) =>
  match goal with
    | wG: wf_env ?G, sub: Sub ?G _ _ |- _ =>
      destruct (sub_wf sub wG); clear sub
  end.

Lemma ptyping_wf {Γ p T Δ} (wt: PTyping Γ p T Δ) :
   wf_env Γ → wfTy Γ T ∧ wf_env (extend Γ Δ).
Proof.
  induction wt using ind_ptyping with
  (P0 := fun Γ pr Tr Δpr wpr => wf_env Γ → wfTRec Γ Tr ∧ wf_env (extend Γ Δpr));
    intros; simpl; repeat specIH; try repeat econstructor; eauto.
  - eauto using strengthen_extend_wftrec.
  - rewrite extend_append; eauto.
Qed.

Lemma typing_wf {Γ t T} (wt: Typing Γ t T) : wf_env Γ → wfTy Γ T.
Proof.
  induction wt using ind_typing with
  (P0 := fun Γ tr Tr wt => wf_env Γ → wfTRec Γ Tr);
    simpl; intros wfΓ; repeat specIH; try econstructor; eauto.
  - assert (wf_env (evar Γ T1)) by eauto; eauto using strengthen_evar_wfty.
  - assert (wf_env (extend Γ Δ)) by apply (ptyping_wf p0 wfΓ);
      eauto using strengthen_extend_wfty.
  - inversion IHwt; subst. eapply (lookup_tref_wf H0); eauto.
Qed.

(******************************************************************************)
(* Weakening lemmas                                                           *)
(******************************************************************************)

Lemma PTyping_bindPat_lengthExt {Γ p T Δ} (wp: PTyping Γ p T Δ) :
  bindPat p = lengthExt Δ.
Proof.
  induction wp using ind_ptyping with
  (P0 := fun Γ pr Tr Δ wpr => bindPRec pr = lengthExt Δ);
    simpl; try rewrite lengthExt_append; try congruence; eauto.
Qed.

Lemma shift_etvar_sub {Γ1 U V} (sub: Sub Γ1 U V) :
  ∀ {c Γ2}, shift_etvar c Γ1 Γ2 → Sub Γ2 (tshiftTy c U) (tshiftTy c V).
Proof.
  induction sub; simpl in *; intros; try (econstructor; eauto; fail).
  - constructor; eauto.
    + intro l; repeat rewrite get_tshifttrec; eauto using incl_map.
    + intros l S2 T2; repeat rewrite get_tshifttrec; intros E1 E2.
      specialize (H l).
      induction (Tr l); try discriminate; inversion E1.
      induction (Sr l); try discriminate; inversion E2.
      eauto.
Qed.
Hint Resolve shift_etvar_sub.

Lemma shift_evar_sub {Γ1 U V} (sub: Sub Γ1 U V) :
  ∀ {c Γ2}, shift_evar c Γ1 Γ2 → Sub Γ2 U V.
Proof.
  induction sub; simpl; econstructor; eauto.
Qed.
Hint Resolve shift_evar_sub.

Lemma weaken_sub {Γ U V} (sub: Sub Γ U V) :
  ∀ Δ, Sub (extend Γ Δ) U V.
Proof.
  induction Δ; simpl; eauto using shift_evar_sub.
Qed.

Lemma shift_etvar_ptyping {Γ1 p T Δ} (wt: PTyping Γ1 p T Δ) :
  ∀ {c Γ2}, shift_etvar c Γ1 Γ2 →
    PTyping Γ2 (tshiftPat c p) (tshiftTy c T) (tshiftExt c Δ).
Proof.
  induction wt using ind_ptyping with
  (P0 := fun Γ1 pr Tr Δ wpr =>
           ∀ {c Γ2}, shift_etvar c Γ1 Γ2 →
             PRTyping Γ2 (tshiftPRec c pr) (tshiftTRec c Tr) (tshiftExt c Δ));
    simpl; intros; autorewrite with interaction; econstructor; simpl; eauto.
  - rewrite get_tshifttrec, e; auto.
Qed.

Lemma shift_etvar_typing {Γ1 t T} (wt: Typing Γ1 t T) :
  ∀ {c Γ2}, shift_etvar c Γ1 Γ2 → Typing Γ2 (tshiftTm c t) (tshiftTy c T).
Proof.
  induction wt using ind_typing with
  (P0 := fun Γ1 tr Tr wtr =>
           ∀ {c Γ2}, shift_etvar c Γ1 Γ2 →
             RTyping Γ2 (tshiftRec c tr) (tshiftTRec c Tr)); simpl; intros;
    try (try rewrite tshiftTy_tsubstTy_comm; simpl;
         econstructor; eauto using shift_etvar_ptyping; fail).
  - apply T_Proj with (Tr := tshiftTRec c Tr); eauto.
    + rewrite get_tshifttrec, lIn; trivial.
Qed.

Lemma shift_evar_ptyping {Γ1 p T Δ} (wp: PTyping Γ1 p T Δ) :
  ∀ {c Γ2}, shift_evar c Γ1 Γ2 → PTyping Γ2 p T Δ.
Proof.
  induction wp using ind_ptyping with
  (P0 := fun Γ1 pr Tr Δ wpr =>
           ∀ {c Γ2}, shift_evar c Γ1 Γ2 →
             PRTyping Γ2 pr Tr Δ);
    simpl; econstructor; eauto.
Qed.

Lemma shift_evar_typing {Γ1 t T} (wt: Typing Γ1 t T) :
  ∀ {c Γ2}, shift_evar c Γ1 Γ2 → Typing Γ2 (shiftTm c t) T.
Proof.
  induction wt using ind_typing with
  (P0 := fun Γ1 tr Tr wtr =>
           ∀ {c Γ2}, shift_evar c Γ1 Γ2 → RTyping Γ2 (shiftRec c tr) Tr);
    simpl; econstructor; eauto using shift_evar_ptyping.
  - rewrite (PTyping_bindPat_lengthExt p0); simpl; eauto.
Qed.

Lemma weaken_typing {Γ} Δ :
  ∀ {t T}, Typing Γ t T → Typing (extend Γ Δ) (weakenTm t (lengthExt Δ)) T.
Proof.
  induction Δ; simpl; eauto using shift_evar_typing.
Qed.

Lemma shift_value_rvalue :
  (∀ t {c}, Value t → Value (shiftTm c t)) ∧
  (∀ ts {c}, RValue ts → RValue (shiftRec c ts)).
Proof.
  apply ind_Tm_Rec; simpl; intros; try contradiction; destruct_conjs; auto.
Qed.

Lemma tshift_value_rvalue :
  (∀ t {c}, Value t → Value (tshiftTm c t)) ∧
  (∀ ts {c}, RValue ts → RValue (tshiftRec c ts)).
Proof.
  apply ind_Tm_Rec; simpl; intros; try contradiction; destruct_conjs; auto.
Qed.

Lemma weaken_value u :
  ∀ {t}, Value t → Value (weakenTm t u).
Proof.
  induction u; simpl; auto; intros.
  - apply shift_value_rvalue; auto.
Qed.

Lemma weaken_rvalue u :
  ∀ {vr}, RValue vr → RValue (weakenRec vr u).
Proof.
  induction u; simpl; auto; intros.
  - apply shift_value_rvalue; auto.
Qed.

(******************************************************************************)
(* Reflexivity, transitivity and narrowing                                    *)
(******************************************************************************)

Lemma sub_refl {Γ T} (wT: wfTy Γ T) : wf_env Γ → Sub Γ T T.
Proof.
  induction wT using ind_wfTy with
  (P0 := fun Γ (Tr : TRec) wTr => wf_env Γ →
             ∀ l T, Tr l = Some T → Sub Γ T T); simpl; intros; subst;
  try discriminate; try specIH; try (econstructor; eauto; fail).
  - constructor; eauto.
    + intros l S T El1 El2; rewrite El1 in El2; inversion El2; subst; eauto.
  - destruct (eq_lab_dec l0 l); eauto.
    + inversion H0; subst; auto.
Qed.

Inductive narrow : nat → Env → Env → Prop :=
  | nw_here {Γ T1 T2} :
      Sub Γ T2 T1 → narrow O (etvar Γ T1) (etvar Γ T2)
  | nw_evar {X Γ1 Γ2 T} :
      narrow X Γ1 Γ2 → narrow X (evar Γ1 T) (evar Γ2 T)
  | nw_etvar {X Γ1 Γ2 T} :
      narrow X Γ1 Γ2 → narrow (S X) (etvar Γ1 T) (etvar Γ2 T).
Hint Constructors narrow.

Lemma weaken_narrow Δ :
  ∀ {X Γ1 Γ2}, narrow X Γ1 Γ2 → narrow X (extend Γ1 Δ) (extend Γ2 Δ).
Proof.
  induction Δ; simpl; intros; try constructor; auto.
Qed.

Lemma narrow_lookup_etvar_eq {X Γ1 Γ2} (nw : narrow X Γ1 Γ2) :
  ∃ T1 T2, lookup_etvar Γ1 X T1 ∧ lookup_etvar Γ2 X T2 ∧ Sub Γ2 T2 T1.
Proof with destruct_conjs; repeat split; eauto.
  induction nw; intros.
  - exists (tshiftTy O T1),(tshiftTy O T2)...
  - destruct IHnw as (T1 & T2 & ?); exists T1, T2...
  - destruct IHnw as (T1 & T2 & ?); exists (tshiftTy O T1),(tshiftTy O T2)...
Qed.

Lemma narrow_lookup_etvar_ne {X Γ1 Γ2} (nw: narrow X Γ1 Γ2) :
  ∀ {T1 Y}, X ≠ Y → lookup_etvar Γ1 Y T1 → lookup_etvar Γ2 Y T1.
Proof.
  induction nw; inversion 2; subst; simpl; intuition.
Qed.

Lemma narrow_wfty_wftrec : ∀ Γ1,
  (∀ T (wT: wfTy Γ1 T), ∀ {X Γ2}, narrow X Γ1 Γ2 → wfTy Γ2 T) ∧
  (∀ Tr (wTr: wfTRec Γ1 Tr), ∀ {X Γ2}, narrow X Γ1 Γ2 → wfTRec Γ2 Tr).
Proof.
  apply ind_wfty_wftrec; simpl; eauto.
  - clear; intros Γ1 X T l Y Γ2 nw.
    case (eq_nat_dec X Y); intros; subst; simpl.
    + pose proof (narrow_lookup_etvar_eq nw); destruct_conjs; eauto.
    + eauto using narrow_lookup_etvar_ne.
Qed.

Lemma narrow_wfty {Γ1 T} (wT: wfTy Γ1 T) :
  ∀ {X Γ2}, narrow X Γ1 Γ2 → wfTy Γ2 T.
Proof. apply narrow_wfty_wftrec; auto. Qed.

Lemma narrow_wftrec {Γ1 Tr} (wT: wfTRec Γ1 Tr) :
  ∀ {X Γ2}, narrow X Γ1 Γ2 → wfTRec Γ2 Tr.
Proof. apply narrow_wfty_wftrec; auto. Qed.

Lemma narrow_wf_env {X Γ1 Γ2} (nw : narrow X Γ1 Γ2) : wf_env Γ1 → wf_env Γ2.
Proof.
  induction nw; inversion 1; subst; simpl; eauto using narrow_wfty.
Qed.

Definition Trans Q :=
  ∀ Γ S T, wf_env Γ → Sub Γ S Q → Sub Γ Q T → Sub Γ S T.

Definition Narrow Q :=
  ∀ Γ1 Γ2 Y, wf_env Γ1 → narrow Y Γ1 Γ2 →
    lookup_etvar Γ1 Y Q → ∀ S T, Sub Γ1 S T → Sub Γ2 S T.

Lemma size_lookup {Tr : TRec} {k T} :
  Tr k = Some T → sizeTy T < S (sizeTRec Tr).
Proof.
  induction Tr; simpl; intros E; try discriminate.
  - destruct (eq_lab_dec k l).
    + inversion E; omega.
    + specialize (IHTr E); omega.
Qed.

Lemma trans_case Q
  (hypT : ∀ R, sizeTy R < sizeTy Q → Trans R)
  (hypN : ∀ R, sizeTy R < sizeTy Q → Narrow R) :
  Trans Q.
Proof with simpl; eauto; try omega.
  intros Γ S T wΓ subSQ subQT; revert T subQT.
  induction subSQ; intros; auto.
  - dependent destruction subQT; constructor; auto.
  - eapply SA_Trans_TVar; eauto.
  - dependent destruction subQT; simpl; constructor...
    + apply (hypT T1)...
    + apply (hypT T2)...
  - assert (wf_env (etvar Γ T1)) by eauto.
    assert (wfTy (etvar Γ T1) S2) by eauto.
    assert (wfTy (etvar Γ S1) S2) by eauto using replace_etvar_wfty.
    dependent destruction subQT; constructor...
    + apply (hypT T1)...
    + apply (hypT T2)...
      refine (hypN _ _ _ _ _ _ (nw_here subQT1) gb_here _ _ subSQ2);
        try rewrite tshift_preserves_size...
  - dependent destruction subQT; simpl; constructor...
    + intros l ? Tr0l; destruct (incl0 _ _ Tr0l) as [? Trl].
      destruct (incl _ _ Trl) as [? Srl]; eauto.
    + intros l S T0 Tr0l Srl; destruct (incl0 _ _ Tr0l) as [T Trl].
      apply (hypT T); eauto using (size_lookup Trl).
Qed.

Lemma narrow_case Q (hyp : ∀ (R : Ty), sizeTy R = sizeTy Q → Trans R) :
  Narrow Q.
Proof.
  intros Γ1 Γ2 Y wΓ nw gb U T subUT; revert Q hyp Γ2 Y wΓ nw gb.
  induction subUT; intros.
  - apply SA_Top; eauto using narrow_wfty.
  - eapply SA_Refl_TVar; eauto using narrow_wfty.
  - case (eq_nat_dec X Y); intros; subst.
    + destruct (narrow_lookup_etvar_eq nw) as [T1 [T2 [gb1 [gb2 sub]]]].
      pose proof (lookup_etvar_functional gb H).
      pose proof (lookup_etvar_functional gb gb1).
      apply (hyp U); repeat subst; eauto using narrow_wf_env.
      apply (SA_Trans_TVar gb2 sub); eauto.
    + apply (SA_Trans_TVar (U := U)); eauto.
      apply (narrow_lookup_etvar_ne nw); auto.
  - constructor; eauto.
  - constructor; eauto.
    refine (IHsubUT2 _ _ _ _ _ (nw_etvar nw) (gb_bound gb));
      try rewrite tshift_preserves_size; auto.
  - constructor; eauto using narrow_wftrec.
Qed.

Lemma sizeind (P : Ty → Prop) (step: ∀ T, (∀ S, sizeTy S < sizeTy T → P S) → P T) :
  ∀ T, P T.
Proof.
  cut (forall n T, sizeTy T <= n → P T); eauto.
  induction n; intros; apply step; intros; try apply IHn; omega.
Qed.

Lemma trans_and_narrow Q : Trans Q ∧ Narrow Q.
Proof.
  induction Q using sizeind; intros; split.
  apply trans_case; apply H; auto.
  apply narrow_case; intros R eq.
  apply trans_case; rewrite eq; apply H; auto.
Qed.

Lemma sub_trans {Q Γ S T} (wΓ: wf_env Γ) : Sub Γ S Q → Sub Γ Q T → Sub Γ S T.
Proof.
  apply (trans_and_narrow); auto.
Qed.

Lemma narrow_sub {X Γ1 Γ2 S T} (wΓ1: wf_env Γ1) (nw: narrow X Γ1 Γ2) :
  Sub Γ1 S T → Sub Γ2 S T.
Proof.
  destruct (narrow_lookup_etvar_eq nw); destruct_conjs.
  eapply trans_and_narrow; eauto.
Qed.

Lemma narrow_ptyping {Γ1 p T Δ} (wp: PTyping Γ1 p T Δ) :
  wf_env Γ1 → ∀ {X Γ2}, narrow X Γ1 Γ2 → PTyping Γ2 p T Δ.
Proof.
  apply ind_ptyping_prtyping with
  (P := fun Γ1 p T Δp wp => wf_env Γ1 →
          ∀ (X : nat) (Γ2 : Env), narrow X Γ1 Γ2 → PTyping Γ2 p T Δp)
  (P0 := fun Γ1 pr Tr Δpr wpr => wf_env Γ1 →
          ∀ {X Γ2}, narrow X Γ1 Γ2 → PRTyping Γ2 pr Tr Δpr);
    simpl; intros; try econstructor; eauto using narrow_wfty.
  - assert (wf_env (extend Γ Δp)) by apply (ptyping_wf p1 H1).
    eapply H0; eauto using weaken_narrow.
Qed.

Lemma narrow_lookup_evar {X Γ1 Γ2} (nw: narrow X Γ1 Γ2) :
  ∀ {T x}, lookup_evar Γ1 x T → lookup_evar Γ2 x T.
Proof.
  induction nw; inversion 1; subst; simpl; intuition.
Qed.

Lemma narrow_typing {Γ1 t T} (wt: Typing Γ1 t T) :
  wf_env Γ1 → ∀ {X Γ2}, narrow X Γ1 Γ2 → Typing Γ2 t T.
Proof.
  apply ind_typing_rtyping with
  (P := fun Γ1 t T wt => wf_env Γ1 → ∀ {X Γ2}, narrow X Γ1 Γ2 → Typing Γ2 t T)
  (P0 := fun Γ1 tr Tr wtr => wf_env Γ1 →
    ∀ {X Γ2}, narrow X Γ1 Γ2 → RTyping Γ2 tr Tr); simpl; intros; auto;
    try (econstructor; eauto using narrow_lookup_evar, narrow_sub,
                       narrow_wfty, narrow_ptyping; fail).
  - assert (wf_env (extend Γ Δ)) by eapply (ptyping_wf p0 H1).
    econstructor; eauto using weaken_narrow, narrow_ptyping.
Qed.

Lemma typing_narrow0 {Γ Q R t T} (sub: Sub Γ R Q) (wΓ: wf_env Γ) :
  Typing (etvar Γ Q) t T → Typing (etvar Γ R) t T.
Proof. eauto using narrow_typing, narrow. Qed.

(******************************************************************************)
(* Substitution lemmas                                                        *)
(******************************************************************************)

Lemma subst_etvar_lookup_etvar {S X Γ1 Γ2}
  (wΓ1: wf_env Γ1) (esub: subst_etvar S X Γ1 Γ2) :
  ∀ {Y T}, lookup_etvar Γ1 Y T → Sub Γ2 (tsubstIndex X S Y) (tsubstTy X S T).
Proof.
  induction esub; inversion wΓ1; inversion 1; subst; simpl; intros;
    try rewrite tsubstTy_tshiftTy_reflection;
    try rewrite tsubstTy_tshiftTy_comm; eauto using Sub, sub_refl.
  refine (SA_Trans_TVar _ (sub_refl _ _)); eauto using subst_etvar_wf_env.
Qed.

Lemma subst_etvar_sub {S Γ1 U V} (wΓ1: wf_env Γ1) (sub: Sub Γ1 U V) :
  ∀ {X Γ2}, subst_etvar S X Γ1 Γ2 → Sub Γ2 (tsubstTy X S U) (tsubstTy X S V).
Proof.
  induction sub; simpl; intros; try (econstructor; eauto; fail).
  - apply sub_refl; eauto using subst_etvar_lookup_etvar_wfty, subst_etvar_wf_env.
  - apply (sub_trans (Q := tsubstTy X0 S U));
      eauto using subst_etvar_wf_env, subst_etvar_lookup_etvar.
  - constructor; eauto.
    + intro l; specialize (incl l).
      repeat rewrite get_tsubsttrec; eauto using incl_map.
    + intros l S2 T2; repeat rewrite get_tsubsttrec; intros E1 E2.
      specialize (H l).
      induction (Tr l); try discriminate; inversion E1.
      induction (Sr l); try discriminate; inversion E2.
      eauto.
Qed.

Lemma subst_etvar_ptyping {Γ B S Γ1 p T Δp}
  (wfS: wfTy Γ S) (sub: Sub Γ S B) (wt: PTyping Γ1 p T Δp) :
    ∀ {X Γ2}, subst_etvar S X Γ1 Γ2 →
    PTyping Γ2 (tsubstPat X S p) (tsubstTy X S T) (tsubstExt X S Δp).
Proof.
  induction wt using ind_ptyping
  with (P0 := fun Γ1 pr Tr Δpr wpr =>
    ∀ {X Γ2}, subst_etvar S X Γ1 Γ2 →
      PRTyping Γ2 (tsubstPRec X S pr) (tsubstTRec X S Tr) (tsubstExt X S Δpr));
      simpl; intros; simpl; try econstructor; eauto using subst_etvar_sub.
  - rewrite tsubstExt_append; econstructor; eauto.
    rewrite get_tsubsttrec, e; auto.
Qed.

Lemma subst_etvar_typing {Γ B S Γ1 t T} (wS: wfTy Γ S) (wΓ1: wf_env Γ1)
  (subS: Sub Γ S B) (wt: Typing Γ1 t T) :
  ∀ {X Γ2}, subst_etvar S X Γ1 Γ2 → Typing Γ2 (tsubstTm X S t) (tsubstTy X S T).
Proof.
  induction wt using ind_typing
  with (P0 := fun Γ1 tr Tr wtr => wf_env Γ1 →
                ∀ {X Γ2}, subst_etvar S X Γ1 Γ2 →
                  RTyping Γ2 (tsubstRec X S tr) (tsubstTRec X S Tr));
    simpl; intros; try (econstructor; eauto using subst_etvar_sub; fail).
  - rewrite tsubstTy_tsubstTy_comm; econstructor; eauto using subst_etvar_sub.
  - eapply T_Let; eauto using subst_etvar_ptyping.
    eapply IHwt2; eauto; eapply ptyping_wf; eauto.
  - refine (@T_Proj Γ2 _ l _ (tsubstTRec X S Tr) _ _); eauto.
    rewrite get_tsubsttrec, lIn; auto.
Qed.

Lemma subst_evar_sub {s Γ1 U V} (sub: Sub Γ1 U V) :
  ∀ {x Γ2}, subst_evar s x Γ1 Γ2 → Sub Γ2 U V.
Proof.
  induction sub; simpl; intros; econstructor;
    eauto using subst_evar_lookup_etvar.
Qed.

Lemma subst_evar_ptyping {Γ s S} (wts: Typing Γ s S)
  {Γ1 p T Δp} (wp: PTyping Γ1 p T Δp) :
  ∀ {x Γ2}, subst_evar s x Γ1 Γ2 → PTyping Γ2 p T Δp.
Proof.
  induction wp using ind_ptyping
  with (P0 := fun Γ1 pr Tr Δpr wpr =>
      ∀ {x Γ2}, subst_evar s x Γ1 Γ2 → PRTyping Γ2 pr Tr Δpr);
    simpl; econstructor; eauto.
Qed.

Lemma subst_evar_lookup_evar {Γ s S x Γ1 Γ2} (ws: Typing Γ s S)
  (esub: subst_evar s x Γ1 Γ2) :
  ∀ {y T}, lookup_evar Γ1 y T → Typing Γ2 (substIndex x s y) T.
Proof.
  induction esub; inversion 1; subst; simpl;
    eauto using T_Var, shift_evar_typing, shift_etvar_typing.
Qed.

Lemma subst_evar_typing {Γ s S Γ1 t T} (ws: Typing Γ s S) (wt: Typing Γ1 t T) :
  ∀ {x Γ2}, subst_evar s x Γ1 Γ2 → Typing Γ2 (substTm x s t) T.
Proof.
  induction wt using ind_typing
  with (P0 := fun Γ1 tr Tr wtr =>
      ∀ {x Γ2}, subst_evar s x Γ1 Γ2 → RTyping Γ2 (substRec x s tr) Tr);
   simpl; eauto using subst_evar_lookup_evar;
  econstructor; eauto using subst_evar_sub, subst_evar_ptyping.
  - rewrite (PTyping_bindPat_lengthExt p0); eauto.
Qed.

(******************************************************************************)
(* Inversion lemmas                                                           *)
(******************************************************************************)

Lemma sub_trec_inversion {Γ S Tr} (subST: Sub Γ S (trec Tr)) :
  (∃ X, S = tvar X) ∨
  (∃ Sr, S = trec Sr ∧
     ∀ l T, Tr l = Some T → ∃ S, Sr l = Some S ∧ Sub Γ S T).
Proof.
  dependent destruction subST; eauto.
  - right; exists Sr; split; auto; intros l T Trl.
    destruct (incl _ _ Trl) as [S Srl]; eauto.
Qed.

Lemma rtyping_trec_inversion {Γ tr Tr k T} (wt: RTyping Γ tr Tr) :
  Tr k = Some T → ∃ t, tr k = Some t ∧ Typing Γ t T.
Proof.
  induction wt; simpl; try discriminate.
  - destruct (eq_lab_dec k l); simpl; intros E; subst.
    + inversion E; subst; exists t; split; auto.
    + apply IHwt; auto.
Qed.

Lemma typing_rec_inversion {Γ tr S} (wΓ: wf_env Γ) (wt: Typing Γ (rec tr) S) :
  ∀ Tr, Sub Γ S (trec Tr) →
    ∀ l T, Tr l = Some T →
      ∃ t, tr l = Some t ∧ Typing Γ t T.
Proof.
  depind wt; simpl; intros Tr' subST l T Trl'.
  - apply sub_trec_inversion in subST; destruct subST as [[X]|[Sr [E subs]]].
    + discriminate.
    + inversion E; subst.
      destruct (subs l T Trl') as [S [Srl subST]].
      destruct (rtyping_trec_inversion H Srl) as [t [Et wt]].
      eauto using T_Sub.
  - eapply IHwt; eauto using sub_trans.
Qed.

(******************************************************************************)
(* Progress                                                                   *)
(******************************************************************************)

Lemma can_form_tvar {Γ t X} (v: Value t) (wt: Typing Γ t (tvar X)) : False.
Proof.
  depind wt; auto.
  depind H; eauto.
Qed.

Lemma can_form_tarr {Γ t T1 T2} (v: Value t) (wt: Typing Γ t (tarr T1 T2)) :
  ∃ T1' t2, t = abs T1' t2.
Proof.
  depind wt; try contradiction.
  - exists T1, t; reflexivity.
  - inversion H; subst; eauto.
    + elimtype False; eauto using can_form_tvar.
Qed.

Lemma can_form_tall {Γ t T1 T2} (v: Value t) (wt: Typing Γ t (tall T1 T2)) :
  ∃ T1' t2, t = tabs T1' t2.
Proof.
  depind wt; try contradiction.
  - exists T1, t; reflexivity.
  - inversion H; subst; eauto.
    + elimtype False; eauto using can_form_tvar.
Qed.

Lemma rtyping_lab_incl {Γ tr Tr} (wt: RTyping Γ tr Tr) :
  ∀ l T, Tr l = Some T → ∃ t, tr l = Some t.
Proof.
  induction wt; simpl; intros k Tk Trk; try discriminate;
     destruct (eq_lab_dec k l); eauto.
Qed.

Lemma can_form_trec {Γ t Sr} (v : Value t) (wt: Typing Γ t (trec Sr)) :
  ∃ vr, t = rec vr ∧ ∀ l T, Sr l = Some T → ∃ s, vr l = Some s.
Proof.
  depind wt; try contradiction.
  - eauto using rtyping_lab_incl.
  - inversion H; subst; eauto.
    + elimtype False; eauto using can_form_tvar.
    + destruct (IHwt Sr0 v eq_refl) as (vr & eq & inc').
      exists vr; split; auto; intros ? ? Srl.
      destruct (incl _ _ Srl) as [S0 Sr0l]; eauto.
Qed.

Lemma get_value {tr : Rec} {k t} :
  tr k = Some t -> RValue tr -> Value t.
Proof.
  induction tr; simpl; intros; try discriminate; destruct (eq_lab_dec k l);
    destruct_conjs; inversion H; subst; auto.
Qed.

Lemma matching_defined {Γ p S Δ} (wΓ: wf_env Γ) (wp: PTyping Γ p S Δ) :
  ∀ {v}, Value v → Typing Γ v S → ∀ t, ∃ t', Match p v t t'.
Proof.
  induction wp using ind_ptyping with
  (P0 := fun Γ pr (Sr : TRec) Δ wpr => wf_env Γ →
           ∀ {vr}, RValue vr →
             (∀ l S, Sr l = Some S → ∃ v, vr l = Some v ∧ Typing Γ v S) →
             ∀ t, ∃ t', RMatch pr vr t t').
  - intros v val_v wv t.
    exists (substTm sub_here v t).
    refine M_Var.
  - intros v val_v wv t.
    destruct (can_form_trec val_v wv) as (vr & eq & inc); subst.
    pose proof (typing_rec_inversion wΓ wv (sub_refl (typing_wf wv wΓ) wΓ)).
    destruct (IHwp wΓ vr val_v H t) as [t' m].
    exists t'; constructor; auto.
  - rename e into Trl; intros vr val_vr wvr t.
    assert ((∀ (l0 : Lab) (S : Ty), Tr l0 = Some S
           → ∃ v : Tm,
             (weakenRec vr (lengthExt Δp)) l0 = Some v
             ∧ Typing (extend Γ Δp) v S)).
    { intros k Sk; generalize (wvr k); simpl.
      case (eq_lab_dec k l); intros; subst.
      - rewrite Trl in H0; discriminate.
      - destruct (H _ H0) as [vk [eqvk wvk]].
        exists (weakenTm vk (lengthExt Δp)); split; auto.
        rewrite get_weaken_rec, eqvk; auto.
        eapply weaken_typing; eauto.
    }
    assert (wf_env (extend Γ Δp)) by apply (ptyping_wf wp wΓ).
    destruct (IHwp0 H0 (weakenRec vr (lengthExt Δp))
                 (weaken_rvalue (lengthExt Δp) val_vr) H t) as [t' mr].
    simpl in wvr; pose proof (wvr l); destruct (eq_lab_dec l l); intuition.
    destruct (H1 _ eq_refl) as (vp, (vrl, wvp)).
    destruct (H2 vp (get_value vrl val_vr) wvp t') as [t'' m].
    exists t''; econstructor; eauto.
    rewrite (PTyping_bindPat_lengthExt wp); eauto.
  - eauto using RMatch.
Qed.

Lemma progress {Γ t U} (wt: Typing Γ t U) :
  Γ = empty → Value t ∨ ∃ t', red t t'.
Proof with destruct_conjs; subst; simpl; eauto using red, rred.
  induction wt using ind_typing with
  (P0 := fun Γ tr Tr wtr => Γ = empty → RValue tr ∨ ∃ tr', rred tr tr');
    intros...
  - inversion l.
  - destruct IHwt1 as [v1|[t1' r1]]...
    destruct IHwt2 as [v2|[t2' r2]]...
    destruct (can_form_tarr v1 wt1)...
  - destruct IHwt as [vt|[t1' r1]]...
    destruct (can_form_tall vt wt)...
  - destruct IHwt1 as [v1|[t1' r1]]...
    destruct (matching_defined wf_empty p0 v1 wt1 t2)...
  - destruct IHwt as [vt|[t1' r1]]...
  - destruct IHwt as [vt|[t1' r1]]...
    destruct (can_form_trec vt wt)...
    destruct (H0 l T lIn) as [v H].
    pose proof (get_value H vt)...
  - destruct IHwt as [vt|[t1' r']]...
    destruct IHwt0 as [vr|[tr' rr]]...
Qed.

(******************************************************************************)
(* Preservation                                                               *)
(******************************************************************************)

Lemma t_abs_inversion {Γ t T0 T1} (wΓ: wf_env Γ) (wt: Typing Γ (abs T1 t) T0) :
  ∀ {T2 T3}, Sub Γ T0 (tarr T2 T3) → Sub Γ T2 T1 ∧ Typing (evar Γ T1) t T3.
Proof.
  depind wt.
  - inversion 1; eauto using T_Sub, shift_evar_sub.
  - eauto using sub_trans.
Qed.

Lemma t_tabs_inversion {Γ t T0 T1} (wΓ: wf_env Γ) (wt: Typing Γ (tabs T1 t) T0) :
  ∀ {T2 T3}, Sub Γ T0 (tall T2 T3) → Sub Γ T2 T1 ∧ Typing (etvar Γ T2) t T3.
Proof.
  depind wt.
  - inversion 1; subst; split; eauto using T_Sub, typing_narrow0.
  - eauto using sub_trans.
Qed.

Lemma local_preservation_lett {Γ p Tp Δp} (wΓ: wf_env Γ) (wp: PTyping Γ p Tp Δp) :
  ∀ {v u u' U}, Typing Γ v Tp → Typing (extend Γ Δp) u U →
    Match p v u u' → Typing Γ u' U.
Proof.
  induction wp using ind_ptyping with (P0 :=
      fun Γ pr Tr Δpr (wpr: PRTyping Γ pr Tr Δpr) =>
        ∀ (wΓ: wf_env Γ) (vr : Rec) u u' U,
          (∀ l S, Tr l = Some S → ∃ v, vr l = Some v ∧ Typing Γ v S) →
          Typing (extend Γ Δpr) u U → RMatch pr vr u u' → Typing Γ u' U).
  - intros v u u' U wv wu m.
    dependent destruction m.
    eauto using subst_evar_typing, subst_evar.
  - intros v u u' U wv wu m.
    dependent destruction m.
    eauto using typing_rec_inversion, sub_refl, typing_wf.
  - intros vr u u' U vtypes wu rm.
    dependent destruction rm.
    assert (∀ (l0 : Lab) (S : Ty), Tr l0 = Some S
        → ∃ v0 : Tm,
          (weakenRec vr (lengthExt Δp)) l0 = Some v0
          ∧ Typing (extend Γ Δp) v0 S) as vtypes'.
    { intros k Sk; specialize (vtypes k); simpl in vtypes.
      rewrite get_weaken_rec.
      destruct (eq_lab_dec k l); simpl; subst.
      * rewrite vIn, e; discriminate.
      * induction (Tr k); try discriminate; simpl.
        intro eqSk; inversion eqSk; subst; clear eqSk.
        destruct (vtypes _ eq_refl) as [vk [eqvk wvk]].
        rewrite eqvk; simpl; eauto using weaken_typing.
    }
    assert (wf_env (extend Γ Δp)) by apply (ptyping_wf wp wΓ).
    rewrite extend_append in wu.
    specialize (vtypes l Tp); simpl in vtypes.
    destruct (eq_lab_dec l l); intuition.
    destruct H3 as [v' [eqv wv']].
    rewrite (PTyping_bindPat_lengthExt wp) in rm.
    rewrite vIn in eqv; inversion eqv; subst; eauto.
  - intros; dependent destruction H1; eauto.
Qed.

Lemma preservation {Γ t U} (wΓ: wf_env Γ) (wt: Typing Γ t U) :
  ∀ {t'}, red t t' → Typing Γ t' U.
Proof.
  induction wt using ind_typing
  with (P0 := fun Γ tr Tr wtr => ∀ (wΓ: wf_env Γ) {tr'},
                rred tr tr' → RTyping Γ tr' Tr);
    intros t' rd; inversion rd; subst; eauto using Typing, RTyping.
  - destruct (t_abs_inversion wΓ wt1 (sub_refl (typing_wf wt1 wΓ) wΓ)).
    eauto using subst_evar_typing, T_Sub.
  - destruct (t_tabs_inversion wΓ wt (sub_refl (typing_wf wt wΓ) wΓ)).
    refine (subst_etvar_typing _ _ _ H0 _); eauto.
  - eauto using local_preservation_lett.
  - destruct (typing_rec_inversion wΓ wt (sub_refl (typing_wf wt wΓ) wΓ) _ lIn)
      as [t'' [eq'' wt'']].
    rewrite H3 in eq''; inversion eq''; subst; auto.
Qed.
