Require Import Prelude.
Require Import Specification.

Module DEBRUIJN (Spec : Specification).

  Export Spec.

  (*  _   _      _                                                        *)
  (* | | | | ___| |_ ___ _ __ ___   __ _  ___ _ __   ___  ___  _   _ ___  *)
  (* | |_| |/ _ \ __/ _ \ '__/ _ \ / _` |/ _ \ '_ \ / _ \/ _ \| | | / __| *)
  (* |  _  |  __/ ||  __/ | | (_) | (_| |  __/ | | |  __/ (_) | |_| \__ \ *)
  (* |_| |_|\___|\__\___|_|  \___/ \__, |\___|_| |_|\___|\___/ \__,_|___/ *)
  (*                               |___/                                  *)

  (* Representation of heterogeneous variable lists *)

  Inductive HVL : Set :=
    | HV0 : HVL
    | HVS (α : Namespace) : HVL → HVL.

  (* Addition in disguise *)
  Fixpoint append (xs ys : HVL) : HVL :=
    match ys with
      | HV0      => xs
      | HVS α ys => HVS α (append xs ys)
    end.

  Lemma append_assoc {h1 h2 h3: HVL} :
    append (append h1 h2) h3 =
    append h1 (append h2 h3).
  Proof.
    revert h1 h2; induction h3; simpl; intros; auto; f_equal; auto.
  Qed.

  Class Weaken (T : Type) :=
    { weaken : T → HVL → T;
      (* weaken_inj (k : HVL) (t1 t2 : T):
        weaken t1 k = weaken t2 k -> t1 = t2; *)
      weaken_append (t : T) (k1 k2 : HVL) :
          weaken (weaken t k1) k2 =
          weaken t (append k1 k2)
    }.

  (*  ___           _            *)
  (* |_ _|_ __   __| | _____  __ *)
  (*  | || '_ \ / _` |/ _ \ \/ / *)
  (*  | || | | | (_| |  __/>  <  *)
  (* |___|_| |_|\__,_|\___/_/\_\ *)

  (* Index with a phantom namespace. *)
  Inductive Index (α : Namespace) : Set :=
    | I0 : Index α
    | IS : Index α → Index α.
  Arguments I0 {α}.
  Arguments IS {α} i.

  Fixpoint weakenIndex {α} (i : Index α) (hvl : HVL)  : Index α :=
    match hvl with
      | HV0 => i
      | HVS β hvl =>
        match α ≟ β with
          | inl _ => IS (weakenIndex i hvl)
          | inr _ => weakenIndex i hvl
        end
    end.

  Lemma weakenIndex_append {α} (i : Index α) (k1 k2 : HVL) :
    weakenIndex (weakenIndex i k1) k2 =
    weakenIndex i (append k1 k2).
  Proof.
    revert i k1; induction k2; simpl; intros.
    reflexivity.
    destruct (α ≟ α0).
    f_equal.
    auto.
    auto.
  Qed.

  Global Instance WeakenIndex {α} : Weaken (Index α) :=
    { weaken := weakenIndex; weaken_append := weakenIndex_append }.

  (*  _____                        *)
  (* |_   _|__ _ __ _ __ ___  ___  *)
  (*   | |/ _ \ '__| '_ ` _ \/ __| *)
  (*   | |  __/ |  | | | | | \__ \ *)
  (*   |_|\___|_|  |_| |_| |_|___/ *)

  (* Use a strong specification of terms to avoid having
    to deal with well-sortedness everywhere. *)
  Inductive Term : Sort → Set :=
    | Var
        (α : Namespace) (y : Index α) :
        Term (sortOf α)
    | Fix
        (ct : Ctor) (ts : ∀ {s}, s ∈ ctorFields ct → Term s) :
        Term (ctorTarget ct).

  Arguments Var [α] y.
  Arguments Fix [ct] ts.

  Definition Terms (ss : list Sort) := ∀ s, s ∈ ss → Term s.

  (*  _____            _   ____  _           _ ____                    *)
  (* | ____|_   ____ _| | | __ )(_)_ __   __| / ___| _ __   ___  ___   *)
  (* |  _| \ \ / / _` | | |  _ \| | '_ \ / _` \___ \| '_ \ / _ \/ __|  *)
  (* | |___ \ V / (_| | | | |_) | | | | | (_| |___) | |_) |  __/ (__   *)
  (* |_____| \_/ \__,_|_| |____/|_|_| |_|\__,_|____/| .__/ \___|\___|  *)
  (*                                                |_|                *)

  (* Binding specifications are evaluated to heterogeneous lists that
     represent the variables which are bound in the bindspec. The relative
     order of the variables is always preserved. *)

  Definition evalVarListItem {Γ : list Sort}
    (evalTs : ∀ s (sΓ : s ∈ Γ) (fn : AuxFn s), HVL)
    {αs : list Namespace} (vl : VarListItem Γ αs) : HVL :=
    match vl with
      | vlOne α αIn          => HVS α HV0
      | vlAuxFn nt fn ntΓ _  => evalTs _ ntΓ fn
    end.

  Fixpoint evalVarListExp {Γ : list Sort}
    (evalTs : ∀ (s: Sort) (sΓ: s ∈ Γ) (fn: AuxFn s), HVL)
    {αs : list Namespace} (vl: VarListExp Γ αs) : HVL :=
      match vl with
        | nil => HV0
        | cons vli vle => append
                           (evalVarListExp evalTs vle)
                           (evalVarListItem evalTs vli)
      end.

  Fixpoint evalFunCall {s} (t : Term s) :
    ∀ (fn : AuxFn s), HVL :=
    match t with
      | Var α _   => λ fn, match auxFnsNoVar fn False with end
      | Fix ct ts => λ fn, evalVarListExp
                            (λ s' s'Γ, evalFunCall (ts _ s'Γ))
                            (ctorAuxFn fn)
    end.

  Definition evalFunCall' {Γ} (ts: Terms Γ) :
    forall s, s ∈ Γ → AuxFn s → HVL := fun s sΓ => evalFunCall (ts s sΓ).

  Fixpoint evalBindSpec {Γ} (fs: Terms Γ) (h: HVL) {αs} (bs: VarListExp Γ αs) : HVL :=
    match bs with
      | []                 => h
      | vli :: bs =>
        append
          (evalBindSpec fs h bs)
          (evalVarListItem (evalFunCall' fs) vli)
    end.

  Lemma evalVarListExp_cong {α Γ}
    (evalTs1 : ∀ s (sΓ: s ∈ Γ) (fn: AuxFn s), HVL)
    (evalTs2 : ∀ s (sΓ: s ∈ Γ) (fn: AuxFn s), HVL)
    (hyp : ∀ (s: Sort) (sΓ: s ∈ Γ) (fn: AuxFn s),
             evalTs1 s sΓ fn = evalTs2 s sΓ fn) :
    ∀ (vl : VarListExp Γ α),
      evalVarListExp evalTs1 vl = evalVarListExp evalTs2 vl.
  Proof.
    induction vl; simpl; intros; auto.
    destruct a; simpl; congruence.
  Qed.

  Inductive SubHVL (αs : list Namespace) : HVL → Prop :=
     | SubHV0 :
         SubHVL αs HV0
     | SubHVS {α h} :
         ∀ (αIn : α ∈ αs) (subh: SubHVL αs h),
           SubHVL αs(HVS α h).

  Lemma subhvl_append {αs h1 h2} :
      SubHVL αs h1 → SubHVL αs h2 → SubHVL αs (append h1 h2).
  Proof.
    intros sh1 sh2; revert h1 sh1.
    induction sh2; simpl; try constructor; auto.
  Qed.

  Lemma subhvl_incl {αs1 αs2 h} (hyp: αs1 ⊆ αs2) :
      SubHVL αs1 h → SubHVL αs2 h.
  Proof.
    induction 1; simpl.
    - constructor.
    - constructor; eauto.
  Qed.

  Lemma subhvl_evalvarlistexp {Γ : list Sort}
    (evalts : ∀ (s: Sort) (sΓ: s ∈ Γ) (fn: AuxFn s), HVL)
    (subhvl_evalts : ∀ (s: Sort) (sΓ: s ∈ Γ) (fn: AuxFn s),
      SubHVL (auxFnsTarget fn) (evalts s sΓ fn))
    {αs : list Namespace} (vl: VarListExp Γ αs) :
      SubHVL αs (evalVarListExp evalts vl).
  Proof.
    induction vl; simpl.
    - constructor.
    - apply subhvl_append.
      + auto.
      + destruct a; simpl.
        * constructor; auto; constructor.
        * eapply (subhvl_incl i); eauto.
  Qed.

  Lemma subhvl_evalfuncall {s} (t : Term s) (fn : AuxFn s) :
    SubHVL (auxFnsTarget fn) (evalFunCall t fn).
  Proof.
    induction t; simpl.
    - destruct (auxFnsNoVar fn False).
    - apply subhvl_evalvarlistexp; eauto.
  Qed.

  (*   ____      _         __  __      *)
  (*  / ___|   _| |_ ___  / _|/ _|___  *)
  (* | |  | | | | __/ _ \| |_| |_/ __| *)
  (* | |__| |_| | || (_) |  _|  _\__ \ *)
  (*  \____\__,_|\__\___/|_| |_| |___/ *)

  Inductive Cutoff (α : Namespace) : Set :=
    | C0 : Cutoff α
    | CS : Cutoff α → Cutoff α.
  Arguments C0 {α}.
  Arguments CS {α} c.

  Definition Cutoffs (α : Namespace) (ss : list Sort) : Set :=
    ∀ s, s ∈ ss → Cutoff α.

  Definition Cs {α Γ} (c : Cutoff α) : Cutoffs α Γ := λ z zΓ, c.

  Fixpoint shiftCutoff {α}
    (c : Cutoff α) (x : Cutoff α) : Cutoff α :=
    match c with
      | C0   => CS x
      | CS c => match x with
                  | C0   => C0
                  | CS x => CS (shiftCutoff c x)
                end
    end.

  Definition shiftCutoff' {α β} (c : Cutoff α) : Cutoff β → Cutoff β :=
    match α ≟ β with
      | inl e =>
        match e in (_ ≡ β) return Cutoff β → Cutoff β with
          | seq_refl => shiftCutoff c
        end
      | inr _ => λ x, x
    end.

  Fixpoint shiftCutoff2 {α}
    (c : Cutoff α) (x : Cutoff α) {struct x} : Cutoff α :=
    match x with
      | C0   => C0
      | CS y => match c with
                  | C0   => CS x
                  | CS c => CS (shiftCutoff2 c y)
                end
    end.

  Definition shiftCutoff2' {α β} (c : Cutoff α) : Cutoff β → Cutoff β :=
    match α ≟ β with
      | inl e =>
        match e in (_ ≡ β) return Cutoff β → Cutoff β with
          | seq_refl => shiftCutoff2 c
        end
      | inr _ => λ x, x
    end.

  Fixpoint weakenCutoff {α} (c : Cutoff α) (hvl : HVL)  : Cutoff α :=
    match hvl with
      | HV0 => c
      | HVS β hvl =>
        match α ≟ β with
          | inl _ => CS (weakenCutoff c hvl)
          | inr _ => weakenCutoff c hvl
        end
    end.

  Lemma weakenCutoff_append {α} (c : Cutoff α) (k1 k2 : HVL) :
    weakenCutoff (weakenCutoff c k1) k2 =
    weakenCutoff c (append k1 k2).
  Proof.
    revert c k1; induction k2; simpl; intros.
    reflexivity.
    destruct (α ≟ α0).
    f_equal.
    auto.
    auto.
  Qed.

  Lemma weakenCutoff_shiftCutoff {α} (hvl : HVL) :
    ∀ (c1 c2 : Cutoff α),
      weakenCutoff (shiftCutoff c1 c2) hvl =
      shiftCutoff (weakenCutoff c1 hvl) (weakenCutoff c2 hvl).
  Proof.
    induction hvl as [|β hvl]; simpl; intros.
    - reflexivity.
    - destruct (α ≟ β).
      + dependent destruction s.
        simpl.
        f_equal.
        auto.
      + auto.
  Qed.

  Lemma weakenCutoff_shiftCutoff2 {α} (hvl : HVL) :
    ∀ (c1 c2 : Cutoff α),
      weakenCutoff (shiftCutoff2 c1 c2) hvl =
      shiftCutoff2 (weakenCutoff c1 hvl) (weakenCutoff c2 hvl).
  Proof.
    induction hvl as [|β hvl]; simpl; intros.
    - reflexivity.
    - destruct (α ≟ β).
      + dependent destruction s.
        simpl.
        f_equal.
        auto.
      + auto.
  Qed.

  Global Instance WeakenCutoff {α} : Weaken (Cutoff α) :=
    { weaken := weakenCutoff; weaken_append := weakenCutoff_append }.
  Global Instance WeakenCutoffs {α} {ss : list Sort} : Weaken (Cutoffs α ss) :=
    { weaken cs hvl := λ s sΓ, weaken (cs s sΓ) hvl }.
  Proof.
    intros.
    extensionality s.
    extensionality sΓ.
    apply weaken_append.
  Defined.

  (*  ____  _     _  __ _   _              *)
  (* / ___|| |__ (_)/ _| |_(_)_ __   __ _  *)
  (* \___ \| '_ \| | |_| __| | '_ \ / _` | *)
  (*  ___) | | | | |  _| |_| | | | | (_| | *)
  (* |____/|_| |_|_|_|  \__|_|_| |_|\__, | *)
  (*                                |___/  *)

  Fixpoint shiftIndex {α}
    (c : Cutoff α) (x : Index α) : Index α :=
    match c , x with
      | C0   , x    => IS x
      | CS c , I0   => I0
      | CS c , IS x => IS (shiftIndex c x)
    end.

  Definition shiftVar {α β} (eq : (α ≡ β) + (α ≢ β))
    (c : Cutoff α) : Index β → Index β :=
    match eq with
      | inl e => match e with
                   | seq_refl => shiftIndex c
                 end
      | inr _ => λ x, x
    end.

  Fixpoint shift {α s} (αIn : α ∈ depsOf s)
           (c : Cutoff α) (t : Term s) : Term s :=
    match t with
      | Var β x   => Var (shiftVar (α ≟ β) c x)
      | Fix ct ts => Fix (λ z zΓ,
          match α ∈? depsOf z with
            | inl αInz => shift αInz
                           (weakenCutoff c (evalBindSpec ts HV0 (ctorBindspec ct z zΓ)))
                           (ts z zΓ)
            | inr _    => ts z zΓ
          end)
    end.

  Definition shift' {α s}
    (c : Cutoff α) (t : Term s) : Term s :=
    match α ∈? depsOf s with
      | inl αIn => shift αIn c t
      | inr _   => t
    end.

  Definition shiftTerms {α ss} (cs : Cutoffs α ss) (ts : Terms ss) : Terms ss :=
    λ z zΓ, shift' (cs z zΓ) (ts z zΓ).

  Fixpoint weakenTerm {s} (t : Term s) (hvl : HVL) : Term s :=
    match hvl with
      | HV0       => t
      | HVS α hvl => match α ∈? depsOf s with
                       | inl αIn => shift αIn C0 (weakenTerm t hvl)
                       | inr _   => weakenTerm t hvl
                     end
    end.

  Definition weakenTerms {ss} (ts : Terms ss) (hvl: HVL) : Terms ss :=
      λ z zΓ, weakenTerm (ts z zΓ) hvl.

  Lemma weakenTerm_append {s} (t : Term s) (k1 k2 : HVL) :
      weakenTerm (weakenTerm t k1) k2 =
      weakenTerm t (append k1 k2) .
  Proof.
    intros; revert s t k1.
    induction k2; simpl; intros.
    reflexivity.
    destruct (α ∈? depsOf s).
    f_equal; auto.
    auto.
  Qed.

  Global Instance WeakenTerm {s} : Weaken (Term s) :=
    { weaken := weakenTerm; weaken_append := @weakenTerm_append _ }.


  (*                         _____                                            *)
  (*                        |_   _| __ __ _  ___ ___                          *)
  (*                          | || '__/ _` |/ __/ _ \                         *)
  (*                          | || | | (_| | (_|  __/                         *)
  (*                          |_||_|  \__,_|\___\___|                         *)

  (* Trace of an α binding through intermediate bindings from namespaces βs. *)
  Inductive Trace (α : Namespace) : Set :=
    | T0 : Trace α
    | TS {β} : β ∈ depsOf (sortOf α) → Trace α → Trace α.
  Arguments T0 {α}.
  Arguments TS {α} {β} p t.

  Fixpoint weakenTrace {α} (x : Trace α) (hvl : HVL) : Trace α :=
    match hvl with
      | HV0       => x
      | HVS β hvl => match β ∈? depsOf (sortOf α) with
                      | inl βInα => TS βInα (weakenTrace x hvl)
                      | inr _    => weakenTrace x hvl
                    end
    end.

  Lemma weakenTrace_append {α} (x : Trace α) (hvl1 hvl2 : HVL) :
    weakenTrace (weakenTrace x hvl1) hvl2 =
    weakenTrace x (append hvl1 hvl2).
  Proof.
    revert x hvl1; induction hvl2 as [|β hvl]; simpl; intros.
    - reflexivity.
    - destruct (β ∈? depsOf (sortOf α)).
      + f_equal; auto.
      + auto.
  Qed.

  Global Instance WeakenTrace {α} : Weaken (Trace α) :=
    { weaken := weakenTrace; weaken_append := weakenTrace_append }.

  (* Each term constructor has a list of fields with bindings specifications.
     For the substitution function we need to hoist the trace under the bindings.
     Traces represents a list of hoisted traces, one for each field *)

  (* Traces for the fields. *)
  Definition Traces (α : Namespace) (ss : list Sort) :=
    ∀ s, s ∈ ss → Trace α.

  Definition Xs {α Γ} (x : Trace α) : Traces α Γ := λ z zΓ, x.

  (*
  Global Instance WeakenTraces {α ss} :
    Weaken α (Traces α ss) :=
    {| weaken1 :=
         mapAll
           (λ s (i : α ∈ depsOf s → Trace α) αIn,
            weaken1 (i αIn))
    |}.
  *)

  (*  ____        _         _   _ _         _   _              *)
  (* / ___| _   _| |__  ___| |_(_) |_ _   _| |_(_) ___  _ __   *)
  (* \___ \| | | | '_ \/ __| __| | __| | | | __| |/ _ \| '_ \  *)
  (*  ___) | |_| | |_) \__ \ |_| | |_| |_| | |_| | (_) | | | | *)
  (* |____/ \__,_|_.__/|___/\__|_|\__|\__,_|\__|_|\___/|_| |_| *)

  Fixpoint substIndex {α}
    (x : Trace α) (t : Term (sortOf α))
    (y : Index α) : Term (sortOf α) :=
    match x , y with
      | T0          , I0   => t
      | T0          , IS y => Var y
      | TS β βInα x , y    =>
        if α ≟ β
        then match y with
               | I0   => Var I0
               | IS y => shift (namespaceDepVar α) C0 (substIndex x t y)
             end
        else shift βInα C0 (substIndex x t y)
    end.

  Definition substVar {α β} (eq : (α ≡ β) + (α ≢ β))
    (x : Trace α) (t : Term (sortOf α))
    (y : Index β) : Term (sortOf β) :=
    match eq with
      | inl e =>
        match e in (_ ≡ β) return (Index β → Term (sortOf β)) with
          | seq_refl => substIndex x t
        end y
      | inr _ => Var y
    end.

  Fixpoint subst {α s} (αIn : α ∈ depsOf s)
    (xα : Trace α) (tα : Term (sortOf α))
    (t : Term s) : Term s :=
    match t in Term s return Term s
    with
      | Var β y   => substVar (α ≟ β) xα tα y
      | Fix ct ts => Fix (λ z zΓ,
          match α ∈? depsOf z with
            | inl αInz => subst αInz
                           (weaken xα (evalBindSpec ts HV0 (ctorBindspec ct z zΓ)))
                           tα (ts z zΓ)
            | inr _    => ts z zΓ
          end)
    end.

  Definition subst' {α s}
    (xα: Trace α) (tα: Term (sortOf α)) (t : Term s) : Term s :=
    match α ∈? depsOf s with
      | inl αIn => subst αIn xα tα t
      | inr _   => t
    end.

  Definition substTerms {α ss} (xs : Traces α ss)
    (t: Term (sortOf α)) (ts : Terms ss) : Terms ss :=
      λ z zΓ, subst' (xs z zΓ) t (ts z zΓ).

  (*  ____  _          *)
  (* / ___|(_)_______  *)
  (* \___ \| |_  / _ \ *)
  (*  ___) | |/ /  __/ *)
  (* |____/|_/___\___| *)

  Fixpoint sizeTerm {s} (t : Term s) : nat :=
    match t with
      | Var β x   => 1
      | Fix ct ts =>
        S (list_rec
             (λ l : list Sort, (∀ z, z ∈ l → nat) → nat)
             (λ _ : ∀ z, z ∈ [] → nat, 0)
             (λ w l (IHl : (∀ z, z ∈ l → nat) → nat) (size_rec : ∀ z, z ∈ w :: l → nat),
              size_rec w (InHere w) + IHl (λ z zΓ, size_rec z (InThere zΓ)))
             (ctorFields ct)
             (fun z zΓ => sizeTerm (ts z zΓ))
          )
    end.

  Ltac nsimpl1 :=
    repeat (
        repeat (unfold evalFunCall', shiftTerms, shift', shiftCutoff', shiftCutoff2',
        substTerms, subst', substVar, shiftVar, Cs, Xs; simpl);
        try match goal with
          | H: _ ≡ _ |- _ =>
            apply seq_eq in H; repeat subst
          | |- context [?α ≟ ?α] =>
            rewrite (seq_namespace_dec_refl α)
          | |- context [?α ∈? depsOf (sortOf ?α)] =>
            rewrite (namespaceDep_In _ _ (namespaceDepVar α))
          | H: ?α ≢ ?α |- _ =>
            destruct (H seq_refl)
          | H: ?α ≡ ?β |- context [?α ≟ ?β] =>
            rewrite (seq_namespace_dec_seq H)
          | H: ?β ≡ ?α |- context [?α ≟ ?β] =>
            rewrite (seq_namespace_dec_seq (sym_seq H))
          | H: ?α ≢ ?β |- context [?α ≟ ?β] =>
            rewrite (seq_namespace_dec_nseq H)
          | H: ?β ≢ ?α |- context [?α ≟ ?β] =>
            rewrite (seq_namespace_dec_nseq (sym_nseq H))
          | H1: ?α ∈ ?αs, H2: ?α ∉ ?αs |- _ =>
            destruct (H2 H1)
          | H: ?α ∈ ?αs |- context [?α ∈? ?αs] =>
            rewrite (namespaceDep_In _ _ H)
          | H: ?α ∉ ?αs |- context [?α ∈? ?αs] =>
            rewrite (namespaceDep_notIn _ _ H)
          | H: ?α ∉ depsOf (sortOf ?α) |- _ =>
            destruct (H (namespaceDepVar α))
          | H: ?α ∈ depsOf (sortOf ?α) |- _ =>
            rewrite (namespaceDepIn _ _ H (namespaceDepVar α)); clear H
          | H1: ?α ∈ ?αs, H2: ?α ∈ ?αs |- _ =>
            rewrite (namespaceDepIn _ _ H2 H1); clear H2
          | H1: ?α ∈ depsOf (sortOf ?β),
            H2: ?b ∈ ?αs |- _ =>
            rewrite (namespaceDep_In _ _ (namespaceDepTransVar H1 H2))
          | H1: ?α ∈ depsOf (sortOf ?β),
            H2: ?α ∉ ?αs,
            H3: ?β ∈ ?αs |- _ =>
            destruct (H2 (namespaceDepTransVar H1 H3))
          | sΓ: ?s ∈ ctorFields ?ct,
            αNotIn: ?α ∉ depsOf (ctorTarget ?ct)
            |- ?a ∉ depsOf ?s =>
            apply (namespaceDepTransFix' αNotIn sΓ)
          | H: ?α ≢ ?β |- ?β ≢ ?α =>
            exact (sym_nseq H)
          | H: ?α ∉ ?αs |- ?α ∉ ?αs =>
            exact H
          | |- context [match ?e in _ ≡ _ return _ with | seq_refl => _ end] =>
            destruct e
          | H: ?a ∉ depsOf (sortOf ?b) |- ?a ≢ ?b =>
            let eq := fresh in
              intro eq; apply H; inversion_clear eq; repeat subst; apply namespaceDepVar
        end); simpl.

  Ltac nsimpl := nsimpl1; intros.

  Ltac ncrush1 :=
    repeat (
        nsimpl1;
        match goal with
          | |- context [?α ≟ ?β] =>
            destruct (α ≟ β); nsimpl1
        end); nsimpl1.

  Ltac ncrush2 :=
    repeat (
        ncrush1;
        match goal with
          | |- context [?α ∈? ?αs] =>
            destruct (α ∈? αs); ncrush1
        end); ncrush1.

  Ltac ncrush := repeat ncrush2.

  (* Evaluation of auxiliary functions is stable under shift. *)
  Lemma evalFunCall_shift {α} :
    (∀ {s : Sort} (t : Term s)
       (αIn : α ∈ depsOf s) (fn : AuxFn s) (c : Cutoff α),
         evalFunCall (shift αIn c t) fn = evalFunCall t fn).
  Proof.
    induction t; nsimpl; auto.
    apply (evalVarListExp_cong); intros; ncrush; auto.
  Qed.

  Lemma evalFunCall_subst {α} :
    (∀ {s : Sort} (t : Term s)
       (αIn : α ∈ depsOf s) (fn : AuxFn s) (x : Trace α) (tα : Term (sortOf α)),
         evalFunCall (subst αIn x tα t) fn = evalFunCall t fn).
  Proof.
    induction t; nsimpl; auto.
    - destruct (auxFnsNoVar fn False).
    - apply (evalVarListExp_cong); intros; ncrush; auto.
  Qed.

  Section NonIndexed.

    Local Set Implicit Arguments.
    Local Set Contextual Implicit.

    Inductive EqHVL {Γ} (ts : Terms Γ)  : HVL → HVL → Prop :=
      | EqhRefl h :
          EqHVL ts h h
      | EqhCongSucc {α c1 c2} :
          EqHVL ts c1 c2 ->
          EqHVL ts (HVS α c1) (HVS α c2)
      | EqhCongAppend {h1 h1' h2 h2'} :
          EqHVL ts h1 h1' → EqHVL ts h2 h2' →
          EqHVL ts (append h1 h2) (append h1' h2')
      | EqhShiftFun {α s} (αIn: α ∈ depsOf s) (sΓ: s ∈ Γ) (fn: AuxFn s) (c: Cutoff α) :
          EqHVL ts (evalFunCall (shift αIn c (ts s sΓ)) fn) (evalFunCall (ts s sΓ) fn)
      | EqhSubstFun {α s} (αIn: α ∈ depsOf s)
          (sΓ: s ∈ Γ) (fn: AuxFn s) (x: Trace α) (tα: Term (sortOf α)) :
          EqHVL ts (evalFunCall (subst αIn x tα (ts s sΓ)) fn) (evalFunCall (ts s sΓ) fn)
      | EqhSym {h1 h2} :
          EqHVL ts h1 h2 → EqHVL ts h2 h1
      | EqhTrans {h1 h2 h3} :
          EqHVL ts h1 h2 → EqHVL ts h2 h3 → EqHVL ts h1 h3.

    Global Arguments EqhRefl {Γ ts} h.
    Global Arguments EqhShiftFun {Γ ts} α {s αIn} sΓ fn c.

    Lemma EqHVL_sound {Γ} (ts : Terms Γ) (h1 h2 : HVL) :
      EqHVL ts h1 h2 → h1 = h2.
    Proof.
      induction 1; simpl; try congruence.
      apply evalFunCall_shift.
      apply evalFunCall_subst.
    Qed.

    Inductive EqCutoff {Γ} (ts : Terms Γ) {α} : Cutoff α → Cutoff α → Prop :=
      | EqcRefl {c : Cutoff α} :
          EqCutoff ts c c
      | EqcCongSucc {c1 c2} :
          EqCutoff ts c1 c2 →
          EqCutoff ts (CS c1) (CS c2)
      | EqcCongShift {c1 c2 c3 c4: Cutoff α} :
          EqCutoff ts c1 c2 → EqCutoff ts c3 c4 →
          EqCutoff ts (shiftCutoff c1 c3) (shiftCutoff c2 c4)
      | EqcCongShift' {c1 c2 c3 c4: Cutoff α} :
          EqCutoff ts c1 c2 → EqCutoff ts c3 c4 →
          EqCutoff ts (shiftCutoff2 c1 c3) (shiftCutoff2 c2 c4)
      | EqcCongWeaken {c1 c2 : Cutoff α} {h1 h2} :
          EqCutoff ts c1 c2 → EqHVL ts h1 h2 →
          EqCutoff ts (weakenCutoff c1 h1) (weakenCutoff c2 h2)
      | EqcWeakenShift {c1 c2 : Cutoff α} {h} :
          EqCutoff ts
            (weakenCutoff (shiftCutoff c1 c2) h)
            (shiftCutoff (weakenCutoff c1 h) (weakenCutoff c2 h))
      | EqcWeakenShift' {c1 c2 : Cutoff α} {h} :
          EqCutoff ts
            (weakenCutoff (shiftCutoff2 c1 c2) h)
            (shiftCutoff2 (weakenCutoff c1 h) (weakenCutoff c2 h))
      | EqcWeakenAppend c h1 h2 :
          EqCutoff ts
            (weakenCutoff (weakenCutoff c h1) h2)
            (weakenCutoff c (append h1 h2))
      | EqcSym {c1 c2} :
          EqCutoff ts c1 c2 →
          EqCutoff ts c2 c1
      | EqcTrans {c1 c2 c3} :
          EqCutoff ts c1 c2 → EqCutoff ts c2 c3 → EqCutoff ts c1 c3.

    Global Arguments EqcRefl {Γ ts} α {c}.
    Global Arguments EqcWeakenShift {Γ ts} α {c1 c2 h}.
    Global Arguments EqcWeakenShift' {Γ ts} α {c1 c2 h}.
    Global Arguments EqcWeakenAppend {Γ ts} α {c h1 h2}.

    Lemma EqCutoff_sound {Γ} (ts : Terms Γ) :
      ∀ {α} (c1 c2 : Cutoff α), EqCutoff ts c1 c2 → c1 = c2.
    Proof.
      induction 1.
      - auto.
      - congruence.
      - congruence.
      - congruence.
      - apply EqHVL_sound in H0; congruence.
      - apply weakenCutoff_shiftCutoff.
      - apply weakenCutoff_shiftCutoff2.
      - apply weakenCutoff_append.
      - now symmetry.
      - congruence.
    Qed.

    Inductive EqTrace {Γ} (ts : Terms Γ) {α} : Trace α → Trace α → Prop :=
      | EqxRefl {x : Trace α} :
          EqTrace ts x x
      | EqxCongSucc {x1 x2} {β} (βIn: β ∈ depsOf (sortOf α)) :
          EqTrace ts x1 x2 →
          EqTrace ts (TS βIn x1) (TS βIn x2)
      | EqxCongWeaken {x1 x2 : Trace α} {h1 h2} :
          EqTrace ts x1 x2 → EqHVL ts h1 h2 →
          EqTrace ts (weaken x1 h1) (weaken x2 h2)
      | EqxWeakenAppend {x : Trace α} {h1 h2} :
          EqTrace ts
            (weakenTrace (weakenTrace x h1) h2)
            (weakenTrace x (append h1 h2))
      | EqxSym {x1 x2} :
          EqTrace ts x1 x2 →
          EqTrace ts x2 x1
      | EqxTrans {x1 x2 x3} :
          EqTrace ts x1 x2 → EqTrace ts x2 x3 → EqTrace ts x1 x3.

    Lemma EqTrace_sound {Γ} (ts : Terms Γ) {α} :
      ∀ {x1 x2 : Trace α}, EqTrace ts x1 x2 → x1 = x2.
    Proof.
      induction 1; intros; try congruence.
      - apply EqHVL_sound in H0; congruence.
      - apply weakenTrace_append; congruence.
    Qed.

    Inductive EqTerm {Γ} (ts : Terms Γ) : ∀ {s}, Term s → Term s → Prop :=
    | EqtRefl {s} {t : Term s} :
        EqTerm ts t t
    | EqtAssumption {s} {t1 t2: Term s} :
        t1 = t2 → EqTerm ts t1 t2
    | EqtCongWeaken {s} {t1 t2: Term s} {h1 h2} :
        EqTerm ts t1 t2 →
        EqHVL ts h1 h2 →
        EqTerm ts (weakenTerm t1 h1) (weakenTerm t2 h2)
    | EqtCongShift {α s} (c1 c2: Cutoff α) (αIn: α ∈ depsOf s) (t1 t2 : Term s) :
        EqCutoff ts c1 c2 →
        EqTerm ts t1 t2 →
        EqTerm ts (shift αIn c1 t1) (shift αIn c2 t2)
    | EqtCongSubst {α s} (x1 x2: Trace α) (αIn: α ∈ depsOf s)
        (tα1 tα2: Term (sortOf α)) (t1 t2: Term s) :
        EqTrace ts x1 x2 →
        EqTerm ts tα1 tα2 →
        EqTerm ts t1 t2 →
        EqTerm ts (subst αIn x1 tα1 t1) (subst αIn x2 tα2 t2)
    | EqtSym {s} {t1 t2: Term s} :
        EqTerm ts t1 t2 →
        EqTerm ts t2 t1
    | EqtTrans {s} {t1 t2 t3: Term s} :
        EqTerm ts t1 t2 →
        EqTerm ts t2 t3 →
        EqTerm ts t1 t3.

    Lemma EqTerm_sound {Γ} (ts : Terms Γ) :
      ∀ {s} {t1 t2 : Term s}, EqTerm ts t1 t2 → t1 = t2.
    Proof.
      induction 1; intros; try congruence.
      - apply EqHVL_sound in H0; congruence.
      - apply EqCutoff_sound in H; congruence.
      - apply EqTrace_sound in H; congruence.
    Qed.

  End NonIndexed.

  Section Indexed.

    Local Set Implicit Arguments.
    Local Set Contextual Implicit.

    Inductive eqhvl (Γ : list Sort) : Set :=
      | eqh_refl
      | eqh_sym (eq : eqhvl Γ)
      | eqh_trans (eq1 : eqhvl Γ) (eq2 : eqhvl Γ)
      | eqh_cong_succ (α: Namespace) (eq : eqhvl Γ)
      | eqh_cong_append (eq1 : eqhvl Γ) (eq2 : eqhvl Γ)
      | eqh_shift_fun {s} (fn : AuxFn s) (sΓ : s ∈ Γ)
      | eqh_subst_fun {s} (fn : AuxFn s) (sΓ : s ∈ Γ).

    Inductive eqcutoff (Γ : list Sort) : Set :=
      | eqc_refl
      | eqc_cong_succ (ec: eqcutoff Γ)
      | eqc_cong_weaken (ec: eqcutoff Γ) (eh: eqhvl Γ)
      | eqc_cong_shift (ec1: eqcutoff Γ) (ec2: eqcutoff Γ)
      | eqc_cong_shift' (ec1: eqcutoff Γ) (ec2: eqcutoff Γ)
      | eqc_weaken_shift
      | eqc_weaken_shift'
      | eqc_weaken_append
      | eqc_sym (ec: eqcutoff Γ)
      | eqc_trans (ec1 ec2: eqcutoff Γ).

    Inductive eqterm (Γ : list Sort) : Set :=
      | eqt_refl
      | eqt_ih
      | eqt_cong_weaken (et: eqterm Γ) (eh: eqhvl Γ)
      | eqt_cong_shift (ec: eqcutoff Γ) (et: eqterm Γ)
      | eqt_sym (et: eqterm Γ)
      | eqt_trans (et1: eqterm Γ) (et2: eqterm Γ).

    Context {Γ : list Sort}.
    Variable (ts : Terms Γ).

    Inductive EQHVL : eqhvl Γ → HVL → HVL → Prop :=
      | EQH_refl {h} :
          EQHVL eqh_refl    h h
      | EQH_cong_succ {α e h h'} :
          EQHVL e h h' →
          EQHVL (eqh_cong_succ α e) (HVS α h) (HVS α h')
      | EQH_cong_append {e1 e2 h1 h1' h2 h2'} :
          EQHVL e1 h1 h1' →
          EQHVL e2 h2 h2' →
          EQHVL (eqh_cong_append e1 e2) (append h1 h2) (append h1' h2')
      | EQH_shift_fun {α s} (αIn : α ∈ depsOf s) (fn : AuxFn s) (sΓ: s ∈ Γ) c :
          EQHVL
                (eqh_shift_fun fn sΓ)
                (evalFunCall (shift αIn c (ts sΓ)) fn)
                (evalFunCall (ts sΓ) fn)
      | EQH_subst_fun {α s} (αIn : α ∈ depsOf s) (fn : AuxFn s) (sΓ: s ∈ Γ) xα tα :
          EQHVL
                (eqh_subst_fun fn sΓ)
                (evalFunCall (subst αIn xα tα (ts sΓ)) fn)
                (evalFunCall (ts sΓ) fn)
      | EQH_sym {e h1 h2} :
          EQHVL e h1 h2 →
          EQHVL (eqh_sym e) h2 h1
      | EQH_trans {e1 e2 h1 h2 h3} :
          EQHVL e1 h1 h2 →
          EQHVL e2 h2 h3 →
          EQHVL (eqh_trans e1 e2) h1 h3.

    Lemma EQH_sound {e : eqhvl Γ} {h1 h2 : HVL} :
      EQHVL e h1 h2 → h1 = h2.
    Proof.
      intro H; induction H; simpl; try congruence.
      apply evalFunCall_shift.
      apply evalFunCall_subst.
    Qed.

    Inductive EQCutoff {α} : eqcutoff Γ → Cutoff α → Cutoff α → Prop :=
      | EQC_refl {c : Cutoff α} :
          EQCutoff eqc_refl c c
      | EQC_cong_succ {c1 c2: Cutoff α} {ec} :
          EQCutoff ec c1 c2 →
          EQCutoff (eqc_cong_succ ec) (CS c1) (CS c2)
      | EQC_cong_shift {c1 c2 c3 c4: Cutoff α} {ec1 ec2} :
          EQCutoff ec1 c1 c2 →
          EQCutoff ec2 c3 c4 →
          EQCutoff (eqc_cong_shift ec1 ec2) (shiftCutoff c1 c3) (shiftCutoff c2 c4)
      | EQC_cong_shift' {c1 c2 c3 c4: Cutoff α} {ec1 ec2} :
          EQCutoff ec1 c1 c2 →
          EQCutoff ec2 c3 c4 →
          EQCutoff (eqc_cong_shift' ec1 ec2) (shiftCutoff2 c1 c3) (shiftCutoff2 c2 c4)
      | EQC_cong_weaken {c1 c2 : Cutoff α} {h1 h2} {ec eh} :
          EQCutoff ec c1 c2 →
          EQHVL eh h1 h2 →
          EQCutoff (eqc_cong_weaken ec eh)
                   (weakenCutoff c1 h1) (weakenCutoff c2 h2)
      | EQC_weaken_shift {c1 c2 : Cutoff α} {h} :
          EQCutoff eqc_weaken_shift
                   (weakenCutoff (shiftCutoff c1 c2) h)
                   (shiftCutoff (weakenCutoff c1 h) (weakenCutoff c2 h))
      | EQC_weaken_shift' {c1 c2 : Cutoff α} {h} :
          EQCutoff eqc_weaken_shift'
                   (weakenCutoff (shiftCutoff2 c1 c2) h)
                   (shiftCutoff2 (weakenCutoff c1 h) (weakenCutoff c2 h))
      | EQC_weaken_append {c : Cutoff α} {h1 h2 : HVL} :
          EQCutoff eqc_weaken_append
                   (weakenCutoff (weakenCutoff c h1) h2)
                   (weakenCutoff c (append h1 h2))
      | EQC_sym {c1 c2} {ec} :
          EQCutoff ec c1 c2 →
          EQCutoff (eqc_sym ec) c2 c1
      | EQC_trans {c1 c2 c3} {ec1 ec2} :
          EQCutoff ec1 c1 c2 →
          EQCutoff ec2 c2 c3 →
          EQCutoff (eqc_trans ec1 ec2) c1 c3.

    Lemma EQC_sound {α} {e} {c1 c2 : Cutoff α} :
      EQCutoff e c1 c2 → c1 = c2.
    Proof.
      induction 1.
      - auto.
      - congruence.
      - congruence.
      - congruence.
      - apply EQH_sound in H0; intuition congruence.
      - apply weakenCutoff_shiftCutoff.
      - apply weakenCutoff_shiftCutoff2.
      - apply weakenCutoff_append.
      - now symmetry.
      - congruence.
    Qed.

    Inductive EQTerm {s} (u v : Term s) : eqterm Γ → Term s → Term s → Prop :=
      | EQT_refl {t} :
          EQTerm u v eqt_refl t t
      | EQT_ih :
          EQTerm u v eqt_ih u v
      | EQT_cong_weaken {t1 t2: Term s} {h1 h2} {et eh} :
          EQTerm u v et t1 t2 →
          EQHVL eh h1 h2 →
          EQTerm u v (eqt_cong_weaken et eh) (weakenTerm t1 h1) (weakenTerm t2 h2)
      | EQT_cong_shift {α} (c1 c2: Cutoff α) (αIn: α ∈ depsOf s) (t1 t2 : Term s) {ec et} :
          EQCutoff ec c1 c2 →
          EQTerm u v et t1 t2 →
          EQTerm u v (eqt_cong_shift ec et) (shift αIn c1 t1) (shift αIn c2 t2)
      | EQT_sym {t1 t2} {et} :
          EQTerm u v et t1 t2 →
          EQTerm u v (eqt_sym et) t2 t1
      | EQT_trans {t1 t2 t3} {et1 et2} :
          EQTerm u v et1 t1 t2 →
          EQTerm u v et2 t2 t3 →
          EQTerm u v (eqt_trans et1 et2) t1 t3.

    Lemma EQT_sound {s} {u v : Term s} (ih: u = v) e {t1 t2 : Term s} :
      EQTerm u v e t1 t2 → t1 = t2.
    Proof.
      induction 1; try intuition congruence.
      apply EQH_sound in H0; intuition congruence.
      apply EQC_sound in H; intuition congruence.
    Qed.

  End Indexed.

  Section LemmaEvalBindSpecShift.

    Lemma Eqh_evalbindspec_shift {Γ} α (cs: Cutoffs α Γ) {ts: Terms Γ} h {αs} (bs: VarListExp Γ αs) :
      EqHVL ts
        (evalBindSpec (shiftTerms cs ts) h bs)
        (evalBindSpec ts                 h bs).
    Proof.
      induction bs as [|vli]; nsimpl.
      eapply EqhRefl.
      - destruct vli; nsimpl.
        + eapply EqhCongSucc.
          eapply IHbs.
        + eapply EqhCongAppend.
          eapply IHbs.
          unfold evalFunCall', shiftTerms, shift'; nsimpl.
          destruct (α ∈? depsOf s).
          * eapply EqhShiftFun.
          * eapply EqhRefl.
    Defined.

    Global Arguments Eqh_evalbindspec_shift {Γ} α {cs ts} h {αs} bs.

    Lemma Eqh_evalbindspec_subst {Γ} α (xs: Traces α Γ) (tα: Term (sortOf α)) {ts: Terms Γ} h {αs} (bs: VarListExp Γ αs) :
      EqHVL ts
        (evalBindSpec (substTerms xs tα ts) h bs)
        (evalBindSpec ts                    h bs).
    Proof.
      induction bs as [|vli]; nsimpl.
      eapply EqhRefl.
      - destruct vli; nsimpl.
        + eapply EqhCongSucc.
          eapply IHbs.
        + eapply EqhCongAppend.
          eapply IHbs.
          unfold evalFunCall', substTerms, subst'; nsimpl.
          destruct (α ∈? depsOf s).
          * eapply EqhSubstFun.
          * eapply EqhRefl.
    Defined.

    (* Print Eqh_evalbindspec_shift. *)

    Fixpoint EQH_evalbindspec_shift {Γ αs} α (bs: VarListExp Γ αs) : eqhvl Γ :=
      match bs with
        | [] => eqh_refl
        | vli :: bs' =>
          match vli with
            | vlOne γ γIn       =>
              eqh_cong_succ γ (EQH_evalbindspec_shift α bs')
            | vlAuxFn s fn sΓ _ =>
              eqh_cong_append
                (EQH_evalbindspec_shift α bs')
                (if α ∈? depsOf s
                 then eqh_shift_fun fn sΓ
                 else eqh_refl)
          end
      end.

    Lemma EQH_evalbindspec_shift_sound {Γ α} (ts: Terms Γ):
      forall (cs: Cutoffs α Γ) (h: HVL) {αs} (bs: VarListExp Γ αs),
        EQHVL ts
          (EQH_evalbindspec_shift α bs)
          (evalBindSpec (shiftTerms cs ts) h bs)
          (evalBindSpec ts h bs).
    Proof.
      induction bs as [|[]]; ncrush; repeat constructor; auto.
    Qed.

    Fixpoint EQH_evalbindspec_subst {Γ αs} α (bs: VarListExp Γ αs) : eqhvl Γ :=
      match bs with
        | [] => eqh_refl
        | vli :: bs' =>
          match vli with
            | vlOne γ γIn       =>
              eqh_cong_succ γ (EQH_evalbindspec_subst α bs')
            | vlAuxFn s fn sΓ _ =>
              eqh_cong_append
                (EQH_evalbindspec_subst α bs')
                (if α ∈? depsOf s
                 then eqh_subst_fun fn sΓ
                 else eqh_refl)
          end
      end.

    Lemma EQH_evalbindspec_subst_sound {Γ α} (ts: Terms Γ):
      forall (xαs: Traces α Γ) (tα: Term (sortOf α)) (h: HVL) {αs} (bs: VarListExp Γ αs),
        EQHVL ts
          (EQH_evalbindspec_subst α bs)
          (evalBindSpec (substTerms xαs tα ts) h bs)
          (evalBindSpec ts h bs).
    Proof.
      induction bs as [|[]]; ncrush; repeat constructor; auto.
    Qed.

    Lemma evalbindspec_shift :
      ∀ (β : Namespace) (Γ : list Sort) (cs : Cutoffs β Γ)
        (fs : Terms Γ) {αs} (bs : VarListExp Γ αs) (h: HVL),
        evalBindSpec (shiftTerms cs fs) h bs = evalBindSpec fs h bs.
    Proof.
      eauto using EQH_sound, EQH_evalbindspec_shift_sound.
    Qed.

    Lemma evalbindspec_subst :
      ∀ (β : Namespace) (Γ : list Sort) (xs: Traces β Γ) (tβ: Term (sortOf β))
        (fs : Terms Γ) {αs} (bs : VarListExp Γ αs) (h: HVL),
        evalBindSpec (substTerms xs tβ fs) h bs = evalBindSpec fs h bs.
    Proof.
      eauto using EqHVL_sound, Eqh_evalbindspec_subst.
    Qed.

  End LemmaEvalBindSpecShift.

End DEBRUIJN.
