
Require Import Prelude.
Require Import Specification.
Require Import Context.
Require Import DeBruijn.
Require Import DeBruijnInteraction.
Require Import DeBruijnWellFormed.

Module Type DeBruijnContext.
  Declare Module Spec : Specification.
  Declare Module Ctx : Context.
  Export Ctx.

  Module WellFormed := WellFormed Spec.
  Export WellFormed.

  Inductive Env : Set :=
    | enil
    | econs (Γ : Env) (α : Namespace) (ts : Terms (namespaceData α)).

  Fixpoint appendEnv (Γ : Env) (Δ : Env) : Env :=
    match Δ with
      | enil         => Γ
      | econs Δ α ts => econs (appendEnv Γ Δ) α ts
    end.

  Lemma appendEnv_assoc (Γ : Env) (Δ1 Δ2 : Env) :
    appendEnv Γ (appendEnv Δ1 Δ2) = appendEnv (appendEnv Γ Δ1) Δ2.
  Proof.
    induction Δ2; simpl; f_equal; auto.
  Qed.

  Fixpoint domain (Δ : Env) : HVL :=
    match Δ with
      | enil         => HV0
      | econs Δ α ts => HVS α (domain Δ)
    end.

  Lemma domain_append (Δ1 Δ2 : Env) :
    domain (appendEnv Δ1 Δ2) = append (domain Δ1) (domain Δ2).
  Proof.
    induction Δ2; simpl; try rewrite IHΔ2; induction Δ1; auto.
  Qed.

  Fixpoint shiftEnv {α} (c : Cutoff α) (Δ : Env) : Env :=
    match Δ with
      | enil         => enil
      | econs Δ β ts => econs (shiftEnv c Δ) β (shiftTerms (Cs (weaken c (domain Δ))) ts)
    end.

  Fixpoint substEnv {α} (xα : Trace α) (tα : Term (sortOf α)) (Δ : Env) : Env :=
    match Δ with
      | enil         => enil
      | econs Δ β ts => econs (substEnv xα tα Δ) β (substTerms (Xs (weaken xα (domain Δ))) tα ts)
    end.

  Lemma shiftEnv_append {α} (c : Cutoff α) (Δ1 Δ2 : Env) :
    shiftEnv c (appendEnv Δ1 Δ2) =
    appendEnv (shiftEnv c Δ1) (shiftEnv (weakenCutoff c (domain Δ1)) Δ2).
  Proof.
    revert c Δ1.
    induction Δ2; simpl; intros.
    - reflexivity.
    - f_equal.
      + apply IHΔ2.
      + rewrite weakenCutoff_append.
        rewrite domain_append.
        reflexivity.
  Qed.

  Lemma substEnv_append {α} (xα : Trace α) (tα : Term (sortOf α)) (Δ1 Δ2 : Env) :
    substEnv xα tα (appendEnv Δ1 Δ2) =
    appendEnv (substEnv xα tα Δ1) (substEnv (weakenTrace xα (domain Δ1)) tα Δ2).
  Proof.
    revert xα Δ1.
    induction Δ2; simpl; intros.
    - reflexivity.
    - f_equal.
      + apply IHΔ2.
      + rewrite weakenTrace_append.
        rewrite domain_append.
        reflexivity.
  Qed.

  Lemma domain_shiftEnv {α} (c : Cutoff α) (Δ : Env) :
    domain (shiftEnv c Δ) = domain Δ.
  Proof.
    induction Δ; simpl; f_equal; auto.
  Qed.

  Lemma domain_substEnv {α} (xα : Trace α) (tα : Term (sortOf α)) (Δ : Env) :
    domain (substEnv xα tα Δ) = domain Δ.
  Proof.
    induction Δ; simpl; f_equal; auto.
  Qed.

  Lemma shiftTerms_shiftTerms0_hom_comm:
    ∀ (α : Namespace) (c : Cutoff α) {Γ} (ts : Terms Γ),
      shiftTerms (Cs (CS c)) (shiftTerms (Cs (C0 (α := α))) ts) =
      shiftTerms (Cs (C0 (α := α))) (shiftTerms (Cs c) ts).
  Proof.
    intros; unfold shiftTerms, shift', Cs.
    extensionality z.
    extensionality zΓ.
    ncrush.
    + apply shift_shift0_hom_comm.
    + reflexivity.
  Qed.

  Lemma shiftTerms_shiftTerms0_het_comm:
    ∀ (α β : Namespace) (αβ : α ≢ β) (c : Cutoff α) {Γ} (ts : Terms Γ),
      shiftTerms (Cs c) (shiftTerms (Cs (C0 (α := β))) ts) =
      shiftTerms (Cs (C0 (α := β))) (shiftTerms (Cs c) ts).
  Proof.
    intros; unfold shiftTerms, shift', Cs.
    extensionality z.
    extensionality zΓ.
    ncrush.
    + apply shift_shift0_het_comm; auto.
    + reflexivity.
    + reflexivity.
    + reflexivity.
  Qed.

  Inductive Insert {α} : Cutoff α → Env → Env → Prop :=
    | ihere
        {Γ : Env} (ts : Terms (namespaceData α)) :
        Insert C0 Γ (econs Γ α ts)
    | ithere_eq
        {Γ1 Γ2 : Env} c (ts' : Terms (namespaceData α)) :
        Insert c Γ1 Γ2 →
        Insert (CS c) (econs Γ1 α ts') (econs Γ2 α (shiftTerms (Cs c) ts'))
    | ithere_neq  {β} (αβ : α ≢ β)
        {Γ1 Γ2 : Env} c (tsβ : Terms (namespaceData β)) :
        Insert c Γ1 Γ2 →
        Insert c (econs Γ1 β tsβ) (econs Γ2 β (shiftTerms (Cs c) tsβ)).

  Lemma weakenInsert {α} (c : Cutoff α) (Γ1 Γ2 : Env) (Δ : Env) :
    Insert c Γ1 Γ2 →
    Insert (weakenCutoff c (domain Δ)) (appendEnv Γ1 Δ) (appendEnv Γ2 (shiftEnv c Δ)).
  Proof.
    revert c Γ2.
    induction Δ; simpl; intros.
    - auto.
    - ncrush.
      + constructor; auto.
      + constructor; auto.
  Qed.

  Lemma Insert_Insertion {α} (c : Cutoff α) (Γ1 Γ2 : Env) :
    Insert c Γ1 Γ2 → Insertion c (domain Γ1) (domain Γ2).
  Proof.
    induction 1; constructor; auto.
  Qed.

  Inductive SubstEnv (Γ : Env) {α} (tα: Term (sortOf α)) (tsα: Terms (namespaceData α)) : Trace α → Env → Env → Prop :=
    | shere :
        SubstEnv Γ tα tsα T0 (econs Γ α tsα) Γ
    | sthere_in {x Γ1 Γ2}
        {β} {βIn: β ∈ depsOf (sortOf α)} {tsβ : Terms (namespaceData β)} :
        SubstEnv Γ tα tsα x Γ1 Γ2 →
        SubstEnv Γ tα tsα (TS βIn x) (econs Γ1 β tsβ) (econs Γ2 β (substTerms (Xs x) tα tsβ))
    | sthere_notin {x Γ1 Γ2}
        {β} {βNotIn: β ∉ depsOf (sortOf α)} {tsβ : Terms (namespaceData β)} :
        SubstEnv Γ tα tsα x Γ1 Γ2 →
        SubstEnv Γ tα tsα x (econs Γ1 β tsβ) (econs Γ2 β (substTerms (Xs x) tα tsβ)).

  Lemma weakenSubstEnv {Γ α tα tsα} (Γ1 Δ : Env) :
    ∀ (Γ2 : Env) (x: Trace α),
      SubstEnv Γ tα tsα x Γ1 Γ2 →
      SubstEnv Γ tα tsα
        (weakenTrace x (domain Δ))
        (appendEnv Γ1 Δ)
        (appendEnv Γ2 (substEnv x tα Δ)).
  Proof.
    induction Δ; simpl; intros.
    - auto.
    - ncrush.
      + constructor; auto.
      + constructor; auto.
  Qed.

  Lemma SubstEnv_Deletion {Γ α tα tsα} {x: Trace α} (Γ1 Γ2 : Env) :
    SubstEnv Γ tα tsα x Γ1 Γ2 → Deletion (domain Γ) x (domain Γ1) (domain Γ2).
  Proof.
    induction 1; constructor; auto.
  Qed.

  Section WellFormedEnvInductive.

    Inductive WfEnv : HVL -> Env -> Prop :=
    | wf_enil {h} :
        WfEnv h enil
    | wf_econs {h} {Γ} {α} {ts : Terms (namespaceData α)} :
        WfEnv h Γ →
        wfterms (append h (domain Γ)) ts →
        WfEnv h (econs Γ α ts).

    Lemma WfEnv_appendEnv {Γ Δ: Env} {h} :
      WfEnv h Γ → WfEnv (append h (domain Γ)) Δ → WfEnv h (appendEnv Γ Δ).
    Proof.
      intros wfΓ wfΔ.
      dependent induction wfΔ; simpl.
      - auto.
      - constructor.
        + auto.
        + rewrite domain_append.
          rewrite <- append_assoc.
          auto.
      Restart.
      induction Δ; simpl; intros wfΓ wfΔ.
      - auto.
      - inversion wfΔ; subst.
        apply inj_pair2 in H3; subst.
        constructor.
        + auto.
        + rewrite domain_append.
          rewrite <- append_assoc.
          auto.
    Qed.

    Lemma shiftEnv_WfEnv {h1 Γ} (wfΓ : WfEnv h1 Γ) :
      forall {α} {c : Cutoff α} {h2},
        Insertion c h1 h2 → WfEnv h2 (shiftEnv c Γ).
    Proof.
      induction wfΓ; simpl; intros.
      - constructor.
      - constructor.
        + apply IHwfΓ; auto.
        + rewrite domain_shiftEnv.
          apply (wfterms_shift _ _ H).
          apply weaken_insertion; auto.
    Qed.

    Lemma substEnv_WfEnv {h h1 Γ} (wfΓ : WfEnv h1 Γ) :
      forall {α} {xα: Trace α} {tα: Term (sortOf α)} (wf_tα: wfterm h tα) {h2},
        Deletion h xα h1 h2 →
        WfEnv h2 (substEnv xα tα Γ).
    Proof.
      induction wfΓ; simpl; intros.
      - constructor.
      - constructor.
        + apply IHwfΓ; auto.
        + rewrite domain_substEnv.
          apply (wfterms_subst wf_tα _ H).
          apply weaken_deletion; auto.
    Qed.

  End WellFormedEnvInductive.

  Section WellFormedEnvRecursive.

    Fixpoint wfenv (h: HVL) (Γ: Env) : Prop :=
      match Γ with
        | enil => True
        | econs Γ α ts =>
          wfenv h Γ /\ wfterms (append h (domain Γ)) ts
      end.

    Lemma wfenv_appendEnv {Γ Δ: Env} {h} :
      wfenv h Γ → wfenv (append h (domain Γ)) Δ → wfenv h (appendEnv Γ Δ).
    Proof.
      induction Δ; simpl; intros wfΓ wfΔ.
      - auto.
      - destruct wfΔ as [wfΔ wfts].
        constructor.
        + auto.
        + rewrite domain_append.
          rewrite <- append_assoc.
          auto.
    Qed.

    Lemma shiftEnv_wfenv {h1 Γ} :
      forall (wfΓ : wfenv h1 Γ) {α} {c : Cutoff α} {h2},
        Insertion c h1 h2 → wfenv h2 (shiftEnv c Γ).
    Proof.
      induction Γ; simpl; intros.
      - auto.
      - destruct wfΓ as [wfΓ wfts].
        split.
        + auto.
        + rewrite domain_shiftEnv.
          apply (wfterms_shift _ _ wfts).
          apply weaken_insertion; auto.
    Qed.

    Lemma substEnv_wfenv {h h1 Γ} (wfΓ : wfenv h1 Γ) :
      forall {α} {xα: Trace α} {tα: Term (sortOf α)} (wf_tα: wfterm h tα) {h2},
        Deletion h xα h1 h2 →
        wfenv h2 (substEnv xα tα Γ).
    Proof.
      induction Γ; simpl; intros.
      - auto.
      - destruct wfΓ as [wfΓ wfts].
        split.
        + auto.
        + rewrite domain_substEnv.
          apply (wfterms_subst wf_tα _ wfts).
          apply weaken_deletion; auto.
    Qed.

  End WellFormedEnvRecursive.

  Section LookupInductive.

    Inductive Lookup {α} : Env → Index α → Terms (namespaceData α) → Prop :=
      | get_here {Γ : Env} ts :
          wfterms (domain Γ) ts →
          Lookup (econs Γ α ts) I0 (shiftTerms (Cs (C0 (α := α))) ts)
      | get_there_eq
          {Γ : Env} X (ts' : Terms (namespaceData α)) ts :
          Lookup Γ X ts →
          Lookup (econs Γ α ts') (IS X) (shiftTerms (Cs (C0 (α := α))) ts)
      | get_there_neq {β} (αβ : α ≢ β)
          {Γ : Env}  X (tsβ : Terms (namespaceData β)) ts :
          Lookup Γ X ts →
          Lookup (econs Γ β tsβ) X (shiftTerms (Cs (C0 (α := β))) ts).

    Lemma Lookup_weaken (Γ:Env) {α} :
      forall (Δ:Env) (xα: Index α) (ts : Terms _),
        Lookup Γ xα ts →
        Lookup (appendEnv Γ Δ) (weaken xα (domain Δ)) (weakenTerms ts (domain Δ)).
    Proof.
      unfold weakenTerms; simpl.
      induction Δ; ncrush; simpl; intros; try constructor; auto.
    Qed.

    Lemma Lookup_inversion_here (Γ:Env) {α} (ts1 ts2: Terms _) :
      Lookup (econs Γ α ts1) I0 ts2 → shiftTerms (Cs (C0 (α := α))) ts1 = ts2.
    Proof.
      intros H; dependent destruction H; ncrush; auto.
    Qed.

    Lemma Lookup_inversion_there_eq (Γ:Env) {α} (ts1 ts2: Terms _) (X : Index α) :
      Lookup (econs Γ α ts1) (IS X) ts2 →
      ∃ ts2',
        shiftTerms (Cs (C0 (α := α))) ts2' = ts2 /\
        Lookup Γ X ts2'.
    Proof.
      intros H; dependent destruction H.
      + exists ts; auto.
      + ncrush.
    Qed.

    Lemma Lookup_inversion_there_neq (Γ:Env) {α β} (αβ : α ≢ β) (ts1 : Terms _)
          (ts2: Terms _) (X : Index α) :
      Lookup (econs Γ β ts1) X ts2 →
      ∃ ts2',
        shiftTerms (Cs (C0 (α := β))) ts2' = ts2 /\
        Lookup Γ X ts2'.
    Proof.
      intros H; dependent destruction H.
      + ncrush.
      + ncrush.
      + exists ts; auto.
    Qed.

    Lemma Lookup_functional {α} (Γ : Env) (X : Index α) (ts1 ts2 : Terms _) :
      Lookup Γ X ts1 → Lookup Γ X ts2 → ts1 = ts2.
    Proof.
      intro get; revert ts2.
      induction get; intros ts2 get2.
      - apply (Lookup_inversion_here _ _ _ get2).
      - apply Lookup_inversion_there_eq in get2.
        destruct get2 as [ts2' [eq_ts2 get2]].
        specialize (IHget ts2' get2).
        congruence.
      - apply Lookup_inversion_there_neq in get2; auto.
        destruct get2 as [ts2' [eq_ts2 get2]].
        specialize (IHget ts2' get2).
        congruence.
    Qed.

    Lemma Lookup_wellformed {α} (Γ : Env) (X : Index α) (ts : Terms _) :
      Lookup Γ X ts → wfterms (domain Γ) ts.
    Proof.
      intro lk; induction lk;
        apply (wfterms_shift(domain Γ)); auto; constructor.
    Qed.

    (* Lemma Lookup_wellformed {α h} (Γ : Env) (X : Index α) (ts : Terms _) : *)
    (*   Lookup Γ X ts → wfenv h Γ → wfterms (append h (domain Γ)) ts. *)
    (* Proof. *)
    (*   intro lk; induction lk; intros (wfΓ,wfts); *)
    (*     apply (wfterms_shift (append h (domain Γ))); auto; constructor. *)
    (* Qed. *)

    Lemma lookup_shift {α} {c : Cutoff α} {Γ1 Γ2 : Env} (ib : Insert c Γ1 Γ2) :
      ∀ (X : Index α) (ts : Terms (namespaceData α)),
        Lookup Γ1 X ts → Lookup Γ2 (shiftIndex c X) (shiftTerms (Cs c) ts).
    Proof.
      induction ib; intros; simpl.
      - (* ihere *)
        constructor.
        assumption.
      - (* ithere_eq *)
        dependent destruction H; simpl.
        + (* get_here *)
          rewrite shiftTerms_shiftTerms0_hom_comm.
          constructor; eauto using wfterms_shift, Insert_Insertion.
        + (* get_there_eq *)
          rewrite shiftTerms_shiftTerms0_hom_comm.
          constructor; auto.
        + (* get_there_neq *)
          destruct αβ; constructor.
        (* ithere_neq *)
      - dependent destruction H; simpl.
        + (* get_here *)
          ncrush.
        + (* get_there_eq *)
          ncrush.
        + (* get_there_neq *)
          rewrite shiftTerms_shiftTerms0_het_comm; auto.
          constructor; auto.
    Qed.

    Lemma lookup_shift_het {α β} (αβ: α ≢ β) {c : Cutoff α} {Γ1 Γ2 : Env} (ib : Insert c Γ1 Γ2) :
      ∀ (X : Index β) (ts : Terms (namespaceData β)),
        Lookup Γ1 X ts → Lookup Γ2 X (shiftTerms (Cs c) ts).
    Proof.
      induction ib; intros; simpl.
      - (* ihere *)
        constructor; ncrush; auto.
      - (* ithere_eq *)
        dependent destruction H; simpl.
        + (* get_here *)
          rewrite shiftTerms_shiftTerms0_hom_comm.
          constructor; eauto using wfterms_shift, Insert_Insertion.
        + (* get_there_eq *)
          rewrite shiftTerms_shiftTerms0_hom_comm.
          constructor; auto.
        + (* get_there_neq *)
          rewrite shiftTerms_shiftTerms0_hom_comm.
          constructor; auto.
        (* ithere_neq *)
      - dependent destruction H; simpl.
        + (* get_here *)
          rewrite shiftTerms_shiftTerms0_het_comm; auto.
          constructor; eauto using wfterms_shift, Insert_Insertion.
        + (* get_there_eq *)
          rewrite shiftTerms_shiftTerms0_het_comm; auto.
          constructor; auto.
        + (* get_there_neq *)
          rewrite shiftTerms_shiftTerms0_het_comm; auto.
          constructor; auto.
    Qed.

    Lemma lookup_subst {α β Γ Γ1 Γ2} (αβ: α ≢ β)
      {tα: Term (sortOf α)} {wftα: wfterm (domain Γ) tα}
      {tsα: Terms (namespaceData α)} {xα: Trace α}
      (sub: SubstEnv Γ tα tsα xα Γ1 Γ2) :
      ∀ (n : Index β) (tsβ : Terms (namespaceData β)),
        Lookup Γ1 n tsβ →
        Lookup Γ2 n (substTerms (Xs xα) tα tsβ).
    Proof.
      induction sub; intros; simpl.
      - dependent destruction H;
          rewrite substTerms0_shiftTerms0_reflection;
          ncrush; auto.
      - dependent destruction H; simpl;
          rewrite substTerms_shiftTerms0_comm1; constructor;
            eauto using wfterms_subst, SubstEnv_Deletion.
      - dependent destruction H; simpl;
          rewrite substTerms_shiftTerms0_comm2; auto; constructor;
            eauto using wfterms_subst, SubstEnv_Deletion.
    Qed.

    Lemma shiftTerms_notin {α Γ} (hyp: ∀ s, s ∈ Γ → α ∉ depsOf s)
      (c: Cutoff α) (ts: Terms Γ) :
      shiftTerms (Cs c) ts = ts.
    Proof.
      extensionality s; extensionality sIn; specialize (hyp s sIn).
      ncrush; auto.
    Qed.

    Lemma substTerms_notin {α Γ} (hyp: ∀ s, s ∈ Γ → α ∉ depsOf s)
      (x: Trace α) (tα: Term (sortOf α)) (ts: Terms Γ) :
      substTerms (Xs x) tα ts = ts.
    Proof.
      extensionality s; extensionality sIn; specialize (hyp s sIn).
      ncrush; auto.
    Qed.

    Lemma lookup_subst_simpl {α β Γ Γ1 Γ2} (αβ: α ≢ β)
      (hyp: ∀ s, s ∈ namespaceData β → α ∉ depsOf s)
      {tα: Term (sortOf α)}
      {tsα: Terms (namespaceData α)} {xα: Trace α}
      (sub: SubstEnv Γ tα tsα xα Γ1 Γ2) :
      ∀ (n : Index β) (tsβ : Terms (namespaceData β)),
        Lookup Γ1 n tsβ →
        Lookup Γ2 n tsβ.
    Proof.
      (* Set Printing Implicit. *)
      induction sub; intros; simpl.
      - dependent destruction H; try (ncrush; fail).
        rewrite shiftTerms_notin; auto.
      - dependent destruction H.
        + rewrite substTerms_notin; auto; constructor.
          eauto using wfterms_strengthen, SubstEnv_Deletion.
        + constructor; eauto.
        + constructor; eauto.
      - dependent destruction H.
        + rewrite substTerms_notin; auto; constructor.
          eauto using wfterms_strengthen, SubstEnv_Deletion.
        + constructor; eauto.
        + constructor; eauto.
    Qed.

  End LookupInductive.

End DeBruijnContext.
