Set Implicit Arguments.

Require Import Prelude.
Require Import Specification.
Require Import DeBruijn.

Module EqSimplifier (Spec : Specification).

  Module DEBRUIJN := DEBRUIJN Spec.
  Export DEBRUIJN.

  Section Simplifier.
    Context {Γ : list Sort}.

    Ltac crush_eqh :=
      simpl; eauto;
      repeat
        (match goal with
           | IH: EQHVL _ ?e _ _
             |- EQHVL _ (match ?e with _ => _ end) _ _ =>
             remember e; inversion IH; repeat subst
         end;
         try (repeat econstructor; eauto; fail)
        );
      first
        [ econstructor;
          [ econstructor; eauto
          | repeat match goal with
              | H: _ = ?r |- EQHVL _ _ ?r _ =>
                rewrite <- H
            end; econstructor; eauto
          ]
        | idtac
        ].

    Fixpoint eqh_simplify_trans (e1 e2: eqhvl Γ) : eqhvl Γ :=
      match e1, e2 with
        | eqh_refl          , e2       => e2
        | e1                , eqh_refl => e1
        | eqh_trans e11 e12 , e2       => eqh_simplify_trans e11 (eqh_simplify_trans e12 e2)
        | e1                , e2       => eqh_trans e1 e2
      end.

    Lemma eqh_simplify_trans_sound {ts e1 h1 h2} (eq1: EQHVL ts e1 h1 h2) :
      ∀ {e2 h3}, EQHVL ts e2 h2 h3 → EQHVL ts (eqh_simplify_trans e1 e2) h1 h3.
    Proof.
      induction eq1; intros ? ? eq2; try (crush_eqh; fail).
    Qed.

    Definition eqh_simplify_cong_succ (α: Namespace) (e: eqhvl Γ) : eqhvl Γ :=
      match  e with
        | eqh_refl => eqh_refl
        | e        => eqh_cong_succ α e
      end.

    Lemma eqh_simplify_cong_succ_sound {ts: Terms Γ} {α e} {h1 h2: HVL} :
      EQHVL ts e h1 h2 →
      EQHVL ts (eqh_simplify_cong_succ α e) (HVS α h1) (HVS α h2).
    Proof.
      intros eq; unfold eqh_simplify_cong_succ; crush_eqh.
    Qed.

    Definition eqh_simplify_cong_append (e1 e2: eqhvl Γ) : eqhvl Γ :=
      match e1, e2 with
        | eqh_refl , eqh_refl => eqh_refl
        | e1       , e2       => eqh_cong_append e1 e2
      end.

    Lemma eqh_simplify_cong_append_sound {ts: Terms Γ} {e1 e2: eqhvl Γ} {h1 h1' h2 h2'} :
      EQHVL ts e1 h1 h1' →
      EQHVL ts e2 h2 h2' →
      EQHVL ts (eqh_simplify_cong_append e1 e2) (append h1 h2) (append h1' h2').
    Proof.
      intros eq1 eq2; unfold eqh_simplify_cong_append; crush_eqh.
    Qed.

    Fixpoint eqh_simplify (eqh: eqhvl Γ) : eqhvl Γ :=
      match eqh with
        | eqh_refl              => eqh_refl
        | eqh_sym e             => eqh_simplify_sym e
        | eqh_trans e1 e2       => eqh_simplify_trans (eqh_simplify e1) (eqh_simplify e2)
        | eqh_cong_succ α e     => eqh_simplify_cong_succ α (eqh_simplify e)
        | eqh_cong_append e1 e2 => eqh_simplify_cong_append (eqh_simplify e1) (eqh_simplify e2)
        | eqh_shift_fun s fn sΓ => eqh_shift_fun fn sΓ
        | eqh_subst_fun s fn sΓ => eqh_subst_fun fn sΓ
      end
    with eqh_simplify_sym (eqh: eqhvl Γ) : eqhvl Γ :=
      match eqh with
        | eqh_refl              => eqh_refl
        | eqh_sym e             => eqh_simplify e
        | eqh_trans e1 e2       => eqh_simplify_trans (eqh_simplify_sym e2) (eqh_simplify_sym e1)
        | eqh_cong_succ α e     => eqh_simplify_cong_succ α (eqh_simplify_sym e)
        | eqh_cong_append e1 e2 => eqh_simplify_cong_append (eqh_simplify_sym e1) (eqh_simplify_sym e2)
        | eqh_shift_fun s fn sΓ => eqh_sym (eqh_shift_fun fn sΓ)
        | eqh_subst_fun s fn sΓ => eqh_sym (eqh_subst_fun fn sΓ)
      end.

    Lemma eqh_simplify_sound_ind {ts: Terms Γ} {e: eqhvl Γ} {h1 h2: HVL} :
      EQHVL ts e h1 h2 →
      EQHVL ts (eqh_simplify e) h1 h2 /\
      EQHVL ts (eqh_simplify_sym e) h2 h1.
    Proof.
      intro eq; induction eq; simpl; split; destruct_conjs;
      eauto; try (repeat constructor; fail); crush_eqh;
      eauto using
        eqh_simplify_cong_succ_sound,
        eqh_simplify_cong_append_sound,
        eqh_simplify_trans_sound.
    Qed.

    Lemma eqh_simplify_sound {ts: Terms Γ} {e: eqhvl Γ} {h1 h2: HVL} :
      EQHVL ts e h1 h2 →
      EQHVL ts (eqh_simplify e) h1 h2.
    Proof.
      eapply eqh_simplify_sound_ind.
    Qed.

    Lemma eqh_simplify_sym_sound {ts: Terms Γ} {e: eqhvl Γ} {h1 h2: HVL} :
      EQHVL ts e h1 h2 →
      EQHVL ts (eqh_simplify_sym e) h2 h1.
    Proof.
      eapply eqh_simplify_sound_ind.
    Qed.

    Fixpoint eqc_simplify_trans (e1 e2: eqcutoff Γ) : eqcutoff Γ :=
      match e1, e2 with
        | eqc_refl          , e2       => e2
        | e1                , eqc_refl => e1
        | eqc_trans e11 e12 , e2       => eqc_simplify_trans e11 (eqc_trans e12 e2)
        | e1                , e2       => eqc_trans e1 e2
      end.

    Ltac crush_eqc :=
      simpl; eauto;
      repeat
        (match goal with
        | IH: EQCutoff _ ?e _ _
          |- EQCutoff _ (match ?e with _ => _ end) _ _ =>
          remember e; inversion IH; repeat subst
         end;
         try (repeat econstructor; eauto using eqh_simplify_sound; fail)
        );
      first
        [ econstructor;
          [ econstructor; eauto
          | match goal with
              | H: _ = ?r |- EQCutoff _ _ ?r _ =>
                rewrite <- H
              | H: ?r = _ |- EQCutoff _ _ ?r _ =>
                rewrite H
            end; econstructor; eauto
          ]
        | idtac
        ].

    Lemma eqc_simplify_trans_sound {ts α e1} {c1 c2: Cutoff α} (eq1: EQCutoff ts e1 c1 c2) :
      ∀ {e2 c3}, EQCutoff ts e2 c2 c3 → EQCutoff ts (eqc_simplify_trans e1 e2) c1 c3.
    Proof.
      induction eq1; intros ? ? eq2; try (crush_eqc; fail).
      remember e2; inversion eq2; repeat subst; try (repeat econstructor; eauto; fail);
        try (eapply IHeq1_1; repeat (econstructor; eauto); fail).
    Qed.

    Definition eqc_simplify_cong_succ (e: eqcutoff Γ) : eqcutoff Γ :=
      match  e with
        | eqc_refl => eqc_refl
        | e        => eqc_cong_succ e
      end.

    Lemma eqc_simplify_cong_succ_sound {ts: Terms Γ} {α e} {c1 c2: Cutoff α} :
      EQCutoff ts e c1 c2 →
      EQCutoff ts (eqc_simplify_cong_succ e) (CS c1) (CS c2).
    Proof.
      intros eq; unfold eqc_simplify_cong_succ; crush_eqc.
    Qed.

    Definition eqc_simplify_cong_weaken (e1: eqcutoff Γ) (e2: eqhvl Γ) : eqcutoff Γ :=
      match e1, e2 with
        | eqc_refl , eqh_refl => eqc_refl
        | e1       , e2       => eqc_cong_weaken e1 e2
      end.

    Lemma eqc_simplify_cong_weaken_sound {ts e1 e2 α} {c1 c1': Cutoff α} {h2 h2'} :
      EQCutoff ts e1 c1 c1' →
      EQHVL ts e2 h2 h2' →
      EQCutoff ts (eqc_simplify_cong_weaken e1 e2) (weakenCutoff c1 h2) (weakenCutoff c1' h2').
    Proof.
      intros eq1 eq2; unfold eqc_simplify_cong_weaken; crush_eqc; crush_eqh.
      remember e2; inversion eq2; repeat subst; try (repeat econstructor; eauto; fail).
    Qed.

    Definition eqc_simplify_cong_shift (e1 e2: eqcutoff Γ) : eqcutoff Γ :=
      match e1, e2 with
        | eqc_refl , eqc_refl => eqc_refl
        | e1       , e2       => eqc_cong_shift e1 e2
      end.

    Lemma eqc_simplify_cong_shift_sound {ts e1 e2 α} {c1 c1' c2 c2': Cutoff α} :
      EQCutoff ts e1 c1 c1' →
      EQCutoff ts e2 c2 c2' →
      EQCutoff ts (eqc_simplify_cong_shift e1 e2) (shiftCutoff c1 c2) (shiftCutoff c1' c2').
    Proof.
      intros eq1 eq2; unfold eqc_simplify_cong_shift; crush_eqc.
    Qed.

    Definition eqc_simplify_cong_shift' (e1 e2: eqcutoff Γ) : eqcutoff Γ :=
      match e1, e2 with
        | eqc_refl , eqc_refl => eqc_refl
        | e1       , e2       => eqc_cong_shift' e1 e2
      end.

    Lemma eqc_simplify_cong_shift'_sound {ts e1 e2 α} {c1 c1' c2 c2': Cutoff α} :
      EQCutoff ts e1 c1 c1' →
      EQCutoff ts e2 c2 c2' →
      EQCutoff ts (eqc_simplify_cong_shift' e1 e2) (shiftCutoff2 c1 c2) (shiftCutoff2 c1' c2').
    Proof.
      intros eq1 eq2; unfold eqc_simplify_cong_shift'; crush_eqc.
    Qed.

    Fixpoint eqc_simplify (eqc: eqcutoff Γ) : eqcutoff Γ :=
      match eqc with
        | eqc_refl              => eqc_refl
        | eqc_sym e             => eqc_simplify_sym e
        | eqc_trans e1 e2       => eqc_simplify_trans (eqc_simplify e1) (eqc_simplify e2)
        | eqc_cong_succ ec      => eqc_simplify_cong_succ (eqc_simplify ec)
        | eqc_cong_weaken ec eh => eqc_simplify_cong_weaken (eqc_simplify ec) (eqh_simplify eh)
        | eqc_cong_shift e1 e2  => eqc_simplify_cong_shift (eqc_simplify e1) (eqc_simplify e2)
        | eqc_cong_shift' e1 e2 => eqc_simplify_cong_shift' (eqc_simplify e1) (eqc_simplify e2)
        | eqc_weaken_shift      => eqc_weaken_shift
        | eqc_weaken_shift'     => eqc_weaken_shift'
        | eqc_weaken_append     => eqc_weaken_append
      end
    with eqc_simplify_sym (eqc: eqcutoff Γ) : eqcutoff Γ :=
      match eqc with
        | eqc_refl              => eqc_refl
        | eqc_sym e             => eqc_simplify e
        | eqc_trans e1 e2       => eqc_simplify_trans (eqc_simplify_sym e2) (eqc_simplify_sym e1)
        | eqc_cong_succ ec      => eqc_simplify_cong_succ (eqc_simplify_sym ec)
        | eqc_cong_weaken ec eh => eqc_simplify_cong_weaken (eqc_simplify_sym ec) (eqh_simplify_sym eh)
        | eqc_cong_shift e1 e2  => eqc_simplify_cong_shift (eqc_simplify_sym e1) (eqc_simplify_sym e2)
        | eqc_cong_shift' e1 e2 => eqc_simplify_cong_shift' (eqc_simplify_sym e1) (eqc_simplify_sym e2)
        | eqc_weaken_shift  => eqc_sym eqc_weaken_shift
        | eqc_weaken_shift' => eqc_sym eqc_weaken_shift'
        | eqc_weaken_append => eqc_sym eqc_weaken_append
      end.

    Lemma eqc_simplify_sound_ind {α} {ts: Terms Γ} {e: eqcutoff Γ} {c1 c2: Cutoff α} :
      EQCutoff ts e c1 c2 →
      EQCutoff ts (eqc_simplify e) c1 c2 /\
      EQCutoff ts (eqc_simplify_sym e) c2 c1.
    Proof.
      intro eq; induction eq; simpl; split; destruct_conjs;
      eauto; try (repeat constructor; fail); crush_eqc; crush_eqh;
      eauto using
        eqc_simplify_cong_shift_sound,
        eqc_simplify_cong_shift'_sound,
        eqc_simplify_cong_weaken_sound,
        eqc_simplify_cong_succ_sound,
        eqh_simplify_sound,
        eqh_simplify_sym_sound,
        eqc_simplify_trans_sound.
    Qed.

    Lemma eqc_simplify_sound {α} {ts: Terms Γ} {e: eqcutoff Γ} {c1 c2: Cutoff α} :
      EQCutoff ts e c1 c2 →
      EQCutoff ts (eqc_simplify e) c1 c2.
    Proof.
      eapply eqc_simplify_sound_ind.
    Qed.

    Lemma eqc_simplify_sym_sound {α} {ts: Terms Γ} {e: eqcutoff Γ} {c1 c2: Cutoff α} :
      EQCutoff ts e c1 c2 →
      EQCutoff ts (eqc_simplify_sym e) c2 c1.
    Proof.
      eapply eqc_simplify_sound_ind.
    Qed.

    Fixpoint eqt_simplify_trans (e1 e2: eqterm Γ) : eqterm Γ :=
      match e1, e2 with
        | eqt_refl          , e2       => e2
        | e1                , eqt_refl => e1
        | eqt_trans e11 e12 , e2       => eqt_simplify_trans e11 (eqt_trans e12 e2)
        | e1                , e2       => eqt_trans e1 e2
      end.

    Ltac crush_eqt :=
      simpl; eauto;
      repeat
        (match goal with
           | IH: EQTerm _ _ _ ?e _ _
             |- EQTerm _ _ _ (match ?e with _ => _ end) _ _ =>
             remember e; inversion IH; repeat subst
         end;
         try (repeat econstructor; eauto using eqh_simplify_sound; fail)
        );
      first
        [ econstructor;
            [ econstructor; eauto
            | match goal with
                | H: _ = ?r |- EQTerm _ _ _ _ ?r _ =>
                  rewrite <- H
                | H: ?r = _ |- EQTerm _ _ _ _ ?r _ =>
                  rewrite H
              end; econstructor; eauto
            ]
        | idtac
        ].

    Lemma eqt_simplify_trans_sound {ts s e1} {u v t1 t2: Term s} (eq1: EQTerm ts u v e1 t1 t2) :
      ∀ {e2 t3}, EQTerm ts u v e2 t2 t3 → EQTerm ts u v (eqt_simplify_trans e1 e2) t1 t3.
    Proof.
      induction eq1; intros ? ? eq2; try (crush_eqt; fail).
      remember e2; inversion eq2; repeat subst; try (repeat econstructor; eauto; fail);
        try (eapply IHeq1_1; repeat (econstructor; eauto); fail).
    Qed.

    Definition eqt_simplify_cong_weaken (e1: eqterm Γ) (e2: eqhvl Γ) : eqterm Γ :=
      match e1, e2 with
        | eqt_refl , eqh_refl => eqt_refl
        | e1       , e2       => eqt_cong_weaken e1 e2
      end.

    Lemma eqt_simplify_cong_weaken_sound {ts e1 e2 s} {u v t1 t1': Term s} {h2 h2'} :
      EQTerm ts u v e1 t1 t1' →
      EQHVL ts e2 h2 h2' →
      EQTerm ts u v (eqt_simplify_cong_weaken e1 e2) (weakenTerm t1 h2) (weakenTerm t1' h2').
    Proof.
      intros eq1 eq2; unfold eqt_simplify_cong_weaken; crush_eqt; crush_eqh.
      remember e2; inversion eq2; repeat subst; try (repeat econstructor; eauto; fail).
    Qed.

    Definition eqt_simplify_cong_shift (e1: eqcutoff Γ) (e2: eqterm Γ) : eqterm Γ :=
      match e1, e2 with
        | eqc_refl , eqt_refl => eqt_refl
        | e1       , e2       => eqt_cong_shift e1 e2
      end.

    Lemma eqt_simplify_cong_shift_sound {ts e1 e2 α s} (αIn: α ∈ depsOf s) {c1 c1': Cutoff α} {u v t2 t2': Term s} :
      EQCutoff ts e1 c1 c1' →
      EQTerm ts u v e2 t2 t2' →
      EQTerm ts u v (eqt_simplify_cong_shift e1 e2) (shift αIn c1 t2) (shift αIn c1' t2').
    Proof.
      intros eq1 eq2; unfold eqt_simplify_cong_shift.
      remember e1; inversion eq1; repeat subst; try (repeat econstructor; eauto; fail).
      remember e2; inversion eq2; repeat subst; try (repeat econstructor; eauto; fail).
    Qed.

    Fixpoint eqt_simplify (eqt: eqterm Γ) : eqterm Γ :=
      match eqt with
        | eqt_refl              => eqt_refl
        | eqt_sym e             => eqt_simplify_sym e
        | eqt_trans e1 e2       => eqt_simplify_trans (eqt_simplify e1) (eqt_simplify e2)
        | eqt_cong_weaken et eh => eqt_simplify_cong_weaken (eqt_simplify et) (eqh_simplify eh)
        | eqt_cong_shift ec et  => eqt_simplify_cong_shift (eqc_simplify ec) (eqt_simplify et)
        | eqt_ih                => eqt_ih
      end
    with eqt_simplify_sym (eqt: eqterm Γ) : eqterm Γ :=
      match eqt with
        | eqt_refl              => eqt_refl
        | eqt_sym e             => eqt_simplify e
        | eqt_trans e1 e2       => eqt_simplify_trans (eqt_simplify_sym e2) (eqt_simplify_sym e1)
        | eqt_cong_weaken et eh => eqt_simplify_cong_weaken (eqt_simplify_sym et) (eqh_simplify_sym eh)
        | eqt_cong_shift ec et  => eqt_simplify_cong_shift (eqc_simplify_sym ec) (eqt_simplify_sym et)
        | eqt_ih                => eqt_sym eqt_ih
      end.

    Lemma eqt_simplify_sound {s} {ts: Terms Γ} {e: eqterm Γ} {u v : Term s} {t1 t2: Term s} :
      EQTerm ts u v e t1 t2 →
      EQTerm ts u v (eqt_simplify e) t1 t2 /\
      EQTerm ts u v (eqt_simplify_sym e) t2 t1.
    Proof.
      intro eq; induction eq; simpl; split; destruct_conjs;
      eauto; try (repeat constructor; fail); crush_eqt; crush_eqh;
      eauto using
        eqt_simplify_cong_shift_sound,
        eqt_simplify_cong_weaken_sound,
        eqh_simplify_sound,
        eqh_simplify_sym_sound,
        eqt_simplify_trans_sound,
        eqc_simplify_sound,
        eqc_simplify_sym_sound.
    Qed.

  End Simplifier.

End EqSimplifier.
