
Require Export Coq.Unicode.Utf8.
Require Import Coq.Program.Tactics.
Require Import Prelude.
Require Import Specification.
Require Import DeBruijn.
Require Import DeBruijnInteraction.

Module WellFormed (Spec : Specification).

  Module Interaction := Interaction Spec.
  Export Interaction.

  Section Def.

    Fixpoint wfindex (h : HVL) {α} (i : Index α) : Prop :=
      match h with
        | HV0 => False
        | HVS b h =>
          if α ≟ b
          then match i with
                 | I0 => True
                 | IS i => wfindex h i
               end
          else wfindex h i
      end.

    (* A well-scopedness predicate for terms. *)
    Inductive wfterm (Γ : HVL) : ∀ {s}, Term s → Prop :=
      | wft_Var {a} {i : Index a} :
        wfindex Γ i →
        wfterm Γ (Var i)
      | wft_Fix {ct : Ctor} {ts : ∀ s, s ∈ ctorFields ct → Term s} :
          (∀ {s} (sΓ: s ∈ ctorFields ct),
             wfterm (append Γ (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))) (ts s sΓ)) →
          wfterm Γ (Fix ts).

    Definition wfterms (Γ : HVL) {ss} (ts : Terms ss) : Prop :=
      ∀ s (sIn: s ∈ ss), wfterm Γ (ts s sIn).

  End Def.

  Section Shifting.

    Inductive Insertion {α} : ∀ (c: Cutoff α) (h1 h2: HVL), Prop :=
      | ins_here {h} :
          Insertion C0 h (HVS α h)
      | ins_there_eq {c h1 h2} :
          Insertion c h1 h2 →
          Insertion (CS c) (HVS α h1) (HVS α h2)
      | ins_there_neq {c h1 h2 β} (αNotβ: α ≢ β) :
          Insertion c h1 h2 →
          Insertion c (HVS β h1) (HVS β h2).

    Lemma Insertion_direct {α} {c: Cutoff α} {h1 h2} (ins: Insertion c h1 h2) :
      ∃ h d,
        h1 = append h d ∧
        h2 = append (HVS α h) d ∧
        c = weaken C0 d.
    Proof.
      induction ins.
      - exists h, HV0; simpl; auto.
      - destruct IHins as [h [d [eqh1 [eqh2 eqc]]]]; subst.
        exists h, (HVS α d); nsimpl; auto.
      - destruct IHins as [h [d [eqh1 [eqh2 eqc]]]]; subst.
        exists h, (HVS β d); nsimpl; auto.
    Qed.

    Lemma weaken_insertion {α} :
      ∀ (h h1 h2 : HVL) (c : Cutoff α),
        Insertion c h1 h2 →
        Insertion (weaken c h) (append h1 h) (append h2 h).
    Proof.
      intros h h1 h2 c ins.
      induction h; nsimpl.
      - auto.
      - ncrush.
        + constructor; eauto.
        + constructor; eauto.
    Qed.

    Lemma wfindex_shift'_hom {α} {c: Cutoff α} {h1 h2} (ins: Insertion c h1 h2) :
      ∀ i, wfindex h1 i →
           wfindex h2 (shiftIndex c i).
    Proof.
      induction ins; nsimpl.
      - auto.
      - destruct i; simpl.
        + constructor.
        + apply IHins; auto.
      - apply IHins; auto.
    Qed.

    Lemma wfindex_shift'_het {α β} (αNotβ: α ≢ β) {c: Cutoff α} {h1 h2} (ins: Insertion c h1 h2) :
      ∀ (i : Index β),
        wfindex h1 i → wfindex h2 i.
    Proof.
      induction ins; nsimpl.
      - auto.
      - auto.
      - revert H; ncrush; nsimpl.
        + destruct i; simpl; auto.
        + auto.
    Qed.

    Lemma wfterm_shift'_notin_ind {α h1 s} {t : Term s} (wf_t: wfterm h1 t) (αNotIn : α ∉ depsOf s) :
      ∀ (c : Cutoff α) h2,
        Insertion c h1 h2 → wfterm h2 t.
    Proof.
      induction wf_t; intros c h2 ins; nsimpl.
      - constructor.
        refine (wfindex_shift'_het _ ins i H).
        ncrush.
      - constructor; intros.
        specialize (H0 s sΓ (namespaceDepTransFix' αNotIn sΓ)).
        eapply (H0
                  (weaken c (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                  (append h2 (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                  (weaken_insertion (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)) Γ h2 c ins)).
    Defined.

    (* Print wfterm_shift'_notin_ind. *)

    Lemma eq_ind2 {A B: Type} (P: A → B → Prop) {a1 a2 b1 b2} (eq_a: a1 = a2) (eq_b: b1 = b2) :
      P a1 b1 → P a2 b2.
    Proof.
      subst; auto.
    Defined.

    Lemma wfterm_shift'_in_ind {α h1 s} {t : Term s} (wf_t: wfterm h1 t) (αIn : α ∈ depsOf s) :
       ∀ (c : Cutoff α) h2,
         Insertion c h1 h2 → wfterm h2 (shift αIn c t).
    Proof.
      induction wf_t; intros c h2 ins; nsimpl.
      - constructor; ncrush.
        + exact (wfindex_shift'_hom ins i H).
        + exact (wfindex_shift'_het n ins i H).
      - constructor; intros; ncrush.
        + refine (
              eq_ind2
                (fun h t => wfterm h t) _ _
                (H0
                   s sΓ i
                   (weakenCutoff c (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                   (append h2 (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                   (weaken_insertion
                      (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)) Γ h2 c ins))).
          apply (@EqHVL_sound _ ts).
          eapply EqhCongAppend.
          eapply EqhRefl.
          eapply EqhSym.
          eapply Eqh_evalbindspec_shift.
          reflexivity.
        + refine (
              eq_ind2
                (fun h t => wfterm h t) _ _
                (wfterm_shift'_notin_ind (H s sΓ) n
                (weaken c (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                (append h2 (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                (weaken_insertion
                   (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)) Γ h2 c ins))).
          apply (@EqHVL_sound _ ts).
          eapply EqhCongAppend.
          eapply EqhRefl.
          eapply EqhSym.
          eapply Eqh_evalbindspec_shift.
          reflexivity.
    Defined.

    (* Print wfterm_shift'_in_ind. *)

    Lemma wfterm_weaken :
      forall k {h s} {t : Term s} {wf_t: wfterm h t},
        wfterm (append h k) (weakenTerm t k).
    Proof.
      intro k; induction k as [|β]; intros; simpl.
      - auto.
      - ncrush.
        + eapply wfterm_shift'_in_ind; eauto; constructor.
        + eapply wfterm_shift'_notin_ind; eauto; constructor.
    Qed.

    Lemma wfterms_shift :
      (forall (h1 : HVL) {ss} (ts : Terms ss) (wf_ts: wfterms h1 ts),
       ∀ {α} (c : Cutoff α) h2,
         Insertion c h1 h2 →
         wfterms h2 (shiftTerms (Cs c) ts)).
    Proof.
      intros; intros s sIn.
      specialize (wf_ts s sIn).
      ncrush.
      - apply (wfterm_shift'_in_ind wf_ts); auto.
      - eapply (wfterm_shift'_notin_ind wf_ts); eauto.
    Qed.

  End Shifting.

  Section Deletion.

    (* Describes the removal of a variable from the domain which models bot,
       strengthening and substitution for contexts.
     *)

    Inductive Deletion (k: HVL) {α} : ∀ (x: Trace α) (h1 h2: HVL), Prop :=
      | del_here :
          Deletion k T0 (HVS α k) k
      | del_there_in {x h1 h2} {β} {βIn: β ∈ depsOf (sortOf α)} :
          Deletion k x h1 h2 →
          Deletion k (TS βIn x) (HVS β h1) (HVS β h2)
      | del_there_notin {x h1 h2 β} {βNotIn: β ∉ depsOf (sortOf α)} :
          Deletion k x h1 h2 →
          Deletion k x (HVS β h1) (HVS β h2).

    Lemma weaken_deletion {α} :
      ∀ (k h h1 h2 : HVL) (x : Trace α),
        Deletion k x h1 h2 →
        Deletion k (weaken x h) (append h1 h) (append h2 h).
    Proof.
      induction h; nsimpl.
      - auto.
      - ncrush.
        + constructor; eauto.
        + constructor; eauto.
    Qed.

  End Deletion.

  Section Strengthening.

    Lemma wfindex_strengthen {α β} (αNotβ: α ≢ β)
      k {x: Trace α} {h1 h2} (del: Deletion k x h1 h2) :
      ∀ (i : Index β), wfindex h1 i → wfindex h2 i.
    Proof.
      induction del; nsimpl.
      - auto.
      - ncrush; nsimpl.
        + destruct i; simpl; auto.
        + auto.
      - ncrush; nsimpl.
        + destruct i; simpl; auto.
        + auto.
    Qed.

    Lemma wfterm_strengthen {α s} (αNotIn : α ∉ depsOf s)
      {k h1} {t : Term s} (wf_t: wfterm h1 t) :
      ∀ {xα : Trace α} {h2} (del: Deletion k xα h1 h2),
        wfterm h2 t.
    Proof.
      induction wf_t; nsimpl; intros.
      - constructor.
        refine (wfindex_strengthen _ _ del i H).
        ncrush.
      - constructor; intros.
        specialize (H0 s sΓ (namespaceDepTransFix' αNotIn sΓ)).
        refine (H0
                  (weaken xα (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                  (append h2 (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                  (weaken_deletion k (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)) Γ
            h2 xα del)).
    Defined.

    (* Print wfterm_strengthen. *)

    Lemma wfterm_strengthen0 {α s} (αNotIn : α ∉ depsOf s) :
      ∀ k {t : Term s} (wf_t: wfterm (HVS α k) t), wfterm k t.
    Proof.
      intros; eapply (wfterm_strengthen αNotIn wf_t); apply del_here.
    Qed.

    Lemma wfterms_strengthen {α ss} (hyp: ∀ s, s ∈ ss → α ∉ depsOf s)
      {k h1} (ts : Terms ss) (wf_ts: wfterms h1 ts) :
       ∀ {xα : Trace α} {h2} (del: Deletion k xα h1 h2),
         wfterms h2 ts.
    Proof.
      intros; intros s sIn.
      eapply (wfterm_strengthen (hyp s sIn) (wf_ts s sIn) del).
    Qed.

    Lemma wfterm_strengthen_append {αs s} (hyp: ∀ α, α ∈ αs → α ∉ depsOf s) :
      ∀ k h, SubHVL αs h →
      ∀ {t: Term s} (wt: wfterm (append k h) t), wfterm k t.
    Proof.
      induction 1; simpl; eauto using wfterm_strengthen0.
    Qed.

    Corollary wfterm_strengthen_fn {s s'} (fn : AuxFn s')
      (hyp: ∀ α, α ∈ auxFnsTarget fn → α ∉ depsOf s) :
      ∀ k (t' : Term s') {t : Term s},
        wfterm (append k (evalFunCall t' fn)) t →
        wfterm k t.
    Proof.
      eauto using wfterm_strengthen_append, subhvl_evalfuncall.
    Qed.

  End Strengthening.

  Section Substitution.

    Lemma wfindex_subst
      {α k t} (wft: wfterm k t)
      {x: Trace α} {h1 h2} (del: Deletion k x h1 h2) :
      ∀ (i : Index α),
        wfindex h1 i → wfterm h2 (substIndex x t i).
    Proof.
      induction del; nsimpl; ncrush.
      - destruct i.
        + auto.
        + constructor; auto.
      - destruct i; simpl.
        + constructor; nsimpl; auto.
        + eapply wfterm_shift'_in_ind; eauto; constructor.
      - eapply wfterm_shift'_in_ind; eauto; constructor.
      - destruct (α ≟ β); nsimpl.
        eapply wfterm_shift'_notin_ind; eauto; constructor.
    Qed.

    Lemma wfterm_subst
      {α k h1 s} {tα: Term (sortOf α)} (wf_tα: wfterm k tα)
      {t : Term s} (wf_t: wfterm h1 t) (αIn : α ∈ depsOf s) :
      ∀ {xα : Trace α} {h2} (del: Deletion k xα h1 h2),
        wfterm h2 (subst αIn xα tα t).
    Proof.
      induction wf_t; intros xα h2 del; simpl.
      - ncrush.
        + eauto using wfindex_subst.
        + constructor; eauto using wfindex_strengthen.
      - constructor; intros; ncrush.
        + specialize (H0 s sΓ i).
          refine (
              eq_ind2
                (fun h t => wfterm h t) _ _
                (H0
                   (weakenTrace xα (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                   (append h2 (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)))
                   (weaken_deletion k
                      (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)) Γ h2 xα del))).
          apply (@EqHVL_sound _ ts).
          eapply EqhCongAppend.
          eapply EqhRefl.
          eapply EqhSym.
          eapply Eqh_evalbindspec_subst.
          reflexivity.
        + refine (
              eq_ind2
                (fun h t => wfterm h t) _ _
                (wfterm_strengthen
                   n (H s sΓ)
                   (weaken_deletion k
                      (evalBindSpec ts HV0 (ctorBindspec ct s sΓ)) Γ h2 xα del))).
          apply (@EqHVL_sound _ ts).
          eapply EqhCongAppend.
          eapply EqhRefl.
          eapply EqhSym.
          eapply Eqh_evalbindspec_subst.
          reflexivity.
    Defined.

    Lemma wfterm_subst0 {α k s} (αIn : α ∈ depsOf s) :
      ∀ {tα: Term (sortOf α)} (wf_tα: wfterm k tα)
        {t : Term s} (wf_t: wfterm (HVS α k) t),
        wfterm k (subst αIn T0 tα t).
    Proof.
      intros; eapply (wfterm_subst wf_tα wf_t); constructor.
    Qed.

    Lemma wfterms_subst {α k h1 ss}
       {tα: Term (sortOf α)} (wf_tα: wfterm k tα)
       (ts : Terms ss) (wf_ts: wfterms h1 ts) :
       ∀ {xα : Trace α} {h2} (del: Deletion k xα h1 h2),
         wfterms h2 (substTerms (Xs xα) tα ts).
    Proof.
      intros; intros s sIn.
      specialize (wf_ts s sIn).
      ncrush.
      - apply (wfterm_subst wf_tα wf_ts); auto.
      - eapply (wfterm_strengthen n wf_ts); eauto.
    Qed.

  End Substitution.

End WellFormed.
