
Require Export Coq.Unicode.Utf8.
Require Import Coq.Program.Tactics.
Require Import Prelude.
Require Import Specification.
Require Import DeBruijn.

Module WellFormedInductive (Spec : Specification).

  Module DEBRUIJN := DEBRUIJN Spec.
  Export DEBRUIJN.

  Section Def.

    Inductive wfindex {a} : HVL → Index a → Prop :=
      | wfi_here {h} :
          wfindex (HVS a h) I0
      | wfi_there_eq {h i} :
          wfindex h i →
          wfindex (HVS a h) (IS i)
      | wfi_there_neq {h β i} :
          wfindex h i → a ≢ β →
          wfindex (HVS β h) i.

    (* A well-scopedness predicate for terms. *)
    Inductive wfterm (Γ : HVL) : ∀ {s}, Term s → Prop :=
      | wft_Var {a} {i : Index a} :
        wfindex Γ i →
        wfterm Γ (Var i)
      | wft_Fix {ct : Ctor} {ts : ∀ s, s ∈ ctorFields ct → Term s} :
          (∀ {s} (sΓ: s ∈ ctorFields ct),
             wfterm (append Γ (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))) (ts s sΓ)) →
          wfterm Γ (Fix ts).

    Lemma wfindex_inversion_hom {α} h (i : Index α) :
      wfindex (HVS α h) (IS i) -> wfindex h i.
    Proof.
      inversion 1; subst; ncrush; auto.
    Qed.

    Lemma wfindex_inversion_het {α β} (αβ: α ≢ β) h (i : Index α) :
      wfindex (HVS β h) i -> wfindex h i.
    Proof.
      inversion 1; subst; ncrush; auto.
    Qed.

  End Def.

  Section DependentInductionBased.

    Lemma wfindex_shift_hom {α} (h1 h2 : HVL) :
      forall (i : Index α),
          wfindex (append h1 h2) i ->
          wfindex (append (HVS α h1) h2) (shiftIndex (weakenCutoff C0 h2) i).
    Proof.
      induction h2; nsimpl.
      - constructor; auto.
      - ncrush.
        + destruct i; simpl.
          constructor.
          apply wfindex_inversion_hom in H.
          constructor; auto.
        + apply (wfindex_inversion_het n) in H.
          constructor; eauto.
    Qed.

    Lemma wfindex_shift_het {α β} (αβ: α ≢ β) (h1 h2 : HVL) :
      forall (y : Index β),
        wfindex (append h1 h2) y →
        wfindex (append (HVS α h1) h2) y.
    Proof.
      induction h2; nsimpl.
      - constructor; auto; ncrush.
      - destruct (β ≟ α0); ncrush.
        + destruct y; simpl.
          constructor.
          constructor.
          eapply (wfindex_inversion_hom) in H.
          apply IHh2.
          auto.
        + constructor; auto.
          apply IHh2.
          eapply (wfindex_inversion_het) in H; auto.
    Qed.

    (* Lemma wfterm_ap {h1 h1' h2 h2': HVL} {s} {t1 t1' t2 t2' : Term s}: *)
    (*   wfterm h1 t1 -> (wfterm h1' t1' -> wfterm h2' t2') -> *)
    (*   h1 = h1' -> h2 = h2' -> t1 = t1' -> t2 = t2' -> *)
    (*   wfterm h2 t2. *)
    (* Proof. *)
    (*   intros; subst; auto. *)
    (* Qed. *)

    (* Lemma wfterm_conv {h h' : HVL} {s} {t t' : Term s}: *)
    (*   wfterm h t -> h = h' -> t = t' -> *)
    (*   wfterm h' t'. *)
    (* Proof. *)
    (*   intros; subst; auto. *)
    (* Qed. *)

    Lemma wfterm_shift_ind {α} :
      (forall {s} (t : Term s) (h: HVL) (wf_t: wfterm h t),
       ∀ (h1 h2 : HVL) (heq: h = append h1 h2),
         (∀ (αIn: α ∈ depsOf s),
            wfterm (append (HVS α h1) h2) (shift αIn (weakenCutoff (C0 (α := α)) h2) t)) /\
         (∀ (αNotIn: α ∉ depsOf s),
            wfterm (append (HVS α h1) h2) t)).
    Proof.
      intros s t h wf_t; induction wf_t; intros; subst.
      - split; ncrush.
        + constructor.
          eapply wfindex_shift_hom; eauto.
        + constructor.
          eapply wfindex_shift_het; eauto.
        + constructor.
          eapply wfindex_shift_het; eauto; ncrush.
      - split; ncrush; intros; constructor; intros; ncrush.
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ sΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))
            by (symmetry; apply evalbindspec_shift).
          rewrite append_assoc.
          rewrite weakenCutoff_append.
          apply H0; auto using append_assoc.
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ sΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))
            by (symmetry; apply evalbindspec_shift).
          rewrite append_assoc.
          apply H0; auto using append_assoc.
        + rewrite append_assoc.
          apply H0; auto using append_assoc; ncrush.
    Qed.

  End DependentInductionBased.

  Section RelationBased.

    Inductive Insertion {α} : ∀ (c: Cutoff α) (h1 h2: HVL), Prop :=
      | ins_here {h} :
          Insertion C0 h (HVS α h)
      | ins_there_eq {c h1 h2} :
          Insertion c h1 h2 →
          Insertion (CS c) (HVS α h1) (HVS α h2)
      | ins_there_neq {c h1 h2 β} (αNotβ: α ≢ β) :
          Insertion c h1 h2 →
          Insertion c (HVS β h1) (HVS β h2).

    Lemma weaken_insertion {α} :
      ∀ (h h1 h2 : HVL) (c : Cutoff α),
        Insertion c h1 h2 →
        Insertion (weaken c h) (append h1 h) (append h2 h).
    Proof.
      induction h; nsimpl.
      - auto.
      - ncrush.
        + constructor; eauto.
        + constructor; eauto.
    Qed.

    Lemma wfindex_shift'_hom {α} {c: Cutoff α} {h1 h2} (ins: Insertion c h1 h2) :
      ∀ i, wfindex h1 i ->
           wfindex h2 (shiftIndex c i).
    Proof.
      induction ins; simpl; intros.
      - constructor; auto.
      - destruct i; simpl.
        + constructor.
        + apply wfindex_inversion_hom in H.
          constructor; auto.
      - apply wfindex_inversion_het in H; auto.
        constructor; auto.
    Qed.

    Lemma wfindex_shift'_het {α β} (αNotβ: α ≢ β) {c: Cutoff α} {h1 h2} (ins: Insertion c h1 h2) :
      ∀ (i : Index β),
        wfindex h1 i -> wfindex h2 i.
    Proof.
      induction ins; simpl; intros.
      - constructor; ncrush; auto.
      - apply wfindex_inversion_het in H; ncrush; auto.
        constructor; ncrush; auto.
      - destruct (β ≟ β0); ncrush.
        + destruct i.
          * constructor.
          * apply wfindex_inversion_hom in H.
            constructor; auto.
        + apply wfindex_inversion_het in H; auto.
          constructor; auto.
    Defined.

    Lemma wfterm_shift'_ind :
      (forall (h1 : HVL) {s} (t : Term s) (wf_t: wfterm h1 t),
       ∀ {α} (c : Cutoff α) h2,
         Insertion c h1 h2 →
         (∀ (αIn : α ∈ depsOf s), wfterm h2 (shift αIn c t)) /\
         (∀ (αNotIn : α ∉ depsOf s), wfterm h2 t)).
    Proof.
      intros hs s t wf_t; induction wf_t; nsimpl.
      - ncrush.
        + split; intros; ncrush; constructor.
          eapply wfindex_shift'_hom; eauto.
        + split; intros; ncrush; constructor.
          * eapply wfindex_shift'_het; eauto.
          * eapply wfindex_shift'_het; eauto.
      - split; intros; ncrush; constructor; intros; ncrush.
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ sΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))
            by (symmetry; apply evalbindspec_shift).
          eapply H0.
          apply weaken_insertion; auto.
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ sΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))
            by (symmetry; apply evalbindspec_shift).
          eapply H0; eauto.
          eapply weaken_insertion; eauto.
        + eapply H0; eauto using weaken_insertion.
          ncrush.
    Qed.

  End RelationBased.

End WellFormedInductive.
