Set Implicit Arguments.
Generalizable All Variables.

Require Export Coq.Unicode.Utf8.
Require Export Coq.Program.Equality.
Require Export Coq.Program.Tactics.
Require Export Coq.Lists.List.
(* Get rid of this. *)
Require Export Coq.Logic.FunctionalExtensionality.
Export ListNotations.

(*  _____                  _ _ _          *)
(* | ____|__ _ _   _  __ _| (_) |_ _   _  *)
(* |  _| / _` | | | |/ _` | | | __| | | | *)
(* | |__| (_| | |_| | (_| | | | |_| |_| | *)
(* |_____\__, |\__,_|\__,_|_|_|\__|\__, | *)
(*          |_|                    |___/  *)

(* We use a copy of eq that lives in set for
  computationally relevant equality proofs.  *)

Inductive seq {A : Type} (x : A) : A → Set := seq_refl : seq x x.
Arguments seq {A} x y.
Arguments seq_refl [A] [x].
Definition neg (A : Set) : Set := A → Empty_set.
Definition nseq (A : Type) (x y : A) : Set := neg (seq x y).
Infix "≡" := seq (at level 80, no associativity).
Infix "≢" := nseq (at level 80, no associativity).

Lemma sym_seq {A : Type} {x y : A} : x ≡ y → y ≡ x.
Proof.
  intro p; destruct p; constructor.
Qed.

Lemma sym_nseq {A : Type} {x y : A} : x ≢ y → y ≢ x.
Proof.
  intros p q; apply p; destruct q; constructor.
Qed.

Lemma seq_eq {A : Type} {x y : A} : x ≡ y -> x = y.
  intro H; inversion H; subst; auto.
Qed.

Lemma nseq_neq {A : Type} {x y : A} : x ≢ y -> x <> y.
  intros H p; subst. destruct (H seq_refl).
Qed.

(*  _     _     _        *)
(* | |   (_)___| |_ ___  *)
(* | |   | / __| __/ __| *)
(* | |___| \__ \ |_\__ \ *)
(* |_____|_|___/\__|___/ *)

(* Some code for list containment, list inclusion
   and heterogeneously typed lists. *)

Require Import Coq.Lists.List.
Import ListNotations.

Section Lists.
  Context {A : Set}.

  (* An inductive variant of the In predicate in Sort Set. *)
  Inductive In (x : A) : list A → Type :=
   | InHere  {xs : list A}         : In x (x :: xs)
   | InThere {xs : list A} {y : A} : In x xs → In x (y :: xs).
  Infix "∈" := In  (at level 80, no associativity).
  Notation "x ∉ xs" := (neg (x ∈ xs)) (at level 80, no associativity).

  Definition In_nil {X : Type} (x : A) (xnil : x ∈ []) : X :=
    match xnil in _ ∈ xs return
          match xs return Type with
            | []     => X
            | _ :: _ => unit
          end
    with
      | InHere _ => tt
      | InThere _ _ _ => tt
    end.

  Definition In_not_cons (x y : A) (xs : list A) :
    (x ≢ y) → (x ∉ xs) → (x ∉ y :: xs) :=
    λ H1 H2 xIn,
    match xIn in _ ∈ xs return
          match xs with
            | []      => unit
            | y :: ys => (x ≢ y) → (x ∉ ys) → Empty_set
          end
    with
      | InHere _        => λ H1 H2, H1 seq_refl
      | InThere _ _ xIn => λ H1 H2, H2 xIn
    end H1 H2.

  Fixpoint In_dec (dec : ∀ (x y : A), (x ≡ y) + (x ≢ y)) (x : A) (xs : list A) :
    (x ∈ xs) + (x ∉ xs) :=
    match xs return (x ∈ xs) + (x ∉ xs) with
      | [] => inr (In_nil (x:=x))
      | y :: xs =>
        match dec x y with
          | inl e => inl
                       match e with
                         | seq_refl => InHere x
                       end
          | inr neq_xy =>
            match In_dec dec x xs with
              | inl i => inl (InThere i)
              | inr e => inr (In_not_cons neq_xy e)
            end
        end
    end.

  Definition Incl (xs ys : list A) :=
    ∀ x, x ∈ xs → x ∈ ys.
  Infix "⊆" := Incl (at level 80, no associativity).

  Definition incl_refl (xs : list A) : xs ⊆ xs :=
    fun x px => px.

  Definition incl_trans {xs ys zs : list A} :
    xs ⊆ ys → ys ⊆ zs → xs ⊆ zs :=
    fun f g x px => g x (f x px).

  Inductive All (D : A → Set) : list A → Set :=
    | anil :
        All D []
    | acons {xs : list A} {x : A} :
        All D xs → D x → All D (x :: xs).

  Fixpoint lookup {D : A → Set} {xs : list A} (axs : All D xs) :
    ∀ (x : A), x ∈ xs → D x :=
    match axs in All _ xs return ∀ x, x ∈ xs → D x with
    | anil => λ (x : A) (H : x ∈ []), In_nil H
    | acons xs0 x axs0 d =>
        λ (y : A) (yIn : y ∈ x :: xs0),
        match
          yIn in (_ ∈ xs1)
          return
            match xs1 with
            | [] => Empty_set
            | x0 :: xs2 => D x0 → (∀ x1 : A, x1 ∈ xs2 → D x1) → D y
            end
        with
        | InHere xs1 => λ (H : D y) (_ : ∀ x0 : A, x0 ∈ xs1 → D x0), H
        | InThere xs1 y0 yIn0 =>
            λ (_ : D y0) (H0 : ∀ x0 : A, x0 ∈ xs1 → D x0), H0 y yIn0
        end d (lookup  axs0)
    end.

  Definition head {D : A → Set} {xs : list A} {x : A} :
    All D (x :: xs) → D x :=
    λ axs, match axs in All _ xs return match xs with
                                        | []      => unit
                                        | x :: xs => D x
                                      end
          with
            | anil              => tt
            | acons xs x axs px => px
          end.

  Definition tail {D : A → Set} {xs : list A} {x : A} :
    All D (x :: xs) → All D xs :=
    λ axs, match axs in All _ xs return match xs with
                                        | []      => unit
                                        | x :: xs => All D xs
                                      end
          with
            | anil              => tt
            | acons xs x axs px => axs
          end.

  Fixpoint mapAll {D1 D2 : A → Set} {xs : list A}
    (f : ∀ x, D1 x → D2 x) (axs : All D1 xs) : All D2 xs :=
    match axs in All _ xs return All D2 xs
    with
      | anil              => anil _
      | acons xs x axs px => acons (mapAll f axs) (f _ px)
    end.

  Lemma lookup_map {D1 D2 : A → Set} {xs : list A}
    (f : ∀ x, D1 x → D2 x) (axs : All D1 xs) :
    ∀ x (xIn: x ∈ xs),
      lookup (mapAll f axs) xIn = f _ (lookup axs xIn).
  Proof.
    induction axs; simpl; intros.
    + apply (In_nil xIn).
    + dependent destruction xIn; auto.
  Qed.

  Fixpoint zipAll {D1 D2 D3 : A → Set} {xs : list A}
    (f : ∀ x, D1 x → D2 x → D3 x)
    (axs1 : All D1 xs) : All D2 xs → All D3 xs :=
    match axs1 in All _ xs
    return All D2 xs → All D3 xs with
      | anil => λ _, anil D3
      | acons xs x axs px =>
        λ axs2,
        match axs2 in (All _ xs1)
        return match xs1 with
                 | [] => unit
                 | x :: xs =>
                     (D2 x → D3 x) →
                     (All D2 xs → All D3 xs) →
                     All D3 (x :: xs)
               end
        with
          | anil => tt
          | acons xs x axs px =>
            λ (HD23 : D2 x → D3 x) (HALL : All D2 xs → All D3 xs),
            acons (HALL axs) (HD23 px)
        end (f x px) (zipAll f axs)
    end.

  Fixpoint pureAll {D : A → Set} (f : ∀ (x : A) , D x)  {xs : list A} : All D xs :=
    match xs return All D xs with
      | []      => anil D
      | x :: xs => acons (pureAll f) (f x)
    end.

  Definition apAll {D1 D2 : A → Set} {xs : list A} :
    All (λ x, D1 x → D2 x) xs → All D1 xs → All D2 xs :=
    zipAll (λ x (f : D1 x → D2 x) (y : D1 x), f y).

  Infix "⊛" := apAll (at level 50, left associativity).

  Definition zipAll3 {D1 D2 D3 D4 : A → Set} {xs : list A}
    (f : ∀ x, D1 x → D2 x → D3 x → D4 x)
    (axs1 : All D1 xs) (axs2 : All D2 xs) (axs3 : All D3 xs) : All D4 xs :=
    pureAll f ⊛ axs1 ⊛ axs2 ⊛ axs3.

End Lists.
Infix "∈"  := In                     (at level 80, no associativity).
Infix "⊆"  := Incl                   (at level 80, no associativity).
Notation "x ∉ xs" := (neg (x ∈ xs)) (at level 80, no associativity).
