Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Ty : Set :=
    | TArr (T1 : Ty) (T2 : Ty)
    | TUnit 
    | TProd (T1 : Ty) (T2 : Ty).
  Scheme ind_Ty := Induction for Ty Sort Prop.
  
  Inductive Pat : Set :=
    | PVar (T : Ty)
    | PProd (p1 : Pat) (p2 : Pat).
  Scheme ind_Pat := Induction for Pat Sort Prop.
  
  Inductive Tm : Set :=
    | Var (x : (Index TmVar))
    | Abs (T : Ty) (t : Tm)
    | App (t1 : Tm) (t2 : Tm)
    | Unit 
    | Prod (t1 : Tm) (t2 : Tm)
    | Case (t1 : Tm) (p : Pat)
        (t2 : Tm).
  Scheme ind_Tm := Induction for Tm Sort Prop.
End Terms.

Section Functions.
  Fixpoint bind (p : Pat) {struct p} : Hvl :=
    match p with
      | (PVar T) => (appendHvl (HS TmVar H0) H0)
      | (PProd p1 p2) => (appendHvl (bind p1) (appendHvl (bind p2) H0))
    end.
  Scheme ind_bind_Pat := Induction for Pat Sort Prop.
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shift_TmVar_Index (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shift_TmVar_Index c x))
      end
    end.
  Fixpoint shift_TmVar_Tm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shift_TmVar_Index c x))
      | (Abs T t) => (Abs T (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TmVar_Tm (weakenCutoffTmVar c H0) t1) (shift_TmVar_Tm (weakenCutoffTmVar c H0) t2))
      | (Unit) => (Unit)
      | (Prod t0 t3) => (Prod (shift_TmVar_Tm (weakenCutoffTmVar c H0) t0) (shift_TmVar_Tm (weakenCutoffTmVar c H0) t3))
      | (Case t4 p t5) => (Case (shift_TmVar_Tm (weakenCutoffTmVar c H0) t4) p (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (bind p) H0)) t5))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (weakenTy S0 k)
    end.
  Fixpoint weakenPat (p : Pat) (k : Hvl) {struct k} : Pat :=
    match k with
      | (H0) => p
      | (HS TmVar k) => (weakenPat p k)
    end.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shift_TmVar_Tm C0 (weakenTm s k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint subst_TmVar_Index (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shift_TmVar_Tm C0 (subst_TmVar_Index d s x))
      end
    end.
  Fixpoint subst_TmVar_Tm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (subst_TmVar_Index d s x)
      | (Abs T t) => (Abs T (subst_TmVar_Tm (XS TmVar d) s t))
      | (App t1 t2) => (App (subst_TmVar_Tm d s t1) (subst_TmVar_Tm d s t2))
      | (Unit) => (Unit)
      | (Prod t0 t3) => (Prod (subst_TmVar_Tm d s t0) (subst_TmVar_Tm d s t3))
      | (Case t4 p t5) => (Case (subst_TmVar_Tm d s t4) p (subst_TmVar_Tm (weakenTrace d (bind p)) s t5))
    end.
End Subst.

Global Hint Resolve (f_equal (shift_TmVar_Tm C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Lemma stability_weaken_bind  :
  (forall (k : Hvl) (p : Pat) ,
    ((bind (weakenPat p k)) =
    (bind p))).
Proof.
  needleGenericStabilityWeaken .
Qed.
 Hint Rewrite stability_weaken_bind : interaction.
 Hint Rewrite stability_weaken_bind : stability_weaken.
Section SubstShiftReflection.
  Lemma subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind (s : Tm) :
    (forall (k0 : Hvl) (x2 : (Index TmVar)) ,
      ((subst_TmVar_Index (weakenTrace X0 k0) s (shift_TmVar_Index (weakenCutoffTmVar C0 k0) x2)) =
      (Var x2))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind := (ind_Tm (fun (s0 : Tm) =>
    (forall (k1 : Hvl) (s1 : Tm) ,
      ((subst_TmVar_Tm (weakenTrace X0 k1) s1 (shift_TmVar_Tm (weakenCutoffTmVar C0 k1) s0)) =
      s0))) (fun (x5 : (Index TmVar)) =>
    (fun (k1 : Hvl) (s0 : Tm) =>
      (subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind s0 k1 x5))) (fun (T : Ty) (t : Tm) IHt13 (k1 : Hvl) (s0 : Tm) =>
    (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k1 (HS TmVar H0)) eq_refl eq_refl) (IHt13 (HS TmVar k1) s0)))) (fun (t1 : Tm) IHt13 (t2 : Tm) IHt14 (k1 : Hvl) (s0 : Tm) =>
    (f_equal2 App (IHt13 k1 s0) (IHt14 k1 s0))) (fun (k1 : Hvl) (s0 : Tm) =>
    eq_refl) (fun (t1 : Tm) IHt13 (t2 : Tm) IHt14 (k1 : Hvl) (s0 : Tm) =>
    (f_equal2 Prod (IHt13 k1 s0) (IHt14 k1 s0))) (fun (t1 : Tm) IHt13 (p : Pat) (t2 : Tm) IHt14 (k1 : Hvl) (s0 : Tm) =>
    (f_equal3 Case (IHt13 k1 s0) eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k1 (bind p)) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k1 (bind p)) eq_refl)) (IHt14 (appendHvl k1 (bind p)) s0))))).
  Definition subst_TmVar_Tm0_shift_TmVar_Tm0_reflection (s1 : Tm) : (forall (s0 : Tm) ,
    ((subst_TmVar_Tm X0 s0 (shift_TmVar_Tm C0 s1)) =
    s1)) := (subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind s1 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shift_TmVar_Index_shift_TmVar_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shift_TmVar_Index (weakenCutoffTmVar (CS c) k) (shift_TmVar_Index (weakenCutoffTmVar C0 k) x)) =
        (shift_TmVar_Index (weakenCutoffTmVar C0 k) (shift_TmVar_Index (weakenCutoffTmVar c k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_TmVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k0 : Hvl) (c0 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar (CS c0) k0) (shift_TmVar_Tm (weakenCutoffTmVar C0 k0) s)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k0) (shift_TmVar_Tm (weakenCutoffTmVar c0 k0) s))))) (fun (x2 : (Index TmVar)) =>
      (fun (k0 : Hvl) (c0 : (Cutoff TmVar)) =>
        (f_equal Var (shift_TmVar_Index_shift_TmVar_Index0_comm_ind k0 c0 x2)))) (fun (T : Ty) (t : Tm) IHt6 (k0 : Hvl) (c0 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt6 (HS TmVar k0) c0))) (fun (t1 : Tm) IHt6 (t2 : Tm) IHt7 (k0 : Hvl) (c0 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt6 k0 c0) (IHt7 k0 c0))) (fun (k0 : Hvl) (c0 : (Cutoff TmVar)) =>
      eq_refl) (fun (t1 : Tm) IHt6 (t2 : Tm) IHt7 (k0 : Hvl) (c0 : (Cutoff TmVar)) =>
      (f_equal2 Prod (IHt6 k0 c0) (IHt7 k0 c0))) (fun (t1 : Tm) IHt6 (p : Pat) (t2 : Tm) IHt7 (k0 : Hvl) (c0 : (Cutoff TmVar)) =>
      (f_equal3 Case (IHt6 k0 c0) eq_refl (eq_trans (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append (CS c0) k0 (bind p)) (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k0 (bind p)) eq_refl)) (eq_trans (IHt7 (appendHvl k0 (bind p)) c0) (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append C0 k0 (bind p))) (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append c0 k0 (bind p))) eq_refl))))))).
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_TmVar__shift_TmVar_0_Tm_comm (s : Tm) : (forall (c0 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm (CS c0) (shift_TmVar_Tm C0 s)) =
      (shift_TmVar_Tm C0 (shift_TmVar_Tm c0 s)))) := (shift_TmVar__shift_TmVar_0_Tm_comm_ind s H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Tm_comm : interaction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Tm_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTm_shift_TmVar_Tm  :
    (forall (k9 : Hvl) (c1 : (Cutoff TmVar)) (s9 : Tm) ,
      ((weakenTm (shift_TmVar_Tm c1 s9) k9) =
      (shift_TmVar_Tm (weakenCutoffTmVar c1 k9) (weakenTm s9 k9)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shift_TmVar_Tm_subst_TmVar_Index0_comm_ind (c0 : (Cutoff TmVar)) (s0 : Tm) :
      (forall (k1 : Hvl) (x5 : (Index TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c0 k1) (subst_TmVar_Index (weakenTrace X0 k1) s0 x5)) =
        (subst_TmVar_Index (weakenTrace X0 k1) (shift_TmVar_Tm c0 s0) (shift_TmVar_Index (weakenCutoffTmVar (CS c0) k1) x5)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_TmVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s1 : Tm) =>
      (forall (k2 : Hvl) (c1 : (Cutoff TmVar)) (s2 : Tm) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c1 k2) (subst_TmVar_Tm (weakenTrace X0 k2) s2 s1)) =
        (subst_TmVar_Tm (weakenTrace X0 k2) (shift_TmVar_Tm c1 s2) (shift_TmVar_Tm (weakenCutoffTmVar (CS c1) k2) s1))))) (fun (x8 : (Index TmVar)) =>
      (fun (k2 : Hvl) (c1 : (Cutoff TmVar)) (s1 : Tm) =>
        (shift_TmVar_Tm_subst_TmVar_Index0_comm_ind c1 s1 k2 x8))) (fun (T : Ty) (t : Tm) IHt20 (k2 : Hvl) (c1 : (Cutoff TmVar)) (s1 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k2 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt20 (HS TmVar k2) c1 s1) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k2 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt20 (t2 : Tm) IHt21 (k2 : Hvl) (c1 : (Cutoff TmVar)) (s1 : Tm) =>
      (f_equal2 App (IHt20 k2 c1 s1) (IHt21 k2 c1 s1))) (fun (k2 : Hvl) (c1 : (Cutoff TmVar)) (s1 : Tm) =>
      eq_refl) (fun (t1 : Tm) IHt20 (t2 : Tm) IHt21 (k2 : Hvl) (c1 : (Cutoff TmVar)) (s1 : Tm) =>
      (f_equal2 Prod (IHt20 k2 c1 s1) (IHt21 k2 c1 s1))) (fun (t1 : Tm) IHt20 (p : Pat) (t2 : Tm) IHt21 (k2 : Hvl) (c1 : (Cutoff TmVar)) (s1 : Tm) =>
      (f_equal3 Case (IHt20 k2 c1 s1) eq_refl (eq_trans (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append c1 k2 (bind p)) (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k2 (bind p)) eq_refl eq_refl)) (eq_trans (IHt21 (appendHvl k2 (bind p)) c1 s1) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k2 (bind p))) eq_refl (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append (CS c1) k2 (bind p))) eq_refl))))))).
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shift_TmVar_Tm_subst_TmVar_Tm0_comm (s2 : Tm) : (forall (c1 : (Cutoff TmVar)) (s1 : Tm) ,
      ((shift_TmVar_Tm c1 (subst_TmVar_Tm X0 s1 s2)) =
      (subst_TmVar_Tm X0 (shift_TmVar_Tm c1 s1) (shift_TmVar_Tm (CS c1) s2)))) := (shift_TmVar__subst_TmVar_0_Tm_comm_ind s2 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma subst_TmVar_Index_shift_TmVar_Tm0_comm_ind (d : (Trace TmVar)) (s1 : Tm) :
      (forall (k2 : Hvl) (x8 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TmVar d) k2) s1 (shift_TmVar_Index (weakenCutoffTmVar C0 k2) x8)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k2) (subst_TmVar_Index (weakenTrace d k2) s1 x8)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_TmVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s7 : Tm) =>
      (forall (k8 : Hvl) (d5 : (Trace TmVar)) (s8 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TmVar d5) k8) s8 (shift_TmVar_Tm (weakenCutoffTmVar C0 k8) s7)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k8) (subst_TmVar_Tm (weakenTrace d5 k8) s8 s7))))) (fun (x11 : (Index TmVar)) =>
      (fun (k8 : Hvl) (d5 : (Trace TmVar)) (s7 : Tm) =>
        (subst_TmVar_Index_shift_TmVar_Tm0_comm_ind d5 s7 k8 x11))) (fun (T : Ty) (t : Tm) IHt27 (k8 : Hvl) (d5 : (Trace TmVar)) (s7 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d5) k8 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt27 (HS TmVar k8) d5 s7) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d5 k8 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt27 (t2 : Tm) IHt28 (k8 : Hvl) (d5 : (Trace TmVar)) (s7 : Tm) =>
      (f_equal2 App (IHt27 k8 d5 s7) (IHt28 k8 d5 s7))) (fun (k8 : Hvl) (d5 : (Trace TmVar)) (s7 : Tm) =>
      eq_refl) (fun (t1 : Tm) IHt27 (t2 : Tm) IHt28 (k8 : Hvl) (d5 : (Trace TmVar)) (s7 : Tm) =>
      (f_equal2 Prod (IHt27 k8 d5 s7) (IHt28 k8 d5 s7))) (fun (t1 : Tm) IHt27 (p : Pat) (t2 : Tm) IHt28 (k8 : Hvl) (d5 : (Trace TmVar)) (s7 : Tm) =>
      (f_equal3 Case (IHt27 k8 d5 s7) eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d5) k8 (bind p)) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k8 (bind p)) eq_refl)) (eq_trans (IHt28 (appendHvl k8 (bind p)) d5 s7) (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append C0 k8 (bind p))) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d5 k8 (bind p))) eq_refl eq_refl))))))).
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition subst_TmVar_Tm_shift_TmVar_Tm0_comm (s8 : Tm) : (forall (d5 : (Trace TmVar)) (s7 : Tm) ,
      ((subst_TmVar_Tm (XS TmVar d5) s7 (shift_TmVar_Tm C0 s8)) =
      (shift_TmVar_Tm C0 (subst_TmVar_Tm d5 s7 s8)))) := (subst_TmVar__shift_TmVar_0_Tm_comm_ind s8 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    
  End SubstSubordInd.
  Section SubstSubord.
    
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite subst_TmVar_Tm0_shift_TmVar_Tm0_reflection : interaction.
 Hint Rewrite subst_TmVar_Tm0_shift_TmVar_Tm0_reflection : reflection.
 Hint Rewrite subst_TmVar_Tm_shift_TmVar_Tm0_comm : interaction.
 Hint Rewrite subst_TmVar_Tm_shift_TmVar_Tm0_comm : subst_shift0.
 Hint Rewrite shift_TmVar_Tm_subst_TmVar_Tm0_comm : interaction.
 Hint Rewrite shift_TmVar_Tm_subst_TmVar_Tm0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma subst_TmVar_Tm_subst_TmVar_Index0_commright_ind (d5 : (Trace TmVar)) (s7 : Tm) (s8 : Tm) :
      (forall (k8 : Hvl) (x11 : (Index TmVar)) ,
        ((subst_TmVar_Tm (weakenTrace d5 k8) s7 (subst_TmVar_Index (weakenTrace X0 k8) s8 x11)) =
        (subst_TmVar_Tm (weakenTrace X0 k8) (subst_TmVar_Tm d5 s7 s8) (subst_TmVar_Index (weakenTrace (XS TmVar d5) k8) s7 x11)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_TmVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s9 : Tm) =>
      (forall (k9 : Hvl) (d6 : (Trace TmVar)) (s10 : Tm) (s11 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace d6 k9) s10 (subst_TmVar_Tm (weakenTrace X0 k9) s11 s9)) =
        (subst_TmVar_Tm (weakenTrace X0 k9) (subst_TmVar_Tm d6 s10 s11) (subst_TmVar_Tm (weakenTrace (XS TmVar d6) k9) s10 s9))))) (fun (x14 : (Index TmVar)) =>
      (fun (k9 : Hvl) (d6 : (Trace TmVar)) (s9 : Tm) (s10 : Tm) =>
        (subst_TmVar_Tm_subst_TmVar_Index0_commright_ind d6 s9 s10 k9 x14))) (fun (T : Ty) (t : Tm) IHt34 (k9 : Hvl) (d6 : (Trace TmVar)) (s9 : Tm) (s10 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d6 k9 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k9 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt34 (HS TmVar k9) d6 s9 s10) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k9 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d6) k9 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt34 (t2 : Tm) IHt35 (k9 : Hvl) (d6 : (Trace TmVar)) (s9 : Tm) (s10 : Tm) =>
      (f_equal2 App (IHt34 k9 d6 s9 s10) (IHt35 k9 d6 s9 s10))) (fun (k9 : Hvl) (d6 : (Trace TmVar)) (s9 : Tm) (s10 : Tm) =>
      eq_refl) (fun (t1 : Tm) IHt34 (t2 : Tm) IHt35 (k9 : Hvl) (d6 : (Trace TmVar)) (s9 : Tm) (s10 : Tm) =>
      (f_equal2 Prod (IHt34 k9 d6 s9 s10) (IHt35 k9 d6 s9 s10))) (fun (t1 : Tm) IHt34 (p : Pat) (t2 : Tm) IHt35 (k9 : Hvl) (d6 : (Trace TmVar)) (s9 : Tm) (s10 : Tm) =>
      (f_equal3 Case (IHt34 k9 d6 s9 s10) eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d6 k9 (bind p)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k9 (bind p)) eq_refl eq_refl)) (eq_trans (IHt35 (appendHvl k9 (bind p)) d6 s9 s10) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k9 (bind p))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d6) k9 (bind p))) eq_refl eq_refl))))))).
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition subst_TmVar_Tm_subst_TmVar_Tm0_comm (s11 : Tm) : (forall (d6 : (Trace TmVar)) (s9 : Tm) (s10 : Tm) ,
      ((subst_TmVar_Tm d6 s9 (subst_TmVar_Tm X0 s10 s11)) =
      (subst_TmVar_Tm X0 (subst_TmVar_Tm d6 s9 s10) (subst_TmVar_Tm (XS TmVar d6) s9 s11)))) := (subst_TmVar__subst_TmVar_0_Tm_comm_ind s11 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTm_subst_TmVar_Tm  :
      (forall (k9 : Hvl) (d6 : (Trace TmVar)) (s9 : Tm) (s10 : Tm) ,
        ((weakenTm (subst_TmVar_Tm d6 s9 s10) k9) =
        (subst_TmVar_Tm (weakenTrace d6 k9) s9 (weakenTm s10 k9)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTy_append  :
      (forall (S0 : Ty) (k9 : Hvl) (k10 : Hvl) ,
        ((weakenTy (weakenTy S0 k9) k10) =
        (weakenTy S0 (appendHvl k9 k10)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenPat_append  :
      (forall (p4 : Pat) (k9 : Hvl) (k10 : Hvl) ,
        ((weakenPat (weakenPat p4 k9) k10) =
        (weakenPat p4 (appendHvl k9 k10)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTm_append  :
      (forall (s9 : Tm) (k9 : Hvl) (k10 : Hvl) ,
        ((weakenTm (weakenTm s9 k9) k10) =
        (weakenTm s9 (appendHvl k9 k10)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite subst_TmVar_Tm_subst_TmVar_Tm0_comm : interaction.
 Hint Rewrite subst_TmVar_Tm_subst_TmVar_Tm0_comm : subst_subst0.
 Hint Rewrite weakenTm_shift_TmVar_Tm : weaken_shift.
 Hint Rewrite weakenTm_subst_TmVar_Tm : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k9 : Hvl) (i : (Index a)) {struct k9} : Prop :=
    match k9 with
      | (H0) => False
      | (HS b k9) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k9 i)
        end
        | (right n) => (wfindex k9 i)
      end
    end.
  Inductive wfTy (k9 : Hvl) : Ty -> Prop :=
    | wfTy_TArr {T5 : Ty}
        (wf : (wfTy k9 T5)) {T6 : Ty}
        (wf0 : (wfTy k9 T6)) :
        (wfTy k9 (TArr T5 T6))
    | wfTy_TUnit : (wfTy k9 (TUnit))
    | wfTy_TProd {T7 : Ty}
        (wf : (wfTy k9 T7)) {T8 : Ty}
        (wf0 : (wfTy k9 T8)) :
        (wfTy k9 (TProd T7 T8)).
  Inductive wfPat (k9 : Hvl) : Pat -> Prop :=
    | wfPat_PVar {T5 : Ty}
        (wf : (wfTy k9 T5)) :
        (wfPat k9 (PVar T5))
    | wfPat_PProd {p5 : Pat}
        (wf : (wfPat k9 p5)) {p6 : Pat}
        (wf0 : (wfPat k9 p6)) :
        (wfPat k9 (PProd p5 p6)).
  Inductive wfTm (k9 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x17 : (Index TmVar)}
        (wfi : (wfindex k9 x17)) :
        (wfTm k9 (Var x17))
    | wfTm_Abs {T5 : Ty}
        (wf : (wfTy k9 T5)) {t41 : Tm}
        (wf0 : (wfTm (HS TmVar k9) t41))
        : (wfTm k9 (Abs T5 t41))
    | wfTm_App {t42 : Tm}
        (wf : (wfTm k9 t42)) {t43 : Tm}
        (wf0 : (wfTm k9 t43)) :
        (wfTm k9 (App t42 t43))
    | wfTm_Unit : (wfTm k9 (Unit))
    | wfTm_Prod {t44 : Tm}
        (wf : (wfTm k9 t44)) {t45 : Tm}
        (wf0 : (wfTm k9 t45)) :
        (wfTm k9 (Prod t44 t45))
    | wfTm_Case {t46 : Tm}
        (wf : (wfTm k9 t46)) {p5 : Pat}
        (wf0 : (wfPat k9 p5)) {t47 : Tm}
        (wf1 : (wfTm (appendHvl k9 (bind p5)) t47))
        : (wfTm k9 (Case t46 p5 t47)).
  Scheme ind_wfTy := Induction for wfTy Sort Prop.
  Scheme ind_wfPat := Induction for wfPat Sort Prop.
  Scheme ind_wfTm := Induction for wfTm Sort Prop.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c2 : (Cutoff TmVar)) (k9 : Hvl) (k10 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k9 : Hvl) :
        (shifthvl_TmVar C0 k9 (HS TmVar k9))
    | shifthvl_TmVar_there_TmVar
        (c2 : (Cutoff TmVar)) (k9 : Hvl)
        (k10 : Hvl) :
        (shifthvl_TmVar c2 k9 k10) -> (shifthvl_TmVar (CS c2) (HS TmVar k9) (HS TmVar k10)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k9 : Hvl) {c2 : (Cutoff TmVar)} {k10 : Hvl} {k11 : Hvl} ,
      (shifthvl_TmVar c2 k10 k11) -> (shifthvl_TmVar (weakenCutoffTmVar c2 k9) (appendHvl k10 k9) (appendHvl k11 k9))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_TmVar__wfindex_TmVar  :
    (forall (c2 : (Cutoff TmVar)) (k9 : Hvl) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) (x17 : (Index TmVar)) ,
      (wfindex k9 x17) -> (wfindex k10 (shift_TmVar_Index c2 x17))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_TmVar__wfTy : (forall (k9 : Hvl) ,
    (forall (S0 : Ty) (wf : (wfTy k9 S0)) ,
      (forall (c2 : (Cutoff TmVar)) (k10 : Hvl) ,
        (shifthvl_TmVar c2 k9 k10) -> (wfTy k10 S0)))) := (fun (k9 : Hvl) =>
    (ind_wfTy k9 (fun (S0 : Ty) (wf : (wfTy k9 S0)) =>
      (forall (c2 : (Cutoff TmVar)) (k10 : Hvl) ,
        (shifthvl_TmVar c2 k9 k10) -> (wfTy k10 S0))) (fun (T1 : Ty) (wf : (wfTy k9 T1)) IHT1 (T2 : Ty) (wf0 : (wfTy k9 T2)) IHT2 (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
      (wfTy_TArr k10 (IHT1 c2 k10 (weaken_shifthvl_TmVar H0 ins)) (IHT2 c2 k10 (weaken_shifthvl_TmVar H0 ins)))) (fun (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
      (wfTy_TUnit k10)) (fun (T1 : Ty) (wf : (wfTy k9 T1)) IHT1 (T2 : Ty) (wf0 : (wfTy k9 T2)) IHT2 (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
      (wfTy_TProd k10 (IHT1 c2 k10 (weaken_shifthvl_TmVar H0 ins)) (IHT2 c2 k10 (weaken_shifthvl_TmVar H0 ins)))))).
  Definition shift_TmVar__wfPat : (forall (k9 : Hvl) ,
    (forall (p5 : Pat) (wf : (wfPat k9 p5)) ,
      (forall (c2 : (Cutoff TmVar)) (k10 : Hvl) ,
        (shifthvl_TmVar c2 k9 k10) -> (wfPat k10 p5)))) := (fun (k9 : Hvl) =>
    (ind_wfPat k9 (fun (p5 : Pat) (wf : (wfPat k9 p5)) =>
      (forall (c2 : (Cutoff TmVar)) (k10 : Hvl) ,
        (shifthvl_TmVar c2 k9 k10) -> (wfPat k10 p5))) (fun (T : Ty) (wf : (wfTy k9 T)) (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
      (wfPat_PVar k10 (shift_TmVar__wfTy k9 T wf c2 k10 (weaken_shifthvl_TmVar H0 ins)))) (fun (p1 : Pat) (wf : (wfPat k9 p1)) IHp1 (p2 : Pat) (wf0 : (wfPat k9 p2)) IHp2 (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
      (wfPat_PProd k10 (IHp1 c2 k10 (weaken_shifthvl_TmVar H0 ins)) (IHp2 c2 k10 (weaken_shifthvl_TmVar H0 ins)))))).
  Definition shift_TmVar__wfTm : (forall (k9 : Hvl) ,
    (forall (s9 : Tm) (wf : (wfTm k9 s9)) ,
      (forall (c2 : (Cutoff TmVar)) (k10 : Hvl) ,
        (shifthvl_TmVar c2 k9 k10) -> (wfTm k10 (shift_TmVar_Tm c2 s9))))) := (ind_wfTm (fun (k9 : Hvl) (s9 : Tm) (wf : (wfTm k9 s9)) =>
    (forall (c2 : (Cutoff TmVar)) (k10 : Hvl) ,
      (shifthvl_TmVar c2 k9 k10) -> (wfTm k10 (shift_TmVar_Tm c2 s9)))) (fun (k9 : Hvl) (x17 : (Index TmVar)) (wfi : (wfindex k9 x17)) (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
    (wfTm_Var k10 (shift_TmVar__wfindex_TmVar c2 k9 k10 ins x17 wfi))) (fun (k9 : Hvl) (T : Ty) (wf : (wfTy k9 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k9) t)) IHt41 (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
    (wfTm_Abs k10 (shift_TmVar__wfTy k9 T wf c2 k10 (weaken_shifthvl_TmVar H0 ins)) (IHt41 (CS c2) (HS TmVar k10) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k9 : Hvl) (t1 : Tm) (wf : (wfTm k9 t1)) IHt41 (t2 : Tm) (wf0 : (wfTm k9 t2)) IHt42 (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
    (wfTm_App k10 (IHt41 c2 k10 (weaken_shifthvl_TmVar H0 ins)) (IHt42 c2 k10 (weaken_shifthvl_TmVar H0 ins)))) (fun (k9 : Hvl) (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
    (wfTm_Unit k10)) (fun (k9 : Hvl) (t1 : Tm) (wf : (wfTm k9 t1)) IHt41 (t2 : Tm) (wf0 : (wfTm k9 t2)) IHt42 (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
    (wfTm_Prod k10 (IHt41 c2 k10 (weaken_shifthvl_TmVar H0 ins)) (IHt42 c2 k10 (weaken_shifthvl_TmVar H0 ins)))) (fun (k9 : Hvl) (t1 : Tm) (wf : (wfTm k9 t1)) IHt41 (p : Pat) (wf0 : (wfPat k9 p)) (t2 : Tm) (wf1 : (wfTm (appendHvl k9 (bind p)) t2)) IHt42 (c2 : (Cutoff TmVar)) (k10 : Hvl) (ins : (shifthvl_TmVar c2 k9 k10)) =>
    (wfTm_Case k10 (IHt41 c2 k10 (weaken_shifthvl_TmVar H0 ins)) (shift_TmVar__wfPat k9 p wf0 c2 k10 (weaken_shifthvl_TmVar H0 ins)) (IHt42 (weakenCutoffTmVar c2 (bind p)) (appendHvl k10 (bind p)) (weaken_shifthvl_TmVar (bind p) ins))))).
End ShiftWellFormed.
 Hint Resolve shift_TmVar__wfindex_TmVar : infra.
 Hint Resolve shift_TmVar__wfindex_TmVar : shift.
 Hint Resolve shift_TmVar__wfindex_TmVar : shift_wf.
 Hint Resolve shift_TmVar__wfindex_TmVar : wf.
 Hint Resolve shift_TmVar__wfPat shift_TmVar__wfTm shift_TmVar__wfTy : infra.
 Hint Resolve shift_TmVar__wfPat shift_TmVar__wfTm shift_TmVar__wfTy : shift.
 Hint Resolve shift_TmVar__wfPat shift_TmVar__wfTm shift_TmVar__wfTy : shift_wf.
 Hint Resolve shift_TmVar__wfPat shift_TmVar__wfTm shift_TmVar__wfTy : wf.
 Hint Constructors shifthvl_TmVar : infra.
 Hint Constructors shifthvl_TmVar : shift.
 Hint Constructors shifthvl_TmVar : shift_wf.
 Hint Constructors shifthvl_TmVar : wf.
 Hint Resolve weaken_shifthvl_TmVar : infra.
 Hint Resolve weaken_shifthvl_TmVar : shift.
 Hint Resolve weaken_shifthvl_TmVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k9 : Hvl) : (forall (d6 : (Trace TmVar)) (k10 : Hvl) (k11 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k9 X0 (HS TmVar k9) k9)
    | substhvl_TmVar_there_TmVar
        {d6 : (Trace TmVar)} {k10 : Hvl}
        {k11 : Hvl} :
        (substhvl_TmVar k9 d6 k10 k11) -> (substhvl_TmVar k9 (XS TmVar d6) (HS TmVar k10) (HS TmVar k11)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k10 : Hvl} (k9 : Hvl) {d6 : (Trace TmVar)} {k11 : Hvl} {k12 : Hvl} ,
      (substhvl_TmVar k10 d6 k11 k12) -> (substhvl_TmVar k10 (weakenTrace d6 k9) (appendHvl k11 k9) (appendHvl k12 k9))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k13 : Hvl} {s9 : Tm} (wft : (wfTm k13 s9)) :
    (forall {d6 : (Trace TmVar)} {k14 : Hvl} {k15 : Hvl} ,
      (substhvl_TmVar k13 d6 k14 k15) -> (forall {x17 : (Index TmVar)} ,
        (wfindex k14 x17) -> (wfTm k15 (subst_TmVar_Index d6 s9 x17)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Definition substhvl_TmVar_wfTy {k13 : Hvl} : (forall (k14 : Hvl) ,
    (forall (S1 : Ty) (wf0 : (wfTy k14 S1)) ,
      (forall {d6 : (Trace TmVar)} {k15 : Hvl} ,
        (substhvl_TmVar k13 d6 k14 k15) -> (wfTy k15 S1)))) := (fun (k14 : Hvl) =>
    (ind_wfTy k14 (fun (S1 : Ty) (wf0 : (wfTy k14 S1)) =>
      (forall {d6 : (Trace TmVar)} {k15 : Hvl} ,
        (substhvl_TmVar k13 d6 k14 k15) -> (wfTy k15 S1))) (fun (T1 : Ty) (wf0 : (wfTy k14 T1)) IHT1 (T2 : Ty) (wf1 : (wfTy k14 T2)) IHT2 {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
      (wfTy_TArr k15 (IHT1 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)) (IHT2 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)))) (fun {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
      (wfTy_TUnit k15)) (fun (T1 : Ty) (wf0 : (wfTy k14 T1)) IHT1 (T2 : Ty) (wf1 : (wfTy k14 T2)) IHT2 {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
      (wfTy_TProd k15 (IHT1 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)) (IHT2 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)))))).
  Definition substhvl_TmVar_wfPat {k13 : Hvl} : (forall (k14 : Hvl) ,
    (forall (p6 : Pat) (wf0 : (wfPat k14 p6)) ,
      (forall {d6 : (Trace TmVar)} {k15 : Hvl} ,
        (substhvl_TmVar k13 d6 k14 k15) -> (wfPat k15 p6)))) := (fun (k14 : Hvl) =>
    (ind_wfPat k14 (fun (p6 : Pat) (wf0 : (wfPat k14 p6)) =>
      (forall {d6 : (Trace TmVar)} {k15 : Hvl} ,
        (substhvl_TmVar k13 d6 k14 k15) -> (wfPat k15 p6))) (fun (T : Ty) (wf0 : (wfTy k14 T)) {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
      (wfPat_PVar k15 (substhvl_TmVar_wfTy k14 T wf0 (weaken_substhvl_TmVar H0 del)))) (fun (p1 : Pat) (wf0 : (wfPat k14 p1)) IHp1 (p2 : Pat) (wf1 : (wfPat k14 p2)) IHp2 {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
      (wfPat_PProd k15 (IHp1 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)) (IHp2 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)))))).
  Definition substhvl_TmVar_wfTm {k13 : Hvl} {s9 : Tm} (wf : (wfTm k13 s9)) : (forall (k14 : Hvl) ,
    (forall (s10 : Tm) (wf0 : (wfTm k14 s10)) ,
      (forall {d6 : (Trace TmVar)} {k15 : Hvl} ,
        (substhvl_TmVar k13 d6 k14 k15) -> (wfTm k15 (subst_TmVar_Tm d6 s9 s10))))) := (ind_wfTm (fun (k14 : Hvl) (s10 : Tm) (wf0 : (wfTm k14 s10)) =>
    (forall {d6 : (Trace TmVar)} {k15 : Hvl} ,
      (substhvl_TmVar k13 d6 k14 k15) -> (wfTm k15 (subst_TmVar_Tm d6 s9 s10)))) (fun (k14 : Hvl) {x17 : (Index TmVar)} (wfi : (wfindex k14 x17)) {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k14 : Hvl) (T : Ty) (wf0 : (wfTy k14 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k14) t)) IHt41 {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
    (wfTm_Abs k15 (substhvl_TmVar_wfTy k14 T wf0 (weaken_substhvl_TmVar H0 del)) (IHt41 (weakenTrace d6 (HS TmVar H0)) (HS TmVar k15) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k14 : Hvl) (t1 : Tm) (wf0 : (wfTm k14 t1)) IHt41 (t2 : Tm) (wf1 : (wfTm k14 t2)) IHt42 {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
    (wfTm_App k15 (IHt41 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)) (IHt42 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)))) (fun (k14 : Hvl) {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
    (wfTm_Unit k15)) (fun (k14 : Hvl) (t1 : Tm) (wf0 : (wfTm k14 t1)) IHt41 (t2 : Tm) (wf1 : (wfTm k14 t2)) IHt42 {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
    (wfTm_Prod k15 (IHt41 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)) (IHt42 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)))) (fun (k14 : Hvl) (t1 : Tm) (wf0 : (wfTm k14 t1)) IHt41 (p : Pat) (wf1 : (wfPat k14 p)) (t2 : Tm) (wf2 : (wfTm (appendHvl k14 (bind p)) t2)) IHt42 {d6 : (Trace TmVar)} {k15 : Hvl} (del : (substhvl_TmVar k13 d6 k14 k15)) =>
    (wfTm_Case k15 (IHt41 (weakenTrace d6 H0) k15 (weaken_substhvl_TmVar H0 del)) (substhvl_TmVar_wfPat k14 p wf1 (weaken_substhvl_TmVar H0 del)) (IHt42 (weakenTrace d6 (bind p)) (appendHvl k15 (bind p)) (weaken_substhvl_TmVar (bind p) del))))).
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar : wf.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy : wf.
 Hint Constructors substhvl_TmVar : infra.
 Hint Constructors substhvl_TmVar : subst.
 Hint Constructors substhvl_TmVar : subst_wf.
 Hint Constructors substhvl_TmVar : wf.
Fixpoint subhvl_TmVar (k13 : Hvl) {struct k13} : Prop :=
  match k13 with
    | (H0) => True
    | (HS a k13) => match a with
      | (TmVar) => (subhvl_TmVar k13)
    end
  end.
Lemma subhvl_TmVar_append  :
  (forall (k13 : Hvl) (k14 : Hvl) ,
    (subhvl_TmVar k13) -> (subhvl_TmVar k14) -> (subhvl_TmVar (appendHvl k13 k14))).
Proof.
  needleGenericSubHvlAppend .
Qed.
 Hint Resolve subhvl_TmVar_append : infra.
 Hint Resolve subhvl_TmVar_append : wf.
Lemma wfPat_strengthen_subhvl_TmVar  :
  (forall (k10 : Hvl) (k9 : Hvl) (p5 : Pat) ,
    (subhvl_TmVar k10) -> (wfPat (appendHvl k9 k10) (weakenPat p5 k10)) -> (wfPat k9 p5)).
Proof.
  needleGenericWellformedStrengthen .
Qed.
Lemma wfTy_strengthen_subhvl_TmVar  :
  (forall (k12 : Hvl) (k11 : Hvl) (S0 : Ty) ,
    (subhvl_TmVar k12) -> (wfTy (appendHvl k11 k12) (weakenTy S0 k12)) -> (wfTy k11 S0)).
Proof.
  needleGenericWellformedStrengthen .
Qed.
 Hint Extern 2 ((wfPat _ _)) => match goal with
  | H : (wfPat (appendHvl _ _) (weakenPat _ _)) |- _ => apply (wfPat_strengthen_subhvl_TmVar) in H
end : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H0 : (wfTy (appendHvl _ _) (weakenTy _ _)) |- _ => apply (wfTy_strengthen_subhvl_TmVar) in H0
end : infra wf.
Section Context.
  Inductive Ctx : Type :=
    | empty 
    | evar (G : Ctx) (T : Ty).
  Fixpoint appendCtx (G : Ctx) (G0 : Ctx) : Ctx :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendCtx G G1) T)
    end.
  Fixpoint domainCtx (G : Ctx) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainCtx G0))
    end.
  Lemma appendCtx_assoc  :
    (forall (G : Ctx) (G0 : Ctx) (G1 : Ctx) ,
      ((appendCtx G (appendCtx G0 G1)) =
      (appendCtx (appendCtx G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainCtx_appendCtx  :
    (forall (G : Ctx) (G0 : Ctx) ,
      ((domainCtx (appendCtx G G0)) =
      (appendHvl (domainCtx G) (domainCtx G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shift_TmVar_Ctx (c2 : (Cutoff TmVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TmVar_Ctx c2 G0) T)
    end.
  Fixpoint subst_TmVar_Ctx (d6 : (Trace TmVar)) (s9 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TmVar_Ctx d6 s9 G0) T)
    end.
  Lemma domainCtx_shift_TmVar_Ctx  :
    (forall (c2 : (Cutoff TmVar)) (G : Ctx) ,
      ((domainCtx (shift_TmVar_Ctx c2 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_subst_TmVar_Ctx  :
    (forall (d6 : (Trace TmVar)) (s9 : Tm) (G : Ctx) ,
      ((domainCtx (subst_TmVar_Ctx d6 s9 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainCtx_appendCtx : interaction.
 Hint Rewrite domainCtx_appendCtx : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shift_TmVar_Ctx_appendCtx  :
      (forall (c2 : (Cutoff TmVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TmVar_Ctx c2 (appendCtx G G0)) =
        (appendCtx (shift_TmVar_Ctx c2 G) (shift_TmVar_Ctx (weakenCutoffTmVar c2 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma subst_TmVar_Ctx_appendCtx  :
      (forall (d6 : (Trace TmVar)) (s9 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((subst_TmVar_Ctx d6 s9 (appendCtx G G0)) =
        (appendCtx (subst_TmVar_Ctx d6 s9 G) (subst_TmVar_Ctx (weakenTrace d6 (domainCtx G)) s9 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Ctx -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Ctx}
          {T5 : Ty} :
          (wfTy (domainCtx G) T5) -> (lookup_evar (evar G T5) I0 T5)
      | lookup_evar_there_evar
          {G : Ctx} {x17 : (Index TmVar)}
          {T5 : Ty} {T6 : Ty} :
          (lookup_evar G x17 T5) -> (lookup_evar (evar G T6) (IS x17) T5).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Ctx) (S1 : Ty) (S2 : Ty) ,
        (lookup_evar (evar G S1) I0 S2) -> (S1 =
        S2)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Ctx} {x17 : (Index TmVar)} ,
        (forall {S1 : Ty} ,
          (lookup_evar G x17 S1) -> (forall {S2 : Ty} ,
            (lookup_evar G x17 S2) -> (S1 =
            S2)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Ctx} {x17 : (Index TmVar)} {S1 : Ty} ,
        (lookup_evar G x17 S1) -> (wfTy (domainCtx G) S1)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Ctx} (G0 : Ctx) {x17 : (Index TmVar)} {T5 : Ty} ,
        (lookup_evar G x17 T5) -> (lookup_evar (appendCtx G G0) (weakenIndexTmVar x17 (domainCtx G0)) (weakenTy T5 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Ctx} {x17 : (Index TmVar)} {S1 : Ty} ,
        (lookup_evar G x17 S1) -> (wfindex (domainCtx G) x17)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Ctx -> Ctx -> Prop :=
    | shift_evar_here {G : Ctx}
        {T5 : Ty} :
        (shift_evar C0 G (evar G T5))
    | shiftevar_there_evar
        {c2 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c2 G G0) -> (shift_evar (CS c2) (evar G T) (evar G0 T)).
  Lemma weaken_shift_evar  :
    (forall (G : Ctx) {c2 : (Cutoff TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evar c2 G0 G1) -> (shift_evar (weakenCutoffTmVar c2 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c2 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evar c2 G G0) -> (shifthvl_TmVar c2 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Ctx) (T5 : Ty) (s9 : Tm) : (Trace TmVar) -> Ctx -> Ctx -> Prop :=
    | subst_evar_here :
        (subst_evar G T5 s9 X0 (evar G T5) G)
    | subst_evar_there_evar
        {d6 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evar G T5 s9 d6 G0 G1) -> (subst_evar G T5 s9 (XS TmVar d6) (evar G0 T) (evar G1 T)).
  Lemma weaken_subst_evar {G : Ctx} {T5 : Ty} {s9 : Tm} :
    (forall (G0 : Ctx) {d6 : (Trace TmVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evar G T5 s9 d6 G1 G2) -> (subst_evar G T5 s9 (weakenTrace d6 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Ctx} {s9 : Tm} {T5 : Ty} :
    (forall {d6 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evar G T5 s9 d6 G0 G1) -> (substhvl_TmVar (domainCtx G) d6 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Constructors lookup_evar : infra.
 Hint Constructors lookup_evar : lookup.
 Hint Constructors lookup_evar : shift.
 Hint Constructors lookup_evar : subst.
 Hint Resolve weaken_lookup_evar : infra.
 Hint Resolve weaken_lookup_evar : lookup.
 Hint Resolve weaken_lookup_evar : shift.
 Hint Resolve weaken_lookup_evar : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Ctx} (G0 : Ctx) {T5 : Ty} (wf : (wfTy (domainCtx G) T5)) ,
    (lookup_evar (appendCtx (evar G T5) G0) (weakenIndexTmVar I0 (domainCtx G0)) (weakenTy T5 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfPat wfTm wfTy : infra.
 Hint Constructors wfPat wfTm wfTy : wf.
 Hint Extern 10 ((wfPat _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (TArr _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TUnit)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TProd _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfPat _ _)) => match goal with
  | H : (wfPat _ (PVar _)) |- _ => inversion H; subst; clear H
  | H : (wfPat _ (PProd _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Unit)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Prod _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Case _ _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf : infra.
 Hint Resolve lookup_evar_wf : wf.
 Hint Resolve lookup_evar_wfindex : infra.
 Hint Resolve lookup_evar_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex : wf.
 Hint Constructors shift_evar : infra.
 Hint Constructors shift_evar : shift.
 Hint Constructors shift_evar : subst.
 Hint Resolve weaken_shift_evar : infra.
 Hint Resolve weaken_shift_evar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar : wf.
 Hint Constructors subst_evar : infra.
 Hint Constructors subst_evar : subst.
 Hint Resolve weaken_subst_evar : infra.
 Hint Resolve weaken_subst_evar : subst.
 Hint Resolve subst_evar_substhvl_TmVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c2 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c2 G G0)) {x17 : (Index TmVar)} {T5 : Ty} ,
    (lookup_evar G x17 T5) -> (lookup_evar G0 (shift_TmVar_Index c2 x17) T5)).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar : infra.
 Hint Resolve shift_evar_lookup_evar : lookup.
 Hint Resolve shift_evar_lookup_evar : shift.
Fixpoint size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (TArr T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (TUnit) => 1
    | (TProd T0 T3) => (plus 1 (plus (size_Ty T0) (size_Ty T3)))
  end.
Fixpoint size_Pat (p : Pat) {struct p} : nat :=
  match p with
    | (PVar T) => (plus 1 (size_Ty T))
    | (PProd p1 p2) => (plus 1 (plus (size_Pat p1) (size_Pat p2)))
  end.
Fixpoint size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (Abs T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (App t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
    | (Unit) => 1
    | (Prod t0 t3) => (plus 1 (plus (size_Tm t0) (size_Tm t3)))
    | (Case t4 p t5) => (plus 1 (plus (size_Tm t4) (plus (size_Pat p) (size_Tm t5))))
  end.
Lemma shift_TmVar__size_Tm  :
  (forall (s9 : Tm) (c1 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c1 s9)) =
    (size_Tm s9))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros x15 c1.
  reflexivity.
  - intros T4 t34 IHt34.
  intros c2; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt34 (CS c2))).
  - intros t35 IHt35 t36 IHt36.
  intros c2; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt35 c2)).
  + exact ((IHt36 c2)).
  - intros c2; simpl; apply ((f_equal S)); reflexivity.
  - intros t37 IHt37 t38 IHt38.
  intros c2; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt37 c2)).
  + exact ((IHt38 c2)).
  - intros t39 IHt39 p4 t40 IHt40.
  intros c2; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt39 c2)).
  + reflexivity.
  + exact ((IHt40 (weakenCutoffTmVar c2 (bind p4)))).
Qed.

 Hint Rewrite shift_TmVar__size_Tm : interaction.
 Hint Rewrite shift_TmVar__size_Tm : shift_size.
Lemma weaken_size_Pat  :
  (forall (k9 : Hvl) (p5 : Pat) ,
    ((size_Pat (weakenPat p5 k9)) =
    (size_Pat p5))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k9 : Hvl) (s9 : Tm) ,
    ((size_Tm (weakenTm s9 k9)) =
    (size_Tm s9))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k9 : Hvl) (S0 : Ty) ,
    ((size_Ty (weakenTy S0 k9)) =
    (size_Ty S0))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Pat weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Pat weaken_size_Tm weaken_size_Ty : weaken_size.