Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar 
    | TyVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
        | _ => (weakenIndexTmVar x k)
      end
    end.
  Fixpoint weakenIndexTyVar (y : (Index TyVar)) (k : Hvl) {struct k} : (Index TyVar) :=
    match k with
      | (H0) => y
      | (HS a k) => match a with
        | (TyVar) => (IS (weakenIndexTyVar y k))
        | _ => (weakenIndexTyVar y k)
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTyVar_append  :
    (forall (y : (Index TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTyVar (weakenIndexTyVar y k) k0) =
      (weakenIndexTyVar y (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Ty : Set :=
    | TVar (y : (Index TyVar))
    | TArr (T1 : Ty) (T2 : Ty).
  Scheme ind_Ty := Induction for Ty Sort Prop.
  
  Inductive Pat : Set :=
    | PVar 
    | PProp (p1 : Pat) (p2 : Pat).
  Scheme ind_Pat := Induction for Pat Sort Prop.
  
  Inductive Tm : Set :=
    | Var (x : (Index TmVar))
    | App (t1 : Tm) (t2 : Tm)
    | Lam (T : Ty) (t : Tm)
    | Let (p : Pat) (t1 : Tm)
        (t2 : Tm).
  Scheme ind_Tm := Induction for Tm Sort Prop.
End Terms.

Section Functions.
  Fixpoint bind (p : Pat) {struct p} : Hvl :=
    match p with
      | (PVar) => (appendHvl (HS TmVar H0) H0)
      | (PProp p1 p2) => (appendHvl (bind p1) (appendHvl (bind p2) H0))
    end.
  Scheme ind_bind_Pat := Induction for Pat Sort Prop.
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
        | _ => (weakenCutoffTmVar c k)
      end
    end.
  Fixpoint weakenCutoffTyVar (c : (Cutoff TyVar)) (k : Hvl) {struct k} : (Cutoff TyVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TyVar) => (CS (weakenCutoffTyVar c k))
        | _ => (weakenCutoffTyVar c k)
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTyVar_append  :
    (forall (c : (Cutoff TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTyVar (weakenCutoffTyVar c k) k0) =
      (weakenCutoffTyVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shiftIndex (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shiftIndex c x))
      end
    end.
  Fixpoint tshiftIndex (c : (Cutoff TyVar)) (y : (Index TyVar)) {struct c} : (Index TyVar) :=
    match c with
      | (C0) => (IS y)
      | (CS c) => match y with
        | (I0) => I0
        | (IS y) => (IS (tshiftIndex c y))
      end
    end.
  Fixpoint tshiftTy (c : (Cutoff TyVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar y) => (TVar (tshiftIndex c y))
      | (TArr T1 T2) => (TArr (tshiftTy (weakenCutoffTyVar c H0) T1) (tshiftTy (weakenCutoffTyVar c H0) T2))
    end.
  Fixpoint shiftTm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shiftIndex c x))
      | (App t1 t2) => (App (shiftTm (weakenCutoffTmVar c H0) t1) (shiftTm (weakenCutoffTmVar c H0) t2))
      | (Lam T t) => (Lam T (shiftTm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (Let p t0 t3) => (Let p (shiftTm (weakenCutoffTmVar c H0) t0) (shiftTm (weakenCutoffTmVar c (appendHvl (bind p) H0)) t3))
    end.
  Fixpoint tshiftTm (c : (Cutoff TyVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (App t1 t2) => (App (tshiftTm (weakenCutoffTyVar c H0) t1) (tshiftTm (weakenCutoffTyVar c H0) t2))
      | (Lam T t) => (Lam (tshiftTy (weakenCutoffTyVar c H0) T) (tshiftTm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t))
      | (Let p t0 t3) => (Let p (tshiftTm (weakenCutoffTyVar c H0) t0) (tshiftTm (weakenCutoffTyVar c (appendHvl (bind p) H0)) t3))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (weakenTy S0 k)
      | (HS TyVar k) => (tshiftTy C0 (weakenTy S0 k))
    end.
  Fixpoint weakenPat (p : Pat) (k : Hvl) {struct k} : Pat :=
    match k with
      | (H0) => p
      | (HS TmVar k) => (weakenPat p k)
      | (HS TyVar k) => (weakenPat p k)
    end.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shiftTm C0 (weakenTm s k))
      | (HS TyVar k) => (tshiftTm C0 (weakenTm s k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint subst_TmVar_Index (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shiftTm C0 (subst_TmVar_Index d s x))
      end
      | (XS TyVar d) => (tshiftTm C0 (subst_TmVar_Index d s x))
    end.
  Fixpoint subst_TyVar_Index (d : (Trace TyVar)) (S0 : Ty) (y : (Index TyVar)) {struct d} : Ty :=
    match d with
      | (X0) => match y with
        | (I0) => S0
        | (IS y) => (TVar y)
      end
      | (XS TmVar d) => (subst_TyVar_Index d S0 y)
      | (XS TyVar d) => match y with
        | (I0) => (TVar I0)
        | (IS y) => (tshiftTy C0 (subst_TyVar_Index d S0 y))
      end
    end.
  Fixpoint subst_TyVar_Ty (d : (Trace TyVar)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (TVar y) => (subst_TyVar_Index d S0 y)
      | (TArr T1 T2) => (TArr (subst_TyVar_Ty d S0 T1) (subst_TyVar_Ty d S0 T2))
    end.
  Fixpoint subst_TmVar_Tm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (subst_TmVar_Index d s x)
      | (App t1 t2) => (App (subst_TmVar_Tm d s t1) (subst_TmVar_Tm d s t2))
      | (Lam T t) => (Lam T (subst_TmVar_Tm (XS TmVar d) s t))
      | (Let p t0 t3) => (Let p (subst_TmVar_Tm d s t0) (subst_TmVar_Tm (weakenTrace d (bind p)) s t3))
    end.
  Fixpoint subst_TyVar_Tm (d : (Trace TyVar)) (S0 : Ty) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (App t1 t2) => (App (subst_TyVar_Tm d S0 t1) (subst_TyVar_Tm d S0 t2))
      | (Lam T t) => (Lam (subst_TyVar_Ty d S0 T) (subst_TyVar_Tm (XS TmVar d) S0 t))
      | (Let p t0 t3) => (Let p (subst_TyVar_Tm d S0 t0) (subst_TyVar_Tm (weakenTrace d (bind p)) S0 t3))
    end.
End Subst.

Global Hint Resolve (f_equal (shiftTm C0)) : cong_shift0.
Global Hint Resolve (f_equal (tshiftTm C0)) : cong_shift0.
Global Hint Resolve (f_equal (tshiftTy C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Lemma stability_weaken_bind  :
  (forall (k : Hvl) (p : Pat) ,
    ((bind (weakenPat p k)) =
    (bind p))).
Proof.
  needleGenericStabilityWeaken .
Qed.
 Hint Rewrite stability_weaken_bind : interaction.
 Hint Rewrite stability_weaken_bind : stability_weaken.
Section SubstShiftReflection.
  Lemma subst_TmVar_Index0_shiftIndex0_reflection_ind (s : Tm) :
    (forall (k4 : Hvl) (x9 : (Index TmVar)) ,
      ((subst_TmVar_Index (weakenTrace X0 k4) s (shiftIndex (weakenCutoffTmVar C0 k4) x9)) =
      (Var x9))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma subst_TyVar_Index0_tshiftIndex0_reflection_ind (S0 : Ty) :
    (forall (k4 : Hvl) (y3 : (Index TyVar)) ,
      ((subst_TyVar_Index (weakenTrace X0 k4) S0 (tshiftIndex (weakenCutoffTyVar C0 k4) y3)) =
      (TVar y3))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst_TyVar_0_tshift0_Ty_reflection_ind := (ind_Ty (fun (S2 : Ty) =>
    (forall (k7 : Hvl) (S3 : Ty) ,
      ((subst_TyVar_Ty (weakenTrace X0 k7) S3 (tshiftTy (weakenCutoffTyVar C0 k7) S2)) =
      S2))) (fun (y6 : (Index TyVar)) =>
    (fun (k7 : Hvl) (S2 : Ty) =>
      (subst_TyVar_Index0_tshiftIndex0_reflection_ind S2 k7 y6))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 TArr (IHT0 k7 S2) (IHT3 k7 S2)))).
  Definition subst_TmVar_0_shift0_Tm_reflection_ind := (ind_Tm (fun (s0 : Tm) =>
    (forall (k7 : Hvl) (s1 : Tm) ,
      ((subst_TmVar_Tm (weakenTrace X0 k7) s1 (shiftTm (weakenCutoffTmVar C0 k7) s0)) =
      s0))) (fun (x14 : (Index TmVar)) =>
    (fun (k7 : Hvl) (s0 : Tm) =>
      (subst_TmVar_Index0_shiftIndex0_reflection_ind s0 k7 x14))) (fun (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 App (IHt29 k7 s0) (IHt30 k7 s0))) (fun (T : Ty) (t : Tm) IHt29 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 Lam eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt29 (HS TmVar k7) s0)))) (fun (p : Pat) (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k7 : Hvl) (s0 : Tm) =>
    (f_equal3 Let eq_refl (IHt29 k7 s0) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k7 (bind p)) eq_refl (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k7 (bind p)) eq_refl)) (IHt30 (appendHvl k7 (bind p)) s0))))).
  Definition subst_TyVar_0_tshift0_Tm_reflection_ind := (ind_Tm (fun (s0 : Tm) =>
    (forall (k7 : Hvl) (S2 : Ty) ,
      ((subst_TyVar_Tm (weakenTrace X0 k7) S2 (tshiftTm (weakenCutoffTyVar C0 k7) s0)) =
      s0))) (fun (x14 : (Index TmVar)) =>
    (fun (k7 : Hvl) (S2 : Ty) =>
      eq_refl)) (fun (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 App (IHt29 k7 S2) (IHt30 k7 S2))) (fun (T : Ty) (t : Tm) IHt29 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 Lam (let IHT := (subst_TyVar_0_tshift0_Ty_reflection_ind T) in
    (IHT k7 S2)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt29 (HS TmVar k7) S2)))) (fun (p : Pat) (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k7 : Hvl) (S2 : Ty) =>
    (f_equal3 Let eq_refl (IHt29 k7 S2) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k7 (bind p)) eq_refl (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k7 (bind p)) eq_refl)) (IHt30 (appendHvl k7 (bind p)) S2))))).
  Definition subst_TyVar_Ty0_tshiftTy0_reflection (S3 : Ty) : (forall (S2 : Ty) ,
    ((subst_TyVar_Ty X0 S2 (tshiftTy C0 S3)) =
    S3)) := (subst_TyVar_0_tshift0_Ty_reflection_ind S3 H0).
  Definition subst_TmVar_Tm0_shiftTm0_reflection (s1 : Tm) : (forall (s0 : Tm) ,
    ((subst_TmVar_Tm X0 s0 (shiftTm C0 s1)) =
    s1)) := (subst_TmVar_0_shift0_Tm_reflection_ind s1 H0).
  Definition subst_TyVar_Tm0_tshiftTm0_reflection (s0 : Tm) : (forall (S2 : Ty) ,
    ((subst_TyVar_Tm X0 S2 (tshiftTm C0 s0)) =
    s0)) := (subst_TyVar_0_tshift0_Tm_reflection_ind s0 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shiftIndex_shiftIndex0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shiftIndex (weakenCutoffTmVar (CS c) k) (shiftIndex (weakenCutoffTmVar C0 k) x)) =
        (shiftIndex (weakenCutoffTmVar C0 k) (shiftIndex (weakenCutoffTmVar c k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma tshiftIndex_tshiftIndex0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TyVar)) (y : (Index TyVar)) ,
        ((tshiftIndex (weakenCutoffTyVar (CS c) k) (tshiftIndex (weakenCutoffTyVar C0 k) y)) =
        (tshiftIndex (weakenCutoffTyVar C0 k) (tshiftIndex (weakenCutoffTyVar c k) y)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition tshift_tshift0_Ty_comm_ind := (ind_Ty (fun (S0 : Ty) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TyVar)) ,
        ((tshiftTy (weakenCutoffTyVar (CS c4) k4) (tshiftTy (weakenCutoffTyVar C0 k4) S0)) =
        (tshiftTy (weakenCutoffTyVar C0 k4) (tshiftTy (weakenCutoffTyVar c4 k4) S0))))) (fun (y3 : (Index TyVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
        (f_equal TVar (tshiftIndex_tshiftIndex0_comm_ind k4 c4 y3)))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 TArr (IHT0 k4 c4) (IHT3 k4 c4)))).
    Definition shift_shift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TmVar)) ,
        ((shiftTm (weakenCutoffTmVar (CS c4) k4) (shiftTm (weakenCutoffTmVar C0 k4) s)) =
        (shiftTm (weakenCutoffTmVar C0 k4) (shiftTm (weakenCutoffTmVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
        (f_equal Var (shiftIndex_shiftIndex0_comm_ind k4 c4 x9)))) (fun (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt19 k4 c4) (IHt20 k4 c4))) (fun (T : Ty) (t : Tm) IHt19 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 Lam eq_refl (IHt19 (HS TmVar k4) c4))) (fun (p : Pat) (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal3 Let eq_refl (IHt19 k4 c4) (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append (CS c4) k4 (bind p)) (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k4 (bind p)) eq_refl)) (eq_trans (IHt20 (appendHvl k4 (bind p)) c4) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k4 (bind p))) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append c4 k4 (bind p))) eq_refl))))))).
    Definition shift_tshift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TmVar)) ,
        ((shiftTm (weakenCutoffTmVar c4 k4) (tshiftTm (weakenCutoffTyVar C0 k4) s)) =
        (tshiftTm (weakenCutoffTyVar C0 k4) (shiftTm (weakenCutoffTmVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
        eq_refl)) (fun (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt19 k4 c4) (IHt20 k4 c4))) (fun (T : Ty) (t : Tm) IHt19 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 Lam eq_refl (IHt19 (HS TmVar k4) c4))) (fun (p : Pat) (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal3 Let eq_refl (IHt19 k4 c4) (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append c4 k4 (bind p)) (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k4 (bind p)) eq_refl)) (eq_trans (IHt20 (appendHvl k4 (bind p)) c4) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k4 (bind p))) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append c4 k4 (bind p))) eq_refl))))))).
    Definition tshift_shift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TyVar)) ,
        ((tshiftTm (weakenCutoffTyVar c4 k4) (shiftTm (weakenCutoffTmVar C0 k4) s)) =
        (shiftTm (weakenCutoffTmVar C0 k4) (tshiftTm (weakenCutoffTyVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
        eq_refl)) (fun (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt19 k4 c4) (IHt20 k4 c4))) (fun (T : Ty) (t : Tm) IHt19 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 Lam eq_refl (IHt19 (HS TmVar k4) c4))) (fun (p : Pat) (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal3 Let eq_refl (IHt19 k4 c4) (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append c4 k4 (bind p)) (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k4 (bind p)) eq_refl)) (eq_trans (IHt20 (appendHvl k4 (bind p)) c4) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k4 (bind p))) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append c4 k4 (bind p))) eq_refl))))))).
    Definition tshift_tshift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TyVar)) ,
        ((tshiftTm (weakenCutoffTyVar (CS c4) k4) (tshiftTm (weakenCutoffTyVar C0 k4) s)) =
        (tshiftTm (weakenCutoffTyVar C0 k4) (tshiftTm (weakenCutoffTyVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
        eq_refl)) (fun (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt19 k4 c4) (IHt20 k4 c4))) (fun (T : Ty) (t : Tm) IHt19 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 Lam (let IHT := (tshift_tshift0_Ty_comm_ind T) in
      (IHT k4 c4)) (IHt19 (HS TmVar k4) c4))) (fun (p : Pat) (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal3 Let eq_refl (IHt19 k4 c4) (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append (CS c4) k4 (bind p)) (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k4 (bind p)) eq_refl)) (eq_trans (IHt20 (appendHvl k4 (bind p)) c4) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k4 (bind p))) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append c4 k4 (bind p))) eq_refl))))))).
  End ShiftCommInd.
  Section ShiftComm.
    Definition tshift_tshift0_Ty_comm (S0 : Ty) : (forall (c4 : (Cutoff TyVar)) ,
      ((tshiftTy (CS c4) (tshiftTy C0 S0)) =
      (tshiftTy C0 (tshiftTy c4 S0)))) := (tshift_tshift0_Ty_comm_ind S0 H0).
    Definition shift_shift0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TmVar)) ,
      ((shiftTm (CS c4) (shiftTm C0 s)) =
      (shiftTm C0 (shiftTm c4 s)))) := (shift_shift0_Tm_comm_ind s H0).
    Definition shift_tshift0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TmVar)) ,
      ((shiftTm c4 (tshiftTm C0 s)) =
      (tshiftTm C0 (shiftTm c4 s)))) := (shift_tshift0_Tm_comm_ind s H0).
    Definition tshift_shift0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TyVar)) ,
      ((tshiftTm c4 (shiftTm C0 s)) =
      (shiftTm C0 (tshiftTm c4 s)))) := (tshift_shift0_Tm_comm_ind s H0).
    Definition tshift_tshift0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TyVar)) ,
      ((tshiftTm (CS c4) (tshiftTm C0 s)) =
      (tshiftTm C0 (tshiftTm c4 s)))) := (tshift_tshift0_Tm_comm_ind s H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_shift0_Tm_comm shift_tshift0_Tm_comm tshift_shift0_Tm_comm tshift_tshift0_Tm_comm tshift_tshift0_Ty_comm : interaction.
 Hint Rewrite shift_shift0_Tm_comm shift_tshift0_Tm_comm tshift_shift0_Tm_comm tshift_tshift0_Tm_comm tshift_tshift0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTy_tshiftTy  :
    (forall (k39 : Hvl) (c9 : (Cutoff TyVar)) (S25 : Ty) ,
      ((weakenTy (tshiftTy c9 S25) k39) =
      (tshiftTy (weakenCutoffTyVar c9 k39) (weakenTy S25 k39)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shiftTm  :
    (forall (k39 : Hvl) (c9 : (Cutoff TmVar)) (s12 : Tm) ,
      ((weakenTm (shiftTm c9 s12) k39) =
      (shiftTm (weakenCutoffTmVar c9 k39) (weakenTm s12 k39)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_tshiftTm  :
    (forall (k39 : Hvl) (c9 : (Cutoff TyVar)) (s12 : Tm) ,
      ((weakenTm (tshiftTm c9 s12) k39) =
      (tshiftTm (weakenCutoffTyVar c9 k39) (weakenTm s12 k39)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shiftTm_subst_TmVar_Index0_comm_ind (c4 : (Cutoff TmVar)) (s0 : Tm) :
      (forall (k7 : Hvl) (x14 : (Index TmVar)) ,
        ((shiftTm (weakenCutoffTmVar c4 k7) (subst_TmVar_Index (weakenTrace X0 k7) s0 x14)) =
        (subst_TmVar_Index (weakenTrace X0 k7) (shiftTm c4 s0) (shiftIndex (weakenCutoffTmVar (CS c4) k7) x14)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tshiftTm_subst_TmVar_Index0_comm_ind (c4 : (Cutoff TyVar)) (s0 : Tm) :
      (forall (k7 : Hvl) (x14 : (Index TmVar)) ,
        ((tshiftTm (weakenCutoffTyVar c4 k7) (subst_TmVar_Index (weakenTrace X0 k7) s0 x14)) =
        (subst_TmVar_Index (weakenTrace X0 k7) (tshiftTm c4 s0) x14))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tshiftTy_subst_TyVar_Index0_comm_ind (c4 : (Cutoff TyVar)) (S2 : Ty) :
      (forall (k7 : Hvl) (y6 : (Index TyVar)) ,
        ((tshiftTy (weakenCutoffTyVar c4 k7) (subst_TyVar_Index (weakenTrace X0 k7) S2 y6)) =
        (subst_TyVar_Index (weakenTrace X0 k7) (tshiftTy c4 S2) (tshiftIndex (weakenCutoffTyVar (CS c4) k7) y6)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition tshift_subst_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S5 : Ty) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TyVar)) (S6 : Ty) ,
        ((tshiftTy (weakenCutoffTyVar c9 k12) (subst_TyVar_Ty (weakenTrace X0 k12) S6 S5)) =
        (subst_TyVar_Ty (weakenTrace X0 k12) (tshiftTy c9 S6) (tshiftTy (weakenCutoffTyVar (CS c9) k12) S5))))) (fun (y8 : (Index TyVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
        (tshiftTy_subst_TyVar_Index0_comm_ind c9 S5 k12 y8))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 TArr (IHT0 k12 c9 S5) (IHT3 k12 c9 S5)))).
    Definition shift_subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TmVar)) (s3 : Tm) ,
        ((shiftTm (weakenCutoffTmVar c9 k12) (subst_TmVar_Tm (weakenTrace X0 k12) s3 s2)) =
        (subst_TmVar_Tm (weakenTrace X0 k12) (shiftTm c9 s3) (shiftTm (weakenCutoffTmVar (CS c9) k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
        (shiftTm_subst_TmVar_Index0_comm_ind c9 s2 k12 x26))) (fun (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 App (IHt49 k12 c9 s2) (IHt50 k12 c9 s2))) (fun (T : Ty) (t : Tm) IHt49 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 Lam eq_refl (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k12) c9 s2) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (p : Pat) (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal3 Let eq_refl (IHt49 k12 c9 s2) (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append c9 k12 (bind p)) (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (bind p)) eq_refl eq_refl)) (eq_trans (IHt50 (appendHvl k12 (bind p)) c9 s2) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (bind p))) eq_refl (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append (CS c9) k12 (bind p))) eq_refl))))))).
    Definition shift_subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) ,
        ((shiftTm (weakenCutoffTmVar c9 k12) (subst_TyVar_Tm (weakenTrace X0 k12) S5 s2)) =
        (subst_TyVar_Tm (weakenTrace X0 k12) S5 (shiftTm (weakenCutoffTmVar c9 k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
        eq_refl)) (fun (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal2 App (IHt49 k12 c9 S5) (IHt50 k12 c9 S5))) (fun (T : Ty) (t : Tm) IHt49 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal2 Lam eq_refl (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k12) c9 S5) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (p : Pat) (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal3 Let eq_refl (IHt49 k12 c9 S5) (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append c9 k12 (bind p)) (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (bind p)) eq_refl eq_refl)) (eq_trans (IHt50 (appendHvl k12 (bind p)) c9 S5) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (bind p))) eq_refl (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append c9 k12 (bind p))) eq_refl))))))).
    Definition tshift_subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TyVar)) (s3 : Tm) ,
        ((tshiftTm (weakenCutoffTyVar c9 k12) (subst_TmVar_Tm (weakenTrace X0 k12) s3 s2)) =
        (subst_TmVar_Tm (weakenTrace X0 k12) (tshiftTm c9 s3) (tshiftTm (weakenCutoffTyVar c9 k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
        (tshiftTm_subst_TmVar_Index0_comm_ind c9 s2 k12 x26))) (fun (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 App (IHt49 k12 c9 s2) (IHt50 k12 c9 s2))) (fun (T : Ty) (t : Tm) IHt49 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 Lam eq_refl (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k12) c9 s2) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (p : Pat) (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal3 Let eq_refl (IHt49 k12 c9 s2) (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append c9 k12 (bind p)) (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (bind p)) eq_refl eq_refl)) (eq_trans (IHt50 (appendHvl k12 (bind p)) c9 s2) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (bind p))) eq_refl (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append c9 k12 (bind p))) eq_refl))))))).
    Definition tshift_subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) ,
        ((tshiftTm (weakenCutoffTyVar c9 k12) (subst_TyVar_Tm (weakenTrace X0 k12) S5 s2)) =
        (subst_TyVar_Tm (weakenTrace X0 k12) (tshiftTy c9 S5) (tshiftTm (weakenCutoffTyVar (CS c9) k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
        eq_refl)) (fun (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 App (IHt49 k12 c9 S5) (IHt50 k12 c9 S5))) (fun (T : Ty) (t : Tm) IHt49 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 Lam (let IHT := (tshift_subst_TyVar_0_Ty_comm_ind T) in
      (IHT k12 c9 S5)) (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k12) c9 S5) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (p : Pat) (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal3 Let eq_refl (IHt49 k12 c9 S5) (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append c9 k12 (bind p)) (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (bind p)) eq_refl eq_refl)) (eq_trans (IHt50 (appendHvl k12 (bind p)) c9 S5) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (bind p))) eq_refl (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append (CS c9) k12 (bind p))) eq_refl))))))).
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition tshiftTy_subst_TyVar_Ty0_comm (S6 : Ty) : (forall (c9 : (Cutoff TyVar)) (S5 : Ty) ,
      ((tshiftTy c9 (subst_TyVar_Ty X0 S5 S6)) =
      (subst_TyVar_Ty X0 (tshiftTy c9 S5) (tshiftTy (CS c9) S6)))) := (tshift_subst_TyVar_0_Ty_comm_ind S6 H0).
    Definition shiftTm_subst_TmVar_Tm0_comm (s3 : Tm) : (forall (c9 : (Cutoff TmVar)) (s2 : Tm) ,
      ((shiftTm c9 (subst_TmVar_Tm X0 s2 s3)) =
      (subst_TmVar_Tm X0 (shiftTm c9 s2) (shiftTm (CS c9) s3)))) := (shift_subst_TmVar_0_Tm_comm_ind s3 H0).
    Definition shiftTm_subst_TyVar_Tm0_comm (s2 : Tm) : (forall (c9 : (Cutoff TmVar)) (S5 : Ty) ,
      ((shiftTm c9 (subst_TyVar_Tm X0 S5 s2)) =
      (subst_TyVar_Tm X0 S5 (shiftTm c9 s2)))) := (shift_subst_TyVar_0_Tm_comm_ind s2 H0).
    Definition tshiftTm_subst_TmVar_Tm0_comm (s3 : Tm) : (forall (c9 : (Cutoff TyVar)) (s2 : Tm) ,
      ((tshiftTm c9 (subst_TmVar_Tm X0 s2 s3)) =
      (subst_TmVar_Tm X0 (tshiftTm c9 s2) (tshiftTm c9 s3)))) := (tshift_subst_TmVar_0_Tm_comm_ind s3 H0).
    Definition tshiftTm_subst_TyVar_Tm0_comm (s2 : Tm) : (forall (c9 : (Cutoff TyVar)) (S5 : Ty) ,
      ((tshiftTm c9 (subst_TyVar_Tm X0 S5 s2)) =
      (subst_TyVar_Tm X0 (tshiftTy c9 S5) (tshiftTm (CS c9) s2)))) := (tshift_subst_TyVar_0_Tm_comm_ind s2 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma subst_TmVar_Index_shiftTm0_comm_ind (d : (Trace TmVar)) (s2 : Tm) :
      (forall (k12 : Hvl) (x26 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TmVar d) k12) s2 (shiftIndex (weakenCutoffTmVar C0 k12) x26)) =
        (shiftTm (weakenCutoffTmVar C0 k12) (subst_TmVar_Index (weakenTrace d k12) s2 x26)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Index_tshiftTm0_comm_ind (d : (Trace TmVar)) (s2 : Tm) :
      (forall (k12 : Hvl) (x26 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TyVar d) k12) s2 x26) =
        (tshiftTm (weakenCutoffTyVar C0 k12) (subst_TmVar_Index (weakenTrace d k12) s2 x26)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shiftTy0_comm_ind (d : (Trace TyVar)) (S5 : Ty) :
      (forall (k12 : Hvl) (y8 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TmVar d) k12) S5 y8) =
        (subst_TyVar_Index (weakenTrace d k12) S5 y8))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_tshiftTy0_comm_ind (d : (Trace TyVar)) (S5 : Ty) :
      (forall (k12 : Hvl) (y8 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TyVar d) k12) S5 (tshiftIndex (weakenCutoffTyVar C0 k12) y8)) =
        (tshiftTy (weakenCutoffTyVar C0 k12) (subst_TyVar_Index (weakenTrace d k12) S5 y8)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_TyVar__tshift0_Ty_comm_ind := (ind_Ty (fun (S15 : Ty) =>
      (forall (k30 : Hvl) (d17 : (Trace TyVar)) (S16 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TyVar d17) k30) S16 (tshiftTy (weakenCutoffTyVar C0 k30) S15)) =
        (tshiftTy (weakenCutoffTyVar C0 k30) (subst_TyVar_Ty (weakenTrace d17 k30) S16 S15))))) (fun (y10 : (Index TyVar)) =>
      (fun (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
        (subst_TyVar_Index_tshiftTy0_comm_ind d17 S15 k30 y10))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 TArr (IHT0 k30 d17 S15) (IHT3 k30 d17 S15)))).
    Definition subst_TmVar__shift0_Tm_comm_ind := (ind_Tm (fun (s10 : Tm) =>
      (forall (k30 : Hvl) (d17 : (Trace TmVar)) (s11 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TmVar d17) k30) s11 (shiftTm (weakenCutoffTmVar C0 k30) s10)) =
        (shiftTm (weakenCutoffTmVar C0 k30) (subst_TmVar_Tm (weakenTrace d17 k30) s11 s10))))) (fun (x38 : (Index TmVar)) =>
      (fun (k30 : Hvl) (d17 : (Trace TmVar)) (s10 : Tm) =>
        (subst_TmVar_Index_shiftTm0_comm_ind d17 s10 k30 x38))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k30 : Hvl) (d17 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal2 App (IHt69 k30 d17 s10) (IHt70 k30 d17 s10))) (fun (T : Ty) (t : Tm) IHt69 (k30 : Hvl) (d17 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal2 Lam eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d17) k30 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k30) d17 s10) (f_equal2 shiftTm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d17 k30 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (p : Pat) (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k30 : Hvl) (d17 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal3 Let eq_refl (IHt69 k30 d17 s10) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d17) k30 (bind p)) eq_refl (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k30 (bind p)) eq_refl)) (eq_trans (IHt70 (appendHvl k30 (bind p)) d17 s10) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k30 (bind p))) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d17 k30 (bind p))) eq_refl eq_refl))))))).
    Definition subst_TmVar__tshift0_Tm_comm_ind := (ind_Tm (fun (s10 : Tm) =>
      (forall (k30 : Hvl) (d17 : (Trace TmVar)) (s11 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TyVar d17) k30) s11 (tshiftTm (weakenCutoffTyVar C0 k30) s10)) =
        (tshiftTm (weakenCutoffTyVar C0 k30) (subst_TmVar_Tm (weakenTrace d17 k30) s11 s10))))) (fun (x38 : (Index TmVar)) =>
      (fun (k30 : Hvl) (d17 : (Trace TmVar)) (s10 : Tm) =>
        (subst_TmVar_Index_tshiftTm0_comm_ind d17 s10 k30 x38))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k30 : Hvl) (d17 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal2 App (IHt69 k30 d17 s10) (IHt70 k30 d17 s10))) (fun (T : Ty) (t : Tm) IHt69 (k30 : Hvl) (d17 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal2 Lam eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TyVar d17) k30 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k30) d17 s10) (f_equal2 tshiftTm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d17 k30 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (p : Pat) (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k30 : Hvl) (d17 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal3 Let eq_refl (IHt69 k30 d17 s10) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TyVar d17) k30 (bind p)) eq_refl (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k30 (bind p)) eq_refl)) (eq_trans (IHt70 (appendHvl k30 (bind p)) d17 s10) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k30 (bind p))) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d17 k30 (bind p))) eq_refl eq_refl))))))).
    Definition subst_TyVar__shift0_Tm_comm_ind := (ind_Tm (fun (s10 : Tm) =>
      (forall (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d17 k30) S15 (shiftTm (weakenCutoffTmVar C0 k30) s10)) =
        (shiftTm (weakenCutoffTmVar C0 k30) (subst_TyVar_Tm (weakenTrace d17 k30) S15 s10))))) (fun (x38 : (Index TmVar)) =>
      (fun (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
        eq_refl)) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 App (IHt69 k30 d17 S15) (IHt70 k30 d17 S15))) (fun (T : Ty) (t : Tm) IHt69 (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 Lam eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d17 k30 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k30) d17 S15) (f_equal2 shiftTm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d17 k30 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (p : Pat) (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal3 Let eq_refl (IHt69 k30 d17 S15) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d17 k30 (bind p)) eq_refl (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k30 (bind p)) eq_refl)) (eq_trans (IHt70 (appendHvl k30 (bind p)) d17 S15) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k30 (bind p))) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d17 k30 (bind p))) eq_refl eq_refl))))))).
    Definition subst_TyVar__tshift0_Tm_comm_ind := (ind_Tm (fun (s10 : Tm) =>
      (forall (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TyVar d17) k30) S15 (tshiftTm (weakenCutoffTyVar C0 k30) s10)) =
        (tshiftTm (weakenCutoffTyVar C0 k30) (subst_TyVar_Tm (weakenTrace d17 k30) S15 s10))))) (fun (x38 : (Index TmVar)) =>
      (fun (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
        eq_refl)) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 App (IHt69 k30 d17 S15) (IHt70 k30 d17 S15))) (fun (T : Ty) (t : Tm) IHt69 (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 Lam (let IHT := (subst_TyVar__tshift0_Ty_comm_ind T) in
      (IHT k30 d17 S15)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TyVar d17) k30 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k30) d17 S15) (f_equal2 tshiftTm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d17 k30 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (p : Pat) (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k30 : Hvl) (d17 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal3 Let eq_refl (IHt69 k30 d17 S15) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TyVar d17) k30 (bind p)) eq_refl (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k30 (bind p)) eq_refl)) (eq_trans (IHt70 (appendHvl k30 (bind p)) d17 S15) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k30 (bind p))) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d17 k30 (bind p))) eq_refl eq_refl))))))).
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition subst_TyVar_Ty_tshiftTy0_comm (S22 : Ty) : (forall (d23 : (Trace TyVar)) (S21 : Ty) ,
      ((subst_TyVar_Ty (XS TyVar d23) S21 (tshiftTy C0 S22)) =
      (tshiftTy C0 (subst_TyVar_Ty d23 S21 S22)))) := (subst_TyVar__tshift0_Ty_comm_ind S22 H0).
    Definition subst_TmVar_Tm_shiftTm0_comm (s11 : Tm) : (forall (d23 : (Trace TmVar)) (s10 : Tm) ,
      ((subst_TmVar_Tm (XS TmVar d23) s10 (shiftTm C0 s11)) =
      (shiftTm C0 (subst_TmVar_Tm d23 s10 s11)))) := (subst_TmVar__shift0_Tm_comm_ind s11 H0).
    Definition subst_TmVar_Tm_tshiftTm0_comm (s11 : Tm) : (forall (d23 : (Trace TmVar)) (s10 : Tm) ,
      ((subst_TmVar_Tm (XS TyVar d23) s10 (tshiftTm C0 s11)) =
      (tshiftTm C0 (subst_TmVar_Tm d23 s10 s11)))) := (subst_TmVar__tshift0_Tm_comm_ind s11 H0).
    Definition subst_TyVar_Tm_shiftTm0_comm (s10 : Tm) : (forall (d23 : (Trace TyVar)) (S21 : Ty) ,
      ((subst_TyVar_Tm d23 S21 (shiftTm C0 s10)) =
      (shiftTm C0 (subst_TyVar_Tm d23 S21 s10)))) := (subst_TyVar__shift0_Tm_comm_ind s10 H0).
    Definition subst_TyVar_Tm_tshiftTm0_comm (s10 : Tm) : (forall (d23 : (Trace TyVar)) (S21 : Ty) ,
      ((subst_TyVar_Tm (XS TyVar d23) S21 (tshiftTm C0 s10)) =
      (tshiftTm C0 (subst_TyVar_Tm d23 S21 s10)))) := (subst_TyVar__tshift0_Tm_comm_ind s10 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    Definition subst_TyVar__TmVar_Ty_ind := (ind_Ty (fun (S21 : Ty) =>
      (forall (k36 : Hvl) (d23 : (Trace TyVar)) (S22 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TmVar d23) k36) S22 S21) =
        (subst_TyVar_Ty (weakenTrace d23 k36) S22 S21)))) (fun (y12 : (Index TyVar)) =>
      (fun (k36 : Hvl) (d23 : (Trace TyVar)) (S21 : Ty) =>
        (subst_TyVar_Index_shiftTy0_comm_ind d23 S21 k36 y12))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k36 : Hvl) (d23 : (Trace TyVar)) (S21 : Ty) =>
      (f_equal2 TArr (IHT0 k36 d23 S21) (IHT3 k36 d23 S21)))).
    Definition subst_TyVar__TmVar_Tm_ind := (ind_Tm (fun (s10 : Tm) =>
      (forall (k36 : Hvl) (d23 : (Trace TyVar)) (S21 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TmVar d23) k36) S21 s10) =
        (subst_TyVar_Tm (weakenTrace d23 k36) S21 s10)))) (fun (x41 : (Index TmVar)) =>
      (fun (k36 : Hvl) (d23 : (Trace TyVar)) (S21 : Ty) =>
        eq_refl)) (fun (t1 : Tm) IHt74 (t2 : Tm) IHt75 (k36 : Hvl) (d23 : (Trace TyVar)) (S21 : Ty) =>
      (f_equal2 App (IHt74 k36 d23 S21) (IHt75 k36 d23 S21))) (fun (T : Ty) (t : Tm) IHt74 (k36 : Hvl) (d23 : (Trace TyVar)) (S21 : Ty) =>
      (f_equal2 Lam (let IHT := (subst_TyVar__TmVar_Ty_ind T) in
      (IHT k36 d23 S21)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d23) k36 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt74 (HS TmVar k36) d23 S21) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d23 k36 (HS TmVar H0))) eq_refl eq_refl))))) (fun (p : Pat) (t1 : Tm) IHt74 (t2 : Tm) IHt75 (k36 : Hvl) (d23 : (Trace TyVar)) (S21 : Ty) =>
      (f_equal3 Let eq_refl (IHt74 k36 d23 S21) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d23) k36 (bind p)) eq_refl eq_refl) (eq_trans (IHt75 (appendHvl k36 (bind p)) d23 S21) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d23 k36 (bind p))) eq_refl eq_refl)))))).
  End SubstSubordInd.
  Section SubstSubord.
    Definition subst_TyVar_Ty_TmVar (S22 : Ty) : (forall (d23 : (Trace TyVar)) (S21 : Ty) ,
      ((subst_TyVar_Ty (XS TmVar d23) S21 S22) =
      (subst_TyVar_Ty d23 S21 S22))) := (subst_TyVar__TmVar_Ty_ind S22 H0).
    Definition subst_TyVar_Tm_TmVar (s10 : Tm) : (forall (d23 : (Trace TyVar)) (S21 : Ty) ,
      ((subst_TyVar_Tm (XS TmVar d23) S21 s10) =
      (subst_TyVar_Tm d23 S21 s10))) := (subst_TyVar__TmVar_Tm_ind s10 H0).
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite subst_TmVar_Tm0_shiftTm0_reflection subst_TyVar_Tm0_tshiftTm0_reflection subst_TyVar_Ty0_tshiftTy0_reflection : interaction.
 Hint Rewrite subst_TmVar_Tm0_shiftTm0_reflection subst_TyVar_Tm0_tshiftTm0_reflection subst_TyVar_Ty0_tshiftTy0_reflection : reflection.
 Hint Rewrite subst_TmVar_Tm_shiftTm0_comm subst_TmVar_Tm_tshiftTm0_comm subst_TyVar_Tm_shiftTm0_comm subst_TyVar_Tm_tshiftTm0_comm subst_TyVar_Ty_tshiftTy0_comm : interaction.
 Hint Rewrite subst_TmVar_Tm_shiftTm0_comm subst_TmVar_Tm_tshiftTm0_comm subst_TyVar_Tm_shiftTm0_comm subst_TyVar_Tm_tshiftTm0_comm subst_TyVar_Ty_tshiftTy0_comm : subst_shift0.
 Hint Rewrite subst_TyVar_Tm_TmVar subst_TyVar_Ty_TmVar : interaction.
 Hint Rewrite subst_TyVar_Tm_TmVar subst_TyVar_Ty_TmVar : subst_shift0.
 Hint Rewrite shiftTm_subst_TmVar_Tm0_comm shiftTm_subst_TyVar_Tm0_comm tshiftTm_subst_TmVar_Tm0_comm tshiftTm_subst_TyVar_Tm0_comm tshiftTy_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite shiftTm_subst_TmVar_Tm0_comm shiftTm_subst_TyVar_Tm0_comm tshiftTm_subst_TmVar_Tm0_comm tshiftTm_subst_TyVar_Tm0_comm tshiftTy_subst_TyVar_Ty0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma subst_TmVar_Tm_subst_TmVar_Index0_commright_ind (d23 : (Trace TmVar)) (s10 : Tm) (s11 : Tm) :
      (forall (k36 : Hvl) (x41 : (Index TmVar)) ,
        ((subst_TmVar_Tm (weakenTrace d23 k36) s10 (subst_TmVar_Index (weakenTrace X0 k36) s11 x41)) =
        (subst_TmVar_Tm (weakenTrace X0 k36) (subst_TmVar_Tm d23 s10 s11) (subst_TmVar_Index (weakenTrace (XS TmVar d23) k36) s10 x41)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Tm_subst_TmVar_Index0_commright_ind (d23 : (Trace TyVar)) (S21 : Ty) (s10 : Tm) :
      (forall (k36 : Hvl) (x41 : (Index TmVar)) ,
        ((subst_TyVar_Tm (weakenTrace d23 k36) S21 (subst_TmVar_Index (weakenTrace X0 k36) s10 x41)) =
        (subst_TmVar_Index (weakenTrace X0 k36) (subst_TyVar_Tm d23 S21 s10) x41))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Ty_subst_TyVar_Index0_commright_ind (d23 : (Trace TyVar)) (S21 : Ty) (S22 : Ty) :
      (forall (k36 : Hvl) (y12 : (Index TyVar)) ,
        ((subst_TyVar_Ty (weakenTrace d23 k36) S21 (subst_TyVar_Index (weakenTrace X0 k36) S22 y12)) =
        (subst_TyVar_Ty (weakenTrace X0 k36) (subst_TyVar_Ty d23 S21 S22) (subst_TyVar_Index (weakenTrace (XS TyVar d23) k36) S21 y12)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind (d23 : (Trace TmVar)) (s10 : Tm) (S21 : Ty) :
      (forall (k36 : Hvl) (x41 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace d23 k36) s10 x41) =
        (subst_TyVar_Tm (weakenTrace X0 k36) S21 (subst_TmVar_Index (weakenTrace (XS TyVar d23) k36) s10 x41)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_TyVar__subst_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S25 : Ty) =>
      (forall (k39 : Hvl) (d26 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace d26 k39) S26 (subst_TyVar_Ty (weakenTrace X0 k39) S27 S25)) =
        (subst_TyVar_Ty (weakenTrace X0 k39) (subst_TyVar_Ty d26 S26 S27) (subst_TyVar_Ty (weakenTrace (XS TyVar d26) k39) S26 S25))))) (fun (y14 : (Index TyVar)) =>
      (fun (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) =>
        (subst_TyVar_Ty_subst_TyVar_Index0_commright_ind d26 S25 S26 k39 y14))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) =>
      (f_equal2 TArr (IHT0 k39 d26 S25 S26) (IHT3 k39 d26 S25 S26)))).
    Definition subst_TmVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s12 : Tm) =>
      (forall (k39 : Hvl) (d26 : (Trace TmVar)) (s13 : Tm) (s14 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace d26 k39) s13 (subst_TmVar_Tm (weakenTrace X0 k39) s14 s12)) =
        (subst_TmVar_Tm (weakenTrace X0 k39) (subst_TmVar_Tm d26 s13 s14) (subst_TmVar_Tm (weakenTrace (XS TmVar d26) k39) s13 s12))))) (fun (x47 : (Index TmVar)) =>
      (fun (k39 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
        (subst_TmVar_Tm_subst_TmVar_Index0_commright_ind d26 s12 s13 k39 x47))) (fun (t1 : Tm) IHt84 (t2 : Tm) IHt85 (k39 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      (f_equal2 App (IHt84 k39 d26 s12 s13) (IHt85 k39 d26 s12 s13))) (fun (T : Ty) (t : Tm) IHt84 (k39 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      (f_equal2 Lam eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d26 k39 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k39 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt84 (HS TmVar k39) d26 s12 s13) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k39 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d26) k39 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (p : Pat) (t1 : Tm) IHt84 (t2 : Tm) IHt85 (k39 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      (f_equal3 Let eq_refl (IHt84 k39 d26 s12 s13) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d26 k39 (bind p)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k39 (bind p)) eq_refl eq_refl)) (eq_trans (IHt85 (appendHvl k39 (bind p)) d26 s12 s13) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k39 (bind p))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d26) k39 (bind p))) eq_refl eq_refl))))))).
    Definition subst_TmVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s12 : Tm) =>
      (forall (k39 : Hvl) (d26 : (Trace TmVar)) (s13 : Tm) (S25 : Ty) ,
        ((subst_TmVar_Tm (weakenTrace d26 k39) s13 (subst_TyVar_Tm (weakenTrace X0 k39) S25 s12)) =
        (subst_TyVar_Tm (weakenTrace X0 k39) S25 (subst_TmVar_Tm (weakenTrace (XS TyVar d26) k39) s13 s12))))) (fun (x47 : (Index TmVar)) =>
      (fun (k39 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (S25 : Ty) =>
        (subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind d26 s12 S25 k39 x47))) (fun (t1 : Tm) IHt84 (t2 : Tm) IHt85 (k39 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (S25 : Ty) =>
      (f_equal2 App (IHt84 k39 d26 s12 S25) (IHt85 k39 d26 s12 S25))) (fun (T : Ty) (t : Tm) IHt84 (k39 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (S25 : Ty) =>
      (f_equal2 Lam eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d26 k39 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k39 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt84 (HS TmVar k39) d26 s12 S25) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k39 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d26) k39 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (p : Pat) (t1 : Tm) IHt84 (t2 : Tm) IHt85 (k39 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (S25 : Ty) =>
      (f_equal3 Let eq_refl (IHt84 k39 d26 s12 S25) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d26 k39 (bind p)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k39 (bind p)) eq_refl eq_refl)) (eq_trans (IHt85 (appendHvl k39 (bind p)) d26 s12 S25) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k39 (bind p))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d26) k39 (bind p))) eq_refl eq_refl))))))).
    Definition subst_TyVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s12 : Tm) =>
      (forall (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (s13 : Tm) ,
        ((subst_TyVar_Tm (weakenTrace d26 k39) S25 (subst_TmVar_Tm (weakenTrace X0 k39) s13 s12)) =
        (subst_TmVar_Tm (weakenTrace X0 k39) (subst_TyVar_Tm d26 S25 s13) (subst_TyVar_Tm (weakenTrace d26 k39) S25 s12))))) (fun (x47 : (Index TmVar)) =>
      (fun (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (s12 : Tm) =>
        (subst_TyVar_Tm_subst_TmVar_Index0_commright_ind d26 S25 s12 k39 x47))) (fun (t1 : Tm) IHt84 (t2 : Tm) IHt85 (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (s12 : Tm) =>
      (f_equal2 App (IHt84 k39 d26 S25 s12) (IHt85 k39 d26 S25 s12))) (fun (T : Ty) (t : Tm) IHt84 (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (s12 : Tm) =>
      (f_equal2 Lam eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d26 k39 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k39 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt84 (HS TmVar k39) d26 S25 s12) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k39 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d26 k39 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (p : Pat) (t1 : Tm) IHt84 (t2 : Tm) IHt85 (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (s12 : Tm) =>
      (f_equal3 Let eq_refl (IHt84 k39 d26 S25 s12) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d26 k39 (bind p)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k39 (bind p)) eq_refl eq_refl)) (eq_trans (IHt85 (appendHvl k39 (bind p)) d26 S25 s12) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k39 (bind p))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d26 k39 (bind p))) eq_refl eq_refl))))))).
    Definition subst_TyVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s12 : Tm) =>
      (forall (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d26 k39) S25 (subst_TyVar_Tm (weakenTrace X0 k39) S26 s12)) =
        (subst_TyVar_Tm (weakenTrace X0 k39) (subst_TyVar_Ty d26 S25 S26) (subst_TyVar_Tm (weakenTrace (XS TyVar d26) k39) S25 s12))))) (fun (x47 : (Index TmVar)) =>
      (fun (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) =>
        eq_refl)) (fun (t1 : Tm) IHt84 (t2 : Tm) IHt85 (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) =>
      (f_equal2 App (IHt84 k39 d26 S25 S26) (IHt85 k39 d26 S25 S26))) (fun (T : Ty) (t : Tm) IHt84 (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) =>
      (f_equal2 Lam (let IHT := (subst_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT k39 d26 S25 S26)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d26 k39 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k39 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt84 (HS TmVar k39) d26 S25 S26) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k39 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d26) k39 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (p : Pat) (t1 : Tm) IHt84 (t2 : Tm) IHt85 (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) =>
      (f_equal3 Let eq_refl (IHt84 k39 d26 S25 S26) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d26 k39 (bind p)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k39 (bind p)) eq_refl eq_refl)) (eq_trans (IHt85 (appendHvl k39 (bind p)) d26 S25 S26) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k39 (bind p))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d26) k39 (bind p))) eq_refl eq_refl))))))).
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition subst_TyVar_Ty_subst_TyVar_Ty0_comm (S27 : Ty) : (forall (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) ,
      ((subst_TyVar_Ty d26 S25 (subst_TyVar_Ty X0 S26 S27)) =
      (subst_TyVar_Ty X0 (subst_TyVar_Ty d26 S25 S26) (subst_TyVar_Ty (XS TyVar d26) S25 S27)))) := (subst_TyVar__subst_TyVar_0_Ty_comm_ind S27 H0).
    Definition subst_TmVar_Tm_subst_TmVar_Tm0_comm (s14 : Tm) : (forall (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) ,
      ((subst_TmVar_Tm d26 s12 (subst_TmVar_Tm X0 s13 s14)) =
      (subst_TmVar_Tm X0 (subst_TmVar_Tm d26 s12 s13) (subst_TmVar_Tm (XS TmVar d26) s12 s14)))) := (subst_TmVar__subst_TmVar_0_Tm_comm_ind s14 H0).
    Definition subst_TmVar_Tm_subst_TyVar_Tm0_comm (s13 : Tm) : (forall (d26 : (Trace TmVar)) (s12 : Tm) (S25 : Ty) ,
      ((subst_TmVar_Tm d26 s12 (subst_TyVar_Tm X0 S25 s13)) =
      (subst_TyVar_Tm X0 S25 (subst_TmVar_Tm (XS TyVar d26) s12 s13)))) := (subst_TmVar__subst_TyVar_0_Tm_comm_ind s13 H0).
    Definition subst_TyVar_Tm_subst_TmVar_Tm0_comm (s13 : Tm) : (forall (d26 : (Trace TyVar)) (S25 : Ty) (s12 : Tm) ,
      ((subst_TyVar_Tm d26 S25 (subst_TmVar_Tm X0 s12 s13)) =
      (subst_TmVar_Tm X0 (subst_TyVar_Tm d26 S25 s12) (subst_TyVar_Tm d26 S25 s13)))) := (subst_TyVar__subst_TmVar_0_Tm_comm_ind s13 H0).
    Definition subst_TyVar_Tm_subst_TyVar_Tm0_comm (s12 : Tm) : (forall (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) ,
      ((subst_TyVar_Tm d26 S25 (subst_TyVar_Tm X0 S26 s12)) =
      (subst_TyVar_Tm X0 (subst_TyVar_Ty d26 S25 S26) (subst_TyVar_Tm (XS TyVar d26) S25 s12)))) := (subst_TyVar__subst_TyVar_0_Tm_comm_ind s12 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTy_subst_TyVar_Ty  :
      (forall (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (S26 : Ty) ,
        ((weakenTy (subst_TyVar_Ty d26 S25 S26) k39) =
        (subst_TyVar_Ty (weakenTrace d26 k39) S25 (weakenTy S26 k39)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TmVar_Tm  :
      (forall (k39 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) ,
        ((weakenTm (subst_TmVar_Tm d26 s12 s13) k39) =
        (subst_TmVar_Tm (weakenTrace d26 k39) s12 (weakenTm s13 k39)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TyVar_Tm  :
      (forall (k39 : Hvl) (d26 : (Trace TyVar)) (S25 : Ty) (s12 : Tm) ,
        ((weakenTm (subst_TyVar_Tm d26 S25 s12) k39) =
        (subst_TyVar_Tm (weakenTrace d26 k39) S25 (weakenTm s12 k39)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTy_append  :
      (forall (S25 : Ty) (k39 : Hvl) (k40 : Hvl) ,
        ((weakenTy (weakenTy S25 k39) k40) =
        (weakenTy S25 (appendHvl k39 k40)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenPat_append  :
      (forall (p16 : Pat) (k39 : Hvl) (k40 : Hvl) ,
        ((weakenPat (weakenPat p16 k39) k40) =
        (weakenPat p16 (appendHvl k39 k40)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTm_append  :
      (forall (s12 : Tm) (k39 : Hvl) (k40 : Hvl) ,
        ((weakenTm (weakenTm s12 k39) k40) =
        (weakenTm s12 (appendHvl k39 k40)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : subst_subst0.
 Hint Rewrite weakenTm_shiftTm weakenTm_tshiftTm weakenTy_tshiftTy : weaken_shift.
 Hint Rewrite weakenTm_subst_TmVar_Tm weakenTm_subst_TyVar_Tm weakenTy_subst_TyVar_Ty : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k39 : Hvl) (i : (Index a)) {struct k39} : Prop :=
    match k39 with
      | (H0) => False
      | (HS b k39) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k39 i)
        end
        | (right n) => (wfindex k39 i)
      end
    end.
  Inductive wfTy (k39 : Hvl) : Ty -> Prop :=
    | wfTy_TVar
        {y17 : (Index TyVar)}
        (wfi : (wfindex k39 y17)) :
        (wfTy k39 (TVar y17))
    | wfTy_TArr {T32 : Ty}
        (wf : (wfTy k39 T32)) {T33 : Ty}
        (wf0 : (wfTy k39 T33)) :
        (wfTy k39 (TArr T32 T33)).
  Inductive wfPat (k39 : Hvl) : Pat -> Prop :=
    | wfPat_PVar :
        (wfPat k39 (PVar))
    | wfPat_PProp {p18 : Pat}
        (wf : (wfPat k39 p18))
        {p19 : Pat}
        (wf0 : (wfPat k39 p19)) :
        (wfPat k39 (PProp p18 p19)).
  Inductive wfTm (k39 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x52 : (Index TmVar)}
        (wfi : (wfindex k39 x52)) :
        (wfTm k39 (Var x52))
    | wfTm_App {t94 : Tm}
        (wf : (wfTm k39 t94)) {t95 : Tm}
        (wf0 : (wfTm k39 t95)) :
        (wfTm k39 (App t94 t95))
    | wfTm_Lam {T32 : Ty}
        (wf : (wfTy k39 T32)) {t96 : Tm}
        (wf0 : (wfTm (HS TmVar k39) t96))
        : (wfTm k39 (Lam T32 t96))
    | wfTm_Let {p18 : Pat}
        (wf : (wfPat k39 p18))
        {t97 : Tm}
        (wf0 : (wfTm k39 t97))
        {t98 : Tm}
        (wf1 : (wfTm (appendHvl k39 (bind p18)) t98))
        : (wfTm k39 (Let p18 t97 t98)).
  Scheme ind_wfTy := Induction for wfTy Sort Prop.
  Scheme ind_wfPat := Induction for wfPat Sort Prop.
  Scheme ind_wfTm := Induction for wfTm Sort Prop.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c12 : (Cutoff TmVar)) (k39 : Hvl) (k40 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k39 : Hvl) :
        (shifthvl_TmVar C0 k39 (HS TmVar k39))
    | shifthvl_TmVar_there_TmVar
        (c12 : (Cutoff TmVar))
        (k39 : Hvl) (k40 : Hvl) :
        (shifthvl_TmVar c12 k39 k40) -> (shifthvl_TmVar (CS c12) (HS TmVar k39) (HS TmVar k40))
    | shifthvl_TmVar_there_TyVar
        (c12 : (Cutoff TmVar))
        (k39 : Hvl) (k40 : Hvl) :
        (shifthvl_TmVar c12 k39 k40) -> (shifthvl_TmVar c12 (HS TyVar k39) (HS TyVar k40)).
  Inductive shifthvl_TyVar : (forall (c12 : (Cutoff TyVar)) (k39 : Hvl) (k40 : Hvl) ,
    Prop) :=
    | shifthvl_TyVar_here
        (k39 : Hvl) :
        (shifthvl_TyVar C0 k39 (HS TyVar k39))
    | shifthvl_TyVar_there_TmVar
        (c12 : (Cutoff TyVar))
        (k39 : Hvl) (k40 : Hvl) :
        (shifthvl_TyVar c12 k39 k40) -> (shifthvl_TyVar c12 (HS TmVar k39) (HS TmVar k40))
    | shifthvl_TyVar_there_TyVar
        (c12 : (Cutoff TyVar))
        (k39 : Hvl) (k40 : Hvl) :
        (shifthvl_TyVar c12 k39 k40) -> (shifthvl_TyVar (CS c12) (HS TyVar k39) (HS TyVar k40)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k39 : Hvl) {c12 : (Cutoff TmVar)} {k40 : Hvl} {k41 : Hvl} ,
      (shifthvl_TmVar c12 k40 k41) -> (shifthvl_TmVar (weakenCutoffTmVar c12 k39) (appendHvl k40 k39) (appendHvl k41 k39))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TyVar  :
    (forall (k39 : Hvl) {c12 : (Cutoff TyVar)} {k40 : Hvl} {k41 : Hvl} ,
      (shifthvl_TyVar c12 k40 k41) -> (shifthvl_TyVar (weakenCutoffTyVar c12 k39) (appendHvl k40 k39) (appendHvl k41 k39))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_wfindex_TmVar  :
    (forall (c12 : (Cutoff TmVar)) (k39 : Hvl) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) (x52 : (Index TmVar)) ,
      (wfindex k39 x52) -> (wfindex k40 (shiftIndex c12 x52))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_wfindex_TyVar  :
    (forall (c12 : (Cutoff TmVar)) (k39 : Hvl) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) (y17 : (Index TyVar)) ,
      (wfindex k39 y17) -> (wfindex k40 y17)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma tshift_wfindex_TmVar  :
    (forall (c12 : (Cutoff TyVar)) (k39 : Hvl) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) (x52 : (Index TmVar)) ,
      (wfindex k39 x52) -> (wfindex k40 x52)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma tshift_wfindex_TyVar  :
    (forall (c12 : (Cutoff TyVar)) (k39 : Hvl) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) (y17 : (Index TyVar)) ,
      (wfindex k39 y17) -> (wfindex k40 (tshiftIndex c12 y17))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_wfTy : (forall (k39 : Hvl) ,
    (forall (S25 : Ty) (wf : (wfTy k39 S25)) ,
      (forall (c12 : (Cutoff TmVar)) (k40 : Hvl) ,
        (shifthvl_TmVar c12 k39 k40) -> (wfTy k40 S25)))) := (fun (k39 : Hvl) =>
    (ind_wfTy k39 (fun (S25 : Ty) (wf : (wfTy k39 S25)) =>
      (forall (c12 : (Cutoff TmVar)) (k40 : Hvl) ,
        (shifthvl_TmVar c12 k39 k40) -> (wfTy k40 S25))) (fun (y17 : (Index TyVar)) (wfi : (wfindex k39 y17)) (c12 : (Cutoff TmVar)) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) =>
      (wfTy_TVar k40 (shift_wfindex_TyVar c12 k39 k40 ins y17 wfi))) (fun (T1 : Ty) (wf : (wfTy k39 T1)) IHT0 (T2 : Ty) (wf0 : (wfTy k39 T2)) IHT3 (c12 : (Cutoff TmVar)) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) =>
      (wfTy_TArr k40 (IHT0 c12 k40 (weaken_shifthvl_TmVar H0 ins)) (IHT3 c12 k40 (weaken_shifthvl_TmVar H0 ins)))))).
  Definition tshift_wfTy : (forall (k39 : Hvl) ,
    (forall (S25 : Ty) (wf : (wfTy k39 S25)) ,
      (forall (c12 : (Cutoff TyVar)) (k40 : Hvl) ,
        (shifthvl_TyVar c12 k39 k40) -> (wfTy k40 (tshiftTy c12 S25))))) := (fun (k39 : Hvl) =>
    (ind_wfTy k39 (fun (S25 : Ty) (wf : (wfTy k39 S25)) =>
      (forall (c12 : (Cutoff TyVar)) (k40 : Hvl) ,
        (shifthvl_TyVar c12 k39 k40) -> (wfTy k40 (tshiftTy c12 S25)))) (fun (y17 : (Index TyVar)) (wfi : (wfindex k39 y17)) (c12 : (Cutoff TyVar)) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) =>
      (wfTy_TVar k40 (tshift_wfindex_TyVar c12 k39 k40 ins y17 wfi))) (fun (T1 : Ty) (wf : (wfTy k39 T1)) IHT0 (T2 : Ty) (wf0 : (wfTy k39 T2)) IHT3 (c12 : (Cutoff TyVar)) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) =>
      (wfTy_TArr k40 (IHT0 c12 k40 (weaken_shifthvl_TyVar H0 ins)) (IHT3 c12 k40 (weaken_shifthvl_TyVar H0 ins)))))).
  Definition shift_wfPat : (forall (k39 : Hvl) ,
    (forall (p18 : Pat) (wf : (wfPat k39 p18)) ,
      (forall (c12 : (Cutoff TmVar)) (k40 : Hvl) ,
        (shifthvl_TmVar c12 k39 k40) -> (wfPat k40 p18)))) := (fun (k39 : Hvl) =>
    (ind_wfPat k39 (fun (p18 : Pat) (wf : (wfPat k39 p18)) =>
      (forall (c12 : (Cutoff TmVar)) (k40 : Hvl) ,
        (shifthvl_TmVar c12 k39 k40) -> (wfPat k40 p18))) (fun (c12 : (Cutoff TmVar)) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) =>
      (wfPat_PVar k40)) (fun (p1 : Pat) (wf : (wfPat k39 p1)) IHp1 (p2 : Pat) (wf0 : (wfPat k39 p2)) IHp2 (c12 : (Cutoff TmVar)) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) =>
      (wfPat_PProp k40 (IHp1 c12 k40 (weaken_shifthvl_TmVar H0 ins)) (IHp2 c12 k40 (weaken_shifthvl_TmVar H0 ins)))))).
  Definition tshift_wfPat : (forall (k39 : Hvl) ,
    (forall (p18 : Pat) (wf : (wfPat k39 p18)) ,
      (forall (c12 : (Cutoff TyVar)) (k40 : Hvl) ,
        (shifthvl_TyVar c12 k39 k40) -> (wfPat k40 p18)))) := (fun (k39 : Hvl) =>
    (ind_wfPat k39 (fun (p18 : Pat) (wf : (wfPat k39 p18)) =>
      (forall (c12 : (Cutoff TyVar)) (k40 : Hvl) ,
        (shifthvl_TyVar c12 k39 k40) -> (wfPat k40 p18))) (fun (c12 : (Cutoff TyVar)) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) =>
      (wfPat_PVar k40)) (fun (p1 : Pat) (wf : (wfPat k39 p1)) IHp1 (p2 : Pat) (wf0 : (wfPat k39 p2)) IHp2 (c12 : (Cutoff TyVar)) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) =>
      (wfPat_PProp k40 (IHp1 c12 k40 (weaken_shifthvl_TyVar H0 ins)) (IHp2 c12 k40 (weaken_shifthvl_TyVar H0 ins)))))).
  Definition shift_wfTm : (forall (k39 : Hvl) ,
    (forall (s12 : Tm) (wf : (wfTm k39 s12)) ,
      (forall (c12 : (Cutoff TmVar)) (k40 : Hvl) ,
        (shifthvl_TmVar c12 k39 k40) -> (wfTm k40 (shiftTm c12 s12))))) := (ind_wfTm (fun (k39 : Hvl) (s12 : Tm) (wf : (wfTm k39 s12)) =>
    (forall (c12 : (Cutoff TmVar)) (k40 : Hvl) ,
      (shifthvl_TmVar c12 k39 k40) -> (wfTm k40 (shiftTm c12 s12)))) (fun (k39 : Hvl) (x52 : (Index TmVar)) (wfi : (wfindex k39 x52)) (c12 : (Cutoff TmVar)) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) =>
    (wfTm_Var k40 (shift_wfindex_TmVar c12 k39 k40 ins x52 wfi))) (fun (k39 : Hvl) (t1 : Tm) (wf : (wfTm k39 t1)) IHt94 (t2 : Tm) (wf0 : (wfTm k39 t2)) IHt95 (c12 : (Cutoff TmVar)) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) =>
    (wfTm_App k40 (IHt94 c12 k40 (weaken_shifthvl_TmVar H0 ins)) (IHt95 c12 k40 (weaken_shifthvl_TmVar H0 ins)))) (fun (k39 : Hvl) (T : Ty) (wf : (wfTy k39 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k39) t)) IHt94 (c12 : (Cutoff TmVar)) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) =>
    (wfTm_Lam k40 (shift_wfTy k39 T wf c12 k40 (weaken_shifthvl_TmVar H0 ins)) (IHt94 (CS c12) (HS TmVar k40) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k39 : Hvl) (p : Pat) (wf : (wfPat k39 p)) (t1 : Tm) (wf0 : (wfTm k39 t1)) IHt94 (t2 : Tm) (wf1 : (wfTm (appendHvl k39 (bind p)) t2)) IHt95 (c12 : (Cutoff TmVar)) (k40 : Hvl) (ins : (shifthvl_TmVar c12 k39 k40)) =>
    (wfTm_Let k40 (shift_wfPat k39 p wf c12 k40 (weaken_shifthvl_TmVar H0 ins)) (IHt94 c12 k40 (weaken_shifthvl_TmVar H0 ins)) (IHt95 (weakenCutoffTmVar c12 (bind p)) (appendHvl k40 (bind p)) (weaken_shifthvl_TmVar (bind p) ins))))).
  Definition tshift_wfTm : (forall (k39 : Hvl) ,
    (forall (s12 : Tm) (wf : (wfTm k39 s12)) ,
      (forall (c12 : (Cutoff TyVar)) (k40 : Hvl) ,
        (shifthvl_TyVar c12 k39 k40) -> (wfTm k40 (tshiftTm c12 s12))))) := (ind_wfTm (fun (k39 : Hvl) (s12 : Tm) (wf : (wfTm k39 s12)) =>
    (forall (c12 : (Cutoff TyVar)) (k40 : Hvl) ,
      (shifthvl_TyVar c12 k39 k40) -> (wfTm k40 (tshiftTm c12 s12)))) (fun (k39 : Hvl) (x52 : (Index TmVar)) (wfi : (wfindex k39 x52)) (c12 : (Cutoff TyVar)) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) =>
    (wfTm_Var k40 (tshift_wfindex_TmVar c12 k39 k40 ins x52 wfi))) (fun (k39 : Hvl) (t1 : Tm) (wf : (wfTm k39 t1)) IHt94 (t2 : Tm) (wf0 : (wfTm k39 t2)) IHt95 (c12 : (Cutoff TyVar)) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) =>
    (wfTm_App k40 (IHt94 c12 k40 (weaken_shifthvl_TyVar H0 ins)) (IHt95 c12 k40 (weaken_shifthvl_TyVar H0 ins)))) (fun (k39 : Hvl) (T : Ty) (wf : (wfTy k39 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k39) t)) IHt94 (c12 : (Cutoff TyVar)) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) =>
    (wfTm_Lam k40 (tshift_wfTy k39 T wf c12 k40 (weaken_shifthvl_TyVar H0 ins)) (IHt94 c12 (HS TmVar k40) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k39 : Hvl) (p : Pat) (wf : (wfPat k39 p)) (t1 : Tm) (wf0 : (wfTm k39 t1)) IHt94 (t2 : Tm) (wf1 : (wfTm (appendHvl k39 (bind p)) t2)) IHt95 (c12 : (Cutoff TyVar)) (k40 : Hvl) (ins : (shifthvl_TyVar c12 k39 k40)) =>
    (wfTm_Let k40 (tshift_wfPat k39 p wf c12 k40 (weaken_shifthvl_TyVar H0 ins)) (IHt94 c12 k40 (weaken_shifthvl_TyVar H0 ins)) (IHt95 (weakenCutoffTyVar c12 (bind p)) (appendHvl k40 (bind p)) (weaken_shifthvl_TyVar (bind p) ins))))).
End ShiftWellFormed.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : infra.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : shift.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : shift_wf.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : wf.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : infra.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : shift.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : shift_wf.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : infra.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift_wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : infra.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k39 : Hvl) : (forall (d26 : (Trace TmVar)) (k40 : Hvl) (k41 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k39 X0 (HS TmVar k39) k39)
    | substhvl_TmVar_there_TmVar
        {d26 : (Trace TmVar)}
        {k40 : Hvl} {k41 : Hvl} :
        (substhvl_TmVar k39 d26 k40 k41) -> (substhvl_TmVar k39 (XS TmVar d26) (HS TmVar k40) (HS TmVar k41))
    | substhvl_TmVar_there_TyVar
        {d26 : (Trace TmVar)}
        {k40 : Hvl} {k41 : Hvl} :
        (substhvl_TmVar k39 d26 k40 k41) -> (substhvl_TmVar k39 (XS TyVar d26) (HS TyVar k40) (HS TyVar k41)).
  Inductive substhvl_TyVar (k39 : Hvl) : (forall (d26 : (Trace TyVar)) (k40 : Hvl) (k41 : Hvl) ,
    Prop) :=
    | substhvl_TyVar_here :
        (substhvl_TyVar k39 X0 (HS TyVar k39) k39)
    | substhvl_TyVar_there_TmVar
        {d26 : (Trace TyVar)}
        {k40 : Hvl} {k41 : Hvl} :
        (substhvl_TyVar k39 d26 k40 k41) -> (substhvl_TyVar k39 (XS TmVar d26) (HS TmVar k40) (HS TmVar k41))
    | substhvl_TyVar_there_TyVar
        {d26 : (Trace TyVar)}
        {k40 : Hvl} {k41 : Hvl} :
        (substhvl_TyVar k39 d26 k40 k41) -> (substhvl_TyVar k39 (XS TyVar d26) (HS TyVar k40) (HS TyVar k41)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k40 : Hvl} (k39 : Hvl) {d26 : (Trace TmVar)} {k41 : Hvl} {k42 : Hvl} ,
      (substhvl_TmVar k40 d26 k41 k42) -> (substhvl_TmVar k40 (weakenTrace d26 k39) (appendHvl k41 k39) (appendHvl k42 k39))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TyVar  :
    (forall {k40 : Hvl} (k39 : Hvl) {d26 : (Trace TyVar)} {k41 : Hvl} {k42 : Hvl} ,
      (substhvl_TyVar k40 d26 k41 k42) -> (substhvl_TyVar k40 (weakenTrace d26 k39) (appendHvl k41 k39) (appendHvl k42 k39))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k43 : Hvl} {s12 : Tm} (wft : (wfTm k43 s12)) :
    (forall {d26 : (Trace TmVar)} {k44 : Hvl} {k45 : Hvl} ,
      (substhvl_TmVar k43 d26 k44 k45) -> (forall {x52 : (Index TmVar)} ,
        (wfindex k44 x52) -> (wfTm k45 (subst_TmVar_Index d26 s12 x52)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TyVar_wfindex_TyVar {k43 : Hvl} {S26 : Ty} (wft : (wfTy k43 S26)) :
    (forall {d26 : (Trace TyVar)} {k44 : Hvl} {k45 : Hvl} ,
      (substhvl_TyVar k43 d26 k44 k45) -> (forall {y17 : (Index TyVar)} ,
        (wfindex k44 y17) -> (wfTy k45 (subst_TyVar_Index d26 S26 y17)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TmVar_wfindex_TyVar {k43 : Hvl} :
    (forall {d26 : (Trace TmVar)} {k44 : Hvl} {k45 : Hvl} ,
      (substhvl_TmVar k43 d26 k44 k45) -> (forall {y17 : (Index TyVar)} ,
        (wfindex k44 y17) -> (wfindex k45 y17))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TmVar {k43 : Hvl} :
    (forall {d26 : (Trace TyVar)} {k44 : Hvl} {k45 : Hvl} ,
      (substhvl_TyVar k43 d26 k44 k45) -> (forall {x52 : (Index TmVar)} ,
        (wfindex k44 x52) -> (wfindex k45 x52))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_TmVar_wfTy {k43 : Hvl} : (forall (k44 : Hvl) ,
    (forall (S26 : Ty) (wf0 : (wfTy k44 S26)) ,
      (forall {d26 : (Trace TmVar)} {k45 : Hvl} ,
        (substhvl_TmVar k43 d26 k44 k45) -> (wfTy k45 S26)))) := (fun (k44 : Hvl) =>
    (ind_wfTy k44 (fun (S26 : Ty) (wf0 : (wfTy k44 S26)) =>
      (forall {d26 : (Trace TmVar)} {k45 : Hvl} ,
        (substhvl_TmVar k43 d26 k44 k45) -> (wfTy k45 S26))) (fun {y17 : (Index TyVar)} (wfi : (wfindex k44 y17)) {d26 : (Trace TmVar)} {k45 : Hvl} (del : (substhvl_TmVar k43 d26 k44 k45)) =>
      (wfTy_TVar k45 (substhvl_TmVar_wfindex_TyVar del wfi))) (fun (T1 : Ty) (wf0 : (wfTy k44 T1)) IHT0 (T2 : Ty) (wf1 : (wfTy k44 T2)) IHT3 {d26 : (Trace TmVar)} {k45 : Hvl} (del : (substhvl_TmVar k43 d26 k44 k45)) =>
      (wfTy_TArr k45 (IHT0 (weakenTrace d26 H0) k45 (weaken_substhvl_TmVar H0 del)) (IHT3 (weakenTrace d26 H0) k45 (weaken_substhvl_TmVar H0 del)))))).
  Definition substhvl_TyVar_wfTy {k43 : Hvl} {S26 : Ty} (wf : (wfTy k43 S26)) : (forall (k44 : Hvl) ,
    (forall (S27 : Ty) (wf0 : (wfTy k44 S27)) ,
      (forall {d26 : (Trace TyVar)} {k45 : Hvl} ,
        (substhvl_TyVar k43 d26 k44 k45) -> (wfTy k45 (subst_TyVar_Ty d26 S26 S27))))) := (fun (k44 : Hvl) =>
    (ind_wfTy k44 (fun (S27 : Ty) (wf0 : (wfTy k44 S27)) =>
      (forall {d26 : (Trace TyVar)} {k45 : Hvl} ,
        (substhvl_TyVar k43 d26 k44 k45) -> (wfTy k45 (subst_TyVar_Ty d26 S26 S27)))) (fun {y17 : (Index TyVar)} (wfi : (wfindex k44 y17)) {d26 : (Trace TyVar)} {k45 : Hvl} (del : (substhvl_TyVar k43 d26 k44 k45)) =>
      (substhvl_TyVar_wfindex_TyVar wf del wfi)) (fun (T1 : Ty) (wf0 : (wfTy k44 T1)) IHT0 (T2 : Ty) (wf1 : (wfTy k44 T2)) IHT3 {d26 : (Trace TyVar)} {k45 : Hvl} (del : (substhvl_TyVar k43 d26 k44 k45)) =>
      (wfTy_TArr k45 (IHT0 (weakenTrace d26 H0) k45 (weaken_substhvl_TyVar H0 del)) (IHT3 (weakenTrace d26 H0) k45 (weaken_substhvl_TyVar H0 del)))))).
  Definition substhvl_TmVar_wfPat {k43 : Hvl} : (forall (k44 : Hvl) ,
    (forall (p19 : Pat) (wf0 : (wfPat k44 p19)) ,
      (forall {d26 : (Trace TmVar)} {k45 : Hvl} ,
        (substhvl_TmVar k43 d26 k44 k45) -> (wfPat k45 p19)))) := (fun (k44 : Hvl) =>
    (ind_wfPat k44 (fun (p19 : Pat) (wf0 : (wfPat k44 p19)) =>
      (forall {d26 : (Trace TmVar)} {k45 : Hvl} ,
        (substhvl_TmVar k43 d26 k44 k45) -> (wfPat k45 p19))) (fun {d26 : (Trace TmVar)} {k45 : Hvl} (del : (substhvl_TmVar k43 d26 k44 k45)) =>
      (wfPat_PVar k45)) (fun (p1 : Pat) (wf0 : (wfPat k44 p1)) IHp1 (p2 : Pat) (wf1 : (wfPat k44 p2)) IHp2 {d26 : (Trace TmVar)} {k45 : Hvl} (del : (substhvl_TmVar k43 d26 k44 k45)) =>
      (wfPat_PProp k45 (IHp1 (weakenTrace d26 H0) k45 (weaken_substhvl_TmVar H0 del)) (IHp2 (weakenTrace d26 H0) k45 (weaken_substhvl_TmVar H0 del)))))).
  Definition substhvl_TyVar_wfPat {k43 : Hvl} : (forall (k44 : Hvl) ,
    (forall (p19 : Pat) (wf0 : (wfPat k44 p19)) ,
      (forall {d26 : (Trace TyVar)} {k45 : Hvl} ,
        (substhvl_TyVar k43 d26 k44 k45) -> (wfPat k45 p19)))) := (fun (k44 : Hvl) =>
    (ind_wfPat k44 (fun (p19 : Pat) (wf0 : (wfPat k44 p19)) =>
      (forall {d26 : (Trace TyVar)} {k45 : Hvl} ,
        (substhvl_TyVar k43 d26 k44 k45) -> (wfPat k45 p19))) (fun {d26 : (Trace TyVar)} {k45 : Hvl} (del : (substhvl_TyVar k43 d26 k44 k45)) =>
      (wfPat_PVar k45)) (fun (p1 : Pat) (wf0 : (wfPat k44 p1)) IHp1 (p2 : Pat) (wf1 : (wfPat k44 p2)) IHp2 {d26 : (Trace TyVar)} {k45 : Hvl} (del : (substhvl_TyVar k43 d26 k44 k45)) =>
      (wfPat_PProp k45 (IHp1 (weakenTrace d26 H0) k45 (weaken_substhvl_TyVar H0 del)) (IHp2 (weakenTrace d26 H0) k45 (weaken_substhvl_TyVar H0 del)))))).
  Definition substhvl_TmVar_wfTm {k43 : Hvl} {s12 : Tm} (wf : (wfTm k43 s12)) : (forall (k44 : Hvl) ,
    (forall (s13 : Tm) (wf0 : (wfTm k44 s13)) ,
      (forall {d26 : (Trace TmVar)} {k45 : Hvl} ,
        (substhvl_TmVar k43 d26 k44 k45) -> (wfTm k45 (subst_TmVar_Tm d26 s12 s13))))) := (ind_wfTm (fun (k44 : Hvl) (s13 : Tm) (wf0 : (wfTm k44 s13)) =>
    (forall {d26 : (Trace TmVar)} {k45 : Hvl} ,
      (substhvl_TmVar k43 d26 k44 k45) -> (wfTm k45 (subst_TmVar_Tm d26 s12 s13)))) (fun (k44 : Hvl) {x52 : (Index TmVar)} (wfi : (wfindex k44 x52)) {d26 : (Trace TmVar)} {k45 : Hvl} (del : (substhvl_TmVar k43 d26 k44 k45)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k44 : Hvl) (t1 : Tm) (wf0 : (wfTm k44 t1)) IHt94 (t2 : Tm) (wf1 : (wfTm k44 t2)) IHt95 {d26 : (Trace TmVar)} {k45 : Hvl} (del : (substhvl_TmVar k43 d26 k44 k45)) =>
    (wfTm_App k45 (IHt94 (weakenTrace d26 H0) k45 (weaken_substhvl_TmVar H0 del)) (IHt95 (weakenTrace d26 H0) k45 (weaken_substhvl_TmVar H0 del)))) (fun (k44 : Hvl) (T : Ty) (wf0 : (wfTy k44 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k44) t)) IHt94 {d26 : (Trace TmVar)} {k45 : Hvl} (del : (substhvl_TmVar k43 d26 k44 k45)) =>
    (wfTm_Lam k45 (substhvl_TmVar_wfTy k44 T wf0 (weaken_substhvl_TmVar H0 del)) (IHt94 (weakenTrace d26 (HS TmVar H0)) (HS TmVar k45) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k44 : Hvl) (p : Pat) (wf0 : (wfPat k44 p)) (t1 : Tm) (wf1 : (wfTm k44 t1)) IHt94 (t2 : Tm) (wf2 : (wfTm (appendHvl k44 (bind p)) t2)) IHt95 {d26 : (Trace TmVar)} {k45 : Hvl} (del : (substhvl_TmVar k43 d26 k44 k45)) =>
    (wfTm_Let k45 (substhvl_TmVar_wfPat k44 p wf0 (weaken_substhvl_TmVar H0 del)) (IHt94 (weakenTrace d26 H0) k45 (weaken_substhvl_TmVar H0 del)) (IHt95 (weakenTrace d26 (bind p)) (appendHvl k45 (bind p)) (weaken_substhvl_TmVar (bind p) del))))).
  Definition substhvl_TyVar_wfTm {k43 : Hvl} {S26 : Ty} (wf : (wfTy k43 S26)) : (forall (k44 : Hvl) ,
    (forall (s12 : Tm) (wf0 : (wfTm k44 s12)) ,
      (forall {d26 : (Trace TyVar)} {k45 : Hvl} ,
        (substhvl_TyVar k43 d26 k44 k45) -> (wfTm k45 (subst_TyVar_Tm d26 S26 s12))))) := (ind_wfTm (fun (k44 : Hvl) (s12 : Tm) (wf0 : (wfTm k44 s12)) =>
    (forall {d26 : (Trace TyVar)} {k45 : Hvl} ,
      (substhvl_TyVar k43 d26 k44 k45) -> (wfTm k45 (subst_TyVar_Tm d26 S26 s12)))) (fun (k44 : Hvl) {x52 : (Index TmVar)} (wfi : (wfindex k44 x52)) {d26 : (Trace TyVar)} {k45 : Hvl} (del : (substhvl_TyVar k43 d26 k44 k45)) =>
    (wfTm_Var k45 (substhvl_TyVar_wfindex_TmVar del wfi))) (fun (k44 : Hvl) (t1 : Tm) (wf0 : (wfTm k44 t1)) IHt94 (t2 : Tm) (wf1 : (wfTm k44 t2)) IHt95 {d26 : (Trace TyVar)} {k45 : Hvl} (del : (substhvl_TyVar k43 d26 k44 k45)) =>
    (wfTm_App k45 (IHt94 (weakenTrace d26 H0) k45 (weaken_substhvl_TyVar H0 del)) (IHt95 (weakenTrace d26 H0) k45 (weaken_substhvl_TyVar H0 del)))) (fun (k44 : Hvl) (T : Ty) (wf0 : (wfTy k44 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k44) t)) IHt94 {d26 : (Trace TyVar)} {k45 : Hvl} (del : (substhvl_TyVar k43 d26 k44 k45)) =>
    (wfTm_Lam k45 (substhvl_TyVar_wfTy wf k44 T wf0 (weaken_substhvl_TyVar H0 del)) (IHt94 (weakenTrace d26 (HS TmVar H0)) (HS TmVar k45) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k44 : Hvl) (p : Pat) (wf0 : (wfPat k44 p)) (t1 : Tm) (wf1 : (wfTm k44 t1)) IHt94 (t2 : Tm) (wf2 : (wfTm (appendHvl k44 (bind p)) t2)) IHt95 {d26 : (Trace TyVar)} {k45 : Hvl} (del : (substhvl_TyVar k43 d26 k44 k45)) =>
    (wfTm_Let k45 (substhvl_TyVar_wfPat k44 p wf0 (weaken_substhvl_TyVar H0 del)) (IHt94 (weakenTrace d26 H0) k45 (weaken_substhvl_TyVar H0 del)) (IHt95 (weakenTrace d26 (bind p)) (appendHvl k45 (bind p)) (weaken_substhvl_TyVar (bind p) del))))).
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : wf.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : infra.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst_wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : wf.
Fixpoint subhvl_TmVar (k43 : Hvl) {struct k43} : Prop :=
  match k43 with
    | (H0) => True
    | (HS a k43) => match a with
      | (TmVar) => (subhvl_TmVar k43)
      | _ => False
    end
  end.
Lemma subhvl_TmVar_append  :
  (forall (k43 : Hvl) (k44 : Hvl) ,
    (subhvl_TmVar k43) -> (subhvl_TmVar k44) -> (subhvl_TmVar (appendHvl k43 k44))).
Proof.
  needleGenericSubHvlAppend .
Qed.
 Hint Resolve subhvl_TmVar_append : infra.
 Hint Resolve subhvl_TmVar_append : wf.
Lemma wfPat_strengthen_subhvl_TmVar  :
  (forall (k40 : Hvl) (k39 : Hvl) (p18 : Pat) ,
    (subhvl_TmVar k40) -> (wfPat (appendHvl k39 k40) (weakenPat p18 k40)) -> (wfPat k39 p18)).
Proof.
  needleGenericWellformedStrengthen .
Qed.
Lemma wfTy_strengthen_subhvl_TmVar  :
  (forall (k42 : Hvl) (k41 : Hvl) (S25 : Ty) ,
    (subhvl_TmVar k42) -> (wfTy (appendHvl k41 k42) (weakenTy S25 k42)) -> (wfTy k41 S25)).
Proof.
  needleGenericWellformedStrengthen .
Qed.
 Hint Extern 2 ((wfPat _ _)) => match goal with
  | H : (wfPat (appendHvl _ _) (weakenPat _ _)) |- _ => apply (wfPat_strengthen_subhvl_TmVar) in H
end : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H0 : (wfTy (appendHvl _ _) (weakenTy _ _)) |- _ => apply (wfTy_strengthen_subhvl_TmVar) in H0
end : infra wf.
Section Context.
  Inductive Ctx : Type :=
    | empty 
    | evar (G : Ctx) (T : Ty)
    | etvar (G : Ctx).
  Fixpoint appendCtx (G : Ctx) (G0 : Ctx) : Ctx :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendCtx G G1) T)
      | (etvar G1) => (etvar (appendCtx G G1))
    end.
  Fixpoint domainCtx (G : Ctx) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainCtx G0))
      | (etvar G0) => (HS TyVar (domainCtx G0))
    end.
  Lemma appendCtx_assoc  :
    (forall (G : Ctx) (G0 : Ctx) (G1 : Ctx) ,
      ((appendCtx G (appendCtx G0 G1)) =
      (appendCtx (appendCtx G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainCtx_appendCtx  :
    (forall (G : Ctx) (G0 : Ctx) ,
      ((domainCtx (appendCtx G G0)) =
      (appendHvl (domainCtx G) (domainCtx G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shiftCtx (c12 : (Cutoff TmVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shiftCtx c12 G0) T)
      | (etvar G0) => (etvar (shiftCtx c12 G0))
    end.
  Fixpoint tshiftCtx (c12 : (Cutoff TyVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (tshiftCtx c12 G0) (tshiftTy (weakenCutoffTyVar c12 (domainCtx G0)) T))
      | (etvar G0) => (etvar (tshiftCtx c12 G0))
    end.
  Fixpoint subst_TmVar_Ctx (d26 : (Trace TmVar)) (s12 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TmVar_Ctx d26 s12 G0) T)
      | (etvar G0) => (etvar (subst_TmVar_Ctx d26 s12 G0))
    end.
  Fixpoint subst_TyVar_Ctx (d26 : (Trace TyVar)) (S26 : Ty) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TyVar_Ctx d26 S26 G0) (subst_TyVar_Ty (weakenTrace d26 (domainCtx G0)) S26 T))
      | (etvar G0) => (etvar (subst_TyVar_Ctx d26 S26 G0))
    end.
  Lemma domainCtx_shiftCtx  :
    (forall (c12 : (Cutoff TmVar)) (G : Ctx) ,
      ((domainCtx (shiftCtx c12 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_tshiftCtx  :
    (forall (c12 : (Cutoff TyVar)) (G : Ctx) ,
      ((domainCtx (tshiftCtx c12 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_subst_TmVar_Ctx  :
    (forall (d26 : (Trace TmVar)) (s12 : Tm) (G : Ctx) ,
      ((domainCtx (subst_TmVar_Ctx d26 s12 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainCtx_subst_TyVar_Ctx  :
    (forall (d26 : (Trace TyVar)) (S26 : Ty) (G : Ctx) ,
      ((domainCtx (subst_TyVar_Ctx d26 S26 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainCtx_appendCtx : interaction.
 Hint Rewrite domainCtx_appendCtx : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shiftCtx_appendCtx  :
      (forall (c12 : (Cutoff TmVar)) (G : Ctx) (G0 : Ctx) ,
        ((shiftCtx c12 (appendCtx G G0)) =
        (appendCtx (shiftCtx c12 G) (shiftCtx (weakenCutoffTmVar c12 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma tshiftCtx_appendCtx  :
      (forall (c12 : (Cutoff TyVar)) (G : Ctx) (G0 : Ctx) ,
        ((tshiftCtx c12 (appendCtx G G0)) =
        (appendCtx (tshiftCtx c12 G) (tshiftCtx (weakenCutoffTyVar c12 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma subst_TmVar_Ctx_appendCtx  :
      (forall (d26 : (Trace TmVar)) (s12 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((subst_TmVar_Ctx d26 s12 (appendCtx G G0)) =
        (appendCtx (subst_TmVar_Ctx d26 s12 G) (subst_TmVar_Ctx (weakenTrace d26 (domainCtx G)) s12 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma subst_TyVar_Ctx_appendCtx  :
      (forall (d26 : (Trace TyVar)) (S26 : Ty) (G : Ctx) (G0 : Ctx) ,
        ((subst_TyVar_Ctx d26 S26 (appendCtx G G0)) =
        (appendCtx (subst_TyVar_Ctx d26 S26 G) (subst_TyVar_Ctx (weakenTrace d26 (domainCtx G)) S26 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Ctx -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Ctx}
          {T32 : Ty} :
          (wfTy (domainCtx G) T32) -> (lookup_evar (evar G T32) I0 T32)
      | lookup_evar_there_evar
          {G : Ctx} {x52 : (Index TmVar)}
          {T32 : Ty} {T33 : Ty} :
          (lookup_evar G x52 T32) -> (lookup_evar (evar G T33) (IS x52) T32)
      | lookup_evar_there_etvar
          {G : Ctx} {x52 : (Index TmVar)}
          {T32 : Ty} :
          (lookup_evar G x52 T32) -> (lookup_evar (etvar G) x52 (tshiftTy C0 T32)).
    Inductive lookup_etvar : Ctx -> (Index TyVar) -> Prop :=
      | lookup_etvar_here {G : Ctx}
          : (lookup_etvar (etvar G) I0)
      | lookup_etvar_there_evar
          {G : Ctx} {y17 : (Index TyVar)}
          {T32 : Ty} :
          (lookup_etvar G y17) -> (lookup_etvar (evar G T32) y17)
      | lookup_etvar_there_etvar
          {G : Ctx} {y17 : (Index TyVar)}
          :
          (lookup_etvar G y17) -> (lookup_etvar (etvar G) (IS y17)).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Ctx) (S26 : Ty) (S27 : Ty) ,
        (lookup_evar (evar G S26) I0 S27) -> (S26 =
        S27)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Ctx} {x52 : (Index TmVar)} ,
        (forall {S26 : Ty} ,
          (lookup_evar G x52 S26) -> (forall {S27 : Ty} ,
            (lookup_evar G x52 S27) -> (S26 =
            S27)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Ctx} {x52 : (Index TmVar)} {S26 : Ty} ,
        (lookup_evar G x52 S26) -> (wfTy (domainCtx G) S26)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Ctx} (G0 : Ctx) {x52 : (Index TmVar)} {T32 : Ty} ,
        (lookup_evar G x52 T32) -> (lookup_evar (appendCtx G G0) (weakenIndexTmVar x52 (domainCtx G0)) (weakenTy T32 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_etvar  :
      (forall {G : Ctx} (G0 : Ctx) {y17 : (Index TyVar)} ,
        (lookup_etvar G y17) -> (lookup_etvar (appendCtx G G0) (weakenIndexTyVar y17 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Ctx} {x52 : (Index TmVar)} {S26 : Ty} ,
        (lookup_evar G x52 S26) -> (wfindex (domainCtx G) x52)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_etvar_wfindex  :
      (forall {G : Ctx} {y17 : (Index TyVar)} ,
        (lookup_etvar G y17) -> (wfindex (domainCtx G) y17)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Ctx -> Ctx -> Prop :=
    | shift_evar_here {G : Ctx}
        {T32 : Ty} :
        (shift_evar C0 G (evar G T32))
    | shiftevar_there_evar
        {c12 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c12 G G0) -> (shift_evar (CS c12) (evar G T) (evar G0 T))
    | shiftevar_there_etvar
        {c12 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} :
        (shift_evar c12 G G0) -> (shift_evar c12 (etvar G) (etvar G0)).
  Lemma weaken_shift_evar  :
    (forall (G : Ctx) {c12 : (Cutoff TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evar c12 G0 G1) -> (shift_evar (weakenCutoffTmVar c12 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_etvar : (Cutoff TyVar) -> Ctx -> Ctx -> Prop :=
    | shift_etvar_here {G : Ctx} :
        (shift_etvar C0 G (etvar G))
    | shiftetvar_there_evar
        {c12 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_etvar c12 G G0) -> (shift_etvar c12 (evar G T) (evar G0 (tshiftTy c12 T)))
    | shiftetvar_there_etvar
        {c12 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} :
        (shift_etvar c12 G G0) -> (shift_etvar (CS c12) (etvar G) (etvar G0)).
  Lemma weaken_shift_etvar  :
    (forall (G : Ctx) {c12 : (Cutoff TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_etvar c12 G0 G1) -> (shift_etvar (weakenCutoffTyVar c12 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (tshiftCtx c12 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c12 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evar c12 G G0) -> (shifthvl_TmVar c12 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_etvar_shifthvl_TyVar  :
    (forall {c12 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_etvar c12 G G0) -> (shifthvl_TyVar c12 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Ctx) (T32 : Ty) (s12 : Tm) : (Trace TmVar) -> Ctx -> Ctx -> Prop :=
    | subst_evar_here :
        (subst_evar G T32 s12 X0 (evar G T32) G)
    | subst_evar_there_evar
        {d26 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evar G T32 s12 d26 G0 G1) -> (subst_evar G T32 s12 (XS TmVar d26) (evar G0 T) (evar G1 T))
    | subst_evar_there_etvar
        {d26 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} :
        (subst_evar G T32 s12 d26 G0 G1) -> (subst_evar G T32 s12 (XS TyVar d26) (etvar G0) (etvar G1)).
  Lemma weaken_subst_evar {G : Ctx} {T32 : Ty} {s12 : Tm} :
    (forall (G0 : Ctx) {d26 : (Trace TmVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evar G T32 s12 d26 G1 G2) -> (subst_evar G T32 s12 (weakenTrace d26 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_etvar (G : Ctx) (S26 : Ty) : (Trace TyVar) -> Ctx -> Ctx -> Prop :=
    | subst_etvar_here :
        (subst_etvar G S26 X0 (etvar G) G)
    | subst_etvar_there_evar
        {d26 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_etvar G S26 d26 G0 G1) -> (subst_etvar G S26 (XS TmVar d26) (evar G0 T) (evar G1 (subst_TyVar_Ty d26 S26 T)))
    | subst_etvar_there_etvar
        {d26 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} :
        (subst_etvar G S26 d26 G0 G1) -> (subst_etvar G S26 (XS TyVar d26) (etvar G0) (etvar G1)).
  Lemma weaken_subst_etvar {G : Ctx} {S26 : Ty} :
    (forall (G0 : Ctx) {d26 : (Trace TyVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_etvar G S26 d26 G1 G2) -> (subst_etvar G S26 (weakenTrace d26 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (subst_TyVar_Ctx d26 S26 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Ctx} {s12 : Tm} {T32 : Ty} :
    (forall {d26 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evar G T32 s12 d26 G0 G1) -> (substhvl_TmVar (domainCtx G) d26 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_etvar_substhvl_TyVar {G : Ctx} {S26 : Ty} :
    (forall {d26 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_etvar G S26 d26 G0 G1) -> (substhvl_TyVar (domainCtx G) d26 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainCtx_tshiftCtx : interaction.
 Hint Rewrite domainCtx_tshiftCtx : env_domain_shift.
 Hint Rewrite domainCtx_subst_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_subst_TyVar_Ctx : env_domain_subst.
 Hint Rewrite tshiftCtx_appendCtx : interaction.
 Hint Rewrite tshiftCtx_appendCtx : env_shift_append.
 Hint Rewrite subst_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite subst_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Constructors lookup_evar lookup_etvar : infra.
 Hint Constructors lookup_evar lookup_etvar : lookup.
 Hint Constructors lookup_evar lookup_etvar : shift.
 Hint Constructors lookup_evar lookup_etvar : subst.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : infra.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : lookup.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : shift.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Ctx} (G0 : Ctx) {T32 : Ty} (wf : (wfTy (domainCtx G) T32)) ,
    (lookup_evar (appendCtx (evar G T32) G0) (weakenIndexTmVar I0 (domainCtx G0)) (weakenTy T32 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_etvar_here  :
  (forall {G : Ctx} (G0 : Ctx) ,
    (lookup_etvar (appendCtx (etvar G) G0) (weakenIndexTyVar I0 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfPat wfTm wfTy : infra.
 Hint Constructors wfPat wfTm wfTy : wf.
 Hint Extern 10 ((wfPat _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (TVar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TArr _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfPat _ _)) => match goal with
  | H : (wfPat _ (PVar)) |- _ => inversion H; subst; clear H
  | H : (wfPat _ (PProp _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Lam _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Let _ _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf : infra.
 Hint Resolve lookup_evar_wf : wf.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : infra.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : wf.
 Hint Constructors shift_evar shift_etvar : infra.
 Hint Constructors shift_evar shift_etvar : shift.
 Hint Constructors shift_evar shift_etvar : subst.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : infra.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : wf.
 Hint Constructors subst_evar subst_etvar : infra.
 Hint Constructors subst_evar subst_etvar : subst.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : infra.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c12 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c12 G G0)) {x52 : (Index TmVar)} {T32 : Ty} ,
    (lookup_evar G x52 T32) -> (lookup_evar G0 (shiftIndex c12 x52) T32)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_etvar  :
  (forall {c12 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c12 G G0)) {y17 : (Index TyVar)} ,
    (lookup_etvar G y17) -> (lookup_etvar G0 y17)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_evar  :
  (forall {c12 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c12 G G0)) {x52 : (Index TmVar)} {T32 : Ty} ,
    (lookup_evar G x52 T32) -> (lookup_evar G0 x52 (tshiftTy c12 T32))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_etvar  :
  (forall {c12 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c12 G G0)) {y17 : (Index TyVar)} ,
    (lookup_etvar G y17) -> (lookup_etvar G0 (tshiftIndex c12 y17))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : infra.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : lookup.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : shift.
Lemma subst_evar_lookup_etvar {G : Ctx} {T33 : Ty} {s12 : Tm} :
  (forall {d26 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evar G T33 s12 d26 G0 G1)) {y17 : (Index TyVar)} ,
    (lookup_etvar G0 y17) -> (lookup_etvar G1 y17)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_etvar_lookup_evar {G : Ctx} {S26 : Ty} (wf : (wfTy (domainCtx G) S26)) :
  (forall {d26 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_etvar G S26 d26 G0 G1)) {x52 : (Index TmVar)} {T33 : Ty} ,
    (lookup_evar G0 x52 T33) -> (lookup_evar G1 x52 (subst_TyVar_Ty d26 S26 T33))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : infra.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : lookup.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : subst.
Fixpoint size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (TVar y) => 1
    | (TArr T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
  end.
Fixpoint size_Pat (p : Pat) {struct p} : nat :=
  match p with
    | (PVar) => 1
    | (PProp p1 p2) => (plus 1 (plus (size_Pat p1) (size_Pat p2)))
  end.
Fixpoint size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (App t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
    | (Lam T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (Let p t0 t3) => (plus 1 (plus (size_Pat p) (plus (size_Tm t0) (size_Tm t3))))
  end.
Lemma tshift_size_Ty  :
  (forall (S25 : Ty) (c9 : (Cutoff TyVar)) ,
    ((size_Ty (tshiftTy c9 S25)) =
    (size_Ty S25))).
Proof.
  apply_mutual_ind ind_Ty.
  - intros y15 c9.
  reflexivity.
  - intros T28 IHT28 T29 IHT29.
  intros c10; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT28 c10)).
  + exact ((IHT29 c10)).
Qed.

Lemma shift_size_Tm  :
  (forall (s12 : Tm) (c10 : (Cutoff TmVar)) ,
    ((size_Tm (shiftTm c10 s12)) =
    (size_Tm s12))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros x48 c10.
  reflexivity.
  - intros t84 IHt84 t85 IHt85.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt84 c11)).
  + exact ((IHt85 c11)).
  - intros T30 t86 IHt86.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt86 (CS c11))).
  - intros p16 t87 IHt87 t88 IHt88.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt87 c11)).
  + exact ((IHt88 (weakenCutoffTmVar c11 (bind p16)))).
Qed.

Lemma tshift_size_Tm  :
  (forall (s12 : Tm) (c11 : (Cutoff TyVar)) ,
    ((size_Tm (tshiftTm c11 s12)) =
    (size_Tm s12))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros y16 c11.
  reflexivity.
  - intros t89 IHt89 t90 IHt90.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt89 c12)).
  + exact ((IHt90 c12)).
  - intros T31 t91 IHt91.
  pose proof ((tshift_size_Ty T31)) as IHT31.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT31 c12)).
  + exact ((IHt91 c12)).
  - intros p17 t92 IHt92 t93 IHt93.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt92 c12)).
  + exact ((IHt93 (weakenCutoffTyVar c12 (bind p17)))).
Qed.

 Hint Rewrite shift_size_Tm tshift_size_Tm tshift_size_Ty : interaction.
 Hint Rewrite shift_size_Tm tshift_size_Tm tshift_size_Ty : shift_size.
Lemma weaken_size_Pat  :
  (forall (k39 : Hvl) (p18 : Pat) ,
    ((size_Pat (weakenPat p18 k39)) =
    (size_Pat p18))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k39 : Hvl) (s12 : Tm) ,
    ((size_Tm (weakenTm s12 k39)) =
    (size_Tm s12))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k39 : Hvl) (S25 : Ty) ,
    ((size_Ty (weakenTy S25 k39)) =
    (size_Ty S25))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Pat weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Pat weaken_size_Tm weaken_size_Ty : weaken_size.