Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar 
    | TyVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
        | _ => (weakenIndexTmVar x k)
      end
    end.
  Fixpoint weakenIndexTyVar (X : (Index TyVar)) (k : Hvl) {struct k} : (Index TyVar) :=
    match k with
      | (H0) => X
      | (HS a k) => match a with
        | (TyVar) => (IS (weakenIndexTyVar X k))
        | _ => (weakenIndexTyVar X k)
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTyVar_append  :
    (forall (X : (Index TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTyVar (weakenIndexTyVar X k) k0) =
      (weakenIndexTyVar X (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Ty : Set :=
    | TVar (X : (Index TyVar))
    | TArr (T1 : Ty) (T2 : Ty)
    | TAll (T : Ty)
    | TProd (T1 : Ty) (T2 : Ty).
  Scheme ind_Ty := Induction for Ty Sort Prop.
  
  Inductive Pat : Set :=
    | PVar (T : Ty)
    | PTVar 
    | PProd (p1 : Pat) (p2 : Pat).
  Scheme ind_Pat := Induction for Pat Sort Prop.
  
  Inductive Tm : Set :=
    | Var (x : (Index TmVar))
    | Abs (T : Ty) (t : Tm)
    | App (t1 : Tm) (t2 : Tm)
    | TAbs (t : Tm)
    | TApp (t : Tm) (T : Ty)
    | Prod (t1 : Tm) (t2 : Tm)
    | Case (t1 : Tm) (p : Pat)
        (t2 : Tm).
  Scheme ind_Tm := Induction for Tm Sort Prop.
End Terms.

Section Functions.
  Fixpoint bind (p : Pat) {struct p} : Hvl :=
    match p with
      | (PVar T) => (appendHvl (HS TmVar H0) H0)
      | (PTVar) => (appendHvl (HS TyVar H0) H0)
      | (PProd p1 p2) => (appendHvl (bind p1) (appendHvl (bind p2) H0))
    end.
  Scheme ind_bind_Pat := Induction for Pat Sort Prop.
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
        | _ => (weakenCutoffTmVar c k)
      end
    end.
  Fixpoint weakenCutoffTyVar (c : (Cutoff TyVar)) (k : Hvl) {struct k} : (Cutoff TyVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TyVar) => (CS (weakenCutoffTyVar c k))
        | _ => (weakenCutoffTyVar c k)
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTyVar_append  :
    (forall (c : (Cutoff TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTyVar (weakenCutoffTyVar c k) k0) =
      (weakenCutoffTyVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shiftIndex (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shiftIndex c x))
      end
    end.
  Fixpoint tshiftIndex (c : (Cutoff TyVar)) (X : (Index TyVar)) {struct c} : (Index TyVar) :=
    match c with
      | (C0) => (IS X)
      | (CS c) => match X with
        | (I0) => I0
        | (IS X) => (IS (tshiftIndex c X))
      end
    end.
  Fixpoint tshiftTy (c : (Cutoff TyVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar (tshiftIndex c X))
      | (TArr T1 T2) => (TArr (tshiftTy (weakenCutoffTyVar c H0) T1) (tshiftTy (weakenCutoffTyVar c H0) T2))
      | (TAll T) => (TAll (tshiftTy (weakenCutoffTyVar c (appendHvl (HS TyVar H0) H0)) T))
      | (TProd T0 T3) => (TProd (tshiftTy (weakenCutoffTyVar c H0) T0) (tshiftTy (weakenCutoffTyVar c H0) T3))
    end.
  Fixpoint tshiftPat (c : (Cutoff TyVar)) (p : Pat) {struct p} : Pat :=
    match p with
      | (PVar T) => (PVar (tshiftTy (weakenCutoffTyVar c H0) T))
      | (PTVar) => (PTVar)
      | (PProd p1 p2) => (PProd (tshiftPat (weakenCutoffTyVar c H0) p1) (tshiftPat (weakenCutoffTyVar c H0) p2))
    end.
  Fixpoint shiftTm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shiftIndex c x))
      | (Abs T t) => (Abs T (shiftTm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shiftTm (weakenCutoffTmVar c H0) t1) (shiftTm (weakenCutoffTmVar c H0) t2))
      | (TAbs t0) => (TAbs (shiftTm (weakenCutoffTmVar c (appendHvl (HS TyVar H0) H0)) t0))
      | (TApp t3 T0) => (TApp (shiftTm (weakenCutoffTmVar c H0) t3) T0)
      | (Prod t4 t5) => (Prod (shiftTm (weakenCutoffTmVar c H0) t4) (shiftTm (weakenCutoffTmVar c H0) t5))
      | (Case t6 p t7) => (Case (shiftTm (weakenCutoffTmVar c H0) t6) p (shiftTm (weakenCutoffTmVar c (appendHvl (bind p) H0)) t7))
    end.
  Fixpoint tshiftTm (c : (Cutoff TyVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (tshiftTy (weakenCutoffTyVar c H0) T) (tshiftTm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (tshiftTm (weakenCutoffTyVar c H0) t1) (tshiftTm (weakenCutoffTyVar c H0) t2))
      | (TAbs t0) => (TAbs (tshiftTm (weakenCutoffTyVar c (appendHvl (HS TyVar H0) H0)) t0))
      | (TApp t3 T0) => (TApp (tshiftTm (weakenCutoffTyVar c H0) t3) (tshiftTy (weakenCutoffTyVar c H0) T0))
      | (Prod t4 t5) => (Prod (tshiftTm (weakenCutoffTyVar c H0) t4) (tshiftTm (weakenCutoffTyVar c H0) t5))
      | (Case t6 p t7) => (Case (tshiftTm (weakenCutoffTyVar c H0) t6) (tshiftPat (weakenCutoffTyVar c H0) p) (tshiftTm (weakenCutoffTyVar c (appendHvl (bind p) H0)) t7))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (weakenTy S0 k)
      | (HS TyVar k) => (tshiftTy C0 (weakenTy S0 k))
    end.
  Fixpoint weakenPat (p : Pat) (k : Hvl) {struct k} : Pat :=
    match k with
      | (H0) => p
      | (HS TmVar k) => (weakenPat p k)
      | (HS TyVar k) => (tshiftPat C0 (weakenPat p k))
    end.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shiftTm C0 (weakenTm s k))
      | (HS TyVar k) => (tshiftTm C0 (weakenTm s k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint substIndex (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shiftTm C0 (substIndex d s x))
      end
      | (XS TyVar d) => (tshiftTm C0 (substIndex d s x))
    end.
  Fixpoint tsubstIndex (d : (Trace TyVar)) (S0 : Ty) (X : (Index TyVar)) {struct d} : Ty :=
    match d with
      | (X0) => match X with
        | (I0) => S0
        | (IS X) => (TVar X)
      end
      | (XS TmVar d) => (tsubstIndex d S0 X)
      | (XS TyVar d) => match X with
        | (I0) => (TVar I0)
        | (IS X) => (tshiftTy C0 (tsubstIndex d S0 X))
      end
    end.
  Fixpoint tsubstTy (d : (Trace TyVar)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (TVar X) => (tsubstIndex d S0 X)
      | (TArr T1 T2) => (TArr (tsubstTy d S0 T1) (tsubstTy d S0 T2))
      | (TAll T) => (TAll (tsubstTy (XS TyVar d) S0 T))
      | (TProd T0 T3) => (TProd (tsubstTy d S0 T0) (tsubstTy d S0 T3))
    end.
  Fixpoint tsubstPat (d : (Trace TyVar)) (S0 : Ty) (p : Pat) {struct p} : Pat :=
    match p with
      | (PVar T) => (PVar (tsubstTy d S0 T))
      | (PTVar) => (PTVar)
      | (PProd p1 p2) => (PProd (tsubstPat d S0 p1) (tsubstPat d S0 p2))
    end.
  Fixpoint substTm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (substIndex d s x)
      | (Abs T t) => (Abs T (substTm (XS TmVar d) s t))
      | (App t1 t2) => (App (substTm d s t1) (substTm d s t2))
      | (TAbs t0) => (TAbs (substTm (XS TyVar d) s t0))
      | (TApp t3 T0) => (TApp (substTm d s t3) T0)
      | (Prod t4 t5) => (Prod (substTm d s t4) (substTm d s t5))
      | (Case t6 p t7) => (Case (substTm d s t6) p (substTm (weakenTrace d (bind p)) s t7))
    end.
  Fixpoint tsubstTm (d : (Trace TyVar)) (S0 : Ty) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (tsubstTy d S0 T) (tsubstTm (XS TmVar d) S0 t))
      | (App t1 t2) => (App (tsubstTm d S0 t1) (tsubstTm d S0 t2))
      | (TAbs t0) => (TAbs (tsubstTm (XS TyVar d) S0 t0))
      | (TApp t3 T0) => (TApp (tsubstTm d S0 t3) (tsubstTy d S0 T0))
      | (Prod t4 t5) => (Prod (tsubstTm d S0 t4) (tsubstTm d S0 t5))
      | (Case t6 p t7) => (Case (tsubstTm d S0 t6) (tsubstPat d S0 p) (tsubstTm (weakenTrace d (bind p)) S0 t7))
    end.
End Subst.

Global Hint Resolve (f_equal (tshiftPat C0)) : cong_shift0.
Global Hint Resolve (f_equal (shiftTm C0)) : cong_shift0.
Global Hint Resolve (f_equal (tshiftTm C0)) : cong_shift0.
Global Hint Resolve (f_equal (tshiftTy C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Lemma stability_tshift_bind  :
  (forall (p : Pat) ,
    (forall (c : (Cutoff TyVar)) ,
      ((bind (tshiftPat c p)) =
      (bind p)))).
Proof.
  apply_mutual_ind (ind_bind_Pat); simpl; intros; congruence .
Qed.
 Hint Rewrite stability_tshift_bind : interaction.
 Hint Rewrite stability_tshift_bind : stability_shift.
Lemma stability_weaken_bind  :
  (forall (k : Hvl) (p0 : Pat) ,
    ((bind (weakenPat p0 k)) =
    (bind p0))).
Proof.
  needleGenericStabilityWeaken .
Qed.
 Hint Rewrite stability_weaken_bind : interaction.
 Hint Rewrite stability_weaken_bind : stability_weaken.
Lemma stability_tsubst_bind  :
  (forall (p0 : Pat) ,
    (forall (d : (Trace TyVar)) (S0 : Ty) ,
      ((bind (tsubstPat d S0 p0)) =
      (bind p0)))).
Proof.
  apply_mutual_ind (ind_bind_Pat); simpl; intros; congruence .
Qed.
 Hint Rewrite stability_tsubst_bind : interaction.
 Hint Rewrite stability_tsubst_bind : stability_subst.
Section SubstShiftReflection.
  Lemma substIndex0_shiftIndex0_reflection_ind (s : Tm) :
    (forall (k4 : Hvl) (x10 : (Index TmVar)) ,
      ((substIndex (weakenTrace X0 k4) s (shiftIndex (weakenCutoffTmVar C0 k4) x10)) =
      (Var x10))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma tsubstIndex0_tshiftIndex0_reflection_ind (S1 : Ty) :
    (forall (k4 : Hvl) (X9 : (Index TyVar)) ,
      ((tsubstIndex (weakenTrace X0 k4) S1 (tshiftIndex (weakenCutoffTyVar C0 k4) X9)) =
      (TVar X9))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition tsubst0_tshift0_Ty_reflection_ind := (ind_Ty (fun (S3 : Ty) =>
    (forall (k7 : Hvl) (S4 : Ty) ,
      ((tsubstTy (weakenTrace X0 k7) S4 (tshiftTy (weakenCutoffTyVar C0 k7) S3)) =
      S3))) (fun (X16 : (Index TyVar)) =>
    (fun (k7 : Hvl) (S3 : Ty) =>
      (tsubstIndex0_tshiftIndex0_reflection_ind S3 k7 X16))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 TArr (IHT5 k7 S3) (IHT6 k7 S3))) (fun (T : Ty) IHT5 (k7 : Hvl) (S3 : Ty) =>
    (f_equal TAll (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k7 (HS TyVar H0)) eq_refl eq_refl) (IHT5 (HS TyVar k7) S3)))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 TProd (IHT5 k7 S3) (IHT6 k7 S3)))).
  Definition tsubst0_tshift0_Pat_reflection_ind := (ind_Pat (fun (p11 : Pat) =>
    (forall (k7 : Hvl) (S3 : Ty) ,
      ((tsubstPat (weakenTrace X0 k7) S3 (tshiftPat (weakenCutoffTyVar C0 k7) p11)) =
      p11))) (fun (T : Ty) (k7 : Hvl) (S3 : Ty) =>
    (f_equal PVar (let IHT5 := (tsubst0_tshift0_Ty_reflection_ind T) in
    (IHT5 k7 S3)))) (fun (k7 : Hvl) (S3 : Ty) =>
    eq_refl) (fun (p1 : Pat) IHp0 (p2 : Pat) IHp3 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 PProd (IHp0 k7 S3) (IHp3 k7 S3)))).
  Definition subst0_shift0_Tm_reflection_ind := (ind_Tm (fun (s0 : Tm) =>
    (forall (k7 : Hvl) (s1 : Tm) ,
      ((substTm (weakenTrace X0 k7) s1 (shiftTm (weakenCutoffTmVar C0 k7) s0)) =
      s0))) (fun (x16 : (Index TmVar)) =>
    (fun (k7 : Hvl) (s0 : Tm) =>
      (substIndex0_shiftIndex0_reflection_ind s0 k7 x16))) (fun (T : Ty) (t : Tm) IHt53 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 Abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt53 (HS TmVar k7) s0)))) (fun (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 App (IHt53 k7 s0) (IHt54 k7 s0))) (fun (t : Tm) IHt53 (k7 : Hvl) (s0 : Tm) =>
    (f_equal TAbs (eq_trans (f_equal3 substTm (weakenTrace_append TmVar X0 k7 (HS TyVar H0)) eq_refl eq_refl) (IHt53 (HS TyVar k7) s0)))) (fun (t : Tm) IHt53 (T : Ty) (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 TApp (IHt53 k7 s0) eq_refl)) (fun (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 Prod (IHt53 k7 s0) (IHt54 k7 s0))) (fun (t1 : Tm) IHt53 (p : Pat) (t2 : Tm) IHt54 (k7 : Hvl) (s0 : Tm) =>
    (f_equal3 Case (IHt53 k7 s0) eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar X0 k7 (bind p)) eq_refl (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k7 (bind p)) eq_refl)) (IHt54 (appendHvl k7 (bind p)) s0))))).
  Definition tsubst0_tshift0_Tm_reflection_ind := (ind_Tm (fun (s0 : Tm) =>
    (forall (k7 : Hvl) (S3 : Ty) ,
      ((tsubstTm (weakenTrace X0 k7) S3 (tshiftTm (weakenCutoffTyVar C0 k7) s0)) =
      s0))) (fun (x16 : (Index TmVar)) =>
    (fun (k7 : Hvl) (S3 : Ty) =>
      eq_refl)) (fun (T : Ty) (t : Tm) IHt53 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 Abs (let IHT5 := (tsubst0_tshift0_Ty_reflection_ind T) in
    (IHT5 k7 S3)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt53 (HS TmVar k7) S3)))) (fun (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 App (IHt53 k7 S3) (IHt54 k7 S3))) (fun (t : Tm) IHt53 (k7 : Hvl) (S3 : Ty) =>
    (f_equal TAbs (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k7 (HS TyVar H0)) eq_refl eq_refl) (IHt53 (HS TyVar k7) S3)))) (fun (t : Tm) IHt53 (T : Ty) (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 TApp (IHt53 k7 S3) (let IHT5 := (tsubst0_tshift0_Ty_reflection_ind T) in
    (IHT5 k7 S3)))) (fun (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 Prod (IHt53 k7 S3) (IHt54 k7 S3))) (fun (t1 : Tm) IHt53 (p : Pat) (t2 : Tm) IHt54 (k7 : Hvl) (S3 : Ty) =>
    (f_equal3 Case (IHt53 k7 S3) (let IHp := (tsubst0_tshift0_Pat_reflection_ind p) in
    (IHp k7 S3)) (eq_trans (f_equal3 tsubstTm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_tshift_bind _ _) (eq_refl H0))) (weakenTrace_append TyVar X0 k7 (bind p))) eq_refl (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k7 (bind p)) eq_refl)) (IHt54 (appendHvl k7 (bind p)) S3))))).
  Definition tsubstTy0_tshiftTy0_reflection (S4 : Ty) : (forall (S3 : Ty) ,
    ((tsubstTy X0 S3 (tshiftTy C0 S4)) =
    S4)) := (tsubst0_tshift0_Ty_reflection_ind S4 H0).
  Definition tsubstPat0_tshiftPat0_reflection (p11 : Pat) : (forall (S3 : Ty) ,
    ((tsubstPat X0 S3 (tshiftPat C0 p11)) =
    p11)) := (tsubst0_tshift0_Pat_reflection_ind p11 H0).
  Definition substTm0_shiftTm0_reflection (s1 : Tm) : (forall (s0 : Tm) ,
    ((substTm X0 s0 (shiftTm C0 s1)) =
    s1)) := (subst0_shift0_Tm_reflection_ind s1 H0).
  Definition tsubstTm0_tshiftTm0_reflection (s0 : Tm) : (forall (S3 : Ty) ,
    ((tsubstTm X0 S3 (tshiftTm C0 s0)) =
    s0)) := (tsubst0_tshift0_Tm_reflection_ind s0 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shiftIndex_shiftIndex0_comm_ind  :
      (forall (k : Hvl) (c0 : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shiftIndex (weakenCutoffTmVar (CS c0) k) (shiftIndex (weakenCutoffTmVar C0 k) x)) =
        (shiftIndex (weakenCutoffTmVar C0 k) (shiftIndex (weakenCutoffTmVar c0 k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma tshiftIndex_tshiftIndex0_comm_ind  :
      (forall (k : Hvl) (c0 : (Cutoff TyVar)) (X : (Index TyVar)) ,
        ((tshiftIndex (weakenCutoffTyVar (CS c0) k) (tshiftIndex (weakenCutoffTyVar C0 k) X)) =
        (tshiftIndex (weakenCutoffTyVar C0 k) (tshiftIndex (weakenCutoffTyVar c0 k) X)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition tshift_tshift0_Ty_comm_ind := (ind_Ty (fun (S1 : Ty) =>
      (forall (k4 : Hvl) (c5 : (Cutoff TyVar)) ,
        ((tshiftTy (weakenCutoffTyVar (CS c5) k4) (tshiftTy (weakenCutoffTyVar C0 k4) S1)) =
        (tshiftTy (weakenCutoffTyVar C0 k4) (tshiftTy (weakenCutoffTyVar c5 k4) S1))))) (fun (X9 : (Index TyVar)) =>
      (fun (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
        (f_equal TVar (tshiftIndex_tshiftIndex0_comm_ind k4 c5 X9)))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 TArr (IHT5 k4 c5) (IHT6 k4 c5))) (fun (T : Ty) IHT5 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal TAll (IHT5 (HS TyVar k4) c5))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 TProd (IHT5 k4 c5) (IHT6 k4 c5)))).
    Definition tshift_tshift0_Pat_comm_ind := (ind_Pat (fun (p7 : Pat) =>
      (forall (k4 : Hvl) (c5 : (Cutoff TyVar)) ,
        ((tshiftPat (weakenCutoffTyVar (CS c5) k4) (tshiftPat (weakenCutoffTyVar C0 k4) p7)) =
        (tshiftPat (weakenCutoffTyVar C0 k4) (tshiftPat (weakenCutoffTyVar c5 k4) p7))))) (fun (T : Ty) (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal PVar (let IHT5 := (tshift_tshift0_Ty_comm_ind T) in
      (IHT5 k4 c5)))) (fun (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      eq_refl) (fun (p1 : Pat) IHp0 (p2 : Pat) IHp3 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 PProd (IHp0 k4 c5) (IHp3 k4 c5)))).
    Definition shift_shift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c5 : (Cutoff TmVar)) ,
        ((shiftTm (weakenCutoffTmVar (CS c5) k4) (shiftTm (weakenCutoffTmVar C0 k4) s)) =
        (shiftTm (weakenCutoffTmVar C0 k4) (shiftTm (weakenCutoffTmVar c5 k4) s))))) (fun (x10 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
        (f_equal Var (shiftIndex_shiftIndex0_comm_ind k4 c5 x10)))) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt35 (HS TmVar k4) c5))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt35 k4 c5) (IHt36 k4 c5))) (fun (t : Tm) IHt35 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal TAbs (IHt35 (HS TyVar k4) c5))) (fun (t : Tm) IHt35 (T : Ty) (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHt35 k4 c5) eq_refl)) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal2 Prod (IHt35 k4 c5) (IHt36 k4 c5))) (fun (t1 : Tm) IHt35 (p : Pat) (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal3 Case (IHt35 k4 c5) eq_refl (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append (CS c5) k4 (bind p)) (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k4 (bind p)) eq_refl)) (eq_trans (IHt36 (appendHvl k4 (bind p)) c5) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k4 (bind p))) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append c5 k4 (bind p))) eq_refl))))))).
    Definition shift_tshift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c5 : (Cutoff TmVar)) ,
        ((shiftTm (weakenCutoffTmVar c5 k4) (tshiftTm (weakenCutoffTyVar C0 k4) s)) =
        (tshiftTm (weakenCutoffTyVar C0 k4) (shiftTm (weakenCutoffTmVar c5 k4) s))))) (fun (x10 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt35 (HS TmVar k4) c5))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt35 k4 c5) (IHt36 k4 c5))) (fun (t : Tm) IHt35 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal TAbs (IHt35 (HS TyVar k4) c5))) (fun (t : Tm) IHt35 (T : Ty) (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHt35 k4 c5) eq_refl)) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal2 Prod (IHt35 k4 c5) (IHt36 k4 c5))) (fun (t1 : Tm) IHt35 (p : Pat) (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TmVar)) =>
      (f_equal3 Case (IHt35 k4 c5) eq_refl (eq_trans (f_equal2 shiftTm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_tshift_bind _ _) (eq_refl H0))) (weakenCutoffTmVar_append c5 k4 (bind p))) (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k4 (bind p)) eq_refl)) (eq_trans (IHt36 (appendHvl k4 (bind p)) c5) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k4 (bind p))) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append c5 k4 (bind p))) eq_refl))))))).
    Definition tshift_shift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c5 : (Cutoff TyVar)) ,
        ((tshiftTm (weakenCutoffTyVar c5 k4) (shiftTm (weakenCutoffTmVar C0 k4) s)) =
        (shiftTm (weakenCutoffTmVar C0 k4) (tshiftTm (weakenCutoffTyVar c5 k4) s))))) (fun (x10 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 Abs eq_refl (IHt35 (HS TmVar k4) c5))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt35 k4 c5) (IHt36 k4 c5))) (fun (t : Tm) IHt35 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal TAbs (IHt35 (HS TyVar k4) c5))) (fun (t : Tm) IHt35 (T : Ty) (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHt35 k4 c5) eq_refl)) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 Prod (IHt35 k4 c5) (IHt36 k4 c5))) (fun (t1 : Tm) IHt35 (p : Pat) (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal3 Case (IHt35 k4 c5) eq_refl (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append c5 k4 (bind p)) (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k4 (bind p)) eq_refl)) (eq_trans (IHt36 (appendHvl k4 (bind p)) c5) (f_equal2 shiftTm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k4 (bind p))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_tshift_bind _ _)) (eq_refl H0)))) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append c5 k4 (bind p))) eq_refl))))))).
    Definition tshift_tshift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c5 : (Cutoff TyVar)) ,
        ((tshiftTm (weakenCutoffTyVar (CS c5) k4) (tshiftTm (weakenCutoffTyVar C0 k4) s)) =
        (tshiftTm (weakenCutoffTyVar C0 k4) (tshiftTm (weakenCutoffTyVar c5 k4) s))))) (fun (x10 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 Abs (let IHT5 := (tshift_tshift0_Ty_comm_ind T) in
      (IHT5 k4 c5)) (IHt35 (HS TmVar k4) c5))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt35 k4 c5) (IHt36 k4 c5))) (fun (t : Tm) IHt35 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal TAbs (IHt35 (HS TyVar k4) c5))) (fun (t : Tm) IHt35 (T : Ty) (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHt35 k4 c5) (let IHT5 := (tshift_tshift0_Ty_comm_ind T) in
      (IHT5 k4 c5)))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal2 Prod (IHt35 k4 c5) (IHt36 k4 c5))) (fun (t1 : Tm) IHt35 (p : Pat) (t2 : Tm) IHt36 (k4 : Hvl) (c5 : (Cutoff TyVar)) =>
      (f_equal3 Case (IHt35 k4 c5) (let IHp := (tshift_tshift0_Pat_comm_ind p) in
      (IHp k4 c5)) (eq_trans (f_equal2 tshiftTm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_tshift_bind _ _) (eq_refl H0))) (weakenCutoffTyVar_append (CS c5) k4 (bind p))) (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k4 (bind p)) eq_refl)) (eq_trans (IHt36 (appendHvl k4 (bind p)) c5) (f_equal2 tshiftTm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k4 (bind p))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_tshift_bind _ _)) (eq_refl H0)))) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append c5 k4 (bind p))) eq_refl))))))).
  End ShiftCommInd.
  Section ShiftComm.
    Definition tshift_tshift0_Ty_comm (S1 : Ty) : (forall (c5 : (Cutoff TyVar)) ,
      ((tshiftTy (CS c5) (tshiftTy C0 S1)) =
      (tshiftTy C0 (tshiftTy c5 S1)))) := (tshift_tshift0_Ty_comm_ind S1 H0).
    Definition tshift_tshift0_Pat_comm (p7 : Pat) : (forall (c5 : (Cutoff TyVar)) ,
      ((tshiftPat (CS c5) (tshiftPat C0 p7)) =
      (tshiftPat C0 (tshiftPat c5 p7)))) := (tshift_tshift0_Pat_comm_ind p7 H0).
    Definition shift_shift0_Tm_comm (s : Tm) : (forall (c5 : (Cutoff TmVar)) ,
      ((shiftTm (CS c5) (shiftTm C0 s)) =
      (shiftTm C0 (shiftTm c5 s)))) := (shift_shift0_Tm_comm_ind s H0).
    Definition shift_tshift0_Tm_comm (s : Tm) : (forall (c5 : (Cutoff TmVar)) ,
      ((shiftTm c5 (tshiftTm C0 s)) =
      (tshiftTm C0 (shiftTm c5 s)))) := (shift_tshift0_Tm_comm_ind s H0).
    Definition tshift_shift0_Tm_comm (s : Tm) : (forall (c5 : (Cutoff TyVar)) ,
      ((tshiftTm c5 (shiftTm C0 s)) =
      (shiftTm C0 (tshiftTm c5 s)))) := (tshift_shift0_Tm_comm_ind s H0).
    Definition tshift_tshift0_Tm_comm (s : Tm) : (forall (c5 : (Cutoff TyVar)) ,
      ((tshiftTm (CS c5) (tshiftTm C0 s)) =
      (tshiftTm C0 (tshiftTm c5 s)))) := (tshift_tshift0_Tm_comm_ind s H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite tshift_tshift0_Pat_comm shift_shift0_Tm_comm shift_tshift0_Tm_comm tshift_shift0_Tm_comm tshift_tshift0_Tm_comm tshift_tshift0_Ty_comm : interaction.
 Hint Rewrite tshift_tshift0_Pat_comm shift_shift0_Tm_comm shift_tshift0_Tm_comm tshift_shift0_Tm_comm tshift_tshift0_Tm_comm tshift_tshift0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTy_tshiftTy  :
    (forall (k64 : Hvl) (c10 : (Cutoff TyVar)) (S45 : Ty) ,
      ((weakenTy (tshiftTy c10 S45) k64) =
      (tshiftTy (weakenCutoffTyVar c10 k64) (weakenTy S45 k64)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenPat_tshiftPat  :
    (forall (k64 : Hvl) (c10 : (Cutoff TyVar)) (p30 : Pat) ,
      ((weakenPat (tshiftPat c10 p30) k64) =
      (tshiftPat (weakenCutoffTyVar c10 k64) (weakenPat p30 k64)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shiftTm  :
    (forall (k64 : Hvl) (c10 : (Cutoff TmVar)) (s18 : Tm) ,
      ((weakenTm (shiftTm c10 s18) k64) =
      (shiftTm (weakenCutoffTmVar c10 k64) (weakenTm s18 k64)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_tshiftTm  :
    (forall (k64 : Hvl) (c10 : (Cutoff TyVar)) (s18 : Tm) ,
      ((weakenTm (tshiftTm c10 s18) k64) =
      (tshiftTm (weakenCutoffTyVar c10 k64) (weakenTm s18 k64)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shiftTm_substIndex0_comm_ind (c5 : (Cutoff TmVar)) (s0 : Tm) :
      (forall (k7 : Hvl) (x16 : (Index TmVar)) ,
        ((shiftTm (weakenCutoffTmVar c5 k7) (substIndex (weakenTrace X0 k7) s0 x16)) =
        (substIndex (weakenTrace X0 k7) (shiftTm c5 s0) (shiftIndex (weakenCutoffTmVar (CS c5) k7) x16)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tshiftTm_substIndex0_comm_ind (c5 : (Cutoff TyVar)) (s0 : Tm) :
      (forall (k7 : Hvl) (x16 : (Index TmVar)) ,
        ((tshiftTm (weakenCutoffTyVar c5 k7) (substIndex (weakenTrace X0 k7) s0 x16)) =
        (substIndex (weakenTrace X0 k7) (tshiftTm c5 s0) x16))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tshiftTy_tsubstIndex0_comm_ind (c5 : (Cutoff TyVar)) (S3 : Ty) :
      (forall (k7 : Hvl) (X16 : (Index TyVar)) ,
        ((tshiftTy (weakenCutoffTyVar c5 k7) (tsubstIndex (weakenTrace X0 k7) S3 X16)) =
        (tsubstIndex (weakenTrace X0 k7) (tshiftTy c5 S3) (tshiftIndex (weakenCutoffTyVar (CS c5) k7) X16)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition tshift_tsubst0_Ty_comm_ind := (ind_Ty (fun (S6 : Ty) =>
      (forall (k12 : Hvl) (c10 : (Cutoff TyVar)) (S7 : Ty) ,
        ((tshiftTy (weakenCutoffTyVar c10 k12) (tsubstTy (weakenTrace X0 k12) S7 S6)) =
        (tsubstTy (weakenTrace X0 k12) (tshiftTy c10 S7) (tshiftTy (weakenCutoffTyVar (CS c10) k12) S6))))) (fun (X24 : (Index TyVar)) =>
      (fun (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
        (tshiftTy_tsubstIndex0_comm_ind c10 S6 k12 X24))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 TArr (IHT5 k12 c10 S6) (IHT6 k12 c10 S6))) (fun (T : Ty) IHT5 (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal TAll (eq_trans (f_equal2 tshiftTy eq_refl (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHT5 (HS TyVar k12) c10 S6) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 TProd (IHT5 k12 c10 S6) (IHT6 k12 c10 S6)))).
    Definition tshift_tsubst0_Pat_comm_ind := (ind_Pat (fun (p17 : Pat) =>
      (forall (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) ,
        ((tshiftPat (weakenCutoffTyVar c10 k12) (tsubstPat (weakenTrace X0 k12) S6 p17)) =
        (tsubstPat (weakenTrace X0 k12) (tshiftTy c10 S6) (tshiftPat (weakenCutoffTyVar (CS c10) k12) p17))))) (fun (T : Ty) (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal PVar (let IHT5 := (tshift_tsubst0_Ty_comm_ind T) in
      (IHT5 k12 c10 S6)))) (fun (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      eq_refl) (fun (p1 : Pat) IHp0 (p2 : Pat) IHp3 (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 PProd (IHp0 k12 c10 S6) (IHp3 k12 c10 S6)))).
    Definition shift_subst0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c10 : (Cutoff TmVar)) (s3 : Tm) ,
        ((shiftTm (weakenCutoffTmVar c10 k12) (substTm (weakenTrace X0 k12) s3 s2)) =
        (substTm (weakenTrace X0 k12) (shiftTm c10 s3) (shiftTm (weakenCutoffTmVar (CS c10) k12) s2))))) (fun (x29 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c10 : (Cutoff TmVar)) (s2 : Tm) =>
        (shiftTm_substIndex0_comm_ind c10 s2 k12 x29))) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c10 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TmVar k12) c10 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 App (IHt89 k12 c10 s2) (IHt90 k12 c10 s2))) (fun (t : Tm) IHt89 (k12 : Hvl) (c10 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TyVar k12) c10 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt89 (T : Ty) (k12 : Hvl) (c10 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 TApp (IHt89 k12 c10 s2) eq_refl)) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 Prod (IHt89 k12 c10 s2) (IHt90 k12 c10 s2))) (fun (t1 : Tm) IHt89 (p : Pat) (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal3 Case (IHt89 k12 c10 s2) eq_refl (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append c10 k12 (bind p)) (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (bind p)) eq_refl eq_refl)) (eq_trans (IHt90 (appendHvl k12 (bind p)) c10 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (bind p))) eq_refl (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append (CS c10) k12 (bind p))) eq_refl))))))).
    Definition shift_tsubst0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c10 : (Cutoff TmVar)) (S6 : Ty) ,
        ((shiftTm (weakenCutoffTmVar c10 k12) (tsubstTm (weakenTrace X0 k12) S6 s2)) =
        (tsubstTm (weakenTrace X0 k12) S6 (shiftTm (weakenCutoffTmVar c10 k12) s2))))) (fun (x29 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c10 : (Cutoff TmVar)) (S6 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c10 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TmVar k12) c10 S6) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 App (IHt89 k12 c10 S6) (IHt90 k12 c10 S6))) (fun (t : Tm) IHt89 (k12 : Hvl) (c10 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TyVar k12) c10 S6) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt89 (T : Ty) (k12 : Hvl) (c10 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 TApp (IHt89 k12 c10 S6) eq_refl)) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 Prod (IHt89 k12 c10 S6) (IHt90 k12 c10 S6))) (fun (t1 : Tm) IHt89 (p : Pat) (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal3 Case (IHt89 k12 c10 S6) eq_refl (eq_trans (f_equal2 shiftTm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_tsubst_bind _ _ _) (eq_refl H0))) (weakenCutoffTmVar_append c10 k12 (bind p))) (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (bind p)) eq_refl eq_refl)) (eq_trans (IHt90 (appendHvl k12 (bind p)) c10 S6) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (bind p))) eq_refl (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append c10 k12 (bind p))) eq_refl))))))).
    Definition tshift_subst0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c10 : (Cutoff TyVar)) (s3 : Tm) ,
        ((tshiftTm (weakenCutoffTyVar c10 k12) (substTm (weakenTrace X0 k12) s3 s2)) =
        (substTm (weakenTrace X0 k12) (tshiftTm c10 s3) (tshiftTm (weakenCutoffTyVar c10 k12) s2))))) (fun (x29 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c10 : (Cutoff TyVar)) (s2 : Tm) =>
        (tshiftTm_substIndex0_comm_ind c10 s2 k12 x29))) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c10 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TmVar k12) c10 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 App (IHt89 k12 c10 s2) (IHt90 k12 c10 s2))) (fun (t : Tm) IHt89 (k12 : Hvl) (c10 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TyVar k12) c10 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt89 (T : Ty) (k12 : Hvl) (c10 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 TApp (IHt89 k12 c10 s2) eq_refl)) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 Prod (IHt89 k12 c10 s2) (IHt90 k12 c10 s2))) (fun (t1 : Tm) IHt89 (p : Pat) (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal3 Case (IHt89 k12 c10 s2) eq_refl (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append c10 k12 (bind p)) (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (bind p)) eq_refl eq_refl)) (eq_trans (IHt90 (appendHvl k12 (bind p)) c10 s2) (f_equal3 substTm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k12 (bind p))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_tshift_bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append c10 k12 (bind p))) eq_refl))))))).
    Definition tshift_tsubst0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) ,
        ((tshiftTm (weakenCutoffTyVar c10 k12) (tsubstTm (weakenTrace X0 k12) S6 s2)) =
        (tsubstTm (weakenTrace X0 k12) (tshiftTy c10 S6) (tshiftTm (weakenCutoffTyVar (CS c10) k12) s2))))) (fun (x29 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 Abs (let IHT5 := (tshift_tsubst0_Ty_comm_ind T) in
      (IHT5 k12 c10 S6)) (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TmVar k12) c10 S6) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 App (IHt89 k12 c10 S6) (IHt90 k12 c10 S6))) (fun (t : Tm) IHt89 (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TyVar k12) c10 S6) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt89 (T : Ty) (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 TApp (IHt89 k12 c10 S6) (let IHT5 := (tshift_tsubst0_Ty_comm_ind T) in
      (IHT5 k12 c10 S6)))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 Prod (IHt89 k12 c10 S6) (IHt90 k12 c10 S6))) (fun (t1 : Tm) IHt89 (p : Pat) (t2 : Tm) IHt90 (k12 : Hvl) (c10 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal3 Case (IHt89 k12 c10 S6) (let IHp := (tshift_tsubst0_Pat_comm_ind p) in
      (IHp k12 c10 S6)) (eq_trans (f_equal2 tshiftTm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_tsubst_bind _ _ _) (eq_refl H0))) (weakenCutoffTyVar_append c10 k12 (bind p))) (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (bind p)) eq_refl eq_refl)) (eq_trans (IHt90 (appendHvl k12 (bind p)) c10 S6) (f_equal3 tsubstTm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k12 (bind p))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_tshift_bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append (CS c10) k12 (bind p))) eq_refl))))))).
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition tshiftTy_tsubstTy0_comm (S7 : Ty) : (forall (c10 : (Cutoff TyVar)) (S6 : Ty) ,
      ((tshiftTy c10 (tsubstTy X0 S6 S7)) =
      (tsubstTy X0 (tshiftTy c10 S6) (tshiftTy (CS c10) S7)))) := (tshift_tsubst0_Ty_comm_ind S7 H0).
    Definition tshiftPat_tsubstPat0_comm (p17 : Pat) : (forall (c10 : (Cutoff TyVar)) (S6 : Ty) ,
      ((tshiftPat c10 (tsubstPat X0 S6 p17)) =
      (tsubstPat X0 (tshiftTy c10 S6) (tshiftPat (CS c10) p17)))) := (tshift_tsubst0_Pat_comm_ind p17 H0).
    Definition shiftTm_substTm0_comm (s3 : Tm) : (forall (c10 : (Cutoff TmVar)) (s2 : Tm) ,
      ((shiftTm c10 (substTm X0 s2 s3)) =
      (substTm X0 (shiftTm c10 s2) (shiftTm (CS c10) s3)))) := (shift_subst0_Tm_comm_ind s3 H0).
    Definition shiftTm_tsubstTm0_comm (s2 : Tm) : (forall (c10 : (Cutoff TmVar)) (S6 : Ty) ,
      ((shiftTm c10 (tsubstTm X0 S6 s2)) =
      (tsubstTm X0 S6 (shiftTm c10 s2)))) := (shift_tsubst0_Tm_comm_ind s2 H0).
    Definition tshiftTm_substTm0_comm (s3 : Tm) : (forall (c10 : (Cutoff TyVar)) (s2 : Tm) ,
      ((tshiftTm c10 (substTm X0 s2 s3)) =
      (substTm X0 (tshiftTm c10 s2) (tshiftTm c10 s3)))) := (tshift_subst0_Tm_comm_ind s3 H0).
    Definition tshiftTm_tsubstTm0_comm (s2 : Tm) : (forall (c10 : (Cutoff TyVar)) (S6 : Ty) ,
      ((tshiftTm c10 (tsubstTm X0 S6 s2)) =
      (tsubstTm X0 (tshiftTy c10 S6) (tshiftTm (CS c10) s2)))) := (tshift_tsubst0_Tm_comm_ind s2 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma substIndex_shiftTm0_comm_ind (d0 : (Trace TmVar)) (s2 : Tm) :
      (forall (k12 : Hvl) (x29 : (Index TmVar)) ,
        ((substIndex (weakenTrace (XS TmVar d0) k12) s2 (shiftIndex (weakenCutoffTmVar C0 k12) x29)) =
        (shiftTm (weakenCutoffTmVar C0 k12) (substIndex (weakenTrace d0 k12) s2 x29)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma substIndex_tshiftTm0_comm_ind (d0 : (Trace TmVar)) (s2 : Tm) :
      (forall (k12 : Hvl) (x29 : (Index TmVar)) ,
        ((substIndex (weakenTrace (XS TyVar d0) k12) s2 x29) =
        (tshiftTm (weakenCutoffTyVar C0 k12) (substIndex (weakenTrace d0 k12) s2 x29)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tsubstIndex_shiftTy0_comm_ind (d0 : (Trace TyVar)) (S6 : Ty) :
      (forall (k12 : Hvl) (X24 : (Index TyVar)) ,
        ((tsubstIndex (weakenTrace (XS TmVar d0) k12) S6 X24) =
        (tsubstIndex (weakenTrace d0 k12) S6 X24))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tsubstIndex_tshiftTy0_comm_ind (d0 : (Trace TyVar)) (S6 : Ty) :
      (forall (k12 : Hvl) (X24 : (Index TyVar)) ,
        ((tsubstIndex (weakenTrace (XS TyVar d0) k12) S6 (tshiftIndex (weakenCutoffTyVar C0 k12) X24)) =
        (tshiftTy (weakenCutoffTyVar C0 k12) (tsubstIndex (weakenTrace d0 k12) S6 X24)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition tsubst_tshift0_Ty_comm_ind := (ind_Ty (fun (S27 : Ty) =>
      (forall (k47 : Hvl) (d35 : (Trace TyVar)) (S28 : Ty) ,
        ((tsubstTy (weakenTrace (XS TyVar d35) k47) S28 (tshiftTy (weakenCutoffTyVar C0 k47) S27)) =
        (tshiftTy (weakenCutoffTyVar C0 k47) (tsubstTy (weakenTrace d35 k47) S28 S27))))) (fun (X32 : (Index TyVar)) =>
      (fun (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
        (tsubstIndex_tshiftTy0_comm_ind d35 S27 k47 X32))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 TArr (IHT5 k47 d35 S27) (IHT6 k47 d35 S27))) (fun (T : Ty) IHT5 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar (XS TyVar d35) k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHT5 (HS TyVar k47) d35 S27) (f_equal2 tshiftTy eq_refl (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar d35 k47 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 TProd (IHT5 k47 d35 S27) (IHT6 k47 d35 S27)))).
    Definition tsubst_tshift0_Pat_comm_ind := (ind_Pat (fun (p23 : Pat) =>
      (forall (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) ,
        ((tsubstPat (weakenTrace (XS TyVar d35) k47) S27 (tshiftPat (weakenCutoffTyVar C0 k47) p23)) =
        (tshiftPat (weakenCutoffTyVar C0 k47) (tsubstPat (weakenTrace d35 k47) S27 p23))))) (fun (T : Ty) (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal PVar (let IHT5 := (tsubst_tshift0_Ty_comm_ind T) in
      (IHT5 k47 d35 S27)))) (fun (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      eq_refl) (fun (p1 : Pat) IHp0 (p2 : Pat) IHp3 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 PProd (IHp0 k47 d35 S27) (IHp3 k47 d35 S27)))).
    Definition subst_shift0_Tm_comm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k47 : Hvl) (d35 : (Trace TmVar)) (s17 : Tm) ,
        ((substTm (weakenTrace (XS TmVar d35) k47) s17 (shiftTm (weakenCutoffTmVar C0 k47) s16)) =
        (shiftTm (weakenCutoffTmVar C0 k47) (substTm (weakenTrace d35 k47) s17 s16))))) (fun (x42 : (Index TmVar)) =>
      (fun (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
        (substIndex_shiftTm0_comm_ind d35 s16 k47 x42))) (fun (T : Ty) (t : Tm) IHt125 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TmVar d35) k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TmVar k47) d35 s16) (f_equal2 shiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d35 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 App (IHt125 k47 d35 s16) (IHt126 k47 d35 s16))) (fun (t : Tm) IHt125 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TmVar d35) k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TyVar k47) d35 s16) (f_equal2 shiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d35 k47 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt125 (T : Ty) (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 TApp (IHt125 k47 d35 s16) eq_refl)) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 Prod (IHt125 k47 d35 s16) (IHt126 k47 d35 s16))) (fun (t1 : Tm) IHt125 (p : Pat) (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal3 Case (IHt125 k47 d35 s16) eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TmVar d35) k47 (bind p)) eq_refl (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k47 (bind p)) eq_refl)) (eq_trans (IHt126 (appendHvl k47 (bind p)) d35 s16) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k47 (bind p))) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d35 k47 (bind p))) eq_refl eq_refl))))))).
    Definition subst_tshift0_Tm_comm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k47 : Hvl) (d35 : (Trace TmVar)) (s17 : Tm) ,
        ((substTm (weakenTrace (XS TyVar d35) k47) s17 (tshiftTm (weakenCutoffTyVar C0 k47) s16)) =
        (tshiftTm (weakenCutoffTyVar C0 k47) (substTm (weakenTrace d35 k47) s17 s16))))) (fun (x42 : (Index TmVar)) =>
      (fun (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
        (substIndex_tshiftTm0_comm_ind d35 s16 k47 x42))) (fun (T : Ty) (t : Tm) IHt125 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TyVar d35) k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TmVar k47) d35 s16) (f_equal2 tshiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d35 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 App (IHt125 k47 d35 s16) (IHt126 k47 d35 s16))) (fun (t : Tm) IHt125 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TyVar d35) k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TyVar k47) d35 s16) (f_equal2 tshiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d35 k47 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt125 (T : Ty) (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 TApp (IHt125 k47 d35 s16) eq_refl)) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 Prod (IHt125 k47 d35 s16) (IHt126 k47 d35 s16))) (fun (t1 : Tm) IHt125 (p : Pat) (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal3 Case (IHt125 k47 d35 s16) eq_refl (eq_trans (f_equal3 substTm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_tshift_bind _ _) (eq_refl H0))) (weakenTrace_append TmVar (XS TyVar d35) k47 (bind p))) eq_refl (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k47 (bind p)) eq_refl)) (eq_trans (IHt126 (appendHvl k47 (bind p)) d35 s16) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k47 (bind p))) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d35 k47 (bind p))) eq_refl eq_refl))))))).
    Definition tsubst_shift0_Tm_comm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) ,
        ((tsubstTm (weakenTrace d35 k47) S27 (shiftTm (weakenCutoffTmVar C0 k47) s16)) =
        (shiftTm (weakenCutoffTmVar C0 k47) (tsubstTm (weakenTrace d35 k47) S27 s16))))) (fun (x42 : (Index TmVar)) =>
      (fun (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt125 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d35 k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TmVar k47) d35 S27) (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d35 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 App (IHt125 k47 d35 S27) (IHt126 k47 d35 S27))) (fun (t : Tm) IHt125 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d35 k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TyVar k47) d35 S27) (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d35 k47 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt125 (T : Ty) (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 TApp (IHt125 k47 d35 S27) eq_refl)) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 Prod (IHt125 k47 d35 S27) (IHt126 k47 d35 S27))) (fun (t1 : Tm) IHt125 (p : Pat) (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal3 Case (IHt125 k47 d35 S27) eq_refl (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d35 k47 (bind p)) eq_refl (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k47 (bind p)) eq_refl)) (eq_trans (IHt126 (appendHvl k47 (bind p)) d35 S27) (f_equal2 shiftTm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k47 (bind p))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_tsubst_bind _ _ _)) (eq_refl H0)))) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d35 k47 (bind p))) eq_refl eq_refl))))))).
    Definition tsubst_tshift0_Tm_comm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) ,
        ((tsubstTm (weakenTrace (XS TyVar d35) k47) S27 (tshiftTm (weakenCutoffTyVar C0 k47) s16)) =
        (tshiftTm (weakenCutoffTyVar C0 k47) (tsubstTm (weakenTrace d35 k47) S27 s16))))) (fun (x42 : (Index TmVar)) =>
      (fun (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt125 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 Abs (let IHT5 := (tsubst_tshift0_Ty_comm_ind T) in
      (IHT5 k47 d35 S27)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TyVar d35) k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TmVar k47) d35 S27) (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d35 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 App (IHt125 k47 d35 S27) (IHt126 k47 d35 S27))) (fun (t : Tm) IHt125 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TyVar d35) k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TyVar k47) d35 S27) (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d35 k47 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt125 (T : Ty) (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 TApp (IHt125 k47 d35 S27) (let IHT5 := (tsubst_tshift0_Ty_comm_ind T) in
      (IHT5 k47 d35 S27)))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal2 Prod (IHt125 k47 d35 S27) (IHt126 k47 d35 S27))) (fun (t1 : Tm) IHt125 (p : Pat) (t2 : Tm) IHt126 (k47 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) =>
      (f_equal3 Case (IHt125 k47 d35 S27) (let IHp := (tsubst_tshift0_Pat_comm_ind p) in
      (IHp k47 d35 S27)) (eq_trans (f_equal3 tsubstTm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_tshift_bind _ _) (eq_refl H0))) (weakenTrace_append TyVar (XS TyVar d35) k47 (bind p))) eq_refl (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k47 (bind p)) eq_refl)) (eq_trans (IHt126 (appendHvl k47 (bind p)) d35 S27) (f_equal2 tshiftTm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k47 (bind p))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_tsubst_bind _ _ _)) (eq_refl H0)))) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d35 k47 (bind p))) eq_refl eq_refl))))))).
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition tsubstTy_tshiftTy0_comm (S42 : Ty) : (forall (d49 : (Trace TyVar)) (S41 : Ty) ,
      ((tsubstTy (XS TyVar d49) S41 (tshiftTy C0 S42)) =
      (tshiftTy C0 (tsubstTy d49 S41 S42)))) := (tsubst_tshift0_Ty_comm_ind S42 H0).
    Definition tsubstPat_tshiftPat0_comm (p26 : Pat) : (forall (d49 : (Trace TyVar)) (S41 : Ty) ,
      ((tsubstPat (XS TyVar d49) S41 (tshiftPat C0 p26)) =
      (tshiftPat C0 (tsubstPat d49 S41 p26)))) := (tsubst_tshift0_Pat_comm_ind p26 H0).
    Definition substTm_shiftTm0_comm (s17 : Tm) : (forall (d49 : (Trace TmVar)) (s16 : Tm) ,
      ((substTm (XS TmVar d49) s16 (shiftTm C0 s17)) =
      (shiftTm C0 (substTm d49 s16 s17)))) := (subst_shift0_Tm_comm_ind s17 H0).
    Definition substTm_tshiftTm0_comm (s17 : Tm) : (forall (d49 : (Trace TmVar)) (s16 : Tm) ,
      ((substTm (XS TyVar d49) s16 (tshiftTm C0 s17)) =
      (tshiftTm C0 (substTm d49 s16 s17)))) := (subst_tshift0_Tm_comm_ind s17 H0).
    Definition tsubstTm_shiftTm0_comm (s16 : Tm) : (forall (d49 : (Trace TyVar)) (S41 : Ty) ,
      ((tsubstTm d49 S41 (shiftTm C0 s16)) =
      (shiftTm C0 (tsubstTm d49 S41 s16)))) := (tsubst_shift0_Tm_comm_ind s16 H0).
    Definition tsubstTm_tshiftTm0_comm (s16 : Tm) : (forall (d49 : (Trace TyVar)) (S41 : Ty) ,
      ((tsubstTm (XS TyVar d49) S41 (tshiftTm C0 s16)) =
      (tshiftTm C0 (tsubstTm d49 S41 s16)))) := (tsubst_tshift0_Tm_comm_ind s16 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    Definition tsubst_TmVar_Ty_ind := (ind_Ty (fun (S41 : Ty) =>
      (forall (k61 : Hvl) (d49 : (Trace TyVar)) (S42 : Ty) ,
        ((tsubstTy (weakenTrace (XS TmVar d49) k61) S42 S41) =
        (tsubstTy (weakenTrace d49 k61) S42 S41)))) (fun (X37 : (Index TyVar)) =>
      (fun (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
        (tsubstIndex_shiftTy0_comm_ind d49 S41 k61 X37))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal2 TArr (IHT5 k61 d49 S41) (IHT6 k61 d49 S41))) (fun (T : Ty) IHT5 (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar (XS TmVar d49) k61 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHT5 (HS TyVar k61) d49 S41) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar d49 k61 (HS TyVar H0))) eq_refl eq_refl))))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal2 TProd (IHT5 k61 d49 S41) (IHT6 k61 d49 S41)))).
    Definition tsubst_TmVar_Pat_ind := (ind_Pat (fun (p26 : Pat) =>
      (forall (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) ,
        ((tsubstPat (weakenTrace (XS TmVar d49) k61) S41 p26) =
        (tsubstPat (weakenTrace d49 k61) S41 p26)))) (fun (T : Ty) (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal PVar (let IHT5 := (tsubst_TmVar_Ty_ind T) in
      (IHT5 k61 d49 S41)))) (fun (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      eq_refl) (fun (p1 : Pat) IHp0 (p2 : Pat) IHp3 (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal2 PProd (IHp0 k61 d49 S41) (IHp3 k61 d49 S41)))).
    Definition tsubst_TmVar_Tm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) ,
        ((tsubstTm (weakenTrace (XS TmVar d49) k61) S41 s16) =
        (tsubstTm (weakenTrace d49 k61) S41 s16)))) (fun (x46 : (Index TmVar)) =>
      (fun (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt134 (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal2 Abs (let IHT5 := (tsubst_TmVar_Ty_ind T) in
      (IHT5 k61 d49 S41)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TmVar d49) k61 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt134 (HS TmVar k61) d49 S41) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d49 k61 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt134 (t2 : Tm) IHt135 (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal2 App (IHt134 k61 d49 S41) (IHt135 k61 d49 S41))) (fun (t : Tm) IHt134 (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TmVar d49) k61 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt134 (HS TyVar k61) d49 S41) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d49 k61 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt134 (T : Ty) (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal2 TApp (IHt134 k61 d49 S41) (let IHT5 := (tsubst_TmVar_Ty_ind T) in
      (IHT5 k61 d49 S41)))) (fun (t1 : Tm) IHt134 (t2 : Tm) IHt135 (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal2 Prod (IHt134 k61 d49 S41) (IHt135 k61 d49 S41))) (fun (t1 : Tm) IHt134 (p : Pat) (t2 : Tm) IHt135 (k61 : Hvl) (d49 : (Trace TyVar)) (S41 : Ty) =>
      (f_equal3 Case (IHt134 k61 d49 S41) (let IHp := (tsubst_TmVar_Pat_ind p) in
      (IHp k61 d49 S41)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TmVar d49) k61 (bind p)) eq_refl eq_refl) (eq_trans (IHt135 (appendHvl k61 (bind p)) d49 S41) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d49 k61 (bind p))) eq_refl eq_refl)))))).
  End SubstSubordInd.
  Section SubstSubord.
    Definition tsubstTy_TmVar (S42 : Ty) : (forall (d49 : (Trace TyVar)) (S41 : Ty) ,
      ((tsubstTy (XS TmVar d49) S41 S42) =
      (tsubstTy d49 S41 S42))) := (tsubst_TmVar_Ty_ind S42 H0).
    Definition tsubstPat_TmVar (p26 : Pat) : (forall (d49 : (Trace TyVar)) (S41 : Ty) ,
      ((tsubstPat (XS TmVar d49) S41 p26) =
      (tsubstPat d49 S41 p26))) := (tsubst_TmVar_Pat_ind p26 H0).
    Definition tsubstTm_TmVar (s16 : Tm) : (forall (d49 : (Trace TyVar)) (S41 : Ty) ,
      ((tsubstTm (XS TmVar d49) S41 s16) =
      (tsubstTm d49 S41 s16))) := (tsubst_TmVar_Tm_ind s16 H0).
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite tsubstPat0_tshiftPat0_reflection substTm0_shiftTm0_reflection tsubstTm0_tshiftTm0_reflection tsubstTy0_tshiftTy0_reflection : interaction.
 Hint Rewrite tsubstPat0_tshiftPat0_reflection substTm0_shiftTm0_reflection tsubstTm0_tshiftTm0_reflection tsubstTy0_tshiftTy0_reflection : reflection.
 Hint Rewrite tsubstPat_tshiftPat0_comm substTm_shiftTm0_comm substTm_tshiftTm0_comm tsubstTm_shiftTm0_comm tsubstTm_tshiftTm0_comm tsubstTy_tshiftTy0_comm : interaction.
 Hint Rewrite tsubstPat_tshiftPat0_comm substTm_shiftTm0_comm substTm_tshiftTm0_comm tsubstTm_shiftTm0_comm tsubstTm_tshiftTm0_comm tsubstTy_tshiftTy0_comm : subst_shift0.
 Hint Rewrite tsubstPat_TmVar tsubstTm_TmVar tsubstTy_TmVar : interaction.
 Hint Rewrite tsubstPat_TmVar tsubstTm_TmVar tsubstTy_TmVar : subst_shift0.
 Hint Rewrite tshiftPat_tsubstPat0_comm shiftTm_substTm0_comm shiftTm_tsubstTm0_comm tshiftTm_substTm0_comm tshiftTm_tsubstTm0_comm tshiftTy_tsubstTy0_comm : interaction.
 Hint Rewrite tshiftPat_tsubstPat0_comm shiftTm_substTm0_comm shiftTm_tsubstTm0_comm tshiftTm_substTm0_comm tshiftTm_tsubstTm0_comm tshiftTy_tsubstTy0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma substTm_substIndex0_commright_ind (d49 : (Trace TmVar)) (s16 : Tm) (s17 : Tm) :
      (forall (k61 : Hvl) (x46 : (Index TmVar)) ,
        ((substTm (weakenTrace d49 k61) s16 (substIndex (weakenTrace X0 k61) s17 x46)) =
        (substTm (weakenTrace X0 k61) (substTm d49 s16 s17) (substIndex (weakenTrace (XS TmVar d49) k61) s16 x46)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma tsubstTm_substIndex0_commright_ind (d49 : (Trace TyVar)) (S41 : Ty) (s16 : Tm) :
      (forall (k61 : Hvl) (x46 : (Index TmVar)) ,
        ((tsubstTm (weakenTrace d49 k61) S41 (substIndex (weakenTrace X0 k61) s16 x46)) =
        (substIndex (weakenTrace X0 k61) (tsubstTm d49 S41 s16) x46))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma tsubstTy_tsubstIndex0_commright_ind (d49 : (Trace TyVar)) (S41 : Ty) (S42 : Ty) :
      (forall (k61 : Hvl) (X37 : (Index TyVar)) ,
        ((tsubstTy (weakenTrace d49 k61) S41 (tsubstIndex (weakenTrace X0 k61) S42 X37)) =
        (tsubstTy (weakenTrace X0 k61) (tsubstTy d49 S41 S42) (tsubstIndex (weakenTrace (XS TyVar d49) k61) S41 X37)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma substTy_tsubstIndex0_commleft_ind (d49 : (Trace TmVar)) (s16 : Tm) (S41 : Ty) :
      (forall (k61 : Hvl) (x46 : (Index TmVar)) ,
        ((substIndex (weakenTrace d49 k61) s16 x46) =
        (tsubstTm (weakenTrace X0 k61) S41 (substIndex (weakenTrace (XS TyVar d49) k61) s16 x46)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition tsubst_tsubst0_Ty_comm_ind := (ind_Ty (fun (S45 : Ty) =>
      (forall (k64 : Hvl) (d52 : (Trace TyVar)) (S46 : Ty) (S47 : Ty) ,
        ((tsubstTy (weakenTrace d52 k64) S46 (tsubstTy (weakenTrace X0 k64) S47 S45)) =
        (tsubstTy (weakenTrace X0 k64) (tsubstTy d52 S46 S47) (tsubstTy (weakenTrace (XS TyVar d52) k64) S46 S45))))) (fun (X43 : (Index TyVar)) =>
      (fun (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
        (tsubstTy_tsubstIndex0_commright_ind d52 S45 S46 k64 X43))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal2 TArr (IHT5 k64 d52 S45 S46) (IHT6 k64 d52 S45 S46))) (fun (T : Ty) IHT5 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar d52 k64 (HS TyVar H0)) eq_refl (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k64 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHT5 (HS TyVar k64) d52 S45 S46) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar X0 k64 (HS TyVar H0))) eq_refl (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar (XS TyVar d52) k64 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal2 TProd (IHT5 k64 d52 S45 S46) (IHT6 k64 d52 S45 S46)))).
    Definition tsubst_tsubst0_Pat_comm_ind := (ind_Pat (fun (p30 : Pat) =>
      (forall (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) ,
        ((tsubstPat (weakenTrace d52 k64) S45 (tsubstPat (weakenTrace X0 k64) S46 p30)) =
        (tsubstPat (weakenTrace X0 k64) (tsubstTy d52 S45 S46) (tsubstPat (weakenTrace (XS TyVar d52) k64) S45 p30))))) (fun (T : Ty) (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal PVar (let IHT5 := (tsubst_tsubst0_Ty_comm_ind T) in
      (IHT5 k64 d52 S45 S46)))) (fun (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      eq_refl) (fun (p1 : Pat) IHp0 (p2 : Pat) IHp3 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal2 PProd (IHp0 k64 d52 S45 S46) (IHp3 k64 d52 S45 S46)))).
    Definition subst_subst0_Tm_comm_ind := (ind_Tm (fun (s18 : Tm) =>
      (forall (k64 : Hvl) (d52 : (Trace TmVar)) (s19 : Tm) (s20 : Tm) ,
        ((substTm (weakenTrace d52 k64) s19 (substTm (weakenTrace X0 k64) s20 s18)) =
        (substTm (weakenTrace X0 k64) (substTm d52 s19 s20) (substTm (weakenTrace (XS TmVar d52) k64) s19 s18))))) (fun (x53 : (Index TmVar)) =>
      (fun (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
        (substTm_substIndex0_commright_ind d52 s18 s19 k64 x53))) (fun (T : Ty) (t : Tm) IHt152 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d52 k64 (HS TmVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k64 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TmVar k64) d52 s18 s19) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k64 (HS TmVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TmVar d52) k64 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal2 App (IHt152 k64 d52 s18 s19) (IHt153 k64 d52 s18 s19))) (fun (t : Tm) IHt152 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d52 k64 (HS TyVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k64 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TyVar k64) d52 s18 s19) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k64 (HS TyVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TmVar d52) k64 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt152 (T : Ty) (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal2 TApp (IHt152 k64 d52 s18 s19) eq_refl)) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal2 Prod (IHt152 k64 d52 s18 s19) (IHt153 k64 d52 s18 s19))) (fun (t1 : Tm) IHt152 (p : Pat) (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal3 Case (IHt152 k64 d52 s18 s19) eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d52 k64 (bind p)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k64 (bind p)) eq_refl eq_refl)) (eq_trans (IHt153 (appendHvl k64 (bind p)) d52 s18 s19) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k64 (bind p))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TmVar d52) k64 (bind p))) eq_refl eq_refl))))))).
    Definition subst_tsubst0_Tm_comm_ind := (ind_Tm (fun (s18 : Tm) =>
      (forall (k64 : Hvl) (d52 : (Trace TmVar)) (s19 : Tm) (S45 : Ty) ,
        ((substTm (weakenTrace d52 k64) s19 (tsubstTm (weakenTrace X0 k64) S45 s18)) =
        (tsubstTm (weakenTrace X0 k64) S45 (substTm (weakenTrace (XS TyVar d52) k64) s19 s18))))) (fun (x53 : (Index TmVar)) =>
      (fun (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (S45 : Ty) =>
        (substTy_tsubstIndex0_commleft_ind d52 s18 S45 k64 x53))) (fun (T : Ty) (t : Tm) IHt152 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (S45 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d52 k64 (HS TmVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k64 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TmVar k64) d52 s18 S45) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k64 (HS TmVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TyVar d52) k64 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (S45 : Ty) =>
      (f_equal2 App (IHt152 k64 d52 s18 S45) (IHt153 k64 d52 s18 S45))) (fun (t : Tm) IHt152 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (S45 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d52 k64 (HS TyVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k64 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TyVar k64) d52 s18 S45) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k64 (HS TyVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TyVar d52) k64 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt152 (T : Ty) (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (S45 : Ty) =>
      (f_equal2 TApp (IHt152 k64 d52 s18 S45) eq_refl)) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (S45 : Ty) =>
      (f_equal2 Prod (IHt152 k64 d52 s18 S45) (IHt153 k64 d52 s18 S45))) (fun (t1 : Tm) IHt152 (p : Pat) (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (S45 : Ty) =>
      (f_equal3 Case (IHt152 k64 d52 s18 S45) eq_refl (eq_trans (f_equal3 substTm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_tsubst_bind _ _ _) (eq_refl H0))) (weakenTrace_append TmVar d52 k64 (bind p))) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k64 (bind p)) eq_refl eq_refl)) (eq_trans (IHt153 (appendHvl k64 (bind p)) d52 s18 S45) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k64 (bind p))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TyVar d52) k64 (bind p))) eq_refl eq_refl))))))).
    Definition tsubst_subst0_Tm_comm_ind := (ind_Tm (fun (s18 : Tm) =>
      (forall (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (s19 : Tm) ,
        ((tsubstTm (weakenTrace d52 k64) S45 (substTm (weakenTrace X0 k64) s19 s18)) =
        (substTm (weakenTrace X0 k64) (tsubstTm d52 S45 s19) (tsubstTm (weakenTrace d52 k64) S45 s18))))) (fun (x53 : (Index TmVar)) =>
      (fun (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (s18 : Tm) =>
        (tsubstTm_substIndex0_commright_ind d52 S45 s18 k64 x53))) (fun (T : Ty) (t : Tm) IHt152 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (s18 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d52 k64 (HS TmVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k64 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TmVar k64) d52 S45 s18) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k64 (HS TmVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d52 k64 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (s18 : Tm) =>
      (f_equal2 App (IHt152 k64 d52 S45 s18) (IHt153 k64 d52 S45 s18))) (fun (t : Tm) IHt152 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (s18 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d52 k64 (HS TyVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k64 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TyVar k64) d52 S45 s18) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k64 (HS TyVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d52 k64 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt152 (T : Ty) (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (s18 : Tm) =>
      (f_equal2 TApp (IHt152 k64 d52 S45 s18) eq_refl)) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (s18 : Tm) =>
      (f_equal2 Prod (IHt152 k64 d52 S45 s18) (IHt153 k64 d52 S45 s18))) (fun (t1 : Tm) IHt152 (p : Pat) (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (s18 : Tm) =>
      (f_equal3 Case (IHt152 k64 d52 S45 s18) eq_refl (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d52 k64 (bind p)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k64 (bind p)) eq_refl eq_refl)) (eq_trans (IHt153 (appendHvl k64 (bind p)) d52 S45 s18) (f_equal3 substTm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k64 (bind p))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_tsubst_bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d52 k64 (bind p))) eq_refl eq_refl))))))).
    Definition tsubst_tsubst0_Tm_comm_ind := (ind_Tm (fun (s18 : Tm) =>
      (forall (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) ,
        ((tsubstTm (weakenTrace d52 k64) S45 (tsubstTm (weakenTrace X0 k64) S46 s18)) =
        (tsubstTm (weakenTrace X0 k64) (tsubstTy d52 S45 S46) (tsubstTm (weakenTrace (XS TyVar d52) k64) S45 s18))))) (fun (x53 : (Index TmVar)) =>
      (fun (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt152 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal2 Abs (let IHT5 := (tsubst_tsubst0_Ty_comm_ind T) in
      (IHT5 k64 d52 S45 S46)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d52 k64 (HS TmVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k64 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TmVar k64) d52 S45 S46) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k64 (HS TmVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TyVar d52) k64 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal2 App (IHt152 k64 d52 S45 S46) (IHt153 k64 d52 S45 S46))) (fun (t : Tm) IHt152 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d52 k64 (HS TyVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k64 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TyVar k64) d52 S45 S46) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k64 (HS TyVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TyVar d52) k64 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt152 (T : Ty) (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal2 TApp (IHt152 k64 d52 S45 S46) (let IHT5 := (tsubst_tsubst0_Ty_comm_ind T) in
      (IHT5 k64 d52 S45 S46)))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal2 Prod (IHt152 k64 d52 S45 S46) (IHt153 k64 d52 S45 S46))) (fun (t1 : Tm) IHt152 (p : Pat) (t2 : Tm) IHt153 (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) =>
      (f_equal3 Case (IHt152 k64 d52 S45 S46) (let IHp := (tsubst_tsubst0_Pat_comm_ind p) in
      (IHp k64 d52 S45 S46)) (eq_trans (f_equal3 tsubstTm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_tsubst_bind _ _ _) (eq_refl H0))) (weakenTrace_append TyVar d52 k64 (bind p))) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k64 (bind p)) eq_refl eq_refl)) (eq_trans (IHt153 (appendHvl k64 (bind p)) d52 S45 S46) (f_equal3 tsubstTm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k64 (bind p))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_tsubst_bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TyVar d52) k64 (bind p))) eq_refl eq_refl))))))).
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition tsubstTy_tsubstTy0_comm (S47 : Ty) : (forall (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) ,
      ((tsubstTy d52 S45 (tsubstTy X0 S46 S47)) =
      (tsubstTy X0 (tsubstTy d52 S45 S46) (tsubstTy (XS TyVar d52) S45 S47)))) := (tsubst_tsubst0_Ty_comm_ind S47 H0).
    Definition tsubstPat_tsubstPat0_comm (p30 : Pat) : (forall (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) ,
      ((tsubstPat d52 S45 (tsubstPat X0 S46 p30)) =
      (tsubstPat X0 (tsubstTy d52 S45 S46) (tsubstPat (XS TyVar d52) S45 p30)))) := (tsubst_tsubst0_Pat_comm_ind p30 H0).
    Definition substTm_substTm0_comm (s20 : Tm) : (forall (d52 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) ,
      ((substTm d52 s18 (substTm X0 s19 s20)) =
      (substTm X0 (substTm d52 s18 s19) (substTm (XS TmVar d52) s18 s20)))) := (subst_subst0_Tm_comm_ind s20 H0).
    Definition substTm_tsubstTm0_comm (s19 : Tm) : (forall (d52 : (Trace TmVar)) (s18 : Tm) (S45 : Ty) ,
      ((substTm d52 s18 (tsubstTm X0 S45 s19)) =
      (tsubstTm X0 S45 (substTm (XS TyVar d52) s18 s19)))) := (subst_tsubst0_Tm_comm_ind s19 H0).
    Definition tsubstTm_substTm0_comm (s19 : Tm) : (forall (d52 : (Trace TyVar)) (S45 : Ty) (s18 : Tm) ,
      ((tsubstTm d52 S45 (substTm X0 s18 s19)) =
      (substTm X0 (tsubstTm d52 S45 s18) (tsubstTm d52 S45 s19)))) := (tsubst_subst0_Tm_comm_ind s19 H0).
    Definition tsubstTm_tsubstTm0_comm (s18 : Tm) : (forall (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) ,
      ((tsubstTm d52 S45 (tsubstTm X0 S46 s18)) =
      (tsubstTm X0 (tsubstTy d52 S45 S46) (tsubstTm (XS TyVar d52) S45 s18)))) := (tsubst_tsubst0_Tm_comm_ind s18 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTy_tsubstTy  :
      (forall (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (S46 : Ty) ,
        ((weakenTy (tsubstTy d52 S45 S46) k64) =
        (tsubstTy (weakenTrace d52 k64) S45 (weakenTy S46 k64)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenPat_tsubstPat  :
      (forall (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (p30 : Pat) ,
        ((weakenPat (tsubstPat d52 S45 p30) k64) =
        (tsubstPat (weakenTrace d52 k64) S45 (weakenPat p30 k64)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_substTm  :
      (forall (k64 : Hvl) (d52 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) ,
        ((weakenTm (substTm d52 s18 s19) k64) =
        (substTm (weakenTrace d52 k64) s18 (weakenTm s19 k64)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_tsubstTm  :
      (forall (k64 : Hvl) (d52 : (Trace TyVar)) (S45 : Ty) (s18 : Tm) ,
        ((weakenTm (tsubstTm d52 S45 s18) k64) =
        (tsubstTm (weakenTrace d52 k64) S45 (weakenTm s18 k64)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTy_append  :
      (forall (S45 : Ty) (k64 : Hvl) (k65 : Hvl) ,
        ((weakenTy (weakenTy S45 k64) k65) =
        (weakenTy S45 (appendHvl k64 k65)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenPat_append  :
      (forall (p30 : Pat) (k64 : Hvl) (k65 : Hvl) ,
        ((weakenPat (weakenPat p30 k64) k65) =
        (weakenPat p30 (appendHvl k64 k65)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTm_append  :
      (forall (s18 : Tm) (k64 : Hvl) (k65 : Hvl) ,
        ((weakenTm (weakenTm s18 k64) k65) =
        (weakenTm s18 (appendHvl k64 k65)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite tsubstPat_tsubstPat0_comm substTm_substTm0_comm tsubstTm_tsubstTm0_comm tsubstTy_tsubstTy0_comm : interaction.
 Hint Rewrite tsubstPat_tsubstPat0_comm substTm_substTm0_comm tsubstTm_tsubstTm0_comm tsubstTy_tsubstTy0_comm : subst_subst0.
 Hint Rewrite weakenPat_tshiftPat weakenTm_shiftTm weakenTm_tshiftTm weakenTy_tshiftTy : weaken_shift.
 Hint Rewrite weakenPat_tsubstPat weakenTm_substTm weakenTm_tsubstTm weakenTy_tsubstTy : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k64 : Hvl) (i : (Index a)) {struct k64} : Prop :=
    match k64 with
      | (H0) => False
      | (HS b k64) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k64 i)
        end
        | (right n) => (wfindex k64 i)
      end
    end.
  Inductive wfTy (k64 : Hvl) : Ty -> Prop :=
    | wfTy_TVar
        {X50 : (Index TyVar)}
        (wfi : (wfindex k64 X50)) :
        (wfTy k64 (TVar X50))
    | wfTy_TArr {T79 : Ty}
        (wf : (wfTy k64 T79)) {T80 : Ty}
        (wf0 : (wfTy k64 T80)) :
        (wfTy k64 (TArr T79 T80))
    | wfTy_TAll {T81 : Ty}
        (wf : (wfTy (HS TyVar k64) T81))
        : (wfTy k64 (TAll T81))
    | wfTy_TProd {T82 : Ty}
        (wf : (wfTy k64 T82)) {T83 : Ty}
        (wf0 : (wfTy k64 T83)) :
        (wfTy k64 (TProd T82 T83)).
  Inductive wfPat (k64 : Hvl) : Pat -> Prop :=
    | wfPat_PVar {T79 : Ty}
        (wf : (wfTy k64 T79)) :
        (wfPat k64 (PVar T79))
    | wfPat_PTVar :
        (wfPat k64 (PTVar))
    | wfPat_PProd {p34 : Pat}
        (wf : (wfPat k64 p34))
        {p35 : Pat}
        (wf0 : (wfPat k64 p35)) :
        (wfPat k64 (PProd p34 p35)).
  Inductive wfTm (k64 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x59 : (Index TmVar)}
        (wfi : (wfindex k64 x59)) :
        (wfTm k64 (Var x59))
    | wfTm_Abs {T79 : Ty}
        (wf : (wfTy k64 T79))
        {t170 : Tm}
        (wf0 : (wfTm (HS TmVar k64) t170))
        : (wfTm k64 (Abs T79 t170))
    | wfTm_App {t171 : Tm}
        (wf : (wfTm k64 t171))
        {t172 : Tm}
        (wf0 : (wfTm k64 t172)) :
        (wfTm k64 (App t171 t172))
    | wfTm_TAbs {t173 : Tm}
        (wf : (wfTm (HS TyVar k64) t173))
        : (wfTm k64 (TAbs t173))
    | wfTm_TApp {t174 : Tm}
        (wf : (wfTm k64 t174))
        {T80 : Ty}
        (wf0 : (wfTy k64 T80)) :
        (wfTm k64 (TApp t174 T80))
    | wfTm_Prod {t175 : Tm}
        (wf : (wfTm k64 t175))
        {t176 : Tm}
        (wf0 : (wfTm k64 t176)) :
        (wfTm k64 (Prod t175 t176))
    | wfTm_Case {t177 : Tm}
        (wf : (wfTm k64 t177))
        {p34 : Pat}
        (wf0 : (wfPat k64 p34))
        {t178 : Tm}
        (wf1 : (wfTm (appendHvl k64 (bind p34)) t178))
        :
        (wfTm k64 (Case t177 p34 t178)).
  Scheme ind_wfTy := Induction for wfTy Sort Prop.
  Scheme ind_wfPat := Induction for wfPat Sort Prop.
  Scheme ind_wfTm := Induction for wfTm Sort Prop.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c13 : (Cutoff TmVar)) (k64 : Hvl) (k65 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k64 : Hvl) :
        (shifthvl_TmVar C0 k64 (HS TmVar k64))
    | shifthvl_TmVar_there_TmVar
        (c13 : (Cutoff TmVar))
        (k64 : Hvl) (k65 : Hvl) :
        (shifthvl_TmVar c13 k64 k65) -> (shifthvl_TmVar (CS c13) (HS TmVar k64) (HS TmVar k65))
    | shifthvl_TmVar_there_TyVar
        (c13 : (Cutoff TmVar))
        (k64 : Hvl) (k65 : Hvl) :
        (shifthvl_TmVar c13 k64 k65) -> (shifthvl_TmVar c13 (HS TyVar k64) (HS TyVar k65)).
  Inductive shifthvl_TyVar : (forall (c13 : (Cutoff TyVar)) (k64 : Hvl) (k65 : Hvl) ,
    Prop) :=
    | shifthvl_TyVar_here
        (k64 : Hvl) :
        (shifthvl_TyVar C0 k64 (HS TyVar k64))
    | shifthvl_TyVar_there_TmVar
        (c13 : (Cutoff TyVar))
        (k64 : Hvl) (k65 : Hvl) :
        (shifthvl_TyVar c13 k64 k65) -> (shifthvl_TyVar c13 (HS TmVar k64) (HS TmVar k65))
    | shifthvl_TyVar_there_TyVar
        (c13 : (Cutoff TyVar))
        (k64 : Hvl) (k65 : Hvl) :
        (shifthvl_TyVar c13 k64 k65) -> (shifthvl_TyVar (CS c13) (HS TyVar k64) (HS TyVar k65)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k64 : Hvl) {c13 : (Cutoff TmVar)} {k65 : Hvl} {k66 : Hvl} ,
      (shifthvl_TmVar c13 k65 k66) -> (shifthvl_TmVar (weakenCutoffTmVar c13 k64) (appendHvl k65 k64) (appendHvl k66 k64))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TyVar  :
    (forall (k64 : Hvl) {c13 : (Cutoff TyVar)} {k65 : Hvl} {k66 : Hvl} ,
      (shifthvl_TyVar c13 k65 k66) -> (shifthvl_TyVar (weakenCutoffTyVar c13 k64) (appendHvl k65 k64) (appendHvl k66 k64))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_wfindex_TmVar  :
    (forall (c13 : (Cutoff TmVar)) (k64 : Hvl) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) (x59 : (Index TmVar)) ,
      (wfindex k64 x59) -> (wfindex k65 (shiftIndex c13 x59))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_wfindex_TyVar  :
    (forall (c13 : (Cutoff TmVar)) (k64 : Hvl) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) (X50 : (Index TyVar)) ,
      (wfindex k64 X50) -> (wfindex k65 X50)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma tshift_wfindex_TmVar  :
    (forall (c13 : (Cutoff TyVar)) (k64 : Hvl) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) (x59 : (Index TmVar)) ,
      (wfindex k64 x59) -> (wfindex k65 x59)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma tshift_wfindex_TyVar  :
    (forall (c13 : (Cutoff TyVar)) (k64 : Hvl) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) (X50 : (Index TyVar)) ,
      (wfindex k64 X50) -> (wfindex k65 (tshiftIndex c13 X50))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_wfTy : (forall (k64 : Hvl) ,
    (forall (S45 : Ty) (wf : (wfTy k64 S45)) ,
      (forall (c13 : (Cutoff TmVar)) (k65 : Hvl) ,
        (shifthvl_TmVar c13 k64 k65) -> (wfTy k65 S45)))) := (ind_wfTy (fun (k64 : Hvl) (S45 : Ty) (wf : (wfTy k64 S45)) =>
    (forall (c13 : (Cutoff TmVar)) (k65 : Hvl) ,
      (shifthvl_TmVar c13 k64 k65) -> (wfTy k65 S45))) (fun (k64 : Hvl) (X50 : (Index TyVar)) (wfi : (wfindex k64 X50)) (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTy_TVar k65 (shift_wfindex_TyVar c13 k64 k65 ins X50 wfi))) (fun (k64 : Hvl) (T1 : Ty) (wf : (wfTy k64 T1)) IHT5 (T2 : Ty) (wf0 : (wfTy k64 T2)) IHT6 (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTy_TArr k65 (IHT5 c13 k65 (weaken_shifthvl_TmVar H0 ins)) (IHT6 c13 k65 (weaken_shifthvl_TmVar H0 ins)))) (fun (k64 : Hvl) (T : Ty) (wf : (wfTy (HS TyVar k64) T)) IHT5 (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTy_TAll k65 (IHT5 c13 (HS TyVar k65) (weaken_shifthvl_TmVar (HS TyVar H0) ins)))) (fun (k64 : Hvl) (T1 : Ty) (wf : (wfTy k64 T1)) IHT5 (T2 : Ty) (wf0 : (wfTy k64 T2)) IHT6 (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTy_TProd k65 (IHT5 c13 k65 (weaken_shifthvl_TmVar H0 ins)) (IHT6 c13 k65 (weaken_shifthvl_TmVar H0 ins))))).
  Definition tshift_wfTy : (forall (k64 : Hvl) ,
    (forall (S45 : Ty) (wf : (wfTy k64 S45)) ,
      (forall (c13 : (Cutoff TyVar)) (k65 : Hvl) ,
        (shifthvl_TyVar c13 k64 k65) -> (wfTy k65 (tshiftTy c13 S45))))) := (ind_wfTy (fun (k64 : Hvl) (S45 : Ty) (wf : (wfTy k64 S45)) =>
    (forall (c13 : (Cutoff TyVar)) (k65 : Hvl) ,
      (shifthvl_TyVar c13 k64 k65) -> (wfTy k65 (tshiftTy c13 S45)))) (fun (k64 : Hvl) (X50 : (Index TyVar)) (wfi : (wfindex k64 X50)) (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTy_TVar k65 (tshift_wfindex_TyVar c13 k64 k65 ins X50 wfi))) (fun (k64 : Hvl) (T1 : Ty) (wf : (wfTy k64 T1)) IHT5 (T2 : Ty) (wf0 : (wfTy k64 T2)) IHT6 (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTy_TArr k65 (IHT5 c13 k65 (weaken_shifthvl_TyVar H0 ins)) (IHT6 c13 k65 (weaken_shifthvl_TyVar H0 ins)))) (fun (k64 : Hvl) (T : Ty) (wf : (wfTy (HS TyVar k64) T)) IHT5 (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTy_TAll k65 (IHT5 (CS c13) (HS TyVar k65) (weaken_shifthvl_TyVar (HS TyVar H0) ins)))) (fun (k64 : Hvl) (T1 : Ty) (wf : (wfTy k64 T1)) IHT5 (T2 : Ty) (wf0 : (wfTy k64 T2)) IHT6 (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTy_TProd k65 (IHT5 c13 k65 (weaken_shifthvl_TyVar H0 ins)) (IHT6 c13 k65 (weaken_shifthvl_TyVar H0 ins))))).
  Definition shift_wfPat : (forall (k64 : Hvl) ,
    (forall (p34 : Pat) (wf : (wfPat k64 p34)) ,
      (forall (c13 : (Cutoff TmVar)) (k65 : Hvl) ,
        (shifthvl_TmVar c13 k64 k65) -> (wfPat k65 p34)))) := (fun (k64 : Hvl) =>
    (ind_wfPat k64 (fun (p34 : Pat) (wf : (wfPat k64 p34)) =>
      (forall (c13 : (Cutoff TmVar)) (k65 : Hvl) ,
        (shifthvl_TmVar c13 k64 k65) -> (wfPat k65 p34))) (fun (T : Ty) (wf : (wfTy k64 T)) (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
      (wfPat_PVar k65 (shift_wfTy k64 T wf c13 k65 (weaken_shifthvl_TmVar H0 ins)))) (fun (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
      (wfPat_PTVar k65)) (fun (p1 : Pat) (wf : (wfPat k64 p1)) IHp0 (p2 : Pat) (wf0 : (wfPat k64 p2)) IHp3 (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
      (wfPat_PProd k65 (IHp0 c13 k65 (weaken_shifthvl_TmVar H0 ins)) (IHp3 c13 k65 (weaken_shifthvl_TmVar H0 ins)))))).
  Definition tshift_wfPat : (forall (k64 : Hvl) ,
    (forall (p34 : Pat) (wf : (wfPat k64 p34)) ,
      (forall (c13 : (Cutoff TyVar)) (k65 : Hvl) ,
        (shifthvl_TyVar c13 k64 k65) -> (wfPat k65 (tshiftPat c13 p34))))) := (fun (k64 : Hvl) =>
    (ind_wfPat k64 (fun (p34 : Pat) (wf : (wfPat k64 p34)) =>
      (forall (c13 : (Cutoff TyVar)) (k65 : Hvl) ,
        (shifthvl_TyVar c13 k64 k65) -> (wfPat k65 (tshiftPat c13 p34)))) (fun (T : Ty) (wf : (wfTy k64 T)) (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
      (wfPat_PVar k65 (tshift_wfTy k64 T wf c13 k65 (weaken_shifthvl_TyVar H0 ins)))) (fun (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
      (wfPat_PTVar k65)) (fun (p1 : Pat) (wf : (wfPat k64 p1)) IHp0 (p2 : Pat) (wf0 : (wfPat k64 p2)) IHp3 (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
      (wfPat_PProd k65 (IHp0 c13 k65 (weaken_shifthvl_TyVar H0 ins)) (IHp3 c13 k65 (weaken_shifthvl_TyVar H0 ins)))))).
  Definition shift_wfTm : (forall (k64 : Hvl) ,
    (forall (s18 : Tm) (wf : (wfTm k64 s18)) ,
      (forall (c13 : (Cutoff TmVar)) (k65 : Hvl) ,
        (shifthvl_TmVar c13 k64 k65) -> (wfTm k65 (shiftTm c13 s18))))) := (ind_wfTm (fun (k64 : Hvl) (s18 : Tm) (wf : (wfTm k64 s18)) =>
    (forall (c13 : (Cutoff TmVar)) (k65 : Hvl) ,
      (shifthvl_TmVar c13 k64 k65) -> (wfTm k65 (shiftTm c13 s18)))) (fun (k64 : Hvl) (x59 : (Index TmVar)) (wfi : (wfindex k64 x59)) (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTm_Var k65 (shift_wfindex_TmVar c13 k64 k65 ins x59 wfi))) (fun (k64 : Hvl) (T : Ty) (wf : (wfTy k64 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k64) t)) IHt170 (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTm_Abs k65 (shift_wfTy k64 T wf c13 k65 (weaken_shifthvl_TmVar H0 ins)) (IHt170 (CS c13) (HS TmVar k65) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k64 : Hvl) (t1 : Tm) (wf : (wfTm k64 t1)) IHt170 (t2 : Tm) (wf0 : (wfTm k64 t2)) IHt171 (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTm_App k65 (IHt170 c13 k65 (weaken_shifthvl_TmVar H0 ins)) (IHt171 c13 k65 (weaken_shifthvl_TmVar H0 ins)))) (fun (k64 : Hvl) (t : Tm) (wf : (wfTm (HS TyVar k64) t)) IHt170 (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTm_TAbs k65 (IHt170 c13 (HS TyVar k65) (weaken_shifthvl_TmVar (HS TyVar H0) ins)))) (fun (k64 : Hvl) (t : Tm) (wf : (wfTm k64 t)) IHt170 (T : Ty) (wf0 : (wfTy k64 T)) (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTm_TApp k65 (IHt170 c13 k65 (weaken_shifthvl_TmVar H0 ins)) (shift_wfTy k64 T wf0 c13 k65 (weaken_shifthvl_TmVar H0 ins)))) (fun (k64 : Hvl) (t1 : Tm) (wf : (wfTm k64 t1)) IHt170 (t2 : Tm) (wf0 : (wfTm k64 t2)) IHt171 (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTm_Prod k65 (IHt170 c13 k65 (weaken_shifthvl_TmVar H0 ins)) (IHt171 c13 k65 (weaken_shifthvl_TmVar H0 ins)))) (fun (k64 : Hvl) (t1 : Tm) (wf : (wfTm k64 t1)) IHt170 (p : Pat) (wf0 : (wfPat k64 p)) (t2 : Tm) (wf1 : (wfTm (appendHvl k64 (bind p)) t2)) IHt171 (c13 : (Cutoff TmVar)) (k65 : Hvl) (ins : (shifthvl_TmVar c13 k64 k65)) =>
    (wfTm_Case k65 (IHt170 c13 k65 (weaken_shifthvl_TmVar H0 ins)) (shift_wfPat k64 p wf0 c13 k65 (weaken_shifthvl_TmVar H0 ins)) (IHt171 (weakenCutoffTmVar c13 (bind p)) (appendHvl k65 (bind p)) (weaken_shifthvl_TmVar (bind p) ins))))).
  Definition tshift_wfTm : (forall (k64 : Hvl) ,
    (forall (s18 : Tm) (wf : (wfTm k64 s18)) ,
      (forall (c13 : (Cutoff TyVar)) (k65 : Hvl) ,
        (shifthvl_TyVar c13 k64 k65) -> (wfTm k65 (tshiftTm c13 s18))))) := (ind_wfTm (fun (k64 : Hvl) (s18 : Tm) (wf : (wfTm k64 s18)) =>
    (forall (c13 : (Cutoff TyVar)) (k65 : Hvl) ,
      (shifthvl_TyVar c13 k64 k65) -> (wfTm k65 (tshiftTm c13 s18)))) (fun (k64 : Hvl) (x59 : (Index TmVar)) (wfi : (wfindex k64 x59)) (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTm_Var k65 (tshift_wfindex_TmVar c13 k64 k65 ins x59 wfi))) (fun (k64 : Hvl) (T : Ty) (wf : (wfTy k64 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k64) t)) IHt170 (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTm_Abs k65 (tshift_wfTy k64 T wf c13 k65 (weaken_shifthvl_TyVar H0 ins)) (IHt170 c13 (HS TmVar k65) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k64 : Hvl) (t1 : Tm) (wf : (wfTm k64 t1)) IHt170 (t2 : Tm) (wf0 : (wfTm k64 t2)) IHt171 (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTm_App k65 (IHt170 c13 k65 (weaken_shifthvl_TyVar H0 ins)) (IHt171 c13 k65 (weaken_shifthvl_TyVar H0 ins)))) (fun (k64 : Hvl) (t : Tm) (wf : (wfTm (HS TyVar k64) t)) IHt170 (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTm_TAbs k65 (IHt170 (CS c13) (HS TyVar k65) (weaken_shifthvl_TyVar (HS TyVar H0) ins)))) (fun (k64 : Hvl) (t : Tm) (wf : (wfTm k64 t)) IHt170 (T : Ty) (wf0 : (wfTy k64 T)) (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTm_TApp k65 (IHt170 c13 k65 (weaken_shifthvl_TyVar H0 ins)) (tshift_wfTy k64 T wf0 c13 k65 (weaken_shifthvl_TyVar H0 ins)))) (fun (k64 : Hvl) (t1 : Tm) (wf : (wfTm k64 t1)) IHt170 (t2 : Tm) (wf0 : (wfTm k64 t2)) IHt171 (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTm_Prod k65 (IHt170 c13 k65 (weaken_shifthvl_TyVar H0 ins)) (IHt171 c13 k65 (weaken_shifthvl_TyVar H0 ins)))) (fun (k64 : Hvl) (t1 : Tm) (wf : (wfTm k64 t1)) IHt170 (p : Pat) (wf0 : (wfPat k64 p)) (t2 : Tm) (wf1 : (wfTm (appendHvl k64 (bind p)) t2)) IHt171 (c13 : (Cutoff TyVar)) (k65 : Hvl) (ins : (shifthvl_TyVar c13 k64 k65)) =>
    (wfTm_Case k65 (IHt170 c13 k65 (weaken_shifthvl_TyVar H0 ins)) (tshift_wfPat k64 p wf0 c13 k65 (weaken_shifthvl_TyVar H0 ins)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k65) (f_equal2 appendHvl (eq_sym (stability_tshift_bind _ _)) (eq_refl H0))) eq_refl (IHt171 (weakenCutoffTyVar c13 (bind p)) (appendHvl k65 (bind p)) (weaken_shifthvl_TyVar (bind p) ins)))))).
End ShiftWellFormed.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : infra.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : shift.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : shift_wf.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : wf.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : infra.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : shift.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : shift_wf.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : infra.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift_wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : infra.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k64 : Hvl) : (forall (d52 : (Trace TmVar)) (k65 : Hvl) (k66 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k64 X0 (HS TmVar k64) k64)
    | substhvl_TmVar_there_TmVar
        {d52 : (Trace TmVar)}
        {k65 : Hvl} {k66 : Hvl} :
        (substhvl_TmVar k64 d52 k65 k66) -> (substhvl_TmVar k64 (XS TmVar d52) (HS TmVar k65) (HS TmVar k66))
    | substhvl_TmVar_there_TyVar
        {d52 : (Trace TmVar)}
        {k65 : Hvl} {k66 : Hvl} :
        (substhvl_TmVar k64 d52 k65 k66) -> (substhvl_TmVar k64 (XS TyVar d52) (HS TyVar k65) (HS TyVar k66)).
  Inductive substhvl_TyVar (k64 : Hvl) : (forall (d52 : (Trace TyVar)) (k65 : Hvl) (k66 : Hvl) ,
    Prop) :=
    | substhvl_TyVar_here :
        (substhvl_TyVar k64 X0 (HS TyVar k64) k64)
    | substhvl_TyVar_there_TmVar
        {d52 : (Trace TyVar)}
        {k65 : Hvl} {k66 : Hvl} :
        (substhvl_TyVar k64 d52 k65 k66) -> (substhvl_TyVar k64 (XS TmVar d52) (HS TmVar k65) (HS TmVar k66))
    | substhvl_TyVar_there_TyVar
        {d52 : (Trace TyVar)}
        {k65 : Hvl} {k66 : Hvl} :
        (substhvl_TyVar k64 d52 k65 k66) -> (substhvl_TyVar k64 (XS TyVar d52) (HS TyVar k65) (HS TyVar k66)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k65 : Hvl} (k64 : Hvl) {d52 : (Trace TmVar)} {k66 : Hvl} {k67 : Hvl} ,
      (substhvl_TmVar k65 d52 k66 k67) -> (substhvl_TmVar k65 (weakenTrace d52 k64) (appendHvl k66 k64) (appendHvl k67 k64))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TyVar  :
    (forall {k65 : Hvl} (k64 : Hvl) {d52 : (Trace TyVar)} {k66 : Hvl} {k67 : Hvl} ,
      (substhvl_TyVar k65 d52 k66 k67) -> (substhvl_TyVar k65 (weakenTrace d52 k64) (appendHvl k66 k64) (appendHvl k67 k64))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k64 : Hvl} {s18 : Tm} (wft : (wfTm k64 s18)) :
    (forall {d52 : (Trace TmVar)} {k65 : Hvl} {k66 : Hvl} ,
      (substhvl_TmVar k64 d52 k65 k66) -> (forall {x59 : (Index TmVar)} ,
        (wfindex k65 x59) -> (wfTm k66 (substIndex d52 s18 x59)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TyVar_wfindex_TyVar {k64 : Hvl} {S45 : Ty} (wft : (wfTy k64 S45)) :
    (forall {d52 : (Trace TyVar)} {k65 : Hvl} {k66 : Hvl} ,
      (substhvl_TyVar k64 d52 k65 k66) -> (forall {X50 : (Index TyVar)} ,
        (wfindex k65 X50) -> (wfTy k66 (tsubstIndex d52 S45 X50)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TmVar_wfindex_TyVar {k64 : Hvl} :
    (forall {d52 : (Trace TmVar)} {k65 : Hvl} {k66 : Hvl} ,
      (substhvl_TmVar k64 d52 k65 k66) -> (forall {X50 : (Index TyVar)} ,
        (wfindex k65 X50) -> (wfindex k66 X50))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TmVar {k64 : Hvl} :
    (forall {d52 : (Trace TyVar)} {k65 : Hvl} {k66 : Hvl} ,
      (substhvl_TyVar k64 d52 k65 k66) -> (forall {x59 : (Index TmVar)} ,
        (wfindex k65 x59) -> (wfindex k66 x59))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_TmVar_wfTy {k64 : Hvl} : (forall (k65 : Hvl) ,
    (forall (S45 : Ty) (wf0 : (wfTy k65 S45)) ,
      (forall {d52 : (Trace TmVar)} {k66 : Hvl} ,
        (substhvl_TmVar k64 d52 k65 k66) -> (wfTy k66 S45)))) := (ind_wfTy (fun (k65 : Hvl) (S45 : Ty) (wf0 : (wfTy k65 S45)) =>
    (forall {d52 : (Trace TmVar)} {k66 : Hvl} ,
      (substhvl_TmVar k64 d52 k65 k66) -> (wfTy k66 S45))) (fun (k65 : Hvl) {X50 : (Index TyVar)} (wfi : (wfindex k65 X50)) {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTy_TVar k66 (substhvl_TmVar_wfindex_TyVar del wfi))) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT5 (T2 : Ty) (wf1 : (wfTy k65 T2)) IHT6 {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTy_TArr k66 (IHT5 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHT6 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)))) (fun (k65 : Hvl) (T : Ty) (wf0 : (wfTy (HS TyVar k65) T)) IHT5 {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTy_TAll k66 (IHT5 (weakenTrace d52 (HS TyVar H0)) (HS TyVar k66) (weaken_substhvl_TmVar (HS TyVar H0) del)))) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT5 (T2 : Ty) (wf1 : (wfTy k65 T2)) IHT6 {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTy_TProd k66 (IHT5 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHT6 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del))))).
  Definition substhvl_TyVar_wfTy {k64 : Hvl} {S45 : Ty} (wf : (wfTy k64 S45)) : (forall (k65 : Hvl) ,
    (forall (S46 : Ty) (wf0 : (wfTy k65 S46)) ,
      (forall {d52 : (Trace TyVar)} {k66 : Hvl} ,
        (substhvl_TyVar k64 d52 k65 k66) -> (wfTy k66 (tsubstTy d52 S45 S46))))) := (ind_wfTy (fun (k65 : Hvl) (S46 : Ty) (wf0 : (wfTy k65 S46)) =>
    (forall {d52 : (Trace TyVar)} {k66 : Hvl} ,
      (substhvl_TyVar k64 d52 k65 k66) -> (wfTy k66 (tsubstTy d52 S45 S46)))) (fun (k65 : Hvl) {X50 : (Index TyVar)} (wfi : (wfindex k65 X50)) {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (substhvl_TyVar_wfindex_TyVar wf del wfi)) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT5 (T2 : Ty) (wf1 : (wfTy k65 T2)) IHT6 {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTy_TArr k66 (IHT5 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHT6 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)))) (fun (k65 : Hvl) (T : Ty) (wf0 : (wfTy (HS TyVar k65) T)) IHT5 {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTy_TAll k66 (IHT5 (weakenTrace d52 (HS TyVar H0)) (HS TyVar k66) (weaken_substhvl_TyVar (HS TyVar H0) del)))) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT5 (T2 : Ty) (wf1 : (wfTy k65 T2)) IHT6 {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTy_TProd k66 (IHT5 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHT6 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del))))).
  Definition substhvl_TmVar_wfPat {k64 : Hvl} : (forall (k65 : Hvl) ,
    (forall (p34 : Pat) (wf0 : (wfPat k65 p34)) ,
      (forall {d52 : (Trace TmVar)} {k66 : Hvl} ,
        (substhvl_TmVar k64 d52 k65 k66) -> (wfPat k66 p34)))) := (fun (k65 : Hvl) =>
    (ind_wfPat k65 (fun (p34 : Pat) (wf0 : (wfPat k65 p34)) =>
      (forall {d52 : (Trace TmVar)} {k66 : Hvl} ,
        (substhvl_TmVar k64 d52 k65 k66) -> (wfPat k66 p34))) (fun (T : Ty) (wf0 : (wfTy k65 T)) {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
      (wfPat_PVar k66 (substhvl_TmVar_wfTy k65 T wf0 (weaken_substhvl_TmVar H0 del)))) (fun {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
      (wfPat_PTVar k66)) (fun (p1 : Pat) (wf0 : (wfPat k65 p1)) IHp0 (p2 : Pat) (wf1 : (wfPat k65 p2)) IHp3 {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
      (wfPat_PProd k66 (IHp0 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHp3 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)))))).
  Definition substhvl_TyVar_wfPat {k64 : Hvl} {S45 : Ty} (wf : (wfTy k64 S45)) : (forall (k65 : Hvl) ,
    (forall (p34 : Pat) (wf0 : (wfPat k65 p34)) ,
      (forall {d52 : (Trace TyVar)} {k66 : Hvl} ,
        (substhvl_TyVar k64 d52 k65 k66) -> (wfPat k66 (tsubstPat d52 S45 p34))))) := (fun (k65 : Hvl) =>
    (ind_wfPat k65 (fun (p34 : Pat) (wf0 : (wfPat k65 p34)) =>
      (forall {d52 : (Trace TyVar)} {k66 : Hvl} ,
        (substhvl_TyVar k64 d52 k65 k66) -> (wfPat k66 (tsubstPat d52 S45 p34)))) (fun (T : Ty) (wf0 : (wfTy k65 T)) {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
      (wfPat_PVar k66 (substhvl_TyVar_wfTy wf k65 T wf0 (weaken_substhvl_TyVar H0 del)))) (fun {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
      (wfPat_PTVar k66)) (fun (p1 : Pat) (wf0 : (wfPat k65 p1)) IHp0 (p2 : Pat) (wf1 : (wfPat k65 p2)) IHp3 {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
      (wfPat_PProd k66 (IHp0 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHp3 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)))))).
  Definition substhvl_TmVar_wfTm {k64 : Hvl} {s18 : Tm} (wf : (wfTm k64 s18)) : (forall (k65 : Hvl) ,
    (forall (s19 : Tm) (wf0 : (wfTm k65 s19)) ,
      (forall {d52 : (Trace TmVar)} {k66 : Hvl} ,
        (substhvl_TmVar k64 d52 k65 k66) -> (wfTm k66 (substTm d52 s18 s19))))) := (ind_wfTm (fun (k65 : Hvl) (s19 : Tm) (wf0 : (wfTm k65 s19)) =>
    (forall {d52 : (Trace TmVar)} {k66 : Hvl} ,
      (substhvl_TmVar k64 d52 k65 k66) -> (wfTm k66 (substTm d52 s18 s19)))) (fun (k65 : Hvl) {x59 : (Index TmVar)} (wfi : (wfindex k65 x59)) {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k65 : Hvl) (T : Ty) (wf0 : (wfTy k65 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k65) t)) IHt170 {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTm_Abs k66 (substhvl_TmVar_wfTy k65 T wf0 (weaken_substhvl_TmVar H0 del)) (IHt170 (weakenTrace d52 (HS TmVar H0)) (HS TmVar k66) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm k65 t2)) IHt171 {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTm_App k66 (IHt170 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHt171 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)))) (fun (k65 : Hvl) (t : Tm) (wf0 : (wfTm (HS TyVar k65) t)) IHt170 {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTm_TAbs k66 (IHt170 (weakenTrace d52 (HS TyVar H0)) (HS TyVar k66) (weaken_substhvl_TmVar (HS TyVar H0) del)))) (fun (k65 : Hvl) (t : Tm) (wf0 : (wfTm k65 t)) IHt170 (T : Ty) (wf1 : (wfTy k65 T)) {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTm_TApp k66 (IHt170 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)) (substhvl_TmVar_wfTy k65 T wf1 (weaken_substhvl_TmVar H0 del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm k65 t2)) IHt171 {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTm_Prod k66 (IHt170 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHt171 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (p : Pat) (wf1 : (wfPat k65 p)) (t2 : Tm) (wf2 : (wfTm (appendHvl k65 (bind p)) t2)) IHt171 {d52 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d52 k65 k66)) =>
    (wfTm_Case k66 (IHt170 (weakenTrace d52 H0) k66 (weaken_substhvl_TmVar H0 del)) (substhvl_TmVar_wfPat k65 p wf1 (weaken_substhvl_TmVar H0 del)) (IHt171 (weakenTrace d52 (bind p)) (appendHvl k66 (bind p)) (weaken_substhvl_TmVar (bind p) del))))).
  Definition substhvl_TyVar_wfTm {k64 : Hvl} {S45 : Ty} (wf : (wfTy k64 S45)) : (forall (k65 : Hvl) ,
    (forall (s18 : Tm) (wf0 : (wfTm k65 s18)) ,
      (forall {d52 : (Trace TyVar)} {k66 : Hvl} ,
        (substhvl_TyVar k64 d52 k65 k66) -> (wfTm k66 (tsubstTm d52 S45 s18))))) := (ind_wfTm (fun (k65 : Hvl) (s18 : Tm) (wf0 : (wfTm k65 s18)) =>
    (forall {d52 : (Trace TyVar)} {k66 : Hvl} ,
      (substhvl_TyVar k64 d52 k65 k66) -> (wfTm k66 (tsubstTm d52 S45 s18)))) (fun (k65 : Hvl) {x59 : (Index TmVar)} (wfi : (wfindex k65 x59)) {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTm_Var k66 (substhvl_TyVar_wfindex_TmVar del wfi))) (fun (k65 : Hvl) (T : Ty) (wf0 : (wfTy k65 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k65) t)) IHt170 {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTm_Abs k66 (substhvl_TyVar_wfTy wf k65 T wf0 (weaken_substhvl_TyVar H0 del)) (IHt170 (weakenTrace d52 (HS TmVar H0)) (HS TmVar k66) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm k65 t2)) IHt171 {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTm_App k66 (IHt170 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHt171 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)))) (fun (k65 : Hvl) (t : Tm) (wf0 : (wfTm (HS TyVar k65) t)) IHt170 {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTm_TAbs k66 (IHt170 (weakenTrace d52 (HS TyVar H0)) (HS TyVar k66) (weaken_substhvl_TyVar (HS TyVar H0) del)))) (fun (k65 : Hvl) (t : Tm) (wf0 : (wfTm k65 t)) IHt170 (T : Ty) (wf1 : (wfTy k65 T)) {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTm_TApp k66 (IHt170 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)) (substhvl_TyVar_wfTy wf k65 T wf1 (weaken_substhvl_TyVar H0 del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm k65 t2)) IHt171 {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTm_Prod k66 (IHt170 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHt171 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (p : Pat) (wf1 : (wfPat k65 p)) (t2 : Tm) (wf2 : (wfTm (appendHvl k65 (bind p)) t2)) IHt171 {d52 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d52 k65 k66)) =>
    (wfTm_Case k66 (IHt170 (weakenTrace d52 H0) k66 (weaken_substhvl_TyVar H0 del)) (substhvl_TyVar_wfPat wf k65 p wf1 (weaken_substhvl_TyVar H0 del)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k66) (f_equal2 appendHvl (eq_sym (stability_tsubst_bind _ _ _)) (eq_refl H0))) eq_refl (IHt171 (weakenTrace d52 (bind p)) (appendHvl k66 (bind p)) (weaken_substhvl_TyVar (bind p) del)))))).
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : wf.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : infra.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst_wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : wf.
Fixpoint subhvl_TmVar_TyVar (k64 : Hvl) {struct k64} : Prop :=
  match k64 with
    | (H0) => True
    | (HS a k64) => match a with
      | (TmVar) => (subhvl_TmVar_TyVar k64)
      | (TyVar) => (subhvl_TmVar_TyVar k64)
    end
  end.
Lemma subhvl_TmVar_TyVar_append  :
  (forall (k64 : Hvl) (k65 : Hvl) ,
    (subhvl_TmVar_TyVar k64) -> (subhvl_TmVar_TyVar k65) -> (subhvl_TmVar_TyVar (appendHvl k64 k65))).
Proof.
  needleGenericSubHvlAppend .
Qed.
 Hint Resolve subhvl_TmVar_TyVar_append : infra.
 Hint Resolve subhvl_TmVar_TyVar_append : wf.
Section Context.
  Inductive Ctx : Type :=
    | empty 
    | evar (G : Ctx) (T : Ty)
    | etvar (G : Ctx).
  Fixpoint appendCtx (G : Ctx) (G0 : Ctx) : Ctx :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendCtx G G1) T)
      | (etvar G1) => (etvar (appendCtx G G1))
    end.
  Fixpoint domainCtx (G : Ctx) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainCtx G0))
      | (etvar G0) => (HS TyVar (domainCtx G0))
    end.
  Lemma appendCtx_assoc  :
    (forall (G : Ctx) (G0 : Ctx) (G1 : Ctx) ,
      ((appendCtx G (appendCtx G0 G1)) =
      (appendCtx (appendCtx G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainCtx_appendCtx  :
    (forall (G : Ctx) (G0 : Ctx) ,
      ((domainCtx (appendCtx G G0)) =
      (appendHvl (domainCtx G) (domainCtx G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shiftCtx (c13 : (Cutoff TmVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shiftCtx c13 G0) T)
      | (etvar G0) => (etvar (shiftCtx c13 G0))
    end.
  Fixpoint tshiftCtx (c13 : (Cutoff TyVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (tshiftCtx c13 G0) (tshiftTy (weakenCutoffTyVar c13 (domainCtx G0)) T))
      | (etvar G0) => (etvar (tshiftCtx c13 G0))
    end.
  Fixpoint substCtx (d52 : (Trace TmVar)) (s18 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (substCtx d52 s18 G0) T)
      | (etvar G0) => (etvar (substCtx d52 s18 G0))
    end.
  Fixpoint tsubstCtx (d52 : (Trace TyVar)) (S45 : Ty) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (tsubstCtx d52 S45 G0) (tsubstTy (weakenTrace d52 (domainCtx G0)) S45 T))
      | (etvar G0) => (etvar (tsubstCtx d52 S45 G0))
    end.
  Lemma domainCtx_shiftCtx  :
    (forall (c13 : (Cutoff TmVar)) (G : Ctx) ,
      ((domainCtx (shiftCtx c13 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_tshiftCtx  :
    (forall (c13 : (Cutoff TyVar)) (G : Ctx) ,
      ((domainCtx (tshiftCtx c13 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_substCtx  :
    (forall (d52 : (Trace TmVar)) (s18 : Tm) (G : Ctx) ,
      ((domainCtx (substCtx d52 s18 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainCtx_tsubstCtx  :
    (forall (d52 : (Trace TyVar)) (S45 : Ty) (G : Ctx) ,
      ((domainCtx (tsubstCtx d52 S45 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainCtx_appendCtx : interaction.
 Hint Rewrite domainCtx_appendCtx : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shiftCtx_appendCtx  :
      (forall (c13 : (Cutoff TmVar)) (G : Ctx) (G0 : Ctx) ,
        ((shiftCtx c13 (appendCtx G G0)) =
        (appendCtx (shiftCtx c13 G) (shiftCtx (weakenCutoffTmVar c13 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma tshiftCtx_appendCtx  :
      (forall (c13 : (Cutoff TyVar)) (G : Ctx) (G0 : Ctx) ,
        ((tshiftCtx c13 (appendCtx G G0)) =
        (appendCtx (tshiftCtx c13 G) (tshiftCtx (weakenCutoffTyVar c13 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma substCtx_appendCtx  :
      (forall (d52 : (Trace TmVar)) (s18 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((substCtx d52 s18 (appendCtx G G0)) =
        (appendCtx (substCtx d52 s18 G) (substCtx (weakenTrace d52 (domainCtx G)) s18 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma tsubstCtx_appendCtx  :
      (forall (d52 : (Trace TyVar)) (S45 : Ty) (G : Ctx) (G0 : Ctx) ,
        ((tsubstCtx d52 S45 (appendCtx G G0)) =
        (appendCtx (tsubstCtx d52 S45 G) (tsubstCtx (weakenTrace d52 (domainCtx G)) S45 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Ctx -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Ctx}
          {T79 : Ty} :
          (wfTy (domainCtx G) T79) -> (lookup_evar (evar G T79) I0 T79)
      | lookup_evar_there_evar
          {G : Ctx} {x59 : (Index TmVar)}
          {T79 : Ty} {T80 : Ty} :
          (lookup_evar G x59 T79) -> (lookup_evar (evar G T80) (IS x59) T79)
      | lookup_evar_there_etvar
          {G : Ctx} {x59 : (Index TmVar)}
          {T79 : Ty} :
          (lookup_evar G x59 T79) -> (lookup_evar (etvar G) x59 (tshiftTy C0 T79)).
    Inductive lookup_etvar : Ctx -> (Index TyVar) -> Prop :=
      | lookup_etvar_here {G : Ctx}
          : (lookup_etvar (etvar G) I0)
      | lookup_etvar_there_evar
          {G : Ctx} {X50 : (Index TyVar)}
          {T79 : Ty} :
          (lookup_etvar G X50) -> (lookup_etvar (evar G T79) X50)
      | lookup_etvar_there_etvar
          {G : Ctx} {X50 : (Index TyVar)}
          :
          (lookup_etvar G X50) -> (lookup_etvar (etvar G) (IS X50)).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Ctx) (S45 : Ty) (S46 : Ty) ,
        (lookup_evar (evar G S45) I0 S46) -> (S45 =
        S46)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Ctx} {x59 : (Index TmVar)} ,
        (forall {S45 : Ty} ,
          (lookup_evar G x59 S45) -> (forall {S46 : Ty} ,
            (lookup_evar G x59 S46) -> (S45 =
            S46)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Ctx} {x59 : (Index TmVar)} {S45 : Ty} ,
        (lookup_evar G x59 S45) -> (wfTy (domainCtx G) S45)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Ctx} (G0 : Ctx) {x59 : (Index TmVar)} {T79 : Ty} ,
        (lookup_evar G x59 T79) -> (lookup_evar (appendCtx G G0) (weakenIndexTmVar x59 (domainCtx G0)) (weakenTy T79 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_etvar  :
      (forall {G : Ctx} (G0 : Ctx) {X50 : (Index TyVar)} ,
        (lookup_etvar G X50) -> (lookup_etvar (appendCtx G G0) (weakenIndexTyVar X50 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Ctx} {x59 : (Index TmVar)} {S45 : Ty} ,
        (lookup_evar G x59 S45) -> (wfindex (domainCtx G) x59)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_etvar_wfindex  :
      (forall {G : Ctx} {X50 : (Index TyVar)} ,
        (lookup_etvar G X50) -> (wfindex (domainCtx G) X50)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Ctx -> Ctx -> Prop :=
    | shift_evar_here {G : Ctx}
        {T79 : Ty} :
        (shift_evar C0 G (evar G T79))
    | shiftevar_there_evar
        {c13 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c13 G G0) -> (shift_evar (CS c13) (evar G T) (evar G0 T))
    | shiftevar_there_etvar
        {c13 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} :
        (shift_evar c13 G G0) -> (shift_evar c13 (etvar G) (etvar G0)).
  Lemma weaken_shift_evar  :
    (forall (G : Ctx) {c13 : (Cutoff TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evar c13 G0 G1) -> (shift_evar (weakenCutoffTmVar c13 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_etvar : (Cutoff TyVar) -> Ctx -> Ctx -> Prop :=
    | shift_etvar_here {G : Ctx} :
        (shift_etvar C0 G (etvar G))
    | shiftetvar_there_evar
        {c13 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_etvar c13 G G0) -> (shift_etvar c13 (evar G T) (evar G0 (tshiftTy c13 T)))
    | shiftetvar_there_etvar
        {c13 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} :
        (shift_etvar c13 G G0) -> (shift_etvar (CS c13) (etvar G) (etvar G0)).
  Lemma weaken_shift_etvar  :
    (forall (G : Ctx) {c13 : (Cutoff TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_etvar c13 G0 G1) -> (shift_etvar (weakenCutoffTyVar c13 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (tshiftCtx c13 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c13 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evar c13 G G0) -> (shifthvl_TmVar c13 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_etvar_shifthvl_TyVar  :
    (forall {c13 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_etvar c13 G G0) -> (shifthvl_TyVar c13 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Ctx) (T79 : Ty) (s18 : Tm) : (Trace TmVar) -> Ctx -> Ctx -> Prop :=
    | subst_evar_here :
        (subst_evar G T79 s18 X0 (evar G T79) G)
    | subst_evar_there_evar
        {d52 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evar G T79 s18 d52 G0 G1) -> (subst_evar G T79 s18 (XS TmVar d52) (evar G0 T) (evar G1 T))
    | subst_evar_there_etvar
        {d52 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} :
        (subst_evar G T79 s18 d52 G0 G1) -> (subst_evar G T79 s18 (XS TyVar d52) (etvar G0) (etvar G1)).
  Lemma weaken_subst_evar {G : Ctx} {T79 : Ty} {s18 : Tm} :
    (forall (G0 : Ctx) {d52 : (Trace TmVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evar G T79 s18 d52 G1 G2) -> (subst_evar G T79 s18 (weakenTrace d52 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_etvar (G : Ctx) (S45 : Ty) : (Trace TyVar) -> Ctx -> Ctx -> Prop :=
    | subst_etvar_here :
        (subst_etvar G S45 X0 (etvar G) G)
    | subst_etvar_there_evar
        {d52 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_etvar G S45 d52 G0 G1) -> (subst_etvar G S45 (XS TmVar d52) (evar G0 T) (evar G1 (tsubstTy d52 S45 T)))
    | subst_etvar_there_etvar
        {d52 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} :
        (subst_etvar G S45 d52 G0 G1) -> (subst_etvar G S45 (XS TyVar d52) (etvar G0) (etvar G1)).
  Lemma weaken_subst_etvar {G : Ctx} {S45 : Ty} :
    (forall (G0 : Ctx) {d52 : (Trace TyVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_etvar G S45 d52 G1 G2) -> (subst_etvar G S45 (weakenTrace d52 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (tsubstCtx d52 S45 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Ctx} {s18 : Tm} {T79 : Ty} :
    (forall {d52 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evar G T79 s18 d52 G0 G1) -> (substhvl_TmVar (domainCtx G) d52 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_etvar_substhvl_TyVar {G : Ctx} {S45 : Ty} :
    (forall {d52 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_etvar G S45 d52 G0 G1) -> (substhvl_TyVar (domainCtx G) d52 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainCtx_tshiftCtx : interaction.
 Hint Rewrite domainCtx_tshiftCtx : env_domain_shift.
 Hint Rewrite domainCtx_tsubstCtx : interaction.
 Hint Rewrite domainCtx_tsubstCtx : env_domain_subst.
 Hint Rewrite tshiftCtx_appendCtx : interaction.
 Hint Rewrite tshiftCtx_appendCtx : env_shift_append.
 Hint Rewrite tsubstCtx_appendCtx : interaction.
 Hint Rewrite tsubstCtx_appendCtx : env_shift_append.
 Hint Constructors lookup_evar lookup_etvar : infra.
 Hint Constructors lookup_evar lookup_etvar : lookup.
 Hint Constructors lookup_evar lookup_etvar : shift.
 Hint Constructors lookup_evar lookup_etvar : subst.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : infra.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : lookup.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : shift.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Ctx} (G0 : Ctx) {T79 : Ty} (wf : (wfTy (domainCtx G) T79)) ,
    (lookup_evar (appendCtx (evar G T79) G0) (weakenIndexTmVar I0 (domainCtx G0)) (weakenTy T79 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_etvar_here  :
  (forall {G : Ctx} (G0 : Ctx) ,
    (lookup_etvar (appendCtx (etvar G) G0) (weakenIndexTyVar I0 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfPat wfTm wfTy : infra.
 Hint Constructors wfPat wfTm wfTy : wf.
 Hint Extern 10 ((wfPat _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (TVar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TArr _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TAll _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TProd _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfPat _ _)) => match goal with
  | H : (wfPat _ (PVar _)) |- _ => inversion H; subst; clear H
  | H : (wfPat _ (PTVar)) |- _ => inversion H; subst; clear H
  | H : (wfPat _ (PProd _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (TAbs _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (TApp _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Prod _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Case _ _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf : infra.
 Hint Resolve lookup_evar_wf : wf.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : infra.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : wf.
 Hint Constructors shift_evar shift_etvar : infra.
 Hint Constructors shift_evar shift_etvar : shift.
 Hint Constructors shift_evar shift_etvar : subst.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : infra.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : wf.
 Hint Constructors subst_evar subst_etvar : infra.
 Hint Constructors subst_evar subst_etvar : subst.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : infra.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c13 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c13 G G0)) {x59 : (Index TmVar)} {T79 : Ty} ,
    (lookup_evar G x59 T79) -> (lookup_evar G0 (shiftIndex c13 x59) T79)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_etvar  :
  (forall {c13 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c13 G G0)) {X50 : (Index TyVar)} ,
    (lookup_etvar G X50) -> (lookup_etvar G0 X50)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_evar  :
  (forall {c13 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c13 G G0)) {x59 : (Index TmVar)} {T79 : Ty} ,
    (lookup_evar G x59 T79) -> (lookup_evar G0 x59 (tshiftTy c13 T79))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_etvar  :
  (forall {c13 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c13 G G0)) {X50 : (Index TyVar)} ,
    (lookup_etvar G X50) -> (lookup_etvar G0 (tshiftIndex c13 X50))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : infra.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : lookup.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : shift.
Lemma subst_evar_lookup_etvar {G : Ctx} {T80 : Ty} {s18 : Tm} :
  (forall {d52 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evar G T80 s18 d52 G0 G1)) {X50 : (Index TyVar)} ,
    (lookup_etvar G0 X50) -> (lookup_etvar G1 X50)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_etvar_lookup_evar {G : Ctx} {S45 : Ty} (wf : (wfTy (domainCtx G) S45)) :
  (forall {d52 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_etvar G S45 d52 G0 G1)) {x59 : (Index TmVar)} {T80 : Ty} ,
    (lookup_evar G0 x59 T80) -> (lookup_evar G1 x59 (tsubstTy d52 S45 T80))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : infra.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : lookup.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : subst.
Fixpoint size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (TVar X) => 1
    | (TArr T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (TAll T) => (plus 1 (size_Ty T))
    | (TProd T0 T3) => (plus 1 (plus (size_Ty T0) (size_Ty T3)))
  end.
Fixpoint size_Pat (p : Pat) {struct p} : nat :=
  match p with
    | (PVar T) => (plus 1 (size_Ty T))
    | (PTVar) => 1
    | (PProd p1 p2) => (plus 1 (plus (size_Pat p1) (size_Pat p2)))
  end.
Fixpoint size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (Abs T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (App t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
    | (TAbs t0) => (plus 1 (size_Tm t0))
    | (TApp t3 T0) => (plus 1 (plus (size_Tm t3) (size_Ty T0)))
    | (Prod t4 t5) => (plus 1 (plus (size_Tm t4) (size_Tm t5)))
    | (Case t6 p t7) => (plus 1 (plus (size_Tm t6) (plus (size_Pat p) (size_Tm t7))))
  end.
Lemma tshift_size_Ty  :
  (forall (S45 : Ty) (c10 : (Cutoff TyVar)) ,
    ((size_Ty (tshiftTy c10 S45)) =
    (size_Ty S45))).
Proof.
  apply_mutual_ind ind_Ty.
  - intros X44 c10.
  reflexivity.
  - intros T69 IHT69 T70 IHT70.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT69 c11)).
  + exact ((IHT70 c11)).
  - intros T71 IHT71.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT71 (CS c11))).
  - intros T72 IHT72 T73 IHT73.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT72 c11)).
  + exact ((IHT73 c11)).
Qed.

Lemma tshift_size_Pat  :
  (forall (p30 : Pat) (c11 : (Cutoff TyVar)) ,
    ((size_Pat (tshiftPat c11 p30)) =
    (size_Pat p30))).
Proof.
  apply_mutual_ind ind_Pat.
  - intros T74.
  pose proof ((tshift_size_Ty T74)) as IHT74.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT74 c11)).
  - intros c11; simpl; apply ((f_equal S)); reflexivity.
  - intros p30 IHp30 p31 IHp31.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHp30 c11)).
  + exact ((IHp31 c11)).
Qed.

Lemma shift_size_Tm  :
  (forall (s18 : Tm) (c11 : (Cutoff TmVar)) ,
    ((size_Tm (shiftTm c11 s18)) =
    (size_Tm s18))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros x55 c11.
  reflexivity.
  - intros T75 t152 IHt152.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt152 (CS c12))).
  - intros t153 IHt153 t154 IHt154.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt153 c12)).
  + exact ((IHt154 c12)).
  - intros t155 IHt155.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt155 c12)).
  - intros t156 IHt156 T76.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt156 c12)).
  + reflexivity.
  - intros t157 IHt157 t158 IHt158.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt157 c12)).
  + exact ((IHt158 c12)).
  - intros t159 IHt159 p32 t160 IHt160.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt159 c12)).
  + reflexivity.
  + exact ((IHt160 (weakenCutoffTmVar c12 (bind p32)))).
Qed.

Lemma tshift_size_Tm  :
  (forall (s18 : Tm) (c12 : (Cutoff TyVar)) ,
    ((size_Tm (tshiftTm c12 s18)) =
    (size_Tm s18))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros X48 c12.
  reflexivity.
  - intros T77 t161 IHt161.
  pose proof ((tshift_size_Ty T77)) as IHT77.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT77 c13)).
  + exact ((IHt161 c13)).
  - intros t162 IHt162 t163 IHt163.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt162 c13)).
  + exact ((IHt163 c13)).
  - intros t164 IHt164.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt164 (CS c13))).
  - intros t165 IHt165 T78.
  pose proof ((tshift_size_Ty T78)) as IHT78.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt165 c13)).
  + exact ((IHT78 c13)).
  - intros t166 IHt166 t167 IHt167.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt166 c13)).
  + exact ((IHt167 c13)).
  - intros t168 IHt168 p33 t169 IHt169.
  pose proof ((tshift_size_Pat p33)) as IHp33.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt168 c13)).
  + exact ((IHp33 c13)).
  + exact ((IHt169 (weakenCutoffTyVar c13 (bind p33)))).
Qed.

 Hint Rewrite tshift_size_Pat shift_size_Tm tshift_size_Tm tshift_size_Ty : interaction.
 Hint Rewrite tshift_size_Pat shift_size_Tm tshift_size_Tm tshift_size_Ty : shift_size.
Lemma weaken_size_Pat  :
  (forall (k64 : Hvl) (p34 : Pat) ,
    ((size_Pat (weakenPat p34 k64)) =
    (size_Pat p34))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k64 : Hvl) (s18 : Tm) ,
    ((size_Tm (weakenTm s18 k64)) =
    (size_Tm s18))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k64 : Hvl) (S45 : Ty) ,
    ((size_Ty (weakenTy S45 k64)) =
    (size_Ty S45))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Pat weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Pat weaken_size_Tm weaken_size_Ty : weaken_size.